//! Module for testing / demonstrating how configuration parameters
//! (knobs) affect the compactor output.
//!
//! See [crate::layout] module for detailed documentation

use crate::layouts::{all_overlapping_l0_files, layout_setup_builder, run_layout_scenario, ONE_MB};

/// This test show the effects of the split_percentage parameter
#[tokio::test]
async fn all_overlapping_l0_split_percentage() {
    test_helpers::maybe_start_logging();

    //
    // configure to leave a smaller part of  data on the leading
    // edge.
    //
    // Expect to see two output files, one having the oldest 95
    // percent of the data, and one having the newest 5 percent.
    let setup = layout_setup_builder()
        .await
        .with_split_percentage(95)
        .build()
        .await;

    // create virtual files
    let setup = all_overlapping_l0_files(setup).await;

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** Simulation run 0, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[190005]). 10 Input Files, 90mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 90mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,190005] 10ns 85.5mb|---------------------------------------L1.?----------------------------------------|     "
    - "L1.?[190006,200000] 10ns 4.5mb                                                                                     |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10"
    - "  Creating 2 files"
    - "**** Final Output Files (90mb written)"
    - "L1                                                                                                                 "
    - "L1.11[100,190005] 10ns 85.5mb|---------------------------------------L1.11---------------------------------------|     "
    - "L1.12[190006,200000] 10ns 4.5mb                                                                                     |L1.12|"
    "###
    );
}

/// This test show the effects of the percentage_max_file_size parameter
#[tokio::test]
async fn all_overlapping_l0_max_file_size() {
    test_helpers::maybe_start_logging();

    //
    // configure the max file size percentage to be 95%
    //
    // This means that the compactor will only apply the "leading
    // edge" optimization for data that is greater than 95 * 100MB =
    // 95MB in total
    //
    // Expect to see a single output file
    let setup = layout_setup_builder()
        .await
        .with_percentage_max_file_size(95)
        .build()
        .await;

    // create virtual files
    let setup = all_overlapping_l0_files(setup).await;

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** Simulation run 0, type=compact(TotalSizeLessThanMaxCompactSize). 10 Input Files, 90mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 90mb total:"
    - "L1, all files 90mb                                                                                                 "
    - "L1.?[100,200000] 10ns    |-----------------------------------------L1.?------------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10"
    - "  Creating 1 files"
    - "**** Final Output Files (90mb written)"
    - "L1, all files 90mb                                                                                                 "
    - "L1.11[100,200000] 10ns   |-----------------------------------------L1.11-----------------------------------------| "
    "###
    );
}

/// This test show how the split percentage and max_file size ineract
#[tokio::test]
async fn all_overlapping_l0_split_percentage_and_max_file_size() {
    test_helpers::maybe_start_logging();

    // configure the max file size percentage to be 95% and the max file size to be 10MB
    //
    // This means that the compactor will only apply the "leading
    // edge" optimization for data that is greater than 95 * 10MB =
    // 9.5 MB in total
    //
    // Expect to see 2 ~40mb files, 1 file that is smaller
    let setup = layout_setup_builder()
        .await
        .with_max_desired_file_size_bytes(40 * ONE_MB)
        .with_percentage_max_file_size(95)
        .build()
        .await;

    // create virtual files
    let setup = all_overlapping_l0_files(setup).await;

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** Simulation run 0, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[88945, 177790]). 10 Input Files, 90mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 90mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,88945] 10ns 40mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[88946,177790] 10ns 40mb                                        |----------------L1.?-----------------|           "
    - "L1.?[177791,200000] 10ns 10mb                                                                                |-L1.?--| "
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10"
    - "  Creating 3 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[177789]). 2 Input Files, 50mb total:"
    - "L1                                                                                                                 "
    - "L1.13[177791,200000] 10ns 10mb                                                                        |-----L1.13-----| "
    - "L1.12[88946,177790] 10ns 40mb|--------------------------------L1.12---------------------------------|                  "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 50mb total:"
    - "L2                                                                                                                 "
    - "L2.?[88946,177789] 10ns 40mb|--------------------------------L2.?---------------------------------|                   "
    - "L2.?[177790,200000] 10ns 10mb                                                                        |-----L2.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.12, L1.13"
    - "  Upgrading 1 files level to CompactionLevel::L2: L1.11"
    - "  Creating 2 files"
    - "**** Final Output Files (140mb written)"
    - "L2                                                                                                                 "
    - "L2.11[100,88945] 10ns 40mb|----------------L2.11-----------------|                                                  "
    - "L2.14[88946,177789] 10ns 40mb                                        |----------------L2.14----------------|           "
    - "L2.15[177790,200000] 10ns 10mb                                                                                |-L2.15-| "
    "###
    );
}

/// This test show how the max_num_files_per_plan setting works
#[tokio::test]
async fn all_overlapping_l0_max_num_files_per_plan() {
    test_helpers::maybe_start_logging();

    // configure the max num files per plan to 3
    let setup = layout_setup_builder()
        .await
        .with_max_num_files_per_plan(3)
        .build()
        .await;

    // create virtual files
    let setup = all_overlapping_l0_files(setup).await;

    // expect that each plan the compactor runs has at most three input files
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** Simulation run 0, type=split(CompactAndSplitOutput(ManySmallFiles))(split_times=[160020]). 3 Input Files, 27mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 27mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,160020] 3ns 21.6mb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[160021,200000] 3ns 5.4mb                                                                        |-----L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.1, L0.2, L0.3"
    - "  Creating 2 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(ManySmallFiles))(split_times=[160020]). 3 Input Files, 27mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 27mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,160020] 6ns 21.6mb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[160021,200000] 6ns 5.4mb                                                                        |-----L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.4, L0.5, L0.6"
    - "  Creating 2 files"
    - "**** Simulation run 2, type=split(CompactAndSplitOutput(ManySmallFiles))(split_times=[160020]). 3 Input Files, 27mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 27mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,160020] 9ns 21.6mb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[160021,200000] 9ns 5.4mb                                                                        |-----L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.7, L0.8, L0.9"
    - "  Creating 2 files"
    - "**** Simulation run 3, type=compact(ManySmallFiles). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.?[100,200000] 10ns    |-----------------------------------------L0.?------------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 1 files: L0.10"
    - "  Creating 1 files"
    - "**** Simulation run 4, type=split(CompactAndSplitOutput(ManySmallFiles))(split_times=[160020]). 3 Input Files, 48.6mb total:"
    - "L0                                                                                                                 "
    - "L0.11[100,160020] 3ns 21.6mb|--------------------------------L0.11---------------------------------|                  "
    - "L0.12[160021,200000] 3ns 5.4mb                                                                        |-----L0.12-----| "
    - "L0.13[100,160020] 6ns 21.6mb|--------------------------------L0.13---------------------------------|                  "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 48.6mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,160020] 6ns 38.88mb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[160021,200000] 6ns 9.72mb                                                                        |-----L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.11, L0.12, L0.13"
    - "  Creating 2 files"
    - "**** Simulation run 5, type=split(CompactAndSplitOutput(ManySmallFiles))(split_times=[160020]). 3 Input Files, 32.4mb total:"
    - "L0                                                                                                                 "
    - "L0.14[160021,200000] 6ns 5.4mb                                                                        |-----L0.14-----| "
    - "L0.15[100,160020] 9ns 21.6mb|--------------------------------L0.15---------------------------------|                  "
    - "L0.16[160021,200000] 9ns 5.4mb                                                                        |-----L0.16-----| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 32.4mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,160020] 9ns 25.92mb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[160021,200000] 9ns 6.48mb                                                                        |-----L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.14, L0.15, L0.16"
    - "  Creating 2 files"
    - "**** Simulation run 6, type=compact(ManySmallFiles). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.17[100,200000] 10ns   |-----------------------------------------L0.17-----------------------------------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.?[100,200000] 10ns    |-----------------------------------------L0.?------------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 1 files: L0.17"
    - "  Creating 1 files"
    - "**** Simulation run 7, type=split(CompactAndSplitOutput(ManySmallFiles))(split_times=[160020]). 3 Input Files, 74.52mb total:"
    - "L0                                                                                                                 "
    - "L0.18[100,160020] 6ns 38.88mb|--------------------------------L0.18---------------------------------|                  "
    - "L0.19[160021,200000] 6ns 9.72mb                                                                        |-----L0.19-----| "
    - "L0.20[100,160020] 9ns 25.92mb|--------------------------------L0.20---------------------------------|                  "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 74.52mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,160020] 9ns 59.62mb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[160021,200000] 9ns 14.9mb                                                                        |-----L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.18, L0.19, L0.20"
    - "  Creating 2 files"
    - "**** Simulation run 8, type=compact(ManySmallFiles). 2 Input Files, 15.48mb total:"
    - "L0                                                                                                                 "
    - "L0.21[160021,200000] 9ns 6.48mb                                                                        |-----L0.21-----| "
    - "L0.22[100,200000] 10ns 9mb|-----------------------------------------L0.22-----------------------------------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 15.48mb total:"
    - "L0, all files 15.48mb                                                                                              "
    - "L0.?[100,200000] 10ns    |-----------------------------------------L0.?------------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.21, L0.22"
    - "  Creating 1 files"
    - "**** Simulation run 9, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[160020]). 3 Input Files, 90mb total:"
    - "L0                                                                                                                 "
    - "L0.25[100,200000] 10ns 15.48mb|-----------------------------------------L0.25-----------------------------------------| "
    - "L0.24[160021,200000] 9ns 14.9mb                                                                        |-----L0.24-----| "
    - "L0.23[100,160020] 9ns 59.62mb|--------------------------------L0.23---------------------------------|                  "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 90mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,160020] 10ns 72mb|---------------------------------L1.?---------------------------------|                  "
    - "L1.?[160021,200000] 10ns 18mb                                                                        |-----L1.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.23, L0.24, L0.25"
    - "  Creating 2 files"
    - "**** Final Output Files (360mb written)"
    - "L1                                                                                                                 "
    - "L1.26[100,160020] 10ns 72mb|--------------------------------L1.26---------------------------------|                  "
    - "L1.27[160021,200000] 10ns 18mb                                                                        |-----L1.27-----| "
    "###
    );
}

#[tokio::test]
async fn all_overlapping_l0_max_input_bytes_per_partition() {
    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder()
        .await
        .with_max_desired_file_size_bytes(10 * ONE_MB)
        .build()
        .await;

    // create virtual files
    let setup = all_overlapping_l0_files(setup).await;

    // expect that each plan the compactor runs has at most three input files
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** Simulation run 0, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 1ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 1ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 1ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 1ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 1ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 1ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 1ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 1, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 2ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 2ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 2ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 2ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 2ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 2ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 2ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 2, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 3ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 3ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 3ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 3ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 3ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 3ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 3ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 3, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 4ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 4ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 4ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 4ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 4ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 4ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 4ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 4, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 5ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 5ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 5ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 5ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 5ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 5ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 5ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 5, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 6ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 6ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 6ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 6ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 6ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 6ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 6ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 6, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 7ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 7ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 7ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 7ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 7ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 7ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 7ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 7, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 8ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 8ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 8ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 8ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 8ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 8ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 8ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 8, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 9ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 9ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 9ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 9ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 9ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 9ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 9ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 9, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 10ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 10ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 10ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 10ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 10ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 10ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 10ns 1.29mb                                                                             |---L0.?---| "
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10"
    - "  Creating 70 files"
    - "**** Simulation run 10, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[67700, 135300]). 23 Input Files, 29.57mb total:"
    - "L0                                                                                                                 "
    - "L0.11[100,28657] 1ns 1.29mb|--L0.11---|                                                                              "
    - "L0.12[28658,57214] 1ns 1.29mb            |--L0.12---|                                                                  "
    - "L0.13[57215,85771] 1ns 1.29mb                         |--L0.13---|                                                     "
    - "L0.14[85772,114328] 1ns 1.29mb                                      |--L0.14---|                                        "
    - "L0.15[114329,142885] 1ns 1.29mb                                                   |--L0.15---|                           "
    - "L0.16[142886,171442] 1ns 1.29mb                                                                |--L0.16---|              "
    - "L0.17[171443,200000] 1ns 1.29mb                                                                             |--L0.17---| "
    - "L0.18[100,28657] 2ns 1.29mb|--L0.18---|                                                                              "
    - "L0.19[28658,57214] 2ns 1.29mb            |--L0.19---|                                                                  "
    - "L0.20[57215,85771] 2ns 1.29mb                         |--L0.20---|                                                     "
    - "L0.21[85772,114328] 2ns 1.29mb                                      |--L0.21---|                                        "
    - "L0.22[114329,142885] 2ns 1.29mb                                                   |--L0.22---|                           "
    - "L0.23[142886,171442] 2ns 1.29mb                                                                |--L0.23---|              "
    - "L0.24[171443,200000] 2ns 1.29mb                                                                             |--L0.24---| "
    - "L0.25[100,28657] 3ns 1.29mb|--L0.25---|                                                                              "
    - "L0.26[28658,57214] 3ns 1.29mb            |--L0.26---|                                                                  "
    - "L0.27[57215,85771] 3ns 1.29mb                         |--L0.27---|                                                     "
    - "L0.28[85772,114328] 3ns 1.29mb                                      |--L0.28---|                                        "
    - "L0.29[114329,142885] 3ns 1.29mb                                                   |--L0.29---|                           "
    - "L0.30[142886,171442] 3ns 1.29mb                                                                |--L0.30---|              "
    - "L0.31[171443,200000] 3ns 1.29mb                                                                             |--L0.31---| "
    - "L0.32[100,28657] 4ns 1.29mb|--L0.32---|                                                                              "
    - "L0.33[28658,57214] 4ns 1.29mb            |--L0.33---|                                                                  "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.57mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,67700] 4ns 10mb |------------L1.?------------|                                                            "
    - "L1.?[67701,135300] 4ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[135301,200000] 4ns 9.57mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 23 files: L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33"
    - "  Creating 3 files"
    - "**** Simulation run 11, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.81[100,67700] 4ns     |-----------------------------------------L1.81------------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,28657] 4ns 4.22mb|----------------L1.?----------------|                                                    "
    - "L1.?[28658,57214] 4ns 4.22mb                                      |----------------L1.?----------------|              "
    - "L1.?[57215,67700] 4ns 1.55mb                                                                            |---L1.?----| "
    - "**** Simulation run 12, type=split(HighL0OverlapTotalBacklog)(split_times=[85771, 114328]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.82[67701,135300] 4ns  |-----------------------------------------L1.82------------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[67701,85771] 4ns 2.67mb|---------L1.?---------|                                                                  "
    - "L1.?[85772,114328] 4ns 4.22mb                        |----------------L1.?----------------|                            "
    - "L1.?[114329,135300] 4ns 3.1mb                                                              |----------L1.?-----------| "
    - "**** Simulation run 13, type=split(HighL0OverlapTotalBacklog)(split_times=[142885, 171442]). 1 Input Files, 9.57mb total:"
    - "L1, all files 9.57mb                                                                                               "
    - "L1.83[135301,200000] 4ns |-----------------------------------------L1.83------------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 9.57mb total:"
    - "L1                                                                                                                 "
    - "L1.?[135301,142885] 4ns 1.12mb|--L1.?--|                                                                                "
    - "L1.?[142886,171442] 4ns 4.22mb          |----------------L1.?-----------------|                                         "
    - "L1.?[171443,200000] 4ns 4.22mb                                                  |----------------L1.?-----------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.81, L1.82, L1.83"
    - "  Creating 9 files"
    - "**** Simulation run 14, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.34[57215,85771] 4ns   |-----------------------------------------L0.34------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 4ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 4ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 15, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.36[114329,142885] 4ns |-----------------------------------------L0.36------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 4ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 4ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 16, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.41[57215,85771] 5ns   |-----------------------------------------L0.41------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 5ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 5ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 17, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.43[114329,142885] 5ns |-----------------------------------------L0.43------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 5ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 5ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 18, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.48[57215,85771] 6ns   |-----------------------------------------L0.48------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 6ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 6ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 19, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.50[114329,142885] 6ns |-----------------------------------------L0.50------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 6ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 6ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 20, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.55[57215,85771] 7ns   |-----------------------------------------L0.55------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 7ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 7ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 21, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.57[114329,142885] 7ns |-----------------------------------------L0.57------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 7ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 7ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 22, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.62[57215,85771] 8ns   |-----------------------------------------L0.62------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 8ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 8ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 23, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.64[114329,142885] 8ns |-----------------------------------------L0.64------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 8ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 8ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 24, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.69[57215,85771] 9ns   |-----------------------------------------L0.69------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 9ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 9ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 25, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.71[114329,142885] 9ns |-----------------------------------------L0.71------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 9ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 9ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 26, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.76[57215,85771] 10ns  |-----------------------------------------L0.76------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 10ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 10ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 27, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.78[114329,142885] 10ns|-----------------------------------------L0.78------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 10ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 10ns 349.69kb                                                                  |--------L0.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.34, L0.36, L0.41, L0.43, L0.48, L0.50, L0.55, L0.57, L0.62, L0.64, L0.69, L0.71, L0.76, L0.78"
    - "  Creating 28 files"
    - "**** Simulation run 28, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[109041, 160867]). 14 Input Files, 27.55mb total:"
    - "L0                                                                                                                 "
    - "L0.93[57215,67700] 4ns 483.39kb|L0.93|                                                                                   "
    - "L0.94[67701,85771] 4ns 833.13kb      |--L0.94--|                                                                         "
    - "L0.35[85772,114328] 4ns 1.29mb                  |-----L0.35-----|                                                       "
    - "L0.95[114329,135300] 4ns 966.83kb                                    |---L0.95---|                                         "
    - "L0.96[135301,142885] 4ns 349.69kb                                                 |L0.96|                                  "
    - "L0.37[142886,171442] 4ns 1.29mb                                                      |-----L0.37-----|                   "
    - "L0.38[171443,200000] 4ns 1.29mb                                                                        |-----L0.38------|"
    - "L1                                                                                                                 "
    - "L1.86[57215,67700] 4ns 1.55mb|L1.86|                                                                                   "
    - "L1.87[67701,85771] 4ns 2.67mb      |--L1.87--|                                                                         "
    - "L1.88[85772,114328] 4ns 4.22mb                  |-----L1.88-----|                                                       "
    - "L1.89[114329,135300] 4ns 3.1mb                                    |---L1.89---|                                         "
    - "L1.90[135301,142885] 4ns 1.12mb                                                 |L1.90|                                  "
    - "L1.91[142886,171442] 4ns 4.22mb                                                      |-----L1.91-----|                   "
    - "L1.92[171443,200000] 4ns 4.22mb                                                                        |-----L1.92------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 27.55mb total:"
    - "L1                                                                                                                 "
    - "L1.?[57215,109041] 4ns 10mb|-------------L1.?-------------|                                                          "
    - "L1.?[109042,160867] 4ns 10mb                                |-------------L1.?-------------|                          "
    - "L1.?[160868,200000] 4ns 7.55mb                                                                 |---------L1.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.35, L0.37, L0.38, L1.86, L1.87, L1.88, L1.89, L1.90, L1.91, L1.92, L0.93, L0.94, L0.95, L0.96"
    - "  Creating 3 files"
    - "**** Simulation run 29, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.121[57215,109041] 4ns |-----------------------------------------L1.121-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[57215,85772] 4ns 5.51mb|---------------------L1.?----------------------|                                         "
    - "L1.?[85773,109041] 4ns 4.49mb                                                 |-----------------L1.?-----------------| "
    - "**** Simulation run 30, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.42[85772,114328] 5ns  |-----------------------------------------L0.42------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 5ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 5ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 31, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.49[85772,114328] 6ns  |-----------------------------------------L0.49------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 6ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 6ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 32, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.56[85772,114328] 7ns  |-----------------------------------------L0.56------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 7ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 7ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 33, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.63[85772,114328] 8ns  |-----------------------------------------L0.63------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 8ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 8ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 34, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.70[85772,114328] 9ns  |-----------------------------------------L0.70------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 9ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 9ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 35, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.77[85772,114328] 10ns |-----------------------------------------L0.77------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[85773,114328] 10ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 36, type=split(HighL0OverlapTotalBacklog)(split_times=[114329, 142886]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.122[109042,160867] 4ns|-----------------------------------------L1.122-----------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[109042,114329] 4ns 1.02mb|-L1.?--|                                                                                 "
    - "L1.?[114330,142886] 4ns 5.51mb         |---------------------L1.?----------------------|                                "
    - "L1.?[142887,160867] 4ns 3.47mb                                                          |------------L1.?-------------| "
    - "**** Simulation run 37, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.99[114329,135300] 5ns |-----------------------------------------L0.99------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 5ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 5ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 38, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.103[114329,135300] 6ns|-----------------------------------------L0.103-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 6ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 6ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 39, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.107[114329,135300] 7ns|-----------------------------------------L0.107-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 7ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 7ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 40, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.111[114329,135300] 8ns|-----------------------------------------L0.111-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 8ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 8ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 41, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.115[114329,135300] 9ns|-----------------------------------------L0.115-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 9ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 9ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 42, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.119[114329,135300] 10ns|-----------------------------------------L0.119-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 10ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 43, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.44[142886,171442] 5ns |-----------------------------------------L0.44------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 5ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 5ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 44, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.51[142886,171442] 6ns |-----------------------------------------L0.51------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 6ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 6ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 45, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.58[142886,171442] 7ns |-----------------------------------------L0.58------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 7ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 7ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 46, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.65[142886,171442] 8ns |-----------------------------------------L0.65------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 8ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 8ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 47, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.72[142886,171442] 9ns |-----------------------------------------L0.72------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 9ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 9ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 48, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.79[142886,171442] 10ns|-----------------------------------------L0.79------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 10ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 49, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 7.55mb total:"
    - "L1, all files 7.55mb                                                                                               "
    - "L1.123[160868,200000] 4ns|-----------------------------------------L1.123-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 7.55mb total:"
    - "L1                                                                                                                 "
    - "L1.?[160868,171443] 4ns 2.04mb|---------L1.?---------|                                                                  "
    - "L1.?[171444,200000] 4ns 5.51mb                        |-----------------------------L1.?------------------------------| "
    - "**** Simulation run 50, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.45[171443,200000] 5ns |-----------------------------------------L0.45------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 5ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 5ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 51, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.66[171443,200000] 8ns |-----------------------------------------L0.66------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 8ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 8ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 52, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.73[171443,200000] 9ns |-----------------------------------------L0.73------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 9ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 9ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 53, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.80[171443,200000] 10ns|-----------------------------------------L0.80------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 10ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 54, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.52[171443,200000] 6ns |-----------------------------------------L0.52------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 6ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 6ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 55, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.59[171443,200000] 7ns |-----------------------------------------L0.59------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 7ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 7ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 27 files: L0.42, L0.44, L0.45, L0.49, L0.51, L0.52, L0.56, L0.58, L0.59, L0.63, L0.65, L0.66, L0.70, L0.72, L0.73, L0.77, L0.79, L0.80, L0.99, L0.103, L0.107, L0.111, L0.115, L0.119, L1.121, L1.122, L1.123"
    - "  Creating 55 files"
    - "**** Simulation run 56, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.127[85773,114328] 5ns |-----------------------------------------L0.127-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 5ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 5ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 57, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.154[142887,171442] 5ns|-----------------------------------------L0.154-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 5ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 5ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 58, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.129[85773,114328] 6ns |-----------------------------------------L0.129-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 6ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 6ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 59, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.156[142887,171442] 6ns|-----------------------------------------L0.156-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 6ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 6ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 60, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.131[85773,114328] 7ns |-----------------------------------------L0.131-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 7ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 7ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 61, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.158[142887,171442] 7ns|-----------------------------------------L0.158-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 7ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 7ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 62, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.133[85773,114328] 8ns |-----------------------------------------L0.133-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 8ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 8ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 63, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.160[142887,171442] 8ns|-----------------------------------------L0.160-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 8ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 8ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 64, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.135[85773,114328] 9ns |-----------------------------------------L0.135-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 9ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 9ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 65, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.162[142887,171442] 9ns|-----------------------------------------L0.162-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 9ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 9ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 66, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.137[85773,114328] 10ns|-----------------------------------------L0.137-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 10ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 10ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 67, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.164[142887,171442] 10ns|-----------------------------------------L0.164-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 10ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 10ns 487.56kb                                                        |-------------L0.?--------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 12 files: L0.127, L0.129, L0.131, L0.133, L0.135, L0.137, L0.154, L0.156, L0.158, L0.160, L0.162, L0.164"
    - "  Creating 24 files"
    - "**** Simulation run 68, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[46513, 92926]). 13 Input Files, 24.61mb total:"
    - "L0                                                                                                                 "
    - "L0.39[100,28657] 5ns 1.29mb|-------L0.39--------|                                                                    "
    - "L0.40[28658,57214] 5ns 1.29mb                      |-------L0.40--------|                                              "
    - "L0.97[57215,67700] 5ns 483.39kb                                             |L0.97-|                                     "
    - "L0.98[67701,85771] 5ns 833.13kb                                                     |---L0.98----|                       "
    - "L0.126[85772,85772] 5ns 0b                                                                   |L0.126|               "
    - "L0.179[85773,109041] 5ns 1.05mb                                                                   |-----L0.179-----|     "
    - "L0.180[109042,114328] 5ns 243.76kb                                                                                     |L0.180|"
    - "L0.141[114329,114329] 5ns 0b                                                                                         |L0.141|"
    - "L1                                                                                                                 "
    - "L1.84[100,28657] 4ns 4.22mb|-------L1.84--------|                                                                    "
    - "L1.85[28658,57214] 4ns 4.22mb                      |-------L1.85--------|                                              "
    - "L1.124[57215,85772] 4ns 5.51mb                                             |-------L1.124-------|                       "
    - "L1.125[85773,109041] 4ns 4.49mb                                                                   |-----L1.125-----|     "
    - "L1.138[109042,114329] 4ns 1.02mb                                                                                     |L1.138|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.61mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,46513] 5ns 10mb |---------------L1.?---------------|                                                      "
    - "L1.?[46514,92926] 5ns 10mb                                    |---------------L1.?---------------|                  "
    - "L1.?[92927,114329] 5ns 4.61mb                                                                         |-----L1.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 13 files: L0.39, L0.40, L1.84, L1.85, L0.97, L0.98, L1.124, L1.125, L0.126, L1.138, L0.141, L0.179, L0.180"
    - "  Creating 3 files"
    - "**** Simulation run 69, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.203[100,46513] 5ns    |----------------------------------------L1.203-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,38176] 5ns 8.2mb|---------------------------------L1.?----------------------------------|                 "
    - "L1.?[38177,46513] 5ns 1.8mb                                                                         |-----L1.?-----| "
    - "**** Simulation run 70, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.47[28658,57214] 6ns   |-----------------------------------------L0.47------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 6ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 6ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 71, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.54[28658,57214] 7ns   |-----------------------------------------L0.54------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 7ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 7ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 72, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.61[28658,57214] 8ns   |-----------------------------------------L0.61------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 8ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 8ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 73, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.68[28658,57214] 9ns   |-----------------------------------------L0.68------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 9ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 9ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 74, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.75[28658,57214] 10ns  |-----------------------------------------L0.75------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 10ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 10ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 75, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.204[46514,92926] 5ns  |-----------------------------------------L1.204-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[46514,76252] 5ns 6.41mb|-------------------------L1.?--------------------------|                                 "
    - "L1.?[76253,92926] 5ns 3.59mb                                                         |-------------L1.?-------------| "
    - "**** Simulation run 76, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.102[67701,85771] 6ns  |-----------------------------------------L0.102-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 6ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 6ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 77, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.106[67701,85771] 7ns  |-----------------------------------------L0.106-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 7ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 7ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 78, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.110[67701,85771] 8ns  |-----------------------------------------L0.110-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 8ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 8ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 79, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.114[67701,85771] 9ns  |-----------------------------------------L0.114-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 9ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 9ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 80, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.118[67701,85771] 10ns |-----------------------------------------L0.118-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 10ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 10ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 12 files: L0.47, L0.54, L0.61, L0.68, L0.75, L0.102, L0.106, L0.110, L0.114, L0.118, L1.203, L1.204"
    - "  Creating 24 files"
    - "**** Simulation run 81, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.209[38177,57214] 6ns  |-----------------------------------------L0.209-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 6ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 6ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 82, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.183[85773,109041] 6ns |----------------------------------------L0.183-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 6ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 6ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 83, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.211[38177,57214] 7ns  |-----------------------------------------L0.211-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 7ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 7ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 84, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.187[85773,109041] 7ns |----------------------------------------L0.187-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 7ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 7ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 85, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.213[38177,57214] 8ns  |-----------------------------------------L0.213-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 8ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 8ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 86, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.191[85773,109041] 8ns |----------------------------------------L0.191-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 8ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 8ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 87, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.215[38177,57214] 9ns  |-----------------------------------------L0.215-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 9ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 9ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 88, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.195[85773,109041] 9ns |----------------------------------------L0.195-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 9ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 9ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 89, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.217[38177,57214] 10ns |-----------------------------------------L0.217-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 10ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 10ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 90, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.199[85773,109041] 10ns|----------------------------------------L0.199-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 10ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 10ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.183, L0.187, L0.191, L0.195, L0.199, L0.209, L0.211, L0.213, L0.215, L0.217"
    - "  Creating 20 files"
    - "**** Simulation run 91, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[156350, 198370]). 11 Input Files, 20.39mb total:"
    - "L0                                                                                                                 "
    - "L0.142[114330,135300] 5ns 966.83kb|-------L0.142-------|                                                                    "
    - "L0.100[135301,142885] 5ns 349.69kb                      |L0.100|                                                            "
    - "L0.153[142886,142886] 5ns 0b                             |L0.153|                                                     "
    - "L0.181[142887,160867] 5ns 828.96kb                              |-----L0.181-----|                                          "
    - "L0.182[160868,171442] 5ns 487.56kb                                                |-L0.182--|                               "
    - "L0.167[171443,171443] 5ns 0b                                                           |L0.167|                       "
    - "L0.168[171444,200000] 5ns 1.29mb                                                            |----------L0.168-----------| "
    - "L1                                                                                                                 "
    - "L1.139[114330,142886] 4ns 5.51mb|----------L1.139-----------|                                                             "
    - "L1.140[142887,160867] 4ns 3.47mb                              |-----L1.140-----|                                          "
    - "L1.165[160868,171443] 4ns 2.04mb                                                |-L1.165--|                               "
    - "L1.166[171444,200000] 4ns 5.51mb                                                            |----------L1.166-----------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 20.39mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,156350] 5ns 10mb|-------------------L1.?-------------------|                                              "
    - "L1.?[156351,198370] 5ns 10mb                                            |-------------------L1.?-------------------|  "
    - "L1.?[198371,200000] 5ns 397.47kb                                                                                        |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 11 files: L0.100, L1.139, L1.140, L0.142, L0.153, L1.165, L1.166, L0.167, L0.168, L0.181, L0.182"
    - "  Creating 3 files"
    - "**** Simulation run 92, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.250[114330,156350] 5ns|-----------------------------------------L1.250-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,142886] 5ns 6.8mb|---------------------------L1.?----------------------------|                             "
    - "L1.?[142887,156350] 5ns 3.2mb                                                             |-----------L1.?-----------| "
    - "**** Simulation run 93, type=split(HighL0OverlapTotalBacklog)(split_times=[171442]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.251[156351,198370] 5ns|-----------------------------------------L1.251-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[156351,171442] 5ns 3.59mb|-------------L1.?-------------|                                                          "
    - "L1.?[171443,198370] 5ns 6.41mb                                |-------------------------L1.?--------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.250, L1.251"
    - "  Creating 4 files"
    - "**** Simulation run 94, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.185[142887,160867] 6ns|-----------------------------------------L0.185-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 6ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 6ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 95, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.176[171444,200000] 6ns|-----------------------------------------L0.176-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 6ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 6ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 96, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.189[142887,160867] 7ns|-----------------------------------------L0.189-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 7ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 7ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 97, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.178[171444,200000] 7ns|-----------------------------------------L0.178-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 7ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 7ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 98, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.193[142887,160867] 8ns|-----------------------------------------L0.193-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 8ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 8ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 99, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.170[171444,200000] 8ns|-----------------------------------------L0.170-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 8ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 8ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 100, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.197[142887,160867] 9ns|-----------------------------------------L0.197-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 9ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 9ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 101, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.172[171444,200000] 9ns|-----------------------------------------L0.172-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 9ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 9ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 102, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.201[142887,160867] 10ns|-----------------------------------------L0.201-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 10ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 10ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 103, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.174[171444,200000] 10ns|-----------------------------------------L0.174-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 10ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 10ns 75.17kb                                                                                    |L0.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.170, L0.172, L0.174, L0.176, L0.178, L0.185, L0.189, L0.193, L0.197, L0.201"
    - "  Creating 20 files"
    - "**** Simulation run 104, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[38491, 76882]). 17 Input Files, 29.75mb total:"
    - "L0                                                                                                                 "
    - "L0.46[100,28657] 6ns 1.29mb|-------L0.46--------|                                                                    "
    - "L0.208[28658,38176] 6ns 438.81kb                      |L0.208|                                                            "
    - "L0.230[38177,46513] 6ns 384.33kb                              |L0.230|                                                    "
    - "L0.231[46514,57214] 6ns 493.38kb                                    |L0.231|                                              "
    - "L0.101[57215,67700] 6ns 483.39kb                                             |L0.101|                                     "
    - "L0.220[67701,76252] 6ns 394.25kb                                                     |L0.220|                             "
    - "L0.221[76253,85771] 6ns 438.88kb                                                            |L0.221|                      "
    - "L0.128[85772,85772] 6ns 0b                                                                   |L0.128|               "
    - "L0.232[85773,92926] 6ns 329.79kb                                                                   |L0.232|               "
    - "L0.233[92927,109041] 6ns 742.98kb                                                                         |--L0.233--|     "
    - "L0.184[109042,114328] 6ns 243.76kb                                                                                     |L0.184|"
    - "L0.143[114329,114329] 6ns 0b                                                                                         |L0.143|"
    - "L1                                                                                                                 "
    - "L1.206[100,38176] 5ns 8.2mb|----------L1.206-----------|                                                             "
    - "L1.207[38177,46513] 5ns 1.8mb                              |L1.207|                                                    "
    - "L1.218[46514,76252] 5ns 6.41mb                                    |-------L1.218--------|                               "
    - "L1.219[76253,92926] 5ns 3.59mb                                                            |--L1.219---|                 "
    - "L1.205[92927,114329] 5ns 4.61mb                                                                         |----L1.205----| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.75mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,38491] 6ns 10mb |------------L1.?------------|                                                            "
    - "L1.?[38492,76882] 6ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[76883,114329] 6ns 9.75mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 17 files: L0.46, L0.101, L0.128, L0.143, L0.184, L1.205, L1.206, L1.207, L0.208, L1.218, L1.219, L0.220, L0.221, L0.230, L0.231, L0.232, L0.233"
    - "  Creating 3 files"
    - "**** Simulation run 105, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.277[100,38491] 6ns    |-----------------------------------------L1.277-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,38176] 6ns 9.92mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[38177,38491] 6ns 84.02kb                                                                                         |L1.?|"
    - "**** Simulation run 106, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.278[38492,76882] 6ns  |-----------------------------------------L1.278-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[38492,76252] 6ns 9.84mb|-----------------------------------------L1.?-----------------------------------------|  "
    - "L1.?[76253,76882] 6ns 168.04kb                                                                                        |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.277, L1.278"
    - "  Creating 4 files"
    - "**** Simulation run 107, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.234[38177,46513] 7ns  |-----------------------------------------L0.234-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 7ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 7ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 108, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.223[76253,85771] 7ns  |-----------------------------------------L0.223-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 7ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 7ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "**** Simulation run 109, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.238[38177,46513] 8ns  |-----------------------------------------L0.238-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 8ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 8ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 110, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.225[76253,85771] 8ns  |-----------------------------------------L0.225-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 8ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 8ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "**** Simulation run 111, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.242[38177,46513] 9ns  |-----------------------------------------L0.242-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 9ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 9ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 112, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.227[76253,85771] 9ns  |-----------------------------------------L0.227-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 9ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 9ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "**** Simulation run 113, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.246[38177,46513] 10ns |-----------------------------------------L0.246-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 10ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 10ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 114, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.229[76253,85771] 10ns |-----------------------------------------L0.229-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 10ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 10ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.223, L0.225, L0.227, L0.229, L0.234, L0.238, L0.242, L0.246"
    - "  Creating 16 files"
    - "**** Simulation run 115, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[149665, 185000]). 14 Input Files, 24.25mb total:"
    - "L0                                                                                                                 "
    - "L0.144[114330,135300] 6ns 966.83kb|-------L0.144-------|                                                                    "
    - "L0.104[135301,142885] 6ns 349.69kb                      |L0.104|                                                            "
    - "L0.155[142886,142886] 6ns 0b                             |L0.155|                                                     "
    - "L0.257[142887,156350] 6ns 620.71kb                              |---L0.257---|                                              "
    - "L0.258[156351,160867] 6ns 208.25kb                                            |L0.258|                                      "
    - "L0.186[160868,171442] 6ns 487.56kb                                                |-L0.186--|                               "
    - "L0.175[171443,171443] 6ns 0b                                                           |L0.175|                       "
    - "L0.259[171444,198370] 6ns 1.21mb                                                            |----------L0.259----------|  "
    - "L0.260[198371,200000] 6ns 75.17kb                                                                                        |L0.260|"
    - "L1                                                                                                                 "
    - "L1.253[114330,142886] 5ns 6.8mb|----------L1.253-----------|                                                             "
    - "L1.254[142887,156350] 5ns 3.2mb                              |---L1.254---|                                              "
    - "L1.255[156351,171442] 5ns 3.59mb                                            |---L1.255----|                               "
    - "L1.256[171443,198370] 5ns 6.41mb                                                           |----------L1.256----------|   "
    - "L1.252[198371,200000] 5ns 397.47kb                                                                                        |L1.252|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.25mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,149665] 6ns 10mb|---------------L1.?----------------|                                                     "
    - "L1.?[149666,185000] 6ns 10mb                                     |---------------L1.?----------------|                "
    - "L1.?[185001,200000] 6ns 4.25mb                                                                          |----L1.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.104, L0.144, L0.155, L0.175, L0.186, L1.252, L1.253, L1.254, L1.255, L1.256, L0.257, L0.258, L0.259, L0.260"
    - "  Creating 3 files"
    - "**** Simulation run 116, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.300[114330,149665] 6ns|-----------------------------------------L1.300-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,142886] 6ns 8.08mb|---------------------------------L1.?---------------------------------|                  "
    - "L1.?[142887,149665] 6ns 1.92mb                                                                        |-----L1.?------| "
    - "**** Simulation run 117, type=split(HighL0OverlapTotalBacklog)(split_times=[171442]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.301[149666,185000] 6ns|-----------------------------------------L1.301-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[149666,171442] 6ns 6.16mb|------------------------L1.?-------------------------|                                   "
    - "L1.?[171443,185000] 6ns 3.84mb                                                       |--------------L1.?--------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.300, L1.301"
    - "  Creating 4 files"
    - "**** Simulation run 118, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.261[142887,156350] 7ns|----------------------------------------L0.261-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 7ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 7ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 119, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.263[171444,198370] 7ns|----------------------------------------L0.263-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 7ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 7ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 120, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.265[142887,156350] 8ns|----------------------------------------L0.265-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 8ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 8ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 121, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.267[171444,198370] 8ns|----------------------------------------L0.267-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 8ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 8ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 122, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.269[142887,156350] 9ns|----------------------------------------L0.269-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 9ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 9ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 123, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.271[171444,198370] 9ns|----------------------------------------L0.271-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 9ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 9ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 124, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.273[142887,156350] 10ns|----------------------------------------L0.273-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 10ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 10ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 125, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.275[171444,198370] 10ns|----------------------------------------L0.275-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 10ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 10ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.261, L0.263, L0.265, L0.267, L0.269, L0.271, L0.273, L0.275"
    - "  Creating 16 files"
    - "**** Simulation run 126, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[32834, 65568]). 12 Input Files, 23.46mb total:"
    - "L0                                                                                                                 "
    - "L0.53[100,28657] 7ns 1.29mb|-------------L0.53-------------|                                                         "
    - "L0.210[28658,38176] 7ns 438.81kb                                 |-L0.210--|                                              "
    - "L0.284[38177,38491] 7ns 14.48kb                                            |L0.284|                                      "
    - "L0.285[38492,46513] 7ns 369.86kb                                             |L0.285-|                                    "
    - "L0.235[46514,57214] 7ns 493.38kb                                                      |--L0.235--|                        "
    - "L0.105[57215,67700] 7ns 483.39kb                                                                  |--L0.105--|            "
    - "L0.222[67701,76252] 7ns 394.25kb                                                                               |-L0.222-| "
    - "L0.286[76253,76882] 7ns 29kb                                                                                         |L0.286|"
    - "L1                                                                                                                 "
    - "L1.280[100,38176] 6ns 9.92mb|------------------L1.280------------------|                                              "
    - "L1.281[38177,38491] 6ns 84.02kb                                            |L1.281|                                      "
    - "L1.282[38492,76252] 6ns 9.84mb                                             |------------------L1.282------------------| "
    - "L1.283[76253,76882] 6ns 168.04kb                                                                                         |L1.283|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 23.46mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,32834] 7ns 10mb |----------------L1.?----------------|                                                    "
    - "L1.?[32835,65568] 7ns 10mb                                      |----------------L1.?----------------|              "
    - "L1.?[65569,76882] 7ns 3.46mb                                                                            |---L1.?----| "
    - "Committing partition 1:"
    - "  Soft Deleting 12 files: L0.53, L0.105, L0.210, L0.222, L0.235, L1.280, L1.281, L1.282, L1.283, L0.284, L0.285, L0.286"
    - "  Creating 3 files"
    - "**** Simulation run 127, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.323[100,32834] 7ns    |-----------------------------------------L1.323-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25694] 7ns 7.82mb|--------------------------------L1.?--------------------------------|                    "
    - "L1.?[25695,32834] 7ns 2.18mb                                                                      |------L1.?-------| "
    - "**** Simulation run 128, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.60[100,28657] 8ns     |-----------------------------------------L0.60------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,25694] 8ns 1.15mb|-------------------------------------L0.?-------------------------------------|          "
    - "L0.?[25695,28657] 8ns 136.6kb                                                                                |-L0.?--| "
    - "**** Simulation run 129, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.67[100,28657] 9ns     |-----------------------------------------L0.67------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,25694] 9ns 1.15mb|-------------------------------------L0.?-------------------------------------|          "
    - "L0.?[25695,28657] 9ns 136.6kb                                                                                |-L0.?--| "
    - "**** Simulation run 130, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.74[100,28657] 10ns    |-----------------------------------------L0.74------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,25694] 10ns 1.15mb|-------------------------------------L0.?-------------------------------------|          "
    - "L0.?[25695,28657] 10ns 136.6kb                                                                                |-L0.?--| "
    - "**** Simulation run 131, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.324[32835,65568] 7ns  |-----------------------------------------L1.324-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[32835,51288] 7ns 5.64mb|----------------------L1.?----------------------|                                        "
    - "L1.?[51289,65568] 7ns 4.36mb                                                  |----------------L1.?-----------------| "
    - "**** Simulation run 132, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 493.38kb total:"
    - "L0, all files 493.38kb                                                                                             "
    - "L0.239[46514,57214] 8ns  |-----------------------------------------L0.239-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 493.38kb total:"
    - "L0                                                                                                                 "
    - "L0.?[46514,51288] 8ns 220.13kb|-----------------L0.?-----------------|                                                  "
    - "L0.?[51289,57214] 8ns 273.25kb                                        |---------------------L0.?----------------------| "
    - "**** Simulation run 133, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 493.38kb total:"
    - "L0, all files 493.38kb                                                                                             "
    - "L0.243[46514,57214] 9ns  |-----------------------------------------L0.243-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 493.38kb total:"
    - "L0                                                                                                                 "
    - "L0.?[46514,51288] 9ns 220.13kb|-----------------L0.?-----------------|                                                  "
    - "L0.?[51289,57214] 9ns 273.25kb                                        |---------------------L0.?----------------------| "
    - "**** Simulation run 134, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 493.38kb total:"
    - "L0, all files 493.38kb                                                                                             "
    - "L0.247[46514,57214] 10ns |-----------------------------------------L0.247-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 493.38kb total:"
    - "L0                                                                                                                 "
    - "L0.?[46514,51288] 10ns 220.13kb|-----------------L0.?-----------------|                                                  "
    - "L0.?[51289,57214] 10ns 273.25kb                                        |---------------------L0.?----------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.60, L0.67, L0.74, L0.239, L0.243, L0.247, L1.323, L1.324"
    - "  Creating 16 files"
    - "**** Simulation run 135, type=split(ReduceOverlap)(split_times=[32834]). 1 Input Files, 438.81kb total:"
    - "L0, all files 438.81kb                                                                                             "
    - "L0.212[28658,38176] 8ns  |-----------------------------------------L0.212-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.81kb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,32834] 8ns 192.53kb|----------------L0.?-----------------|                                                   "
    - "L0.?[32835,38176] 8ns 246.28kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 136, type=split(ReduceOverlap)(split_times=[65568]). 1 Input Files, 483.39kb total:"
    - "L0, all files 483.39kb                                                                                             "
    - "L0.109[57215,67700] 8ns  |-----------------------------------------L0.109-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 483.39kb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,65568] 8ns 385.1kb|--------------------------------L0.?---------------------------------|                   "
    - "L0.?[65569,67700] 8ns 98.29kb                                                                       |------L0.?------| "
    - "**** Simulation run 137, type=split(ReduceOverlap)(split_times=[32834]). 1 Input Files, 438.81kb total:"
    - "L0, all files 438.81kb                                                                                             "
    - "L0.214[28658,38176] 9ns  |-----------------------------------------L0.214-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.81kb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,32834] 9ns 192.53kb|----------------L0.?-----------------|                                                   "
    - "L0.?[32835,38176] 9ns 246.28kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 138, type=split(ReduceOverlap)(split_times=[65568]). 1 Input Files, 483.39kb total:"
    - "L0, all files 483.39kb                                                                                             "
    - "L0.113[57215,67700] 9ns  |-----------------------------------------L0.113-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 483.39kb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,65568] 9ns 385.1kb|--------------------------------L0.?---------------------------------|                   "
    - "L0.?[65569,67700] 9ns 98.29kb                                                                       |------L0.?------| "
    - "**** Simulation run 139, type=split(ReduceOverlap)(split_times=[32834]). 1 Input Files, 438.81kb total:"
    - "L0, all files 438.81kb                                                                                             "
    - "L0.216[28658,38176] 10ns |-----------------------------------------L0.216-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.81kb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,32834] 10ns 192.53kb|----------------L0.?-----------------|                                                   "
    - "L0.?[32835,38176] 10ns 246.28kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 140, type=split(ReduceOverlap)(split_times=[65568]). 1 Input Files, 483.39kb total:"
    - "L0, all files 483.39kb                                                                                             "
    - "L0.117[57215,67700] 10ns |-----------------------------------------L0.117-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 483.39kb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,65568] 10ns 385.1kb|--------------------------------L0.?---------------------------------|                   "
    - "L0.?[65569,67700] 10ns 98.29kb                                                                       |------L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 6 files: L0.109, L0.113, L0.117, L0.212, L0.214, L0.216"
    - "  Creating 12 files"
    - "**** Simulation run 141, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[108723, 140563]). 16 Input Files, 29.7mb total:"
    - "L0                                                                                                                 "
    - "L0.287[76883,85771] 7ns 409.88kb|L0.287|                                                                                  "
    - "L0.130[85772,85772] 7ns 0b        |L0.130|                                                                          "
    - "L0.236[85773,92926] 7ns 329.79kb        |L0.236|                                                                          "
    - "L0.237[92927,109041] 7ns 742.98kb               |---L0.237----|                                                            "
    - "L0.188[109042,114328] 7ns 243.76kb                              |L0.188|                                                    "
    - "L0.145[114329,114329] 7ns 0b                                   |L0.145|                                               "
    - "L0.146[114330,135300] 7ns 966.83kb                                   |-----L0.146------|                                    "
    - "L0.108[135301,142885] 7ns 349.69kb                                                       |L0.108|                           "
    - "L0.157[142886,142886] 7ns 0b                                                              |L0.157|                    "
    - "L0.307[142887,149665] 7ns 312.5kb                                                              |L0.307|                    "
    - "L0.308[149666,156350] 7ns 308.21kb                                                                     |L0.308|             "
    - "L0.262[156351,160867] 7ns 208.25kb                                                                           |L0.262|       "
    - "L1                                                                                                                 "
    - "L1.279[76883,114329] 6ns 9.75mb|-------------L1.279--------------|                                                       "
    - "L1.303[114330,142886] 6ns 8.08mb                                   |---------L1.303----------|                            "
    - "L1.304[142887,149665] 6ns 1.92mb                                                              |L1.304|                    "
    - "L1.305[149666,171442] 6ns 6.16mb                                                                     |------L1.305------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.7mb total:"
    - "L1                                                                                                                 "
    - "L1.?[76883,108723] 7ns 10mb|------------L1.?------------|                                                            "
    - "L1.?[108724,140563] 7ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[140564,171442] 7ns 9.7mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 16 files: L0.108, L0.130, L0.145, L0.146, L0.157, L0.188, L0.236, L0.237, L0.262, L1.279, L0.287, L1.303, L1.304, L1.305, L0.307, L0.308"
    - "  Creating 3 files"
    - "**** Simulation run 142, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.354[76883,108723] 7ns |-----------------------------------------L1.354-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[76883,108402] 7ns 9.9mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[108403,108723] 7ns 103.24kb                                                                                         |L1.?|"
    - "**** Simulation run 143, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 742.98kb total:"
    - "L0, all files 742.98kb                                                                                             "
    - "L0.241[92927,109041] 8ns |-----------------------------------------L0.241-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 742.98kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,108402] 8ns 713.51kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[108403,109041] 8ns 29.46kb                                                                                      |L0.?|"
    - "**** Simulation run 144, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 742.98kb total:"
    - "L0, all files 742.98kb                                                                                             "
    - "L0.245[92927,109041] 9ns |-----------------------------------------L0.245-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 742.98kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,108402] 9ns 713.51kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[108403,109041] 9ns 29.46kb                                                                                      |L0.?|"
    - "**** Simulation run 145, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 742.98kb total:"
    - "L0, all files 742.98kb                                                                                             "
    - "L0.249[92927,109041] 10ns|-----------------------------------------L0.249-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 742.98kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,108402] 10ns 713.51kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[108403,109041] 10ns 29.46kb                                                                                      |L0.?|"
    - "**** Simulation run 146, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.355[108724,140563] 7ns|-----------------------------------------L1.355-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[108724,139921] 7ns 9.8mb|-----------------------------------------L1.?-----------------------------------------|  "
    - "L1.?[139922,140563] 7ns 206.48kb                                                                                        |L1.?|"
    - "**** Simulation run 147, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 349.69kb total:"
    - "L0, all files 349.69kb                                                                                             "
    - "L0.112[135301,142885] 8ns|-----------------------------------------L0.112-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 349.69kb total:"
    - "L0                                                                                                                 "
    - "L0.?[135301,139921] 8ns 213.02kb|------------------------L0.?------------------------|                                    "
    - "L0.?[139922,142885] 8ns 136.67kb                                                      |--------------L0.?---------------| "
    - "**** Simulation run 148, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 349.69kb total:"
    - "L0, all files 349.69kb                                                                                             "
    - "L0.116[135301,142885] 9ns|-----------------------------------------L0.116-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 349.69kb total:"
    - "L0                                                                                                                 "
    - "L0.?[135301,139921] 9ns 213.02kb|------------------------L0.?------------------------|                                    "
    - "L0.?[139922,142885] 9ns 136.67kb                                                      |--------------L0.?---------------| "
    - "**** Simulation run 149, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 349.69kb total:"
    - "L0, all files 349.69kb                                                                                             "
    - "L0.120[135301,142885] 10ns|-----------------------------------------L0.120-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 349.69kb total:"
    - "L0                                                                                                                 "
    - "L0.?[135301,139921] 10ns 213.02kb|------------------------L0.?------------------------|                                    "
    - "L0.?[139922,142885] 10ns 136.67kb                                                      |--------------L0.?---------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.112, L0.116, L0.120, L0.241, L0.245, L0.249, L1.354, L1.355"
    - "  Creating 16 files"
    - "**** Simulation run 150, type=split(ReduceOverlap)(split_times=[108723]). 1 Input Files, 29.46kb total:"
    - "L0, all files 29.46kb                                                                                              "
    - "L0.360[108403,109041] 8ns|-----------------------------------------L0.360-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 29.46kb total:"
    - "L0                                                                                                                 "
    - "L0.?[108403,108723] 8ns 14.78kb|-------------------L0.?--------------------|                                             "
    - "L0.?[108724,109041] 8ns 14.69kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 151, type=split(ReduceOverlap)(split_times=[140563]). 1 Input Files, 136.67kb total:"
    - "L0, all files 136.67kb                                                                                             "
    - "L0.368[139922,142885] 8ns|-----------------------------------------L0.368-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.67kb total:"
    - "L0                                                                                                                 "
    - "L0.?[139922,140563] 8ns 29.57kb|------L0.?-------|                                                                       "
    - "L0.?[140564,142885] 8ns 107.1kb                   |--------------------------------L0.?--------------------------------| "
    - "**** Simulation run 152, type=split(ReduceOverlap)(split_times=[108723]). 1 Input Files, 29.46kb total:"
    - "L0, all files 29.46kb                                                                                              "
    - "L0.362[108403,109041] 9ns|-----------------------------------------L0.362-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 29.46kb total:"
    - "L0                                                                                                                 "
    - "L0.?[108403,108723] 9ns 14.78kb|-------------------L0.?--------------------|                                             "
    - "L0.?[108724,109041] 9ns 14.69kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 153, type=split(ReduceOverlap)(split_times=[140563]). 1 Input Files, 136.67kb total:"
    - "L0, all files 136.67kb                                                                                             "
    - "L0.370[139922,142885] 9ns|-----------------------------------------L0.370-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.67kb total:"
    - "L0                                                                                                                 "
    - "L0.?[139922,140563] 9ns 29.57kb|------L0.?-------|                                                                       "
    - "L0.?[140564,142885] 9ns 107.1kb                   |--------------------------------L0.?--------------------------------| "
    - "**** Simulation run 154, type=split(ReduceOverlap)(split_times=[108723]). 1 Input Files, 29.46kb total:"
    - "L0, all files 29.46kb                                                                                              "
    - "L0.364[108403,109041] 10ns|-----------------------------------------L0.364-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 29.46kb total:"
    - "L0                                                                                                                 "
    - "L0.?[108403,108723] 10ns 14.78kb|-------------------L0.?--------------------|                                             "
    - "L0.?[108724,109041] 10ns 14.69kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 155, type=split(ReduceOverlap)(split_times=[140563]). 1 Input Files, 136.67kb total:"
    - "L0, all files 136.67kb                                                                                             "
    - "L0.372[139922,142885] 10ns|-----------------------------------------L0.372-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.67kb total:"
    - "L0                                                                                                                 "
    - "L0.?[139922,140563] 10ns 29.57kb|------L0.?-------|                                                                       "
    - "L0.?[140564,142885] 10ns 107.1kb                   |--------------------------------L0.?--------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 6 files: L0.360, L0.362, L0.364, L0.368, L0.370, L0.372"
    - "  Creating 12 files"
    - "**** Simulation run 156, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[170977]). 8 Input Files, 19.54mb total:"
    - "L0                                                                                                                 "
    - "L0.190[160868,171442] 7ns 487.56kb                              |----L0.190----|                                            "
    - "L0.177[171443,171443] 7ns 0b                                              |L0.177|                                    "
    - "L0.309[171444,185000] 7ns 625.13kb                                              |------L0.309------|                        "
    - "L0.310[185001,198370] 7ns 616.55kb                                                                   |------L0.310------|   "
    - "L0.264[198371,200000] 7ns 75.17kb                                                                                       |L0.264|"
    - "L1                                                                                                                 "
    - "L1.356[140564,171442] 7ns 9.7mb|-------------------L1.356-------------------|                                            "
    - "L1.306[171443,185000] 6ns 3.84mb                                              |------L1.306------|                        "
    - "L1.302[185001,200000] 6ns 4.25mb                                                                   |-------L1.302-------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 19.54mb total:"
    - "L1                                                                                                                 "
    - "L1.?[140564,170977] 7ns 10mb|--------------------L1.?--------------------|                                            "
    - "L1.?[170978,200000] 7ns 9.54mb                                              |------------------L1.?-------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.177, L0.190, L0.264, L1.302, L1.306, L0.309, L0.310, L1.356"
    - "  Creating 2 files"
    - "**** Simulation run 157, type=split(ReduceOverlap)(split_times=[170977]). 1 Input Files, 487.56kb total:"
    - "L0, all files 487.56kb                                                                                             "
    - "L0.194[160868,171442] 8ns|-----------------------------------------L0.194-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 487.56kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,170977] 8ns 466.12kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[170978,171442] 8ns 21.44kb                                                                                      |L0.?|"
    - "**** Simulation run 158, type=split(ReduceOverlap)(split_times=[170977]). 1 Input Files, 487.56kb total:"
    - "L0, all files 487.56kb                                                                                             "
    - "L0.198[160868,171442] 9ns|-----------------------------------------L0.198-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 487.56kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,170977] 9ns 466.12kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[170978,171442] 9ns 21.44kb                                                                                      |L0.?|"
    - "**** Simulation run 159, type=split(ReduceOverlap)(split_times=[170977]). 1 Input Files, 487.56kb total:"
    - "L0, all files 487.56kb                                                                                             "
    - "L0.202[160868,171442] 10ns|-----------------------------------------L0.202-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 487.56kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,170977] 10ns 466.12kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[170978,171442] 10ns 21.44kb                                                                                      |L0.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.194, L0.198, L0.202"
    - "  Creating 6 files"
    - "**** Simulation run 160, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[28630, 57160]). 17 Input Files, 26.91mb total:"
    - "L0                                                                                                                 "
    - "L0.328[100,25694] 8ns 1.15mb|-----------L0.328-----------|                                                            "
    - "L0.329[25695,28657] 8ns 136.6kb                              |L0.329|                                                    "
    - "L0.342[28658,32834] 8ns 192.53kb                                 |L0.342|                                                 "
    - "L0.343[32835,38176] 8ns 246.28kb                                      |L0.343|                                            "
    - "L0.288[38177,38491] 8ns 14.48kb                                            |L0.288|                                      "
    - "L0.289[38492,46513] 8ns 369.86kb                                             |L0.289-|                                    "
    - "L0.336[46514,51288] 8ns 220.13kb                                                      |L0.336|                            "
    - "L0.337[51289,57214] 8ns 273.25kb                                                            |L0.337|                      "
    - "L0.344[57215,65568] 8ns 385.1kb                                                                  |L0.344-|               "
    - "L0.345[65569,67700] 8ns 98.29kb                                                                            |L0.345|      "
    - "L0.224[67701,76252] 8ns 394.25kb                                                                               |-L0.224-| "
    - "L0.290[76253,76882] 8ns 29kb                                                                                         |L0.290|"
    - "L1                                                                                                                 "
    - "L1.326[100,25694] 7ns 7.82mb|-----------L1.326-----------|                                                            "
    - "L1.327[25695,32834] 7ns 2.18mb                              |L1.327|                                                    "
    - "L1.334[32835,51288] 7ns 5.64mb                                      |------L1.334-------|                               "
    - "L1.335[51289,65568] 7ns 4.36mb                                                            |----L1.335----|              "
    - "L1.325[65569,76882] 7ns 3.46mb                                                                            |--L1.325---| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 26.91mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,28630] 8ns 10mb |-------------L1.?--------------|                                                         "
    - "L1.?[28631,57160] 8ns 10mb                                 |-------------L1.?--------------|                        "
    - "L1.?[57161,76882] 8ns 6.91mb                                                                  |--------L1.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 17 files: L0.224, L0.288, L0.289, L0.290, L1.325, L1.326, L1.327, L0.328, L0.329, L1.334, L1.335, L0.336, L0.337, L0.342, L0.343, L0.344, L0.345"
    - "  Creating 3 files"
    - "**** Simulation run 161, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.393[100,28630] 8ns    |-----------------------------------------L1.393-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25694] 8ns 8.97mb|-------------------------------------L1.?-------------------------------------|          "
    - "L1.?[25695,28630] 8ns 1.03mb                                                                                |-L1.?--| "
    - "**** Simulation run 162, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.394[28631,57160] 8ns  |-----------------------------------------L1.394-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[28631,51288] 8ns 7.94mb|--------------------------------L1.?---------------------------------|                   "
    - "L1.?[51289,57160] 8ns 2.06mb                                                                       |------L1.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.393, L1.394"
    - "  Creating 4 files"
    - "**** Simulation run 163, type=split(ReduceOverlap)(split_times=[28630]). 1 Input Files, 136.6kb total:"
    - "L0, all files 136.6kb                                                                                              "
    - "L0.331[25695,28657] 9ns  |-----------------------------------------L0.331-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.6kb total:"
    - "L0                                                                                                                 "
    - "L0.?[25695,28630] 9ns 135.36kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[28631,28657] 9ns 1.25kb                                                                                         |L0.?|"
    - "**** Simulation run 164, type=split(ReduceOverlap)(split_times=[57160]). 1 Input Files, 273.25kb total:"
    - "L0, all files 273.25kb                                                                                             "
    - "L0.339[51289,57214] 9ns  |-----------------------------------------L0.339-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 273.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[51289,57160] 9ns 270.76kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[57161,57214] 9ns 2.49kb                                                                                         |L0.?|"
    - "**** Simulation run 165, type=split(ReduceOverlap)(split_times=[28630]). 1 Input Files, 136.6kb total:"
    - "L0, all files 136.6kb                                                                                              "
    - "L0.333[25695,28657] 10ns |-----------------------------------------L0.333-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.6kb total:"
    - "L0                                                                                                                 "
    - "L0.?[25695,28630] 10ns 135.36kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[28631,28657] 10ns 1.25kb                                                                                         |L0.?|"
    - "**** Simulation run 166, type=split(ReduceOverlap)(split_times=[57160]). 1 Input Files, 273.25kb total:"
    - "L0, all files 273.25kb                                                                                             "
    - "L0.341[51289,57214] 10ns |-----------------------------------------L0.341-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 273.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[51289,57160] 10ns 270.76kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[57161,57214] 10ns 2.49kb                                                                                         |L0.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L0.331, L0.333, L0.339, L0.341"
    - "  Creating 8 files"
    - "**** Simulation run 167, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[104731, 132579]). 15 Input Files, 22.87mb total:"
    - "L0                                                                                                                 "
    - "L0.291[76883,85771] 8ns 409.88kb|--L0.291--|                                                                              "
    - "L0.132[85772,85772] 8ns 0b            |L0.132|                                                                      "
    - "L0.240[85773,92926] 8ns 329.79kb            |-L0.240-|                                                                    "
    - "L0.359[92927,108402] 8ns 713.51kb                      |------L0.359-------|                                               "
    - "L0.373[108403,108723] 8ns 14.78kb                                            |L0.373|                                      "
    - "L0.374[108724,109041] 8ns 14.69kb                                             |L0.374|                                     "
    - "L0.192[109042,114328] 8ns 243.76kb                                             |L0.192|                                     "
    - "L0.147[114329,114329] 8ns 0b                                                    |L0.147|                              "
    - "L0.148[114330,135300] 8ns 966.83kb                                                    |----------L0.148-----------|         "
    - "L0.367[135301,139921] 8ns 213.02kb                                                                                  |L0.367|"
    - "L0.375[139922,140563] 8ns 29.57kb                                                                                         |L0.375|"
    - "L1                                                                                                                 "
    - "L1.357[76883,108402] 7ns 9.9mb|------------------L1.357------------------|                                              "
    - "L1.358[108403,108723] 7ns 103.24kb                                            |L1.358|                                      "
    - "L1.365[108724,139921] 7ns 9.8mb                                             |------------------L1.365------------------| "
    - "L1.366[139922,140563] 7ns 206.48kb                                                                                         |L1.366|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.87mb total:"
    - "L1                                                                                                                 "
    - "L1.?[76883,104731] 8ns 10mb|----------------L1.?-----------------|                                                   "
    - "L1.?[104732,132579] 8ns 10mb                                       |----------------L1.?-----------------|            "
    - "L1.?[132580,140563] 8ns 2.87mb                                                                              |--L1.?---| "
    - "Committing partition 1:"
    - "  Soft Deleting 15 files: L0.132, L0.147, L0.148, L0.192, L0.240, L0.291, L1.357, L1.358, L0.359, L1.365, L1.366, L0.367, L0.373, L0.374, L0.375"
    - "  Creating 3 files"
    - "**** Simulation run 168, type=split(ReduceOverlap)(split_times=[104731]). 1 Input Files, 713.51kb total:"
    - "L0, all files 713.51kb                                                                                             "
    - "L0.361[92927,108402] 9ns |-----------------------------------------L0.361-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 713.51kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,104731] 9ns 544.25kb|-------------------------------L0.?-------------------------------|                      "
    - "L0.?[104732,108402] 9ns 169.26kb                                                                    |-------L0.?--------| "
    - "**** Simulation run 169, type=split(ReduceOverlap)(split_times=[132579]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.150[114330,135300] 9ns|-----------------------------------------L0.150-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114330,132579] 9ns 841.37kb|------------------------------------L0.?------------------------------------|            "
    - "L0.?[132580,135300] 9ns 125.45kb                                                                              |--L0.?---| "
    - "**** Simulation run 170, type=split(ReduceOverlap)(split_times=[104731]). 1 Input Files, 713.51kb total:"
    - "L0, all files 713.51kb                                                                                             "
    - "L0.363[92927,108402] 10ns|-----------------------------------------L0.363-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 713.51kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,104731] 10ns 544.25kb|-------------------------------L0.?-------------------------------|                      "
    - "L0.?[104732,108402] 10ns 169.26kb                                                                    |-------L0.?--------| "
    - "**** Simulation run 171, type=split(ReduceOverlap)(split_times=[132579]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.152[114330,135300] 10ns|-----------------------------------------L0.152-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114330,132579] 10ns 841.37kb|------------------------------------L0.?------------------------------------|            "
    - "L0.?[132580,135300] 10ns 125.45kb                                                                              |--L0.?---| "
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L0.150, L0.152, L0.361, L0.363"
    - "  Creating 8 files"
    - "**** Simulation run 172, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[167314, 194064]). 13 Input Files, 22.22mb total:"
    - "L0                                                                                                                 "
    - "L0.376[140564,142885] 8ns 107.1kb|L0.376|                                                                                  "
    - "L0.159[142886,142886] 8ns 0b   |L0.159|                                                                               "
    - "L0.311[142887,149665] 8ns 312.5kb   |-L0.311-|                                                                             "
    - "L0.312[149666,156350] 8ns 308.21kb             |-L0.312-|                                                                   "
    - "L0.266[156351,160867] 8ns 208.25kb                       |L0.266|                                                           "
    - "L0.387[160868,170977] 8ns 466.12kb                              |---L0.387----|                                             "
    - "L0.388[170978,171442] 8ns 21.44kb                                              |L0.388|                                    "
    - "L0.169[171443,171443] 8ns 0b                                              |L0.169|                                    "
    - "L0.313[171444,185000] 8ns 625.13kb                                              |------L0.313------|                        "
    - "L0.314[185001,198370] 8ns 616.55kb                                                                   |------L0.314------|   "
    - "L0.268[198371,200000] 8ns 75.17kb                                                                                       |L0.268|"
    - "L1                                                                                                                 "
    - "L1.385[140564,170977] 7ns 10mb|-------------------L1.385-------------------|                                            "
    - "L1.386[170978,200000] 7ns 9.54mb                                              |-----------------L1.386------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.22mb total:"
    - "L1                                                                                                                 "
    - "L1.?[140564,167314] 8ns 10mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[167315,194064] 8ns 10mb                                        |-----------------L1.?-----------------|          "
    - "L1.?[194065,200000] 8ns 2.22mb                                                                                 |-L1.?-| "
    - "Committing partition 1:"
    - "  Soft Deleting 13 files: L0.159, L0.169, L0.266, L0.268, L0.311, L0.312, L0.313, L0.314, L0.376, L1.385, L1.386, L0.387, L0.388"
    - "  Creating 3 files"
    - "**** Simulation run 173, type=split(ReduceOverlap)(split_times=[167314]). 1 Input Files, 466.12kb total:"
    - "L0, all files 466.12kb                                                                                             "
    - "L0.389[160868,170977] 9ns|-----------------------------------------L0.389-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 466.12kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,167314] 9ns 297.22kb|-------------------------L0.?--------------------------|                                 "
    - "L0.?[167315,170977] 9ns 168.9kb                                                         |-------------L0.?-------------| "
    - "**** Simulation run 174, type=split(ReduceOverlap)(split_times=[194064]). 1 Input Files, 616.55kb total:"
    - "L0, all files 616.55kb                                                                                             "
    - "L0.318[185001,198370] 9ns|-----------------------------------------L0.318-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 616.55kb total:"
    - "L0                                                                                                                 "
    - "L0.?[185001,194064] 9ns 417.97kb|---------------------------L0.?----------------------------|                             "
    - "L0.?[194065,198370] 9ns 198.58kb                                                             |-----------L0.?-----------| "
    - "**** Simulation run 175, type=split(ReduceOverlap)(split_times=[167314]). 1 Input Files, 466.12kb total:"
    - "L0, all files 466.12kb                                                                                             "
    - "L0.391[160868,170977] 10ns|-----------------------------------------L0.391-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 466.12kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,167314] 10ns 297.22kb|-------------------------L0.?--------------------------|                                 "
    - "L0.?[167315,170977] 10ns 168.9kb                                                         |-------------L0.?-------------| "
    - "**** Simulation run 176, type=split(ReduceOverlap)(split_times=[194064]). 1 Input Files, 616.55kb total:"
    - "L0, all files 616.55kb                                                                                             "
    - "L0.322[185001,198370] 10ns|-----------------------------------------L0.322-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 616.55kb total:"
    - "L0                                                                                                                 "
    - "L0.?[185001,194064] 10ns 417.97kb|---------------------------L0.?----------------------------|                             "
    - "L0.?[194065,198370] 10ns 198.58kb                                                             |-----------L0.?-----------| "
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L0.318, L0.322, L0.389, L0.391"
    - "  Creating 8 files"
    - "**** Simulation run 177, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[25731, 51362]). 17 Input Files, 29.96mb total:"
    - "L0                                                                                                                 "
    - "L0.330[100,25694] 9ns 1.15mb|-----------L0.330-----------|                                                            "
    - "L0.400[25695,28630] 9ns 135.36kb                              |L0.400|                                                    "
    - "L0.401[28631,28657] 9ns 1.25kb                                 |L0.401|                                                 "
    - "L0.346[28658,32834] 9ns 192.53kb                                 |L0.346|                                                 "
    - "L0.347[32835,38176] 9ns 246.28kb                                      |L0.347|                                            "
    - "L0.292[38177,38491] 9ns 14.48kb                                            |L0.292|                                      "
    - "L0.293[38492,46513] 9ns 369.86kb                                             |L0.293-|                                    "
    - "L0.338[46514,51288] 9ns 220.13kb                                                      |L0.338|                            "
    - "L0.402[51289,57160] 9ns 270.76kb                                                            |L0.402|                      "
    - "L0.403[57161,57214] 9ns 2.49kb                                                                  |L0.403|                "
    - "L0.348[57215,65568] 9ns 385.1kb                                                                  |L0.348-|               "
    - "L0.349[65569,67700] 9ns 98.29kb                                                                            |L0.349|      "
    - "L1                                                                                                                 "
    - "L1.396[100,25694] 8ns 8.97mb|-----------L1.396-----------|                                                            "
    - "L1.397[25695,28630] 8ns 1.03mb                              |L1.397|                                                    "
    - "L1.398[28631,51288] 8ns 7.94mb                                 |---------L1.398---------|                               "
    - "L1.399[51289,57160] 8ns 2.06mb                                                            |L1.399|                      "
    - "L1.395[57161,76882] 8ns 6.91mb                                                                  |-------L1.395--------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.96mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25731] 9ns 10mb |------------L1.?------------|                                                            "
    - "L1.?[25732,51362] 9ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[51363,76882] 9ns 9.96mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 17 files: L0.292, L0.293, L0.330, L0.338, L0.346, L0.347, L0.348, L0.349, L1.395, L1.396, L1.397, L1.398, L1.399, L0.400, L0.401, L0.402, L0.403"
    - "  Creating 3 files"
    - "**** Simulation run 178, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.430[100,25731] 9ns    |-----------------------------------------L1.430-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25694] 9ns 9.99mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[25695,25731] 9ns 14.78kb                                                                                         |L1.?|"
    - "**** Simulation run 179, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.431[25732,51362] 9ns  |-----------------------------------------L1.431-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[25732,51288] 9ns 9.97mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[51289,51362] 9ns 29.56kb                                                                                         |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.430, L1.431"
    - "  Creating 4 files"
    - "**** Simulation run 180, type=split(ReduceOverlap)(split_times=[25731]). 1 Input Files, 135.36kb total:"
    - "L0, all files 135.36kb                                                                                             "
    - "L0.404[25695,28630] 10ns |----------------------------------------L0.404-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 135.36kb total:"
    - "L0                                                                                                                 "
    - "L0.?[25695,25731] 10ns 1.66kb|L0.?|                                                                                    "
    - "L0.?[25732,28630] 10ns 133.7kb |-----------------------------------------L0.?-----------------------------------------| "
    - "**** Simulation run 181, type=split(ReduceOverlap)(split_times=[51362]). 1 Input Files, 270.76kb total:"
    - "L0, all files 270.76kb                                                                                             "
    - "L0.406[51289,57160] 10ns |-----------------------------------------L0.406-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 270.76kb total:"
    - "L0                                                                                                                 "
    - "L0.?[51289,51362] 10ns 3.37kb|L0.?|                                                                                    "
    - "L0.?[51363,57160] 10ns 267.39kb |-----------------------------------------L0.?-----------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.404, L0.406"
    - "  Creating 4 files"
    - "**** Simulation run 182, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[76043, 100723]). 8 Input Files, 21.62mb total:"
    - "L0                                                                                                                 "
    - "L0.226[67701,76252] 9ns 394.25kb                           |---L0.226---|                                                 "
    - "L0.294[76253,76882] 9ns 29kb                                         |L0.294|                                         "
    - "L0.295[76883,85771] 9ns 409.88kb                                           |---L0.295---|                                 "
    - "L0.134[85772,85772] 9ns 0b                                                          |L0.134|                        "
    - "L0.244[85773,92926] 9ns 329.79kb                                                          |--L0.244--|                    "
    - "L0.411[92927,104731] 9ns 544.25kb                                                                      |-----L0.411------| "
    - "L1                                                                                                                 "
    - "L1.432[51363,76882] 9ns 9.96mb|-----------------L1.432------------------|                                               "
    - "L1.408[76883,104731] 8ns 10mb                                           |-------------------L1.408-------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 21.62mb total:"
    - "L1                                                                                                                 "
    - "L1.?[51363,76043] 9ns 10mb|-----------------L1.?------------------|                                                 "
    - "L1.?[76044,100723] 9ns 10mb                                         |-----------------L1.?------------------|        "
    - "L1.?[100724,104731] 9ns 1.62mb                                                                                   |L1.?| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.134, L0.226, L0.244, L0.294, L0.295, L1.408, L0.411, L1.432"
    - "  Creating 3 files"
    - "**** Simulation run 183, type=split(ReduceOverlap)(split_times=[76043]). 1 Input Files, 394.25kb total:"
    - "L0, all files 394.25kb                                                                                             "
    - "L0.228[67701,76252] 10ns |-----------------------------------------L0.228-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 394.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76043] 10ns 384.61kb|----------------------------------------L0.?-----------------------------------------|   "
    - "L0.?[76044,76252] 10ns 9.64kb                                                                                       |L0.?|"
    - "**** Simulation run 184, type=split(ReduceOverlap)(split_times=[100723]). 1 Input Files, 544.25kb total:"
    - "L0, all files 544.25kb                                                                                             "
    - "L0.415[92927,104731] 10ns|-----------------------------------------L0.415-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 544.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,100723] 10ns 359.45kb|--------------------------L0.?---------------------------|                               "
    - "L0.?[100724,104731] 10ns 184.8kb                                                           |------------L0.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.228, L0.415"
    - "  Creating 4 files"
    - "**** Simulation run 185, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[129098, 153464]). 18 Input Files, 25.68mb total:"
    - "L0                                                                                                                 "
    - "L0.412[104732,108402] 9ns 169.26kb|L0.412|                                                                                  "
    - "L0.377[108403,108723] 9ns 14.78kb     |L0.377|                                                                             "
    - "L0.378[108724,109041] 9ns 14.69kb     |L0.378|                                                                             "
    - "L0.196[109042,114328] 9ns 243.76kb      |L0.196|                                                                            "
    - "L0.149[114329,114329] 9ns 0b             |L0.149|                                                                     "
    - "L0.413[114330,132579] 9ns 841.37kb             |---------L0.413---------|                                                   "
    - "L0.414[132580,135300] 9ns 125.45kb                                        |L0.414|                                          "
    - "L0.369[135301,139921] 9ns 213.02kb                                           |L0.369|                                       "
    - "L0.379[139922,140563] 9ns 29.57kb                                                  |L0.379|                                "
    - "L0.380[140564,142885] 9ns 107.1kb                                                   |L0.380|                               "
    - "L0.161[142886,142886] 9ns 0b                                                      |L0.161|                            "
    - "L0.315[142887,149665] 9ns 312.5kb                                                      |L0.315-|                           "
    - "L0.316[149666,156350] 9ns 308.21kb                                                                |L0.316-|                 "
    - "L0.270[156351,160867] 9ns 208.25kb                                                                          |L0.270|        "
    - "L0.422[160868,167314] 9ns 297.22kb                                                                                |L0.422-| "
    - "L1                                                                                                                 "
    - "L1.409[104732,132579] 8ns 10mb|----------------L1.409----------------|                                                  "
    - "L1.410[132580,140563] 8ns 2.87mb                                        |-L1.410--|                                       "
    - "L1.419[140564,167314] 8ns 10mb                                                   |---------------L1.419---------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 25.68mb total:"
    - "L1                                                                                                                 "
    - "L1.?[104732,129098] 9ns 10mb|--------------L1.?---------------|                                                       "
    - "L1.?[129099,153464] 9ns 10mb                                   |--------------L1.?---------------|                    "
    - "L1.?[153465,167314] 9ns 5.68mb                                                                      |------L1.?-------| "
    - "Committing partition 1:"
    - "  Soft Deleting 18 files: L0.149, L0.161, L0.196, L0.270, L0.315, L0.316, L0.369, L0.377, L0.378, L0.379, L0.380, L1.409, L1.410, L0.412, L0.413, L0.414, L1.419, L0.422"
    - "  Creating 3 files"
    - "**** Simulation run 186, type=split(ReduceOverlap)(split_times=[129098]). 1 Input Files, 841.37kb total:"
    - "L0, all files 841.37kb                                                                                             "
    - "L0.417[114330,132579] 10ns|-----------------------------------------L0.417-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 841.37kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114330,129098] 10ns 680.88kb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[129099,132579] 10ns 160.49kb                                                                        |-----L0.?------| "
    - "**** Simulation run 187, type=split(ReduceOverlap)(split_times=[153464]). 1 Input Files, 308.21kb total:"
    - "L0, all files 308.21kb                                                                                             "
    - "L0.320[149666,156350] 10ns|-----------------------------------------L0.320-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 308.21kb total:"
    - "L0                                                                                                                 "
    - "L0.?[149666,153464] 10ns 175.13kb|----------------------L0.?-----------------------|                                       "
    - "L0.?[153465,156350] 10ns 133.08kb                                                   |----------------L0.?----------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.320, L0.417"
    - "  Creating 4 files"
    - "**** Simulation run 188, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[191189]). 9 Input Files, 13.69mb total:"
    - "L0                                                                                                                 "
    - "L0.423[167315,170977] 9ns 168.9kb|-L0.423-|                                                                                "
    - "L0.390[170978,171442] 9ns 21.44kb          |L0.390|                                                                        "
    - "L0.171[171443,171443] 9ns 0b           |L0.171|                                                                       "
    - "L0.317[171444,185000] 9ns 625.13kb           |--------------L0.317---------------|                                          "
    - "L0.424[185001,194064] 9ns 417.97kb                                                |--------L0.424--------|                  "
    - "L0.425[194065,198370] 9ns 198.58kb                                                                         |-L0.425--|      "
    - "L0.272[198371,200000] 9ns 75.17kb                                                                                     |L0.272|"
    - "L1                                                                                                                 "
    - "L1.420[167315,194064] 8ns 10mb|--------------------------------L1.420---------------------------------|                 "
    - "L1.421[194065,200000] 8ns 2.22mb                                                                         |----L1.421----| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 13.69mb total:"
    - "L1                                                                                                                 "
    - "L1.?[167315,191189] 9ns 10mb|-----------------------------L1.?------------------------------|                         "
    - "L1.?[191190,200000] 9ns 3.69mb                                                                 |---------L1.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 9 files: L0.171, L0.272, L0.317, L0.390, L1.420, L1.421, L0.423, L0.424, L0.425"
    - "  Creating 2 files"
    - "**** Simulation run 189, type=split(ReduceOverlap)(split_times=[191189]). 1 Input Files, 417.97kb total:"
    - "L0, all files 417.97kb                                                                                             "
    - "L0.428[185001,194064] 10ns|-----------------------------------------L0.428-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 417.97kb total:"
    - "L0                                                                                                                 "
    - "L0.?[185001,191189] 10ns 285.38kb|---------------------------L0.?----------------------------|                             "
    - "L0.?[191190,194064] 10ns 132.59kb                                                             |-----------L0.?-----------| "
    - "Committing partition 1:"
    - "  Soft Deleting 1 files: L0.428"
    - "  Creating 2 files"
    - "**** Simulation run 190, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[23080, 46060]). 14 Input Files, 22.31mb total:"
    - "L0                                                                                                                 "
    - "L0.332[100,25694] 10ns 1.15mb|------------------L0.332------------------|                                              "
    - "L0.437[25695,25731] 10ns 1.66kb                                            |L0.437|                                      "
    - "L0.438[25732,28630] 10ns 133.7kb                                             |L0.438|                                     "
    - "L0.405[28631,28657] 10ns 1.25kb                                                  |L0.405|                                "
    - "L0.350[28658,32834] 10ns 192.53kb                                                  |L0.350|                                "
    - "L0.351[32835,38176] 10ns 246.28kb                                                         |L0.351-|                        "
    - "L0.296[38177,38491] 10ns 14.48kb                                                                  |L0.296|                "
    - "L0.297[38492,46513] 10ns 369.86kb                                                                   |---L0.297---|         "
    - "L0.340[46514,51288] 10ns 220.13kb                                                                                 |L0.340| "
    - "L0.439[51289,51362] 10ns 3.37kb                                                                                         |L0.439|"
    - "L1                                                                                                                 "
    - "L1.433[100,25694] 9ns 9.99mb|------------------L1.433------------------|                                              "
    - "L1.434[25695,25731] 9ns 14.78kb                                            |L1.434|                                      "
    - "L1.435[25732,51288] 9ns 9.97mb                                             |------------------L1.435------------------| "
    - "L1.436[51289,51362] 9ns 29.56kb                                                                                         |L1.436|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.31mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,23080] 10ns 10mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[23081,46060] 10ns 10mb                                        |-----------------L1.?-----------------|          "
    - "L1.?[46061,51362] 10ns 2.31mb                                                                                |-L1.?--| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.296, L0.297, L0.332, L0.340, L0.350, L0.351, L0.405, L1.433, L1.434, L1.435, L1.436, L0.437, L0.438, L0.439"
    - "  Creating 3 files"
    - "**** Simulation run 191, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[73575, 95787]). 15 Input Files, 24.03mb total:"
    - "L0                                                                                                                 "
    - "L0.440[51363,57160] 10ns 267.39kb|L0.440-|                                                                                 "
    - "L0.407[57161,57214] 10ns 2.49kb         |L0.407|                                                                         "
    - "L0.352[57215,65568] 10ns 385.1kb         |---L0.352---|                                                                   "
    - "L0.353[65569,67700] 10ns 98.29kb                       |L0.353|                                                           "
    - "L0.444[67701,76043] 10ns 384.61kb                           |---L0.444---|                                                 "
    - "L0.445[76044,76252] 10ns 9.64kb                                         |L0.445|                                         "
    - "L0.298[76253,76882] 10ns 29kb                                         |L0.298|                                         "
    - "L0.299[76883,85771] 10ns 409.88kb                                           |---L0.299---|                                 "
    - "L0.136[85772,85772] 10ns 0b                                                          |L0.136|                        "
    - "L0.248[85773,92926] 10ns 329.79kb                                                          |--L0.248--|                    "
    - "L0.446[92927,100723] 10ns 359.45kb                                                                      |--L0.446---|       "
    - "L0.447[100724,104731] 10ns 184.8kb                                                                                   |L0.447|"
    - "L1                                                                                                                 "
    - "L1.441[51363,76043] 9ns 10mb|----------------L1.441-----------------|                                                 "
    - "L1.442[76044,100723] 9ns 10mb                                         |----------------L1.442-----------------|        "
    - "L1.443[100724,104731] 9ns 1.62mb                                                                                   |L1.443|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.03mb total:"
    - "L1                                                                                                                 "
    - "L1.?[51363,73575] 10ns 10mb|---------------L1.?----------------|                                                     "
    - "L1.?[73576,95787] 10ns 10mb                                     |---------------L1.?----------------|                "
    - "L1.?[95788,104731] 10ns 4.03mb                                                                          |----L1.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 15 files: L0.136, L0.248, L0.298, L0.299, L0.352, L0.353, L0.407, L0.440, L1.441, L1.442, L1.443, L0.444, L0.445, L0.446, L0.447"
    - "  Creating 3 files"
    - "**** Simulation run 192, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[175139, 196813]). 14 Input Files, 21.47mb total:"
    - "L0                                                                                                                 "
    - "L0.276[198371,200000] 10ns 75.17kb                                                                                      |L0.276|"
    - "L0.429[194065,198370] 10ns 198.58kb                                                                              |L0.429|    "
    - "L0.458[191190,194064] 10ns 132.59kb                                                                        |L0.458|          "
    - "L0.457[185001,191189] 10ns 285.38kb                                                            |-L0.457--|                   "
    - "L0.321[171444,185000] 10ns 625.13kb                                  |---------L0.321---------|                              "
    - "L0.173[171443,171443] 10ns 0b                                  |L0.173|                                                "
    - "L0.392[170978,171442] 10ns 21.44kb                                 |L0.392|                                                 "
    - "L0.427[167315,170977] 10ns 168.9kb                          |L0.427|                                                        "
    - "L0.426[160868,167314] 10ns 297.22kb              |--L0.426--|                                                                "
    - "L0.274[156351,160867] 10ns 208.25kb     |L0.274|                                                                             "
    - "L0.454[153465,156350] 10ns 133.08kb|L0.454|                                                                                  "
    - "L1                                                                                                                 "
    - "L1.456[191190,200000] 9ns 3.69mb                                                                        |----L1.456-----| "
    - "L1.455[167315,191189] 9ns 10mb                          |-------------------L1.455-------------------|                  "
    - "L1.450[153465,167314] 9ns 5.68mb|---------L1.450---------|                                                                "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 21.47mb total:"
    - "L1                                                                                                                 "
    - "L1.?[153465,175139] 10ns 10mb|-----------------L1.?------------------|                                                 "
    - "L1.?[175140,196813] 10ns 10mb                                         |-----------------L1.?------------------|        "
    - "L1.?[196814,200000] 10ns 1.47mb                                                                                   |L1.?| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.173, L0.274, L0.276, L0.321, L0.392, L0.426, L0.427, L0.429, L1.450, L0.454, L1.455, L1.456, L0.457, L0.458"
    - "  Creating 3 files"
    - "**** Simulation run 193, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[126690, 148648]). 16 Input Files, 22.19mb total:"
    - "L0                                                                                                                 "
    - "L0.416[104732,108402] 10ns 169.26kb|L0.416|                                                                                  "
    - "L0.381[108403,108723] 10ns 14.78kb      |L0.381|                                                                            "
    - "L0.382[108724,109041] 10ns 14.69kb       |L0.382|                                                                           "
    - "L0.200[109042,114328] 10ns 243.76kb       |L0.200-|                                                                          "
    - "L0.151[114329,114329] 10ns 0b                 |L0.151|                                                                 "
    - "L0.451[114330,129098] 10ns 680.88kb                 |---------L0.451----------|                                              "
    - "L0.452[129099,132579] 10ns 160.49kb                                             |L0.452|                                     "
    - "L0.418[132580,135300] 10ns 125.45kb                                                   |L0.418|                               "
    - "L0.371[135301,139921] 10ns 213.02kb                                                        |L0.371|                          "
    - "L0.383[139922,140563] 10ns 29.57kb                                                                |L0.383|                  "
    - "L0.384[140564,142885] 10ns 107.1kb                                                                  |L0.384|                "
    - "L0.163[142886,142886] 10ns 0b                                                                      |L0.163|            "
    - "L0.319[142887,149665] 10ns 312.5kb                                                                      |--L0.319--|        "
    - "L0.453[149666,153464] 10ns 175.13kb                                                                                  |L0.453|"
    - "L1                                                                                                                 "
    - "L1.448[104732,129098] 9ns 10mb|------------------L1.448-------------------|                                             "
    - "L1.449[129099,153464] 9ns 10mb                                             |------------------L1.449------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.19mb total:"
    - "L1                                                                                                                 "
    - "L1.?[104732,126690] 10ns 10mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[126691,148648] 10ns 10mb                                        |-----------------L1.?-----------------|          "
    - "L1.?[148649,153464] 10ns 2.19mb                                                                                 |-L1.?-| "
    - "Committing partition 1:"
    - "  Soft Deleting 16 files: L0.151, L0.163, L0.200, L0.319, L0.371, L0.381, L0.382, L0.383, L0.384, L0.416, L0.418, L1.448, L1.449, L0.451, L0.452, L0.453"
    - "  Creating 3 files"
    - "**** Simulation run 194, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[45717, 68353]). 3 Input Files, 22.31mb total:"
    - "L1                                                                                                                 "
    - "L1.460[23081,46060] 10ns 10mb|----------------L1.460----------------|                                                  "
    - "L1.461[46061,51362] 10ns 2.31mb                                        |L1.461-|                                         "
    - "L1.462[51363,73575] 10ns 10mb                                                  |---------------L1.462----------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.31mb total:"
    - "L2                                                                                                                 "
    - "L2.?[23081,45717] 10ns 10mb|-----------------L2.?-----------------|                                                  "
    - "L2.?[45718,68353] 10ns 10mb                                        |-----------------L2.?-----------------|          "
    - "L2.?[68354,73575] 10ns 2.31mb                                                                                |-L2.?--| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.460, L1.461, L1.462"
    - "  Upgrading 1 files level to CompactionLevel::L2: L1.459"
    - "  Creating 3 files"
    - "**** Simulation run 195, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[95682, 117788]). 3 Input Files, 24.03mb total:"
    - "L1                                                                                                                 "
    - "L1.463[73576,95787] 10ns 10mb|--------------L1.463---------------|                                                     "
    - "L1.464[95788,104731] 10ns 4.03mb                                     |---L1.464----|                                      "
    - "L1.468[104732,126690] 10ns 10mb                                                    |--------------L1.468---------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.03mb total:"
    - "L2                                                                                                                 "
    - "L2.?[73576,95682] 10ns 10mb|---------------L2.?----------------|                                                     "
    - "L2.?[95683,117788] 10ns 10mb                                     |---------------L2.?----------------|                "
    - "L2.?[117789,126690] 10ns 4.03mb                                                                          |----L2.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.463, L1.464, L1.468"
    - "  Creating 3 files"
    - "**** Simulation run 196, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[148521, 170351]). 3 Input Files, 22.19mb total:"
    - "L1                                                                                                                 "
    - "L1.469[126691,148648] 10ns 10mb|----------------L1.469----------------|                                                  "
    - "L1.470[148649,153464] 10ns 2.19mb                                        |L1.470|                                          "
    - "L1.465[153465,175139] 10ns 10mb                                                 |----------------L1.465----------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.19mb total:"
    - "L2                                                                                                                 "
    - "L2.?[126691,148521] 10ns 10mb|-----------------L2.?-----------------|                                                  "
    - "L2.?[148522,170351] 10ns 10mb                                        |-----------------L2.?-----------------|          "
    - "L2.?[170352,175139] 10ns 2.19mb                                                                                 |-L2.?-| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.465, L1.469, L1.470"
    - "  Creating 3 files"
    - "**** Simulation run 197, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[195028]). 2 Input Files, 11.47mb total:"
    - "L1                                                                                                                 "
    - "L1.467[196814,200000] 10ns 1.47mb                                                                              |-L1.467--| "
    - "L1.466[175140,196813] 10ns 10mb|-----------------------------------L1.466-----------------------------------|            "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 11.47mb total:"
    - "L2                                                                                                                 "
    - "L2.?[175140,195028] 10ns 9.18mb|---------------------------------L2.?---------------------------------|                  "
    - "L2.?[195029,200000] 10ns 2.29mb                                                                        |-----L2.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.466, L1.467"
    - "  Creating 2 files"
    - "**** Final Output Files (994.29mb written)"
    - "L2                                                                                                                 "
    - "L2.459[100,23080] 10ns 10mb|-L2.459-|                                                                                "
    - "L2.471[23081,45717] 10ns 10mb          |-L2.471-|                                                                      "
    - "L2.472[45718,68353] 10ns 10mb                    |-L2.472-|                                                            "
    - "L2.473[68354,73575] 10ns 2.31mb                              |L2.473|                                                    "
    - "L2.474[73576,95682] 10ns 10mb                                 |L2.474-|                                                "
    - "L2.475[95683,117788] 10ns 10mb                                           |L2.475-|                                      "
    - "L2.476[117789,126690] 10ns 4.03mb                                                    |L2.476|                              "
    - "L2.477[126691,148521] 10ns 10mb                                                        |L2.477-|                         "
    - "L2.478[148522,170351] 10ns 10mb                                                                  |L2.478-|               "
    - "L2.479[170352,175139] 10ns 2.19mb                                                                            |L2.479|      "
    - "L2.480[175140,195028] 10ns 9.18mb                                                                              |L2.480|    "
    - "L2.481[195029,200000] 10ns 2.29mb                                                                                       |L2.481|"
    "###
    );
}

#[tokio::test]
async fn all_overlapping_l0_max_input_bytes_per_partition_small_max_desired_file_size() {
    test_helpers::maybe_start_logging();

    // limit the plan to combining 30 mb at a time
    let setup = layout_setup_builder()
        .await
        .with_max_desired_file_size_bytes(10 * ONE_MB)
        .build()
        .await;

    // create virtual files
    let setup = all_overlapping_l0_files(setup).await;

    // expect that each plan the compactor runs has at most three input files
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** Simulation run 0, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.1[100,200000] 1ns     |-----------------------------------------L0.1------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 1ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 1ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 1ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 1ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 1ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 1ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 1ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 1, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.2[100,200000] 2ns     |-----------------------------------------L0.2------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 2ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 2ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 2ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 2ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 2ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 2ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 2ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 2, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.3[100,200000] 3ns     |-----------------------------------------L0.3------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 3ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 3ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 3ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 3ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 3ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 3ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 3ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 3, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.4[100,200000] 4ns     |-----------------------------------------L0.4------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 4ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 4ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 4ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 4ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 4ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 4ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 4ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 4, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.5[100,200000] 5ns     |-----------------------------------------L0.5------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 5ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 5ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 5ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 5ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 5ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 5ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 5ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 5, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.6[100,200000] 6ns     |-----------------------------------------L0.6------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 6ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 6ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 6ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 6ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 6ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 6ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 6ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 6, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.7[100,200000] 7ns     |-----------------------------------------L0.7------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 7ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 7ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 7ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 7ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 7ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 7ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 7ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 7, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.8[100,200000] 8ns     |-----------------------------------------L0.8------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 8ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 8ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 8ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 8ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 8ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 8ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 8ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 8, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.9[100,200000] 9ns     |-----------------------------------------L0.9------------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 9ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 9ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 9ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 9ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 9ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 9ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 9ns 1.29mb                                                                             |---L0.?---| "
    - "**** Simulation run 9, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214, 85771, 114328, 142885, 171442]). 1 Input Files, 9mb total:"
    - "L0, all files 9mb                                                                                                  "
    - "L0.10[100,200000] 10ns   |-----------------------------------------L0.10-----------------------------------------| "
    - "**** 7 Output Files (parquet_file_id not yet assigned), 9mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,28657] 10ns 1.29mb|---L0.?---|                                                                              "
    - "L0.?[28658,57214] 10ns 1.29mb            |---L0.?---|                                                                  "
    - "L0.?[57215,85771] 10ns 1.29mb                         |---L0.?---|                                                     "
    - "L0.?[85772,114328] 10ns 1.29mb                                      |---L0.?---|                                        "
    - "L0.?[114329,142885] 10ns 1.29mb                                                   |---L0.?---|                           "
    - "L0.?[142886,171442] 10ns 1.29mb                                                                |---L0.?---|              "
    - "L0.?[171443,200000] 10ns 1.29mb                                                                             |---L0.?---| "
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10"
    - "  Creating 70 files"
    - "**** Simulation run 10, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[67700, 135300]). 23 Input Files, 29.57mb total:"
    - "L0                                                                                                                 "
    - "L0.11[100,28657] 1ns 1.29mb|--L0.11---|                                                                              "
    - "L0.12[28658,57214] 1ns 1.29mb            |--L0.12---|                                                                  "
    - "L0.13[57215,85771] 1ns 1.29mb                         |--L0.13---|                                                     "
    - "L0.14[85772,114328] 1ns 1.29mb                                      |--L0.14---|                                        "
    - "L0.15[114329,142885] 1ns 1.29mb                                                   |--L0.15---|                           "
    - "L0.16[142886,171442] 1ns 1.29mb                                                                |--L0.16---|              "
    - "L0.17[171443,200000] 1ns 1.29mb                                                                             |--L0.17---| "
    - "L0.18[100,28657] 2ns 1.29mb|--L0.18---|                                                                              "
    - "L0.19[28658,57214] 2ns 1.29mb            |--L0.19---|                                                                  "
    - "L0.20[57215,85771] 2ns 1.29mb                         |--L0.20---|                                                     "
    - "L0.21[85772,114328] 2ns 1.29mb                                      |--L0.21---|                                        "
    - "L0.22[114329,142885] 2ns 1.29mb                                                   |--L0.22---|                           "
    - "L0.23[142886,171442] 2ns 1.29mb                                                                |--L0.23---|              "
    - "L0.24[171443,200000] 2ns 1.29mb                                                                             |--L0.24---| "
    - "L0.25[100,28657] 3ns 1.29mb|--L0.25---|                                                                              "
    - "L0.26[28658,57214] 3ns 1.29mb            |--L0.26---|                                                                  "
    - "L0.27[57215,85771] 3ns 1.29mb                         |--L0.27---|                                                     "
    - "L0.28[85772,114328] 3ns 1.29mb                                      |--L0.28---|                                        "
    - "L0.29[114329,142885] 3ns 1.29mb                                                   |--L0.29---|                           "
    - "L0.30[142886,171442] 3ns 1.29mb                                                                |--L0.30---|              "
    - "L0.31[171443,200000] 3ns 1.29mb                                                                             |--L0.31---| "
    - "L0.32[100,28657] 4ns 1.29mb|--L0.32---|                                                                              "
    - "L0.33[28658,57214] 4ns 1.29mb            |--L0.33---|                                                                  "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.57mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,67700] 4ns 10mb |------------L1.?------------|                                                            "
    - "L1.?[67701,135300] 4ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[135301,200000] 4ns 9.57mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 23 files: L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33"
    - "  Creating 3 files"
    - "**** Simulation run 11, type=split(HighL0OverlapTotalBacklog)(split_times=[28657, 57214]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.81[100,67700] 4ns     |-----------------------------------------L1.81------------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,28657] 4ns 4.22mb|----------------L1.?----------------|                                                    "
    - "L1.?[28658,57214] 4ns 4.22mb                                      |----------------L1.?----------------|              "
    - "L1.?[57215,67700] 4ns 1.55mb                                                                            |---L1.?----| "
    - "**** Simulation run 12, type=split(HighL0OverlapTotalBacklog)(split_times=[85771, 114328]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.82[67701,135300] 4ns  |-----------------------------------------L1.82------------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[67701,85771] 4ns 2.67mb|---------L1.?---------|                                                                  "
    - "L1.?[85772,114328] 4ns 4.22mb                        |----------------L1.?----------------|                            "
    - "L1.?[114329,135300] 4ns 3.1mb                                                              |----------L1.?-----------| "
    - "**** Simulation run 13, type=split(HighL0OverlapTotalBacklog)(split_times=[142885, 171442]). 1 Input Files, 9.57mb total:"
    - "L1, all files 9.57mb                                                                                               "
    - "L1.83[135301,200000] 4ns |-----------------------------------------L1.83------------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 9.57mb total:"
    - "L1                                                                                                                 "
    - "L1.?[135301,142885] 4ns 1.12mb|--L1.?--|                                                                                "
    - "L1.?[142886,171442] 4ns 4.22mb          |----------------L1.?-----------------|                                         "
    - "L1.?[171443,200000] 4ns 4.22mb                                                  |----------------L1.?-----------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.81, L1.82, L1.83"
    - "  Creating 9 files"
    - "**** Simulation run 14, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.34[57215,85771] 4ns   |-----------------------------------------L0.34------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 4ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 4ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 15, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.36[114329,142885] 4ns |-----------------------------------------L0.36------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 4ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 4ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 16, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.41[57215,85771] 5ns   |-----------------------------------------L0.41------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 5ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 5ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 17, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.43[114329,142885] 5ns |-----------------------------------------L0.43------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 5ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 5ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 18, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.48[57215,85771] 6ns   |-----------------------------------------L0.48------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 6ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 6ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 19, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.50[114329,142885] 6ns |-----------------------------------------L0.50------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 6ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 6ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 20, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.55[57215,85771] 7ns   |-----------------------------------------L0.55------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 7ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 7ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 21, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.57[114329,142885] 7ns |-----------------------------------------L0.57------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 7ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 7ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 22, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.62[57215,85771] 8ns   |-----------------------------------------L0.62------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 8ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 8ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 23, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.64[114329,142885] 8ns |-----------------------------------------L0.64------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 8ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 8ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 24, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.69[57215,85771] 9ns   |-----------------------------------------L0.69------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 9ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 9ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 25, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.71[114329,142885] 9ns |-----------------------------------------L0.71------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 9ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 9ns 349.69kb                                                                  |--------L0.?---------| "
    - "**** Simulation run 26, type=split(ReduceOverlap)(split_times=[67700]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.76[57215,85771] 10ns  |-----------------------------------------L0.76------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,67700] 10ns 483.39kb|-------------L0.?--------------|                                                         "
    - "L0.?[67701,85771] 10ns 833.13kb                                 |-------------------------L0.?-------------------------| "
    - "**** Simulation run 27, type=split(ReduceOverlap)(split_times=[135300]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.78[114329,142885] 10ns|-----------------------------------------L0.78------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,135300] 10ns 966.83kb|------------------------------L0.?------------------------------|                        "
    - "L0.?[135301,142885] 10ns 349.69kb                                                                  |--------L0.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.34, L0.36, L0.41, L0.43, L0.48, L0.50, L0.55, L0.57, L0.62, L0.64, L0.69, L0.71, L0.76, L0.78"
    - "  Creating 28 files"
    - "**** Simulation run 28, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[109041, 160867]). 14 Input Files, 27.55mb total:"
    - "L0                                                                                                                 "
    - "L0.93[57215,67700] 4ns 483.39kb|L0.93|                                                                                   "
    - "L0.94[67701,85771] 4ns 833.13kb      |--L0.94--|                                                                         "
    - "L0.35[85772,114328] 4ns 1.29mb                  |-----L0.35-----|                                                       "
    - "L0.95[114329,135300] 4ns 966.83kb                                    |---L0.95---|                                         "
    - "L0.96[135301,142885] 4ns 349.69kb                                                 |L0.96|                                  "
    - "L0.37[142886,171442] 4ns 1.29mb                                                      |-----L0.37-----|                   "
    - "L0.38[171443,200000] 4ns 1.29mb                                                                        |-----L0.38------|"
    - "L1                                                                                                                 "
    - "L1.86[57215,67700] 4ns 1.55mb|L1.86|                                                                                   "
    - "L1.87[67701,85771] 4ns 2.67mb      |--L1.87--|                                                                         "
    - "L1.88[85772,114328] 4ns 4.22mb                  |-----L1.88-----|                                                       "
    - "L1.89[114329,135300] 4ns 3.1mb                                    |---L1.89---|                                         "
    - "L1.90[135301,142885] 4ns 1.12mb                                                 |L1.90|                                  "
    - "L1.91[142886,171442] 4ns 4.22mb                                                      |-----L1.91-----|                   "
    - "L1.92[171443,200000] 4ns 4.22mb                                                                        |-----L1.92------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 27.55mb total:"
    - "L1                                                                                                                 "
    - "L1.?[57215,109041] 4ns 10mb|-------------L1.?-------------|                                                          "
    - "L1.?[109042,160867] 4ns 10mb                                |-------------L1.?-------------|                          "
    - "L1.?[160868,200000] 4ns 7.55mb                                                                 |---------L1.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.35, L0.37, L0.38, L1.86, L1.87, L1.88, L1.89, L1.90, L1.91, L1.92, L0.93, L0.94, L0.95, L0.96"
    - "  Creating 3 files"
    - "**** Simulation run 29, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.121[57215,109041] 4ns |-----------------------------------------L1.121-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[57215,85772] 4ns 5.51mb|---------------------L1.?----------------------|                                         "
    - "L1.?[85773,109041] 4ns 4.49mb                                                 |-----------------L1.?-----------------| "
    - "**** Simulation run 30, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.42[85772,114328] 5ns  |-----------------------------------------L0.42------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 5ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 5ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 31, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.49[85772,114328] 6ns  |-----------------------------------------L0.49------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 6ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 6ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 32, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.56[85772,114328] 7ns  |-----------------------------------------L0.56------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 7ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 7ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 33, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.63[85772,114328] 8ns  |-----------------------------------------L0.63------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 8ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 8ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 34, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.70[85772,114328] 9ns  |-----------------------------------------L0.70------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 9ns 0b |L0.?|                                                                                    "
    - "L0.?[85773,114328] 9ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 35, type=split(HighL0OverlapTotalBacklog)(split_times=[85772]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.77[85772,114328] 10ns |-----------------------------------------L0.77------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85772,85772] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[85773,114328] 10ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 36, type=split(HighL0OverlapTotalBacklog)(split_times=[114329, 142886]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.122[109042,160867] 4ns|-----------------------------------------L1.122-----------------------------------------|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[109042,114329] 4ns 1.02mb|-L1.?--|                                                                                 "
    - "L1.?[114330,142886] 4ns 5.51mb         |---------------------L1.?----------------------|                                "
    - "L1.?[142887,160867] 4ns 3.47mb                                                          |------------L1.?-------------| "
    - "**** Simulation run 37, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.99[114329,135300] 5ns |-----------------------------------------L0.99------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 5ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 5ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 38, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.103[114329,135300] 6ns|-----------------------------------------L0.103-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 6ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 6ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 39, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.107[114329,135300] 7ns|-----------------------------------------L0.107-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 7ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 7ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 40, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.111[114329,135300] 8ns|-----------------------------------------L0.111-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 8ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 8ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 41, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.115[114329,135300] 9ns|-----------------------------------------L0.115-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 9ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 9ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 42, type=split(HighL0OverlapTotalBacklog)(split_times=[114329]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.119[114329,135300] 10ns|-----------------------------------------L0.119-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114329,114329] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[114330,135300] 10ns 966.83kb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 43, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.44[142886,171442] 5ns |-----------------------------------------L0.44------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 5ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 5ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 44, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.51[142886,171442] 6ns |-----------------------------------------L0.51------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 6ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 6ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 45, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.58[142886,171442] 7ns |-----------------------------------------L0.58------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 7ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 7ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 46, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.65[142886,171442] 8ns |-----------------------------------------L0.65------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 8ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 8ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 47, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.72[142886,171442] 9ns |-----------------------------------------L0.72------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 9ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 9ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 48, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.79[142886,171442] 10ns|-----------------------------------------L0.79------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142886,142886] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[142887,171442] 10ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 49, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 7.55mb total:"
    - "L1, all files 7.55mb                                                                                               "
    - "L1.123[160868,200000] 4ns|-----------------------------------------L1.123-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 7.55mb total:"
    - "L1                                                                                                                 "
    - "L1.?[160868,171443] 4ns 2.04mb|---------L1.?---------|                                                                  "
    - "L1.?[171444,200000] 4ns 5.51mb                        |-----------------------------L1.?------------------------------| "
    - "**** Simulation run 50, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.45[171443,200000] 5ns |-----------------------------------------L0.45------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 5ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 5ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 51, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.66[171443,200000] 8ns |-----------------------------------------L0.66------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 8ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 8ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 52, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.73[171443,200000] 9ns |-----------------------------------------L0.73------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 9ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 9ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 53, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.80[171443,200000] 10ns|-----------------------------------------L0.80------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 10ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 10ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 54, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.52[171443,200000] 6ns |-----------------------------------------L0.52------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 6ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 6ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "**** Simulation run 55, type=split(HighL0OverlapTotalBacklog)(split_times=[171443]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.59[171443,200000] 7ns |-----------------------------------------L0.59------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171443,171443] 7ns 0b|L0.?|                                                                                    "
    - "L0.?[171444,200000] 7ns 1.29mb|-----------------------------------------L0.?------------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 27 files: L0.42, L0.44, L0.45, L0.49, L0.51, L0.52, L0.56, L0.58, L0.59, L0.63, L0.65, L0.66, L0.70, L0.72, L0.73, L0.77, L0.79, L0.80, L0.99, L0.103, L0.107, L0.111, L0.115, L0.119, L1.121, L1.122, L1.123"
    - "  Creating 55 files"
    - "**** Simulation run 56, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.127[85773,114328] 5ns |-----------------------------------------L0.127-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 5ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 5ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 57, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.154[142887,171442] 5ns|-----------------------------------------L0.154-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 5ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 5ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 58, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.129[85773,114328] 6ns |-----------------------------------------L0.129-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 6ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 6ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 59, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.156[142887,171442] 6ns|-----------------------------------------L0.156-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 6ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 6ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 60, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.131[85773,114328] 7ns |-----------------------------------------L0.131-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 7ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 7ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 61, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.158[142887,171442] 7ns|-----------------------------------------L0.158-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 7ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 7ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 62, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.133[85773,114328] 8ns |-----------------------------------------L0.133-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 8ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 8ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 63, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.160[142887,171442] 8ns|-----------------------------------------L0.160-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 8ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 8ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 64, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.135[85773,114328] 9ns |-----------------------------------------L0.135-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 9ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 9ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 65, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.162[142887,171442] 9ns|-----------------------------------------L0.162-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 9ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 9ns 487.56kb                                                        |-------------L0.?--------------| "
    - "**** Simulation run 66, type=split(ReduceOverlap)(split_times=[109041]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.137[85773,114328] 10ns|-----------------------------------------L0.137-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,109041] 10ns 1.05mb|---------------------------------L0.?----------------------------------|                 "
    - "L0.?[109042,114328] 10ns 243.76kb                                                                         |-----L0.?-----| "
    - "**** Simulation run 67, type=split(ReduceOverlap)(split_times=[160867]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.164[142887,171442] 10ns|-----------------------------------------L0.164-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,160867] 10ns 828.96kb|-------------------------L0.?-------------------------|                                  "
    - "L0.?[160868,171442] 10ns 487.56kb                                                        |-------------L0.?--------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 12 files: L0.127, L0.129, L0.131, L0.133, L0.135, L0.137, L0.154, L0.156, L0.158, L0.160, L0.162, L0.164"
    - "  Creating 24 files"
    - "**** Simulation run 68, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[46513, 92926]). 13 Input Files, 24.61mb total:"
    - "L0                                                                                                                 "
    - "L0.39[100,28657] 5ns 1.29mb|-------L0.39--------|                                                                    "
    - "L0.40[28658,57214] 5ns 1.29mb                      |-------L0.40--------|                                              "
    - "L0.97[57215,67700] 5ns 483.39kb                                             |L0.97-|                                     "
    - "L0.98[67701,85771] 5ns 833.13kb                                                     |---L0.98----|                       "
    - "L0.126[85772,85772] 5ns 0b                                                                   |L0.126|               "
    - "L0.179[85773,109041] 5ns 1.05mb                                                                   |-----L0.179-----|     "
    - "L0.180[109042,114328] 5ns 243.76kb                                                                                     |L0.180|"
    - "L0.141[114329,114329] 5ns 0b                                                                                         |L0.141|"
    - "L1                                                                                                                 "
    - "L1.84[100,28657] 4ns 4.22mb|-------L1.84--------|                                                                    "
    - "L1.85[28658,57214] 4ns 4.22mb                      |-------L1.85--------|                                              "
    - "L1.124[57215,85772] 4ns 5.51mb                                             |-------L1.124-------|                       "
    - "L1.125[85773,109041] 4ns 4.49mb                                                                   |-----L1.125-----|     "
    - "L1.138[109042,114329] 4ns 1.02mb                                                                                     |L1.138|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.61mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,46513] 5ns 10mb |---------------L1.?---------------|                                                      "
    - "L1.?[46514,92926] 5ns 10mb                                    |---------------L1.?---------------|                  "
    - "L1.?[92927,114329] 5ns 4.61mb                                                                         |-----L1.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 13 files: L0.39, L0.40, L1.84, L1.85, L0.97, L0.98, L1.124, L1.125, L0.126, L1.138, L0.141, L0.179, L0.180"
    - "  Creating 3 files"
    - "**** Simulation run 69, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.203[100,46513] 5ns    |----------------------------------------L1.203-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,38176] 5ns 8.2mb|---------------------------------L1.?----------------------------------|                 "
    - "L1.?[38177,46513] 5ns 1.8mb                                                                         |-----L1.?-----| "
    - "**** Simulation run 70, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.47[28658,57214] 6ns   |-----------------------------------------L0.47------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 6ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 6ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 71, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.54[28658,57214] 7ns   |-----------------------------------------L0.54------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 7ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 7ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 72, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.61[28658,57214] 8ns   |-----------------------------------------L0.61------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 8ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 8ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 73, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.68[28658,57214] 9ns   |-----------------------------------------L0.68------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 9ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 9ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 74, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.75[28658,57214] 10ns  |-----------------------------------------L0.75------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,38176] 10ns 438.81kb|-----------L0.?------------|                                                             "
    - "L0.?[38177,57214] 10ns 877.71kb                              |--------------------------L0.?---------------------------| "
    - "**** Simulation run 75, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.204[46514,92926] 5ns  |-----------------------------------------L1.204-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[46514,76252] 5ns 6.41mb|-------------------------L1.?--------------------------|                                 "
    - "L1.?[76253,92926] 5ns 3.59mb                                                         |-------------L1.?-------------| "
    - "**** Simulation run 76, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.102[67701,85771] 6ns  |-----------------------------------------L0.102-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 6ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 6ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 77, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.106[67701,85771] 7ns  |-----------------------------------------L0.106-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 7ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 7ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 78, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.110[67701,85771] 8ns  |-----------------------------------------L0.110-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 8ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 8ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 79, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.114[67701,85771] 9ns  |-----------------------------------------L0.114-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 9ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 9ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "**** Simulation run 80, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 833.13kb total:"
    - "L0, all files 833.13kb                                                                                             "
    - "L0.118[67701,85771] 10ns |-----------------------------------------L0.118-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 833.13kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76252] 10ns 394.25kb|------------------L0.?------------------|                                                "
    - "L0.?[76253,85771] 10ns 438.88kb                                          |--------------------L0.?---------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 12 files: L0.47, L0.54, L0.61, L0.68, L0.75, L0.102, L0.106, L0.110, L0.114, L0.118, L1.203, L1.204"
    - "  Creating 24 files"
    - "**** Simulation run 81, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.209[38177,57214] 6ns  |-----------------------------------------L0.209-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 6ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 6ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 82, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.183[85773,109041] 6ns |----------------------------------------L0.183-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 6ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 6ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 83, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.211[38177,57214] 7ns  |-----------------------------------------L0.211-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 7ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 7ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 84, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.187[85773,109041] 7ns |----------------------------------------L0.187-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 7ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 7ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 85, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.213[38177,57214] 8ns  |-----------------------------------------L0.213-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 8ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 8ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 86, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.191[85773,109041] 8ns |----------------------------------------L0.191-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 8ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 8ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 87, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.215[38177,57214] 9ns  |-----------------------------------------L0.215-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 9ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 9ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 88, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.195[85773,109041] 9ns |----------------------------------------L0.195-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 9ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 9ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "**** Simulation run 89, type=split(ReduceOverlap)(split_times=[46513]). 1 Input Files, 877.71kb total:"
    - "L0, all files 877.71kb                                                                                             "
    - "L0.217[38177,57214] 10ns |-----------------------------------------L0.217-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 877.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,46513] 10ns 384.33kb|----------------L0.?-----------------|                                                   "
    - "L0.?[46514,57214] 10ns 493.38kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 90, type=split(ReduceOverlap)(split_times=[92926]). 1 Input Files, 1.05mb total:"
    - "L0, all files 1.05mb                                                                                               "
    - "L0.199[85773,109041] 10ns|----------------------------------------L0.199-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.05mb total:"
    - "L0                                                                                                                 "
    - "L0.?[85773,92926] 10ns 329.79kb|----------L0.?-----------|                                                               "
    - "L0.?[92927,109041] 10ns 742.98kb                           |----------------------------L0.?----------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.183, L0.187, L0.191, L0.195, L0.199, L0.209, L0.211, L0.213, L0.215, L0.217"
    - "  Creating 20 files"
    - "**** Simulation run 91, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[156350, 198370]). 11 Input Files, 20.39mb total:"
    - "L0                                                                                                                 "
    - "L0.142[114330,135300] 5ns 966.83kb|-------L0.142-------|                                                                    "
    - "L0.100[135301,142885] 5ns 349.69kb                      |L0.100|                                                            "
    - "L0.153[142886,142886] 5ns 0b                             |L0.153|                                                     "
    - "L0.181[142887,160867] 5ns 828.96kb                              |-----L0.181-----|                                          "
    - "L0.182[160868,171442] 5ns 487.56kb                                                |-L0.182--|                               "
    - "L0.167[171443,171443] 5ns 0b                                                           |L0.167|                       "
    - "L0.168[171444,200000] 5ns 1.29mb                                                            |----------L0.168-----------| "
    - "L1                                                                                                                 "
    - "L1.139[114330,142886] 4ns 5.51mb|----------L1.139-----------|                                                             "
    - "L1.140[142887,160867] 4ns 3.47mb                              |-----L1.140-----|                                          "
    - "L1.165[160868,171443] 4ns 2.04mb                                                |-L1.165--|                               "
    - "L1.166[171444,200000] 4ns 5.51mb                                                            |----------L1.166-----------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 20.39mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,156350] 5ns 10mb|-------------------L1.?-------------------|                                              "
    - "L1.?[156351,198370] 5ns 10mb                                            |-------------------L1.?-------------------|  "
    - "L1.?[198371,200000] 5ns 397.47kb                                                                                        |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 11 files: L0.100, L1.139, L1.140, L0.142, L0.153, L1.165, L1.166, L0.167, L0.168, L0.181, L0.182"
    - "  Creating 3 files"
    - "**** Simulation run 92, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.250[114330,156350] 5ns|-----------------------------------------L1.250-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,142886] 5ns 6.8mb|---------------------------L1.?----------------------------|                             "
    - "L1.?[142887,156350] 5ns 3.2mb                                                             |-----------L1.?-----------| "
    - "**** Simulation run 93, type=split(HighL0OverlapTotalBacklog)(split_times=[171442]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.251[156351,198370] 5ns|-----------------------------------------L1.251-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[156351,171442] 5ns 3.59mb|-------------L1.?-------------|                                                          "
    - "L1.?[171443,198370] 5ns 6.41mb                                |-------------------------L1.?--------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.250, L1.251"
    - "  Creating 4 files"
    - "**** Simulation run 94, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.185[142887,160867] 6ns|-----------------------------------------L0.185-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 6ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 6ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 95, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.176[171444,200000] 6ns|-----------------------------------------L0.176-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 6ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 6ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 96, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.189[142887,160867] 7ns|-----------------------------------------L0.189-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 7ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 7ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 97, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.178[171444,200000] 7ns|-----------------------------------------L0.178-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 7ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 7ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 98, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.193[142887,160867] 8ns|-----------------------------------------L0.193-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 8ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 8ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 99, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.170[171444,200000] 8ns|-----------------------------------------L0.170-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 8ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 8ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 100, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.197[142887,160867] 9ns|-----------------------------------------L0.197-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 9ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 9ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 101, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.172[171444,200000] 9ns|-----------------------------------------L0.172-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 9ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 9ns 75.17kb                                                                                    |L0.?|"
    - "**** Simulation run 102, type=split(ReduceOverlap)(split_times=[156350]). 1 Input Files, 828.96kb total:"
    - "L0, all files 828.96kb                                                                                             "
    - "L0.201[142887,160867] 10ns|-----------------------------------------L0.201-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 828.96kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,156350] 10ns 620.71kb|------------------------------L0.?-------------------------------|                       "
    - "L0.?[156351,160867] 10ns 208.25kb                                                                   |--------L0.?--------| "
    - "**** Simulation run 103, type=split(ReduceOverlap)(split_times=[198370]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.174[171444,200000] 10ns|-----------------------------------------L0.174-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,198370] 10ns 1.21mb|---------------------------------------L0.?---------------------------------------|      "
    - "L0.?[198371,200000] 10ns 75.17kb                                                                                    |L0.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.170, L0.172, L0.174, L0.176, L0.178, L0.185, L0.189, L0.193, L0.197, L0.201"
    - "  Creating 20 files"
    - "**** Simulation run 104, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[38491, 76882]). 17 Input Files, 29.75mb total:"
    - "L0                                                                                                                 "
    - "L0.46[100,28657] 6ns 1.29mb|-------L0.46--------|                                                                    "
    - "L0.208[28658,38176] 6ns 438.81kb                      |L0.208|                                                            "
    - "L0.230[38177,46513] 6ns 384.33kb                              |L0.230|                                                    "
    - "L0.231[46514,57214] 6ns 493.38kb                                    |L0.231|                                              "
    - "L0.101[57215,67700] 6ns 483.39kb                                             |L0.101|                                     "
    - "L0.220[67701,76252] 6ns 394.25kb                                                     |L0.220|                             "
    - "L0.221[76253,85771] 6ns 438.88kb                                                            |L0.221|                      "
    - "L0.128[85772,85772] 6ns 0b                                                                   |L0.128|               "
    - "L0.232[85773,92926] 6ns 329.79kb                                                                   |L0.232|               "
    - "L0.233[92927,109041] 6ns 742.98kb                                                                         |--L0.233--|     "
    - "L0.184[109042,114328] 6ns 243.76kb                                                                                     |L0.184|"
    - "L0.143[114329,114329] 6ns 0b                                                                                         |L0.143|"
    - "L1                                                                                                                 "
    - "L1.206[100,38176] 5ns 8.2mb|----------L1.206-----------|                                                             "
    - "L1.207[38177,46513] 5ns 1.8mb                              |L1.207|                                                    "
    - "L1.218[46514,76252] 5ns 6.41mb                                    |-------L1.218--------|                               "
    - "L1.219[76253,92926] 5ns 3.59mb                                                            |--L1.219---|                 "
    - "L1.205[92927,114329] 5ns 4.61mb                                                                         |----L1.205----| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.75mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,38491] 6ns 10mb |------------L1.?------------|                                                            "
    - "L1.?[38492,76882] 6ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[76883,114329] 6ns 9.75mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 17 files: L0.46, L0.101, L0.128, L0.143, L0.184, L1.205, L1.206, L1.207, L0.208, L1.218, L1.219, L0.220, L0.221, L0.230, L0.231, L0.232, L0.233"
    - "  Creating 3 files"
    - "**** Simulation run 105, type=split(HighL0OverlapTotalBacklog)(split_times=[38176]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.277[100,38491] 6ns    |-----------------------------------------L1.277-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,38176] 6ns 9.92mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[38177,38491] 6ns 84.02kb                                                                                         |L1.?|"
    - "**** Simulation run 106, type=split(HighL0OverlapTotalBacklog)(split_times=[76252]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.278[38492,76882] 6ns  |-----------------------------------------L1.278-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[38492,76252] 6ns 9.84mb|-----------------------------------------L1.?-----------------------------------------|  "
    - "L1.?[76253,76882] 6ns 168.04kb                                                                                        |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.277, L1.278"
    - "  Creating 4 files"
    - "**** Simulation run 107, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.234[38177,46513] 7ns  |-----------------------------------------L0.234-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 7ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 7ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 108, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.223[76253,85771] 7ns  |-----------------------------------------L0.223-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 7ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 7ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "**** Simulation run 109, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.238[38177,46513] 8ns  |-----------------------------------------L0.238-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 8ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 8ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 110, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.225[76253,85771] 8ns  |-----------------------------------------L0.225-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 8ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 8ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "**** Simulation run 111, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.242[38177,46513] 9ns  |-----------------------------------------L0.242-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 9ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 9ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 112, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.227[76253,85771] 9ns  |-----------------------------------------L0.227-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 9ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 9ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "**** Simulation run 113, type=split(ReduceOverlap)(split_times=[38491]). 1 Input Files, 384.33kb total:"
    - "L0, all files 384.33kb                                                                                             "
    - "L0.246[38177,46513] 10ns |-----------------------------------------L0.246-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 384.33kb total:"
    - "L0                                                                                                                 "
    - "L0.?[38177,38491] 10ns 14.48kb|L0.?|                                                                                    "
    - "L0.?[38492,46513] 10ns 369.86kb   |----------------------------------------L0.?----------------------------------------| "
    - "**** Simulation run 114, type=split(ReduceOverlap)(split_times=[76882]). 1 Input Files, 438.88kb total:"
    - "L0, all files 438.88kb                                                                                             "
    - "L0.229[76253,85771] 10ns |-----------------------------------------L0.229-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.88kb total:"
    - "L0                                                                                                                 "
    - "L0.?[76253,76882] 10ns 29kb|L0.?|                                                                                    "
    - "L0.?[76883,85771] 10ns 409.88kb     |---------------------------------------L0.?---------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.223, L0.225, L0.227, L0.229, L0.234, L0.238, L0.242, L0.246"
    - "  Creating 16 files"
    - "**** Simulation run 115, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[149665, 185000]). 14 Input Files, 24.25mb total:"
    - "L0                                                                                                                 "
    - "L0.144[114330,135300] 6ns 966.83kb|-------L0.144-------|                                                                    "
    - "L0.104[135301,142885] 6ns 349.69kb                      |L0.104|                                                            "
    - "L0.155[142886,142886] 6ns 0b                             |L0.155|                                                     "
    - "L0.257[142887,156350] 6ns 620.71kb                              |---L0.257---|                                              "
    - "L0.258[156351,160867] 6ns 208.25kb                                            |L0.258|                                      "
    - "L0.186[160868,171442] 6ns 487.56kb                                                |-L0.186--|                               "
    - "L0.175[171443,171443] 6ns 0b                                                           |L0.175|                       "
    - "L0.259[171444,198370] 6ns 1.21mb                                                            |----------L0.259----------|  "
    - "L0.260[198371,200000] 6ns 75.17kb                                                                                        |L0.260|"
    - "L1                                                                                                                 "
    - "L1.253[114330,142886] 5ns 6.8mb|----------L1.253-----------|                                                             "
    - "L1.254[142887,156350] 5ns 3.2mb                              |---L1.254---|                                              "
    - "L1.255[156351,171442] 5ns 3.59mb                                            |---L1.255----|                               "
    - "L1.256[171443,198370] 5ns 6.41mb                                                           |----------L1.256----------|   "
    - "L1.252[198371,200000] 5ns 397.47kb                                                                                        |L1.252|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.25mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,149665] 6ns 10mb|---------------L1.?----------------|                                                     "
    - "L1.?[149666,185000] 6ns 10mb                                     |---------------L1.?----------------|                "
    - "L1.?[185001,200000] 6ns 4.25mb                                                                          |----L1.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.104, L0.144, L0.155, L0.175, L0.186, L1.252, L1.253, L1.254, L1.255, L1.256, L0.257, L0.258, L0.259, L0.260"
    - "  Creating 3 files"
    - "**** Simulation run 116, type=split(HighL0OverlapTotalBacklog)(split_times=[142886]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.300[114330,149665] 6ns|-----------------------------------------L1.300-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[114330,142886] 6ns 8.08mb|---------------------------------L1.?---------------------------------|                  "
    - "L1.?[142887,149665] 6ns 1.92mb                                                                        |-----L1.?------| "
    - "**** Simulation run 117, type=split(HighL0OverlapTotalBacklog)(split_times=[171442]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.301[149666,185000] 6ns|-----------------------------------------L1.301-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[149666,171442] 6ns 6.16mb|------------------------L1.?-------------------------|                                   "
    - "L1.?[171443,185000] 6ns 3.84mb                                                       |--------------L1.?--------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.300, L1.301"
    - "  Creating 4 files"
    - "**** Simulation run 118, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.261[142887,156350] 7ns|----------------------------------------L0.261-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 7ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 7ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 119, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.263[171444,198370] 7ns|----------------------------------------L0.263-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 7ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 7ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 120, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.265[142887,156350] 8ns|----------------------------------------L0.265-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 8ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 8ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 121, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.267[171444,198370] 8ns|----------------------------------------L0.267-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 8ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 8ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 122, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.269[142887,156350] 9ns|----------------------------------------L0.269-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 9ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 9ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 123, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.271[171444,198370] 9ns|----------------------------------------L0.271-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 9ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 9ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 124, type=split(ReduceOverlap)(split_times=[149665]). 1 Input Files, 620.71kb total:"
    - "L0, all files 620.71kb                                                                                             "
    - "L0.273[142887,156350] 10ns|----------------------------------------L0.273-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 620.71kb total:"
    - "L0                                                                                                                 "
    - "L0.?[142887,149665] 10ns 312.5kb|-------------------L0.?--------------------|                                             "
    - "L0.?[149666,156350] 10ns 308.21kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 125, type=split(ReduceOverlap)(split_times=[185000]). 1 Input Files, 1.21mb total:"
    - "L0, all files 1.21mb                                                                                               "
    - "L0.275[171444,198370] 10ns|----------------------------------------L0.275-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.21mb total:"
    - "L0                                                                                                                 "
    - "L0.?[171444,185000] 10ns 625.13kb|-------------------L0.?--------------------|                                             "
    - "L0.?[185001,198370] 10ns 616.55kb                                             |-------------------L0.?-------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.261, L0.263, L0.265, L0.267, L0.269, L0.271, L0.273, L0.275"
    - "  Creating 16 files"
    - "**** Simulation run 126, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[32834, 65568]). 12 Input Files, 23.46mb total:"
    - "L0                                                                                                                 "
    - "L0.53[100,28657] 7ns 1.29mb|-------------L0.53-------------|                                                         "
    - "L0.210[28658,38176] 7ns 438.81kb                                 |-L0.210--|                                              "
    - "L0.284[38177,38491] 7ns 14.48kb                                            |L0.284|                                      "
    - "L0.285[38492,46513] 7ns 369.86kb                                             |L0.285-|                                    "
    - "L0.235[46514,57214] 7ns 493.38kb                                                      |--L0.235--|                        "
    - "L0.105[57215,67700] 7ns 483.39kb                                                                  |--L0.105--|            "
    - "L0.222[67701,76252] 7ns 394.25kb                                                                               |-L0.222-| "
    - "L0.286[76253,76882] 7ns 29kb                                                                                         |L0.286|"
    - "L1                                                                                                                 "
    - "L1.280[100,38176] 6ns 9.92mb|------------------L1.280------------------|                                              "
    - "L1.281[38177,38491] 6ns 84.02kb                                            |L1.281|                                      "
    - "L1.282[38492,76252] 6ns 9.84mb                                             |------------------L1.282------------------| "
    - "L1.283[76253,76882] 6ns 168.04kb                                                                                         |L1.283|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 23.46mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,32834] 7ns 10mb |----------------L1.?----------------|                                                    "
    - "L1.?[32835,65568] 7ns 10mb                                      |----------------L1.?----------------|              "
    - "L1.?[65569,76882] 7ns 3.46mb                                                                            |---L1.?----| "
    - "Committing partition 1:"
    - "  Soft Deleting 12 files: L0.53, L0.105, L0.210, L0.222, L0.235, L1.280, L1.281, L1.282, L1.283, L0.284, L0.285, L0.286"
    - "  Creating 3 files"
    - "**** Simulation run 127, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.323[100,32834] 7ns    |-----------------------------------------L1.323-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25694] 7ns 7.82mb|--------------------------------L1.?--------------------------------|                    "
    - "L1.?[25695,32834] 7ns 2.18mb                                                                      |------L1.?-------| "
    - "**** Simulation run 128, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.60[100,28657] 8ns     |-----------------------------------------L0.60------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,25694] 8ns 1.15mb|-------------------------------------L0.?-------------------------------------|          "
    - "L0.?[25695,28657] 8ns 136.6kb                                                                                |-L0.?--| "
    - "**** Simulation run 129, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.67[100,28657] 9ns     |-----------------------------------------L0.67------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,25694] 9ns 1.15mb|-------------------------------------L0.?-------------------------------------|          "
    - "L0.?[25695,28657] 9ns 136.6kb                                                                                |-L0.?--| "
    - "**** Simulation run 130, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 1.29mb total:"
    - "L0, all files 1.29mb                                                                                               "
    - "L0.74[100,28657] 10ns    |-----------------------------------------L0.74------------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 1.29mb total:"
    - "L0                                                                                                                 "
    - "L0.?[100,25694] 10ns 1.15mb|-------------------------------------L0.?-------------------------------------|          "
    - "L0.?[25695,28657] 10ns 136.6kb                                                                                |-L0.?--| "
    - "**** Simulation run 131, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.324[32835,65568] 7ns  |-----------------------------------------L1.324-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[32835,51288] 7ns 5.64mb|----------------------L1.?----------------------|                                        "
    - "L1.?[51289,65568] 7ns 4.36mb                                                  |----------------L1.?-----------------| "
    - "**** Simulation run 132, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 493.38kb total:"
    - "L0, all files 493.38kb                                                                                             "
    - "L0.239[46514,57214] 8ns  |-----------------------------------------L0.239-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 493.38kb total:"
    - "L0                                                                                                                 "
    - "L0.?[46514,51288] 8ns 220.13kb|-----------------L0.?-----------------|                                                  "
    - "L0.?[51289,57214] 8ns 273.25kb                                        |---------------------L0.?----------------------| "
    - "**** Simulation run 133, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 493.38kb total:"
    - "L0, all files 493.38kb                                                                                             "
    - "L0.243[46514,57214] 9ns  |-----------------------------------------L0.243-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 493.38kb total:"
    - "L0                                                                                                                 "
    - "L0.?[46514,51288] 9ns 220.13kb|-----------------L0.?-----------------|                                                  "
    - "L0.?[51289,57214] 9ns 273.25kb                                        |---------------------L0.?----------------------| "
    - "**** Simulation run 134, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 493.38kb total:"
    - "L0, all files 493.38kb                                                                                             "
    - "L0.247[46514,57214] 10ns |-----------------------------------------L0.247-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 493.38kb total:"
    - "L0                                                                                                                 "
    - "L0.?[46514,51288] 10ns 220.13kb|-----------------L0.?-----------------|                                                  "
    - "L0.?[51289,57214] 10ns 273.25kb                                        |---------------------L0.?----------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.60, L0.67, L0.74, L0.239, L0.243, L0.247, L1.323, L1.324"
    - "  Creating 16 files"
    - "**** Simulation run 135, type=split(ReduceOverlap)(split_times=[32834]). 1 Input Files, 438.81kb total:"
    - "L0, all files 438.81kb                                                                                             "
    - "L0.212[28658,38176] 8ns  |-----------------------------------------L0.212-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.81kb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,32834] 8ns 192.53kb|----------------L0.?-----------------|                                                   "
    - "L0.?[32835,38176] 8ns 246.28kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 136, type=split(ReduceOverlap)(split_times=[65568]). 1 Input Files, 483.39kb total:"
    - "L0, all files 483.39kb                                                                                             "
    - "L0.109[57215,67700] 8ns  |-----------------------------------------L0.109-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 483.39kb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,65568] 8ns 385.1kb|--------------------------------L0.?---------------------------------|                   "
    - "L0.?[65569,67700] 8ns 98.29kb                                                                       |------L0.?------| "
    - "**** Simulation run 137, type=split(ReduceOverlap)(split_times=[32834]). 1 Input Files, 438.81kb total:"
    - "L0, all files 438.81kb                                                                                             "
    - "L0.214[28658,38176] 9ns  |-----------------------------------------L0.214-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.81kb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,32834] 9ns 192.53kb|----------------L0.?-----------------|                                                   "
    - "L0.?[32835,38176] 9ns 246.28kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 138, type=split(ReduceOverlap)(split_times=[65568]). 1 Input Files, 483.39kb total:"
    - "L0, all files 483.39kb                                                                                             "
    - "L0.113[57215,67700] 9ns  |-----------------------------------------L0.113-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 483.39kb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,65568] 9ns 385.1kb|--------------------------------L0.?---------------------------------|                   "
    - "L0.?[65569,67700] 9ns 98.29kb                                                                       |------L0.?------| "
    - "**** Simulation run 139, type=split(ReduceOverlap)(split_times=[32834]). 1 Input Files, 438.81kb total:"
    - "L0, all files 438.81kb                                                                                             "
    - "L0.216[28658,38176] 10ns |-----------------------------------------L0.216-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 438.81kb total:"
    - "L0                                                                                                                 "
    - "L0.?[28658,32834] 10ns 192.53kb|----------------L0.?-----------------|                                                   "
    - "L0.?[32835,38176] 10ns 246.28kb                                       |----------------------L0.?----------------------| "
    - "**** Simulation run 140, type=split(ReduceOverlap)(split_times=[65568]). 1 Input Files, 483.39kb total:"
    - "L0, all files 483.39kb                                                                                             "
    - "L0.117[57215,67700] 10ns |-----------------------------------------L0.117-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 483.39kb total:"
    - "L0                                                                                                                 "
    - "L0.?[57215,65568] 10ns 385.1kb|--------------------------------L0.?---------------------------------|                   "
    - "L0.?[65569,67700] 10ns 98.29kb                                                                       |------L0.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 6 files: L0.109, L0.113, L0.117, L0.212, L0.214, L0.216"
    - "  Creating 12 files"
    - "**** Simulation run 141, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[108723, 140563]). 16 Input Files, 29.7mb total:"
    - "L0                                                                                                                 "
    - "L0.287[76883,85771] 7ns 409.88kb|L0.287|                                                                                  "
    - "L0.130[85772,85772] 7ns 0b        |L0.130|                                                                          "
    - "L0.236[85773,92926] 7ns 329.79kb        |L0.236|                                                                          "
    - "L0.237[92927,109041] 7ns 742.98kb               |---L0.237----|                                                            "
    - "L0.188[109042,114328] 7ns 243.76kb                              |L0.188|                                                    "
    - "L0.145[114329,114329] 7ns 0b                                   |L0.145|                                               "
    - "L0.146[114330,135300] 7ns 966.83kb                                   |-----L0.146------|                                    "
    - "L0.108[135301,142885] 7ns 349.69kb                                                       |L0.108|                           "
    - "L0.157[142886,142886] 7ns 0b                                                              |L0.157|                    "
    - "L0.307[142887,149665] 7ns 312.5kb                                                              |L0.307|                    "
    - "L0.308[149666,156350] 7ns 308.21kb                                                                     |L0.308|             "
    - "L0.262[156351,160867] 7ns 208.25kb                                                                           |L0.262|       "
    - "L1                                                                                                                 "
    - "L1.279[76883,114329] 6ns 9.75mb|-------------L1.279--------------|                                                       "
    - "L1.303[114330,142886] 6ns 8.08mb                                   |---------L1.303----------|                            "
    - "L1.304[142887,149665] 6ns 1.92mb                                                              |L1.304|                    "
    - "L1.305[149666,171442] 6ns 6.16mb                                                                     |------L1.305------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.7mb total:"
    - "L1                                                                                                                 "
    - "L1.?[76883,108723] 7ns 10mb|------------L1.?------------|                                                            "
    - "L1.?[108724,140563] 7ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[140564,171442] 7ns 9.7mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 16 files: L0.108, L0.130, L0.145, L0.146, L0.157, L0.188, L0.236, L0.237, L0.262, L1.279, L0.287, L1.303, L1.304, L1.305, L0.307, L0.308"
    - "  Creating 3 files"
    - "**** Simulation run 142, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.354[76883,108723] 7ns |-----------------------------------------L1.354-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[76883,108402] 7ns 9.9mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[108403,108723] 7ns 103.24kb                                                                                         |L1.?|"
    - "**** Simulation run 143, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 742.98kb total:"
    - "L0, all files 742.98kb                                                                                             "
    - "L0.241[92927,109041] 8ns |-----------------------------------------L0.241-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 742.98kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,108402] 8ns 713.51kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[108403,109041] 8ns 29.46kb                                                                                      |L0.?|"
    - "**** Simulation run 144, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 742.98kb total:"
    - "L0, all files 742.98kb                                                                                             "
    - "L0.245[92927,109041] 9ns |-----------------------------------------L0.245-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 742.98kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,108402] 9ns 713.51kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[108403,109041] 9ns 29.46kb                                                                                      |L0.?|"
    - "**** Simulation run 145, type=split(HighL0OverlapTotalBacklog)(split_times=[108402]). 1 Input Files, 742.98kb total:"
    - "L0, all files 742.98kb                                                                                             "
    - "L0.249[92927,109041] 10ns|-----------------------------------------L0.249-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 742.98kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,108402] 10ns 713.51kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[108403,109041] 10ns 29.46kb                                                                                      |L0.?|"
    - "**** Simulation run 146, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.355[108724,140563] 7ns|-----------------------------------------L1.355-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[108724,139921] 7ns 9.8mb|-----------------------------------------L1.?-----------------------------------------|  "
    - "L1.?[139922,140563] 7ns 206.48kb                                                                                        |L1.?|"
    - "**** Simulation run 147, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 349.69kb total:"
    - "L0, all files 349.69kb                                                                                             "
    - "L0.112[135301,142885] 8ns|-----------------------------------------L0.112-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 349.69kb total:"
    - "L0                                                                                                                 "
    - "L0.?[135301,139921] 8ns 213.02kb|------------------------L0.?------------------------|                                    "
    - "L0.?[139922,142885] 8ns 136.67kb                                                      |--------------L0.?---------------| "
    - "**** Simulation run 148, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 349.69kb total:"
    - "L0, all files 349.69kb                                                                                             "
    - "L0.116[135301,142885] 9ns|-----------------------------------------L0.116-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 349.69kb total:"
    - "L0                                                                                                                 "
    - "L0.?[135301,139921] 9ns 213.02kb|------------------------L0.?------------------------|                                    "
    - "L0.?[139922,142885] 9ns 136.67kb                                                      |--------------L0.?---------------| "
    - "**** Simulation run 149, type=split(HighL0OverlapTotalBacklog)(split_times=[139921]). 1 Input Files, 349.69kb total:"
    - "L0, all files 349.69kb                                                                                             "
    - "L0.120[135301,142885] 10ns|-----------------------------------------L0.120-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 349.69kb total:"
    - "L0                                                                                                                 "
    - "L0.?[135301,139921] 10ns 213.02kb|------------------------L0.?------------------------|                                    "
    - "L0.?[139922,142885] 10ns 136.67kb                                                      |--------------L0.?---------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.112, L0.116, L0.120, L0.241, L0.245, L0.249, L1.354, L1.355"
    - "  Creating 16 files"
    - "**** Simulation run 150, type=split(ReduceOverlap)(split_times=[108723]). 1 Input Files, 29.46kb total:"
    - "L0, all files 29.46kb                                                                                              "
    - "L0.360[108403,109041] 8ns|-----------------------------------------L0.360-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 29.46kb total:"
    - "L0                                                                                                                 "
    - "L0.?[108403,108723] 8ns 14.78kb|-------------------L0.?--------------------|                                             "
    - "L0.?[108724,109041] 8ns 14.69kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 151, type=split(ReduceOverlap)(split_times=[140563]). 1 Input Files, 136.67kb total:"
    - "L0, all files 136.67kb                                                                                             "
    - "L0.368[139922,142885] 8ns|-----------------------------------------L0.368-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.67kb total:"
    - "L0                                                                                                                 "
    - "L0.?[139922,140563] 8ns 29.57kb|------L0.?-------|                                                                       "
    - "L0.?[140564,142885] 8ns 107.1kb                   |--------------------------------L0.?--------------------------------| "
    - "**** Simulation run 152, type=split(ReduceOverlap)(split_times=[108723]). 1 Input Files, 29.46kb total:"
    - "L0, all files 29.46kb                                                                                              "
    - "L0.362[108403,109041] 9ns|-----------------------------------------L0.362-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 29.46kb total:"
    - "L0                                                                                                                 "
    - "L0.?[108403,108723] 9ns 14.78kb|-------------------L0.?--------------------|                                             "
    - "L0.?[108724,109041] 9ns 14.69kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 153, type=split(ReduceOverlap)(split_times=[140563]). 1 Input Files, 136.67kb total:"
    - "L0, all files 136.67kb                                                                                             "
    - "L0.370[139922,142885] 9ns|-----------------------------------------L0.370-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.67kb total:"
    - "L0                                                                                                                 "
    - "L0.?[139922,140563] 9ns 29.57kb|------L0.?-------|                                                                       "
    - "L0.?[140564,142885] 9ns 107.1kb                   |--------------------------------L0.?--------------------------------| "
    - "**** Simulation run 154, type=split(ReduceOverlap)(split_times=[108723]). 1 Input Files, 29.46kb total:"
    - "L0, all files 29.46kb                                                                                              "
    - "L0.364[108403,109041] 10ns|-----------------------------------------L0.364-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 29.46kb total:"
    - "L0                                                                                                                 "
    - "L0.?[108403,108723] 10ns 14.78kb|-------------------L0.?--------------------|                                             "
    - "L0.?[108724,109041] 10ns 14.69kb                                             |-------------------L0.?-------------------| "
    - "**** Simulation run 155, type=split(ReduceOverlap)(split_times=[140563]). 1 Input Files, 136.67kb total:"
    - "L0, all files 136.67kb                                                                                             "
    - "L0.372[139922,142885] 10ns|-----------------------------------------L0.372-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.67kb total:"
    - "L0                                                                                                                 "
    - "L0.?[139922,140563] 10ns 29.57kb|------L0.?-------|                                                                       "
    - "L0.?[140564,142885] 10ns 107.1kb                   |--------------------------------L0.?--------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 6 files: L0.360, L0.362, L0.364, L0.368, L0.370, L0.372"
    - "  Creating 12 files"
    - "**** Simulation run 156, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[170977]). 8 Input Files, 19.54mb total:"
    - "L0                                                                                                                 "
    - "L0.190[160868,171442] 7ns 487.56kb                              |----L0.190----|                                            "
    - "L0.177[171443,171443] 7ns 0b                                              |L0.177|                                    "
    - "L0.309[171444,185000] 7ns 625.13kb                                              |------L0.309------|                        "
    - "L0.310[185001,198370] 7ns 616.55kb                                                                   |------L0.310------|   "
    - "L0.264[198371,200000] 7ns 75.17kb                                                                                       |L0.264|"
    - "L1                                                                                                                 "
    - "L1.356[140564,171442] 7ns 9.7mb|-------------------L1.356-------------------|                                            "
    - "L1.306[171443,185000] 6ns 3.84mb                                              |------L1.306------|                        "
    - "L1.302[185001,200000] 6ns 4.25mb                                                                   |-------L1.302-------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 19.54mb total:"
    - "L1                                                                                                                 "
    - "L1.?[140564,170977] 7ns 10mb|--------------------L1.?--------------------|                                            "
    - "L1.?[170978,200000] 7ns 9.54mb                                              |------------------L1.?-------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.177, L0.190, L0.264, L1.302, L1.306, L0.309, L0.310, L1.356"
    - "  Creating 2 files"
    - "**** Simulation run 157, type=split(ReduceOverlap)(split_times=[170977]). 1 Input Files, 487.56kb total:"
    - "L0, all files 487.56kb                                                                                             "
    - "L0.194[160868,171442] 8ns|-----------------------------------------L0.194-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 487.56kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,170977] 8ns 466.12kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[170978,171442] 8ns 21.44kb                                                                                      |L0.?|"
    - "**** Simulation run 158, type=split(ReduceOverlap)(split_times=[170977]). 1 Input Files, 487.56kb total:"
    - "L0, all files 487.56kb                                                                                             "
    - "L0.198[160868,171442] 9ns|-----------------------------------------L0.198-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 487.56kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,170977] 9ns 466.12kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[170978,171442] 9ns 21.44kb                                                                                      |L0.?|"
    - "**** Simulation run 159, type=split(ReduceOverlap)(split_times=[170977]). 1 Input Files, 487.56kb total:"
    - "L0, all files 487.56kb                                                                                             "
    - "L0.202[160868,171442] 10ns|-----------------------------------------L0.202-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 487.56kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,170977] 10ns 466.12kb|----------------------------------------L0.?----------------------------------------|    "
    - "L0.?[170978,171442] 10ns 21.44kb                                                                                      |L0.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.194, L0.198, L0.202"
    - "  Creating 6 files"
    - "**** Simulation run 160, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[28630, 57160]). 17 Input Files, 26.91mb total:"
    - "L0                                                                                                                 "
    - "L0.328[100,25694] 8ns 1.15mb|-----------L0.328-----------|                                                            "
    - "L0.329[25695,28657] 8ns 136.6kb                              |L0.329|                                                    "
    - "L0.342[28658,32834] 8ns 192.53kb                                 |L0.342|                                                 "
    - "L0.343[32835,38176] 8ns 246.28kb                                      |L0.343|                                            "
    - "L0.288[38177,38491] 8ns 14.48kb                                            |L0.288|                                      "
    - "L0.289[38492,46513] 8ns 369.86kb                                             |L0.289-|                                    "
    - "L0.336[46514,51288] 8ns 220.13kb                                                      |L0.336|                            "
    - "L0.337[51289,57214] 8ns 273.25kb                                                            |L0.337|                      "
    - "L0.344[57215,65568] 8ns 385.1kb                                                                  |L0.344-|               "
    - "L0.345[65569,67700] 8ns 98.29kb                                                                            |L0.345|      "
    - "L0.224[67701,76252] 8ns 394.25kb                                                                               |-L0.224-| "
    - "L0.290[76253,76882] 8ns 29kb                                                                                         |L0.290|"
    - "L1                                                                                                                 "
    - "L1.326[100,25694] 7ns 7.82mb|-----------L1.326-----------|                                                            "
    - "L1.327[25695,32834] 7ns 2.18mb                              |L1.327|                                                    "
    - "L1.334[32835,51288] 7ns 5.64mb                                      |------L1.334-------|                               "
    - "L1.335[51289,65568] 7ns 4.36mb                                                            |----L1.335----|              "
    - "L1.325[65569,76882] 7ns 3.46mb                                                                            |--L1.325---| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 26.91mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,28630] 8ns 10mb |-------------L1.?--------------|                                                         "
    - "L1.?[28631,57160] 8ns 10mb                                 |-------------L1.?--------------|                        "
    - "L1.?[57161,76882] 8ns 6.91mb                                                                  |--------L1.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 17 files: L0.224, L0.288, L0.289, L0.290, L1.325, L1.326, L1.327, L0.328, L0.329, L1.334, L1.335, L0.336, L0.337, L0.342, L0.343, L0.344, L0.345"
    - "  Creating 3 files"
    - "**** Simulation run 161, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.393[100,28630] 8ns    |-----------------------------------------L1.393-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25694] 8ns 8.97mb|-------------------------------------L1.?-------------------------------------|          "
    - "L1.?[25695,28630] 8ns 1.03mb                                                                                |-L1.?--| "
    - "**** Simulation run 162, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.394[28631,57160] 8ns  |-----------------------------------------L1.394-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[28631,51288] 8ns 7.94mb|--------------------------------L1.?---------------------------------|                   "
    - "L1.?[51289,57160] 8ns 2.06mb                                                                       |------L1.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.393, L1.394"
    - "  Creating 4 files"
    - "**** Simulation run 163, type=split(ReduceOverlap)(split_times=[28630]). 1 Input Files, 136.6kb total:"
    - "L0, all files 136.6kb                                                                                              "
    - "L0.331[25695,28657] 9ns  |-----------------------------------------L0.331-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.6kb total:"
    - "L0                                                                                                                 "
    - "L0.?[25695,28630] 9ns 135.36kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[28631,28657] 9ns 1.25kb                                                                                         |L0.?|"
    - "**** Simulation run 164, type=split(ReduceOverlap)(split_times=[57160]). 1 Input Files, 273.25kb total:"
    - "L0, all files 273.25kb                                                                                             "
    - "L0.339[51289,57214] 9ns  |-----------------------------------------L0.339-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 273.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[51289,57160] 9ns 270.76kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[57161,57214] 9ns 2.49kb                                                                                         |L0.?|"
    - "**** Simulation run 165, type=split(ReduceOverlap)(split_times=[28630]). 1 Input Files, 136.6kb total:"
    - "L0, all files 136.6kb                                                                                              "
    - "L0.333[25695,28657] 10ns |-----------------------------------------L0.333-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 136.6kb total:"
    - "L0                                                                                                                 "
    - "L0.?[25695,28630] 10ns 135.36kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[28631,28657] 10ns 1.25kb                                                                                         |L0.?|"
    - "**** Simulation run 166, type=split(ReduceOverlap)(split_times=[57160]). 1 Input Files, 273.25kb total:"
    - "L0, all files 273.25kb                                                                                             "
    - "L0.341[51289,57214] 10ns |-----------------------------------------L0.341-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 273.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[51289,57160] 10ns 270.76kb|-----------------------------------------L0.?------------------------------------------| "
    - "L0.?[57161,57214] 10ns 2.49kb                                                                                         |L0.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L0.331, L0.333, L0.339, L0.341"
    - "  Creating 8 files"
    - "**** Simulation run 167, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[104731, 132579]). 15 Input Files, 22.87mb total:"
    - "L0                                                                                                                 "
    - "L0.291[76883,85771] 8ns 409.88kb|--L0.291--|                                                                              "
    - "L0.132[85772,85772] 8ns 0b            |L0.132|                                                                      "
    - "L0.240[85773,92926] 8ns 329.79kb            |-L0.240-|                                                                    "
    - "L0.359[92927,108402] 8ns 713.51kb                      |------L0.359-------|                                               "
    - "L0.373[108403,108723] 8ns 14.78kb                                            |L0.373|                                      "
    - "L0.374[108724,109041] 8ns 14.69kb                                             |L0.374|                                     "
    - "L0.192[109042,114328] 8ns 243.76kb                                             |L0.192|                                     "
    - "L0.147[114329,114329] 8ns 0b                                                    |L0.147|                              "
    - "L0.148[114330,135300] 8ns 966.83kb                                                    |----------L0.148-----------|         "
    - "L0.367[135301,139921] 8ns 213.02kb                                                                                  |L0.367|"
    - "L0.375[139922,140563] 8ns 29.57kb                                                                                         |L0.375|"
    - "L1                                                                                                                 "
    - "L1.357[76883,108402] 7ns 9.9mb|------------------L1.357------------------|                                              "
    - "L1.358[108403,108723] 7ns 103.24kb                                            |L1.358|                                      "
    - "L1.365[108724,139921] 7ns 9.8mb                                             |------------------L1.365------------------| "
    - "L1.366[139922,140563] 7ns 206.48kb                                                                                         |L1.366|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.87mb total:"
    - "L1                                                                                                                 "
    - "L1.?[76883,104731] 8ns 10mb|----------------L1.?-----------------|                                                   "
    - "L1.?[104732,132579] 8ns 10mb                                       |----------------L1.?-----------------|            "
    - "L1.?[132580,140563] 8ns 2.87mb                                                                              |--L1.?---| "
    - "Committing partition 1:"
    - "  Soft Deleting 15 files: L0.132, L0.147, L0.148, L0.192, L0.240, L0.291, L1.357, L1.358, L0.359, L1.365, L1.366, L0.367, L0.373, L0.374, L0.375"
    - "  Creating 3 files"
    - "**** Simulation run 168, type=split(ReduceOverlap)(split_times=[104731]). 1 Input Files, 713.51kb total:"
    - "L0, all files 713.51kb                                                                                             "
    - "L0.361[92927,108402] 9ns |-----------------------------------------L0.361-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 713.51kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,104731] 9ns 544.25kb|-------------------------------L0.?-------------------------------|                      "
    - "L0.?[104732,108402] 9ns 169.26kb                                                                    |-------L0.?--------| "
    - "**** Simulation run 169, type=split(ReduceOverlap)(split_times=[132579]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.150[114330,135300] 9ns|-----------------------------------------L0.150-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114330,132579] 9ns 841.37kb|------------------------------------L0.?------------------------------------|            "
    - "L0.?[132580,135300] 9ns 125.45kb                                                                              |--L0.?---| "
    - "**** Simulation run 170, type=split(ReduceOverlap)(split_times=[104731]). 1 Input Files, 713.51kb total:"
    - "L0, all files 713.51kb                                                                                             "
    - "L0.363[92927,108402] 10ns|-----------------------------------------L0.363-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 713.51kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,104731] 10ns 544.25kb|-------------------------------L0.?-------------------------------|                      "
    - "L0.?[104732,108402] 10ns 169.26kb                                                                    |-------L0.?--------| "
    - "**** Simulation run 171, type=split(ReduceOverlap)(split_times=[132579]). 1 Input Files, 966.83kb total:"
    - "L0, all files 966.83kb                                                                                             "
    - "L0.152[114330,135300] 10ns|-----------------------------------------L0.152-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 966.83kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114330,132579] 10ns 841.37kb|------------------------------------L0.?------------------------------------|            "
    - "L0.?[132580,135300] 10ns 125.45kb                                                                              |--L0.?---| "
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L0.150, L0.152, L0.361, L0.363"
    - "  Creating 8 files"
    - "**** Simulation run 172, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[167314, 194064]). 13 Input Files, 22.22mb total:"
    - "L0                                                                                                                 "
    - "L0.376[140564,142885] 8ns 107.1kb|L0.376|                                                                                  "
    - "L0.159[142886,142886] 8ns 0b   |L0.159|                                                                               "
    - "L0.311[142887,149665] 8ns 312.5kb   |-L0.311-|                                                                             "
    - "L0.312[149666,156350] 8ns 308.21kb             |-L0.312-|                                                                   "
    - "L0.266[156351,160867] 8ns 208.25kb                       |L0.266|                                                           "
    - "L0.387[160868,170977] 8ns 466.12kb                              |---L0.387----|                                             "
    - "L0.388[170978,171442] 8ns 21.44kb                                              |L0.388|                                    "
    - "L0.169[171443,171443] 8ns 0b                                              |L0.169|                                    "
    - "L0.313[171444,185000] 8ns 625.13kb                                              |------L0.313------|                        "
    - "L0.314[185001,198370] 8ns 616.55kb                                                                   |------L0.314------|   "
    - "L0.268[198371,200000] 8ns 75.17kb                                                                                       |L0.268|"
    - "L1                                                                                                                 "
    - "L1.385[140564,170977] 7ns 10mb|-------------------L1.385-------------------|                                            "
    - "L1.386[170978,200000] 7ns 9.54mb                                              |-----------------L1.386------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.22mb total:"
    - "L1                                                                                                                 "
    - "L1.?[140564,167314] 8ns 10mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[167315,194064] 8ns 10mb                                        |-----------------L1.?-----------------|          "
    - "L1.?[194065,200000] 8ns 2.22mb                                                                                 |-L1.?-| "
    - "Committing partition 1:"
    - "  Soft Deleting 13 files: L0.159, L0.169, L0.266, L0.268, L0.311, L0.312, L0.313, L0.314, L0.376, L1.385, L1.386, L0.387, L0.388"
    - "  Creating 3 files"
    - "**** Simulation run 173, type=split(ReduceOverlap)(split_times=[167314]). 1 Input Files, 466.12kb total:"
    - "L0, all files 466.12kb                                                                                             "
    - "L0.389[160868,170977] 9ns|-----------------------------------------L0.389-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 466.12kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,167314] 9ns 297.22kb|-------------------------L0.?--------------------------|                                 "
    - "L0.?[167315,170977] 9ns 168.9kb                                                         |-------------L0.?-------------| "
    - "**** Simulation run 174, type=split(ReduceOverlap)(split_times=[194064]). 1 Input Files, 616.55kb total:"
    - "L0, all files 616.55kb                                                                                             "
    - "L0.318[185001,198370] 9ns|-----------------------------------------L0.318-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 616.55kb total:"
    - "L0                                                                                                                 "
    - "L0.?[185001,194064] 9ns 417.97kb|---------------------------L0.?----------------------------|                             "
    - "L0.?[194065,198370] 9ns 198.58kb                                                             |-----------L0.?-----------| "
    - "**** Simulation run 175, type=split(ReduceOverlap)(split_times=[167314]). 1 Input Files, 466.12kb total:"
    - "L0, all files 466.12kb                                                                                             "
    - "L0.391[160868,170977] 10ns|-----------------------------------------L0.391-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 466.12kb total:"
    - "L0                                                                                                                 "
    - "L0.?[160868,167314] 10ns 297.22kb|-------------------------L0.?--------------------------|                                 "
    - "L0.?[167315,170977] 10ns 168.9kb                                                         |-------------L0.?-------------| "
    - "**** Simulation run 176, type=split(ReduceOverlap)(split_times=[194064]). 1 Input Files, 616.55kb total:"
    - "L0, all files 616.55kb                                                                                             "
    - "L0.322[185001,198370] 10ns|-----------------------------------------L0.322-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 616.55kb total:"
    - "L0                                                                                                                 "
    - "L0.?[185001,194064] 10ns 417.97kb|---------------------------L0.?----------------------------|                             "
    - "L0.?[194065,198370] 10ns 198.58kb                                                             |-----------L0.?-----------| "
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L0.318, L0.322, L0.389, L0.391"
    - "  Creating 8 files"
    - "**** Simulation run 177, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[25731, 51362]). 17 Input Files, 29.96mb total:"
    - "L0                                                                                                                 "
    - "L0.330[100,25694] 9ns 1.15mb|-----------L0.330-----------|                                                            "
    - "L0.400[25695,28630] 9ns 135.36kb                              |L0.400|                                                    "
    - "L0.401[28631,28657] 9ns 1.25kb                                 |L0.401|                                                 "
    - "L0.346[28658,32834] 9ns 192.53kb                                 |L0.346|                                                 "
    - "L0.347[32835,38176] 9ns 246.28kb                                      |L0.347|                                            "
    - "L0.292[38177,38491] 9ns 14.48kb                                            |L0.292|                                      "
    - "L0.293[38492,46513] 9ns 369.86kb                                             |L0.293-|                                    "
    - "L0.338[46514,51288] 9ns 220.13kb                                                      |L0.338|                            "
    - "L0.402[51289,57160] 9ns 270.76kb                                                            |L0.402|                      "
    - "L0.403[57161,57214] 9ns 2.49kb                                                                  |L0.403|                "
    - "L0.348[57215,65568] 9ns 385.1kb                                                                  |L0.348-|               "
    - "L0.349[65569,67700] 9ns 98.29kb                                                                            |L0.349|      "
    - "L1                                                                                                                 "
    - "L1.396[100,25694] 8ns 8.97mb|-----------L1.396-----------|                                                            "
    - "L1.397[25695,28630] 8ns 1.03mb                              |L1.397|                                                    "
    - "L1.398[28631,51288] 8ns 7.94mb                                 |---------L1.398---------|                               "
    - "L1.399[51289,57160] 8ns 2.06mb                                                            |L1.399|                      "
    - "L1.395[57161,76882] 8ns 6.91mb                                                                  |-------L1.395--------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 29.96mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25731] 9ns 10mb |------------L1.?------------|                                                            "
    - "L1.?[25732,51362] 9ns 10mb                              |------------L1.?------------|                              "
    - "L1.?[51363,76882] 9ns 9.96mb                                                            |-----------L1.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 17 files: L0.292, L0.293, L0.330, L0.338, L0.346, L0.347, L0.348, L0.349, L1.395, L1.396, L1.397, L1.398, L1.399, L0.400, L0.401, L0.402, L0.403"
    - "  Creating 3 files"
    - "**** Simulation run 178, type=split(HighL0OverlapTotalBacklog)(split_times=[25694]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.430[100,25731] 9ns    |-----------------------------------------L1.430-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,25694] 9ns 9.99mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[25695,25731] 9ns 14.78kb                                                                                         |L1.?|"
    - "**** Simulation run 179, type=split(HighL0OverlapTotalBacklog)(split_times=[51288]). 1 Input Files, 10mb total:"
    - "L1, all files 10mb                                                                                                 "
    - "L1.431[25732,51362] 9ns  |-----------------------------------------L1.431-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 10mb total:"
    - "L1                                                                                                                 "
    - "L1.?[25732,51288] 9ns 9.97mb|-----------------------------------------L1.?------------------------------------------| "
    - "L1.?[51289,51362] 9ns 29.56kb                                                                                         |L1.?|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.430, L1.431"
    - "  Creating 4 files"
    - "**** Simulation run 180, type=split(ReduceOverlap)(split_times=[25731]). 1 Input Files, 135.36kb total:"
    - "L0, all files 135.36kb                                                                                             "
    - "L0.404[25695,28630] 10ns |----------------------------------------L0.404-----------------------------------------| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 135.36kb total:"
    - "L0                                                                                                                 "
    - "L0.?[25695,25731] 10ns 1.66kb|L0.?|                                                                                    "
    - "L0.?[25732,28630] 10ns 133.7kb |-----------------------------------------L0.?-----------------------------------------| "
    - "**** Simulation run 181, type=split(ReduceOverlap)(split_times=[51362]). 1 Input Files, 270.76kb total:"
    - "L0, all files 270.76kb                                                                                             "
    - "L0.406[51289,57160] 10ns |-----------------------------------------L0.406-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 270.76kb total:"
    - "L0                                                                                                                 "
    - "L0.?[51289,51362] 10ns 3.37kb|L0.?|                                                                                    "
    - "L0.?[51363,57160] 10ns 267.39kb |-----------------------------------------L0.?-----------------------------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.404, L0.406"
    - "  Creating 4 files"
    - "**** Simulation run 182, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[76043, 100723]). 8 Input Files, 21.62mb total:"
    - "L0                                                                                                                 "
    - "L0.226[67701,76252] 9ns 394.25kb                           |---L0.226---|                                                 "
    - "L0.294[76253,76882] 9ns 29kb                                         |L0.294|                                         "
    - "L0.295[76883,85771] 9ns 409.88kb                                           |---L0.295---|                                 "
    - "L0.134[85772,85772] 9ns 0b                                                          |L0.134|                        "
    - "L0.244[85773,92926] 9ns 329.79kb                                                          |--L0.244--|                    "
    - "L0.411[92927,104731] 9ns 544.25kb                                                                      |-----L0.411------| "
    - "L1                                                                                                                 "
    - "L1.432[51363,76882] 9ns 9.96mb|-----------------L1.432------------------|                                               "
    - "L1.408[76883,104731] 8ns 10mb                                           |-------------------L1.408-------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 21.62mb total:"
    - "L1                                                                                                                 "
    - "L1.?[51363,76043] 9ns 10mb|-----------------L1.?------------------|                                                 "
    - "L1.?[76044,100723] 9ns 10mb                                         |-----------------L1.?------------------|        "
    - "L1.?[100724,104731] 9ns 1.62mb                                                                                   |L1.?| "
    - "Committing partition 1:"
    - "  Soft Deleting 8 files: L0.134, L0.226, L0.244, L0.294, L0.295, L1.408, L0.411, L1.432"
    - "  Creating 3 files"
    - "**** Simulation run 183, type=split(ReduceOverlap)(split_times=[76043]). 1 Input Files, 394.25kb total:"
    - "L0, all files 394.25kb                                                                                             "
    - "L0.228[67701,76252] 10ns |-----------------------------------------L0.228-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 394.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[67701,76043] 10ns 384.61kb|----------------------------------------L0.?-----------------------------------------|   "
    - "L0.?[76044,76252] 10ns 9.64kb                                                                                       |L0.?|"
    - "**** Simulation run 184, type=split(ReduceOverlap)(split_times=[100723]). 1 Input Files, 544.25kb total:"
    - "L0, all files 544.25kb                                                                                             "
    - "L0.415[92927,104731] 10ns|-----------------------------------------L0.415-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 544.25kb total:"
    - "L0                                                                                                                 "
    - "L0.?[92927,100723] 10ns 359.45kb|--------------------------L0.?---------------------------|                               "
    - "L0.?[100724,104731] 10ns 184.8kb                                                           |------------L0.?------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.228, L0.415"
    - "  Creating 4 files"
    - "**** Simulation run 185, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[129098, 153464]). 18 Input Files, 25.68mb total:"
    - "L0                                                                                                                 "
    - "L0.412[104732,108402] 9ns 169.26kb|L0.412|                                                                                  "
    - "L0.377[108403,108723] 9ns 14.78kb     |L0.377|                                                                             "
    - "L0.378[108724,109041] 9ns 14.69kb     |L0.378|                                                                             "
    - "L0.196[109042,114328] 9ns 243.76kb      |L0.196|                                                                            "
    - "L0.149[114329,114329] 9ns 0b             |L0.149|                                                                     "
    - "L0.413[114330,132579] 9ns 841.37kb             |---------L0.413---------|                                                   "
    - "L0.414[132580,135300] 9ns 125.45kb                                        |L0.414|                                          "
    - "L0.369[135301,139921] 9ns 213.02kb                                           |L0.369|                                       "
    - "L0.379[139922,140563] 9ns 29.57kb                                                  |L0.379|                                "
    - "L0.380[140564,142885] 9ns 107.1kb                                                   |L0.380|                               "
    - "L0.161[142886,142886] 9ns 0b                                                      |L0.161|                            "
    - "L0.315[142887,149665] 9ns 312.5kb                                                      |L0.315-|                           "
    - "L0.316[149666,156350] 9ns 308.21kb                                                                |L0.316-|                 "
    - "L0.270[156351,160867] 9ns 208.25kb                                                                          |L0.270|        "
    - "L0.422[160868,167314] 9ns 297.22kb                                                                                |L0.422-| "
    - "L1                                                                                                                 "
    - "L1.409[104732,132579] 8ns 10mb|----------------L1.409----------------|                                                  "
    - "L1.410[132580,140563] 8ns 2.87mb                                        |-L1.410--|                                       "
    - "L1.419[140564,167314] 8ns 10mb                                                   |---------------L1.419---------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 25.68mb total:"
    - "L1                                                                                                                 "
    - "L1.?[104732,129098] 9ns 10mb|--------------L1.?---------------|                                                       "
    - "L1.?[129099,153464] 9ns 10mb                                   |--------------L1.?---------------|                    "
    - "L1.?[153465,167314] 9ns 5.68mb                                                                      |------L1.?-------| "
    - "Committing partition 1:"
    - "  Soft Deleting 18 files: L0.149, L0.161, L0.196, L0.270, L0.315, L0.316, L0.369, L0.377, L0.378, L0.379, L0.380, L1.409, L1.410, L0.412, L0.413, L0.414, L1.419, L0.422"
    - "  Creating 3 files"
    - "**** Simulation run 186, type=split(ReduceOverlap)(split_times=[129098]). 1 Input Files, 841.37kb total:"
    - "L0, all files 841.37kb                                                                                             "
    - "L0.417[114330,132579] 10ns|-----------------------------------------L0.417-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 841.37kb total:"
    - "L0                                                                                                                 "
    - "L0.?[114330,129098] 10ns 680.88kb|---------------------------------L0.?---------------------------------|                  "
    - "L0.?[129099,132579] 10ns 160.49kb                                                                        |-----L0.?------| "
    - "**** Simulation run 187, type=split(ReduceOverlap)(split_times=[153464]). 1 Input Files, 308.21kb total:"
    - "L0, all files 308.21kb                                                                                             "
    - "L0.320[149666,156350] 10ns|-----------------------------------------L0.320-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 308.21kb total:"
    - "L0                                                                                                                 "
    - "L0.?[149666,153464] 10ns 175.13kb|----------------------L0.?-----------------------|                                       "
    - "L0.?[153465,156350] 10ns 133.08kb                                                   |----------------L0.?----------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.320, L0.417"
    - "  Creating 4 files"
    - "**** Simulation run 188, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[191189]). 9 Input Files, 13.69mb total:"
    - "L0                                                                                                                 "
    - "L0.423[167315,170977] 9ns 168.9kb|-L0.423-|                                                                                "
    - "L0.390[170978,171442] 9ns 21.44kb          |L0.390|                                                                        "
    - "L0.171[171443,171443] 9ns 0b           |L0.171|                                                                       "
    - "L0.317[171444,185000] 9ns 625.13kb           |--------------L0.317---------------|                                          "
    - "L0.424[185001,194064] 9ns 417.97kb                                                |--------L0.424--------|                  "
    - "L0.425[194065,198370] 9ns 198.58kb                                                                         |-L0.425--|      "
    - "L0.272[198371,200000] 9ns 75.17kb                                                                                     |L0.272|"
    - "L1                                                                                                                 "
    - "L1.420[167315,194064] 8ns 10mb|--------------------------------L1.420---------------------------------|                 "
    - "L1.421[194065,200000] 8ns 2.22mb                                                                         |----L1.421----| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 13.69mb total:"
    - "L1                                                                                                                 "
    - "L1.?[167315,191189] 9ns 10mb|-----------------------------L1.?------------------------------|                         "
    - "L1.?[191190,200000] 9ns 3.69mb                                                                 |---------L1.?---------| "
    - "Committing partition 1:"
    - "  Soft Deleting 9 files: L0.171, L0.272, L0.317, L0.390, L1.420, L1.421, L0.423, L0.424, L0.425"
    - "  Creating 2 files"
    - "**** Simulation run 189, type=split(ReduceOverlap)(split_times=[191189]). 1 Input Files, 417.97kb total:"
    - "L0, all files 417.97kb                                                                                             "
    - "L0.428[185001,194064] 10ns|-----------------------------------------L0.428-----------------------------------------|"
    - "**** 2 Output Files (parquet_file_id not yet assigned), 417.97kb total:"
    - "L0                                                                                                                 "
    - "L0.?[185001,191189] 10ns 285.38kb|---------------------------L0.?----------------------------|                             "
    - "L0.?[191190,194064] 10ns 132.59kb                                                             |-----------L0.?-----------| "
    - "Committing partition 1:"
    - "  Soft Deleting 1 files: L0.428"
    - "  Creating 2 files"
    - "**** Simulation run 190, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[23080, 46060]). 14 Input Files, 22.31mb total:"
    - "L0                                                                                                                 "
    - "L0.332[100,25694] 10ns 1.15mb|------------------L0.332------------------|                                              "
    - "L0.437[25695,25731] 10ns 1.66kb                                            |L0.437|                                      "
    - "L0.438[25732,28630] 10ns 133.7kb                                             |L0.438|                                     "
    - "L0.405[28631,28657] 10ns 1.25kb                                                  |L0.405|                                "
    - "L0.350[28658,32834] 10ns 192.53kb                                                  |L0.350|                                "
    - "L0.351[32835,38176] 10ns 246.28kb                                                         |L0.351-|                        "
    - "L0.296[38177,38491] 10ns 14.48kb                                                                  |L0.296|                "
    - "L0.297[38492,46513] 10ns 369.86kb                                                                   |---L0.297---|         "
    - "L0.340[46514,51288] 10ns 220.13kb                                                                                 |L0.340| "
    - "L0.439[51289,51362] 10ns 3.37kb                                                                                         |L0.439|"
    - "L1                                                                                                                 "
    - "L1.433[100,25694] 9ns 9.99mb|------------------L1.433------------------|                                              "
    - "L1.434[25695,25731] 9ns 14.78kb                                            |L1.434|                                      "
    - "L1.435[25732,51288] 9ns 9.97mb                                             |------------------L1.435------------------| "
    - "L1.436[51289,51362] 9ns 29.56kb                                                                                         |L1.436|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.31mb total:"
    - "L1                                                                                                                 "
    - "L1.?[100,23080] 10ns 10mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[23081,46060] 10ns 10mb                                        |-----------------L1.?-----------------|          "
    - "L1.?[46061,51362] 10ns 2.31mb                                                                                |-L1.?--| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.296, L0.297, L0.332, L0.340, L0.350, L0.351, L0.405, L1.433, L1.434, L1.435, L1.436, L0.437, L0.438, L0.439"
    - "  Creating 3 files"
    - "**** Simulation run 191, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[73575, 95787]). 15 Input Files, 24.03mb total:"
    - "L0                                                                                                                 "
    - "L0.440[51363,57160] 10ns 267.39kb|L0.440-|                                                                                 "
    - "L0.407[57161,57214] 10ns 2.49kb         |L0.407|                                                                         "
    - "L0.352[57215,65568] 10ns 385.1kb         |---L0.352---|                                                                   "
    - "L0.353[65569,67700] 10ns 98.29kb                       |L0.353|                                                           "
    - "L0.444[67701,76043] 10ns 384.61kb                           |---L0.444---|                                                 "
    - "L0.445[76044,76252] 10ns 9.64kb                                         |L0.445|                                         "
    - "L0.298[76253,76882] 10ns 29kb                                         |L0.298|                                         "
    - "L0.299[76883,85771] 10ns 409.88kb                                           |---L0.299---|                                 "
    - "L0.136[85772,85772] 10ns 0b                                                          |L0.136|                        "
    - "L0.248[85773,92926] 10ns 329.79kb                                                          |--L0.248--|                    "
    - "L0.446[92927,100723] 10ns 359.45kb                                                                      |--L0.446---|       "
    - "L0.447[100724,104731] 10ns 184.8kb                                                                                   |L0.447|"
    - "L1                                                                                                                 "
    - "L1.441[51363,76043] 9ns 10mb|----------------L1.441-----------------|                                                 "
    - "L1.442[76044,100723] 9ns 10mb                                         |----------------L1.442-----------------|        "
    - "L1.443[100724,104731] 9ns 1.62mb                                                                                   |L1.443|"
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.03mb total:"
    - "L1                                                                                                                 "
    - "L1.?[51363,73575] 10ns 10mb|---------------L1.?----------------|                                                     "
    - "L1.?[73576,95787] 10ns 10mb                                     |---------------L1.?----------------|                "
    - "L1.?[95788,104731] 10ns 4.03mb                                                                          |----L1.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 15 files: L0.136, L0.248, L0.298, L0.299, L0.352, L0.353, L0.407, L0.440, L1.441, L1.442, L1.443, L0.444, L0.445, L0.446, L0.447"
    - "  Creating 3 files"
    - "**** Simulation run 192, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[175139, 196813]). 14 Input Files, 21.47mb total:"
    - "L0                                                                                                                 "
    - "L0.276[198371,200000] 10ns 75.17kb                                                                                      |L0.276|"
    - "L0.429[194065,198370] 10ns 198.58kb                                                                              |L0.429|    "
    - "L0.458[191190,194064] 10ns 132.59kb                                                                        |L0.458|          "
    - "L0.457[185001,191189] 10ns 285.38kb                                                            |-L0.457--|                   "
    - "L0.321[171444,185000] 10ns 625.13kb                                  |---------L0.321---------|                              "
    - "L0.173[171443,171443] 10ns 0b                                  |L0.173|                                                "
    - "L0.392[170978,171442] 10ns 21.44kb                                 |L0.392|                                                 "
    - "L0.427[167315,170977] 10ns 168.9kb                          |L0.427|                                                        "
    - "L0.426[160868,167314] 10ns 297.22kb              |--L0.426--|                                                                "
    - "L0.274[156351,160867] 10ns 208.25kb     |L0.274|                                                                             "
    - "L0.454[153465,156350] 10ns 133.08kb|L0.454|                                                                                  "
    - "L1                                                                                                                 "
    - "L1.456[191190,200000] 9ns 3.69mb                                                                        |----L1.456-----| "
    - "L1.455[167315,191189] 9ns 10mb                          |-------------------L1.455-------------------|                  "
    - "L1.450[153465,167314] 9ns 5.68mb|---------L1.450---------|                                                                "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 21.47mb total:"
    - "L1                                                                                                                 "
    - "L1.?[153465,175139] 10ns 10mb|-----------------L1.?------------------|                                                 "
    - "L1.?[175140,196813] 10ns 10mb                                         |-----------------L1.?------------------|        "
    - "L1.?[196814,200000] 10ns 1.47mb                                                                                   |L1.?| "
    - "Committing partition 1:"
    - "  Soft Deleting 14 files: L0.173, L0.274, L0.276, L0.321, L0.392, L0.426, L0.427, L0.429, L1.450, L0.454, L1.455, L1.456, L0.457, L0.458"
    - "  Creating 3 files"
    - "**** Simulation run 193, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[126690, 148648]). 16 Input Files, 22.19mb total:"
    - "L0                                                                                                                 "
    - "L0.416[104732,108402] 10ns 169.26kb|L0.416|                                                                                  "
    - "L0.381[108403,108723] 10ns 14.78kb      |L0.381|                                                                            "
    - "L0.382[108724,109041] 10ns 14.69kb       |L0.382|                                                                           "
    - "L0.200[109042,114328] 10ns 243.76kb       |L0.200-|                                                                          "
    - "L0.151[114329,114329] 10ns 0b                 |L0.151|                                                                 "
    - "L0.451[114330,129098] 10ns 680.88kb                 |---------L0.451----------|                                              "
    - "L0.452[129099,132579] 10ns 160.49kb                                             |L0.452|                                     "
    - "L0.418[132580,135300] 10ns 125.45kb                                                   |L0.418|                               "
    - "L0.371[135301,139921] 10ns 213.02kb                                                        |L0.371|                          "
    - "L0.383[139922,140563] 10ns 29.57kb                                                                |L0.383|                  "
    - "L0.384[140564,142885] 10ns 107.1kb                                                                  |L0.384|                "
    - "L0.163[142886,142886] 10ns 0b                                                                      |L0.163|            "
    - "L0.319[142887,149665] 10ns 312.5kb                                                                      |--L0.319--|        "
    - "L0.453[149666,153464] 10ns 175.13kb                                                                                  |L0.453|"
    - "L1                                                                                                                 "
    - "L1.448[104732,129098] 9ns 10mb|------------------L1.448-------------------|                                             "
    - "L1.449[129099,153464] 9ns 10mb                                             |------------------L1.449------------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.19mb total:"
    - "L1                                                                                                                 "
    - "L1.?[104732,126690] 10ns 10mb|-----------------L1.?-----------------|                                                  "
    - "L1.?[126691,148648] 10ns 10mb                                        |-----------------L1.?-----------------|          "
    - "L1.?[148649,153464] 10ns 2.19mb                                                                                 |-L1.?-| "
    - "Committing partition 1:"
    - "  Soft Deleting 16 files: L0.151, L0.163, L0.200, L0.319, L0.371, L0.381, L0.382, L0.383, L0.384, L0.416, L0.418, L1.448, L1.449, L0.451, L0.452, L0.453"
    - "  Creating 3 files"
    - "**** Simulation run 194, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[45717, 68353]). 3 Input Files, 22.31mb total:"
    - "L1                                                                                                                 "
    - "L1.460[23081,46060] 10ns 10mb|----------------L1.460----------------|                                                  "
    - "L1.461[46061,51362] 10ns 2.31mb                                        |L1.461-|                                         "
    - "L1.462[51363,73575] 10ns 10mb                                                  |---------------L1.462----------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.31mb total:"
    - "L2                                                                                                                 "
    - "L2.?[23081,45717] 10ns 10mb|-----------------L2.?-----------------|                                                  "
    - "L2.?[45718,68353] 10ns 10mb                                        |-----------------L2.?-----------------|          "
    - "L2.?[68354,73575] 10ns 2.31mb                                                                                |-L2.?--| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.460, L1.461, L1.462"
    - "  Upgrading 1 files level to CompactionLevel::L2: L1.459"
    - "  Creating 3 files"
    - "**** Simulation run 195, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[95682, 117788]). 3 Input Files, 24.03mb total:"
    - "L1                                                                                                                 "
    - "L1.463[73576,95787] 10ns 10mb|--------------L1.463---------------|                                                     "
    - "L1.464[95788,104731] 10ns 4.03mb                                     |---L1.464----|                                      "
    - "L1.468[104732,126690] 10ns 10mb                                                    |--------------L1.468---------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 24.03mb total:"
    - "L2                                                                                                                 "
    - "L2.?[73576,95682] 10ns 10mb|---------------L2.?----------------|                                                     "
    - "L2.?[95683,117788] 10ns 10mb                                     |---------------L2.?----------------|                "
    - "L2.?[117789,126690] 10ns 4.03mb                                                                          |----L2.?-----| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.463, L1.464, L1.468"
    - "  Creating 3 files"
    - "**** Simulation run 196, type=split(CompactAndSplitOutput(FoundSubsetLessThanMaxCompactSize))(split_times=[148521, 170351]). 3 Input Files, 22.19mb total:"
    - "L1                                                                                                                 "
    - "L1.469[126691,148648] 10ns 10mb|----------------L1.469----------------|                                                  "
    - "L1.470[148649,153464] 10ns 2.19mb                                        |L1.470|                                          "
    - "L1.465[153465,175139] 10ns 10mb                                                 |----------------L1.465----------------| "
    - "**** 3 Output Files (parquet_file_id not yet assigned), 22.19mb total:"
    - "L2                                                                                                                 "
    - "L2.?[126691,148521] 10ns 10mb|-----------------L2.?-----------------|                                                  "
    - "L2.?[148522,170351] 10ns 10mb                                        |-----------------L2.?-----------------|          "
    - "L2.?[170352,175139] 10ns 2.19mb                                                                                 |-L2.?-| "
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L1.465, L1.469, L1.470"
    - "  Creating 3 files"
    - "**** Simulation run 197, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[195028]). 2 Input Files, 11.47mb total:"
    - "L1                                                                                                                 "
    - "L1.467[196814,200000] 10ns 1.47mb                                                                              |-L1.467--| "
    - "L1.466[175140,196813] 10ns 10mb|-----------------------------------L1.466-----------------------------------|            "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 11.47mb total:"
    - "L2                                                                                                                 "
    - "L2.?[175140,195028] 10ns 9.18mb|---------------------------------L2.?---------------------------------|                  "
    - "L2.?[195029,200000] 10ns 2.29mb                                                                        |-----L2.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.466, L1.467"
    - "  Creating 2 files"
    - "**** Final Output Files (994.29mb written)"
    - "L2                                                                                                                 "
    - "L2.459[100,23080] 10ns 10mb|-L2.459-|                                                                                "
    - "L2.471[23081,45717] 10ns 10mb          |-L2.471-|                                                                      "
    - "L2.472[45718,68353] 10ns 10mb                    |-L2.472-|                                                            "
    - "L2.473[68354,73575] 10ns 2.31mb                              |L2.473|                                                    "
    - "L2.474[73576,95682] 10ns 10mb                                 |L2.474-|                                                "
    - "L2.475[95683,117788] 10ns 10mb                                           |L2.475-|                                      "
    - "L2.476[117789,126690] 10ns 4.03mb                                                    |L2.476|                              "
    - "L2.477[126691,148521] 10ns 10mb                                                        |L2.477-|                         "
    - "L2.478[148522,170351] 10ns 10mb                                                                  |L2.478-|               "
    - "L2.479[170352,175139] 10ns 2.19mb                                                                            |L2.479|      "
    - "L2.480[175140,195028] 10ns 9.18mb                                                                              |L2.480|    "
    - "L2.481[195029,200000] 10ns 2.29mb                                                                                       |L2.481|"
    "###
    );
}
