//! layout tests for scenarios with large numbers of files
//!
//! See [crate::layout] module for detailed documentation

use data_types::CompactionLevel;
use iox_time::Time;

use crate::layouts::{layout_setup_builder, parquet_builder, run_layout_scenario, ONE_MB};

#[tokio::test]
async fn many_l0_files_different_created_order() {
    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder()
        .await
        // compact at most 2 L0 files per plan
        .with_max_num_files_per_plan(2)
        .build()
        .await;

    // L0.1
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(10)
                .with_max_time(22)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(1)),
        )
        .await;

    // L0.2
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(30)
                .with_max_time(42)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(2)),
        )
        .await;

    // L0.3
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(20)
                .with_max_time(32)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(3)),
        )
        .await;

    // L0.4
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(40)
                .with_max_time(52)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(4)),
        )
        .await;

    // L0 files must be compacted in their max_l0_created_at order to have the right deduplication
    // L0.1 and L0.2 must be compacted together before they are compacted with L0.3 and L0.4
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 2.55kb                                                                                               "
    - "L0.1[10,22] 1ns          |---------L0.1----------|                                                                 "
    - "L0.2[30,42] 2ns                                                    |---------L0.2----------|                       "
    - "L0.3[20,32] 3ns                               |---------L0.3----------|                                            "
    - "L0.4[40,52] 4ns                                                                          |---------L0.4----------| "
    - "**** Simulation run 0, type=compact(ManySmallFiles). 2 Input Files, 5.1kb total:"
    - "L0, all files 2.55kb                                                                                               "
    - "L0.1[10,22] 1ns          |-------------L0.1--------------|                                                         "
    - "L0.2[30,42] 2ns                                                                  |-------------L0.2--------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 5.1kb total:"
    - "L0, all files 5.1kb                                                                                                "
    - "L0.?[10,42] 2ns          |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.1, L0.2"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=compact(ManySmallFiles). 2 Input Files, 5.1kb total:"
    - "L0, all files 2.55kb                                                                                               "
    - "L0.3[20,32] 3ns          |-------------L0.3--------------|                                                         "
    - "L0.4[40,52] 4ns                                                                  |-------------L0.4--------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 5.1kb total:"
    - "L0, all files 5.1kb                                                                                                "
    - "L0.?[20,52] 4ns          |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.3, L0.4"
    - "  Creating 1 files"
    - "**** Simulation run 2, type=compact(TotalSizeLessThanMaxCompactSize). 2 Input Files, 10.2kb total:"
    - "L0, all files 5.1kb                                                                                                "
    - "L0.6[20,52] 4ns                               |-------------------------------L0.6-------------------------------| "
    - "L0.5[10,42] 2ns          |-------------------------------L0.5-------------------------------|                      "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 10.2kb total:"
    - "L1, all files 10.2kb                                                                                               "
    - "L1.?[10,52] 4ns          |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.5, L0.6"
    - "  Creating 1 files"
    - "**** Final Output Files (20.39kb written)"
    - "L1, all files 10.2kb                                                                                               "
    - "L1.7[10,52] 4ns          |------------------------------------------L1.7------------------------------------------|"
    "###
    );
}

#[tokio::test]
async fn many_l1_files_different_created_order() {
    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder()
        .await
        .with_min_num_l1_files_to_compact(2)
        .with_max_num_files_per_plan(2)
        .build()
        .await;

    // L1.1
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(11)
                .with_max_time(20)
                .with_compaction_level(CompactionLevel::FileNonOverlapped)
                .with_max_l0_created_at(Time::from_timestamp_nanos(1)),
        )
        .await;

    // L1.2
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(31)
                .with_max_time(40)
                .with_compaction_level(CompactionLevel::FileNonOverlapped)
                .with_max_l0_created_at(Time::from_timestamp_nanos(2)),
        )
        .await;

    // L1.3
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(21)
                .with_max_time(30)
                .with_compaction_level(CompactionLevel::FileNonOverlapped)
                .with_max_l0_created_at(Time::from_timestamp_nanos(3)),
        )
        .await;

    // L1.4
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(41)
                .with_max_time(50)
                .with_compaction_level(CompactionLevel::FileNonOverlapped)
                .with_max_l0_created_at(Time::from_timestamp_nanos(4)),
        )
        .await;

    // All L1 files are non-overlapped. Compacting them together must keep them non-overlapped.
    // If we cannot compact them together in one run, they must be split in the order of their min_time (or max_time)
    // to ensure that the output files are still non-overlapped.
    // L1.1 must be compacted with L1.3 instead of L1.2
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L1, all files 2.55kb                                                                                               "
    - "L1.1[11,20] 1ns          |-------L1.1-------|                                                                      "
    - "L1.2[31,40] 2ns                                                        |-------L1.2-------|                        "
    - "L1.3[21,30] 3ns                                 |-------L1.3-------|                                               "
    - "L1.4[41,50] 4ns                                                                               |-------L1.4-------| "
    - "**** Simulation run 0, type=compact(ManySmallFiles). 2 Input Files, 5.1kb total:"
    - "L1, all files 2.55kb                                                                                               "
    - "L1.1[11,20] 1ns          |------------------L1.1------------------|                                                "
    - "L1.3[21,30] 3ns                                                         |------------------L1.3------------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 5.1kb total:"
    - "L1, all files 5.1kb                                                                                                "
    - "L1.?[11,30] 3ns          |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.1, L1.3"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=compact(ManySmallFiles). 2 Input Files, 5.1kb total:"
    - "L1, all files 2.55kb                                                                                               "
    - "L1.2[31,40] 2ns          |------------------L1.2------------------|                                                "
    - "L1.4[41,50] 4ns                                                         |------------------L1.4------------------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 5.1kb total:"
    - "L1, all files 5.1kb                                                                                                "
    - "L1.?[31,50] 4ns          |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.2, L1.4"
    - "  Creating 1 files"
    - "**** Simulation run 2, type=compact(TotalSizeLessThanMaxCompactSize). 2 Input Files, 10.2kb total:"
    - "L1, all files 5.1kb                                                                                                "
    - "L1.6[31,50] 4ns                                                        |------------------L1.6-------------------| "
    - "L1.5[11,30] 3ns          |------------------L1.5-------------------|                                               "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 10.2kb total:"
    - "L2, all files 10.2kb                                                                                               "
    - "L2.?[11,50] 4ns          |------------------------------------------L2.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L1.5, L1.6"
    - "  Creating 1 files"
    - "**** Final Output Files (20.39kb written)"
    - "L2, all files 10.2kb                                                                                               "
    - "L2.7[11,50] 4ns          |------------------------------------------L2.7------------------------------------------|"
    "###
    );
}

#[tokio::test]
async fn many_l0_files_different_created_order_non_overlap() {
    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder()
        .await
        .with_min_num_l1_files_to_compact(2)
        .with_max_num_files_per_plan(2)
        .build()
        .await;

    // L0.1
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(11)
                .with_max_time(20)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(1)),
        )
        .await;

    // L0.2
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(31)
                .with_max_time(40)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(2)),
        )
        .await;

    // L0.3
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(21)
                .with_max_time(30)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(3)),
        )
        .await;

    // L0.4
    setup
        .partition
        .create_parquet_file(
            parquet_builder()
                .with_min_time(41)
                .with_max_time(50)
                .with_compaction_level(CompactionLevel::Initial)
                .with_max_l0_created_at(Time::from_timestamp_nanos(4)),
        )
        .await;

    // All L0 files do not overlap but their time ranges are in different order of their createion time
    // They still need to be sorted on their creation time before being compacted
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 2.55kb                                                                                               "
    - "L0.1[11,20] 1ns          |-------L0.1-------|                                                                      "
    - "L0.2[31,40] 2ns                                                        |-------L0.2-------|                        "
    - "L0.3[21,30] 3ns                                 |-------L0.3-------|                                               "
    - "L0.4[41,50] 4ns                                                                               |-------L0.4-------| "
    - "**** Simulation run 0, type=compact(ManySmallFiles). 2 Input Files, 5.1kb total:"
    - "L0, all files 2.55kb                                                                                               "
    - "L0.1[11,20] 1ns          |----------L0.1-----------|                                                               "
    - "L0.2[31,40] 2ns                                                                        |----------L0.2-----------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 5.1kb total:"
    - "L0, all files 5.1kb                                                                                                "
    - "L0.?[11,40] 2ns          |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.1, L0.2"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=compact(ManySmallFiles). 2 Input Files, 5.1kb total:"
    - "L0, all files 2.55kb                                                                                               "
    - "L0.3[21,30] 3ns          |----------L0.3-----------|                                                               "
    - "L0.4[41,50] 4ns                                                                        |----------L0.4-----------| "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 5.1kb total:"
    - "L0, all files 5.1kb                                                                                                "
    - "L0.?[21,50] 4ns          |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.3, L0.4"
    - "  Creating 1 files"
    - "**** Simulation run 2, type=compact(TotalSizeLessThanMaxCompactSize). 2 Input Files, 10.2kb total:"
    - "L0, all files 5.1kb                                                                                                "
    - "L0.6[21,50] 4ns                                 |------------------------------L0.6------------------------------| "
    - "L0.5[11,40] 2ns          |------------------------------L0.5------------------------------|                        "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 10.2kb total:"
    - "L1, all files 10.2kb                                                                                               "
    - "L1.?[11,50] 4ns          |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.5, L0.6"
    - "  Creating 1 files"
    - "**** Final Output Files (20.39kb written)"
    - "L1, all files 10.2kb                                                                                               "
    - "L1.7[11,50] 4ns          |------------------------------------------L1.7------------------------------------------|"
    "###
    );
}

#[tokio::test]
async fn many_l1_files() {
    test_helpers::maybe_start_logging();

    let setup = layout_setup_builder().await.build().await;

    // L1: 20 non overlapping files (more than with_min_num_l1_files_to_compact)
    for i in 0..20 {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i * 2)
                    .with_max_time(i * 2 + 1)
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    // content in L1 files ingested before all available L0
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(10 * ONE_MB),
            )
            .await;
    }
    // L0: a few small files that overlap
    for i in 0..3 {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(24)
                    .with_max_time(25)
                    // L0 files created after all L1 files
                    .with_max_l0_created_at(Time::from_timestamp_nanos(20 + i + 1))
                    .with_file_size_bytes(ONE_MB),
            )
            .await;
    }

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0                                                                                                                 "
    - "L0.21[24,25] 21ns 1mb                                                           |L0.21|                            "
    - "L0.22[24,25] 22ns 1mb                                                           |L0.22|                            "
    - "L0.23[24,25] 23ns 1mb                                                           |L0.23|                            "
    - "L1                                                                                                                 "
    - "L1.1[0,1] 1ns 10mb       |L1.1|                                                                                    "
    - "L1.2[2,3] 2ns 10mb           |L1.2|                                                                                "
    - "L1.3[4,5] 3ns 10mb                |L1.3|                                                                           "
    - "L1.4[6,7] 4ns 10mb                    |L1.4|                                                                       "
    - "L1.5[8,9] 5ns 10mb                         |L1.5|                                                                  "
    - "L1.6[10,11] 6ns 10mb                            |L1.6|                                                             "
    - "L1.7[12,13] 7ns 10mb                                |L1.7|                                                         "
    - "L1.8[14,15] 8ns 10mb                                     |L1.8|                                                    "
    - "L1.9[16,17] 9ns 10mb                                         |L1.9|                                                "
    - "L1.10[18,19] 10ns 10mb                                            |L1.10|                                          "
    - "L1.11[20,21] 11ns 10mb                                                 |L1.11|                                     "
    - "L1.12[22,23] 12ns 10mb                                                     |L1.12|                                 "
    - "L1.13[24,25] 13ns 10mb                                                          |L1.13|                            "
    - "L1.14[26,27] 14ns 10mb                                                               |L1.14|                       "
    - "L1.15[28,29] 15ns 10mb                                                                   |L1.15|                   "
    - "L1.16[30,31] 16ns 10mb                                                                        |L1.16|              "
    - "L1.17[32,33] 17ns 10mb                                                                            |L1.17|          "
    - "L1.18[34,35] 18ns 10mb                                                                                 |L1.18|     "
    - "L1.19[36,37] 19ns 10mb                                                                                      |L1.19|"
    - "L1.20[38,39] 20ns 10mb                                                                                          |L1.20|"
    - "**** Simulation run 0, type=compact(TotalSizeLessThanMaxCompactSize). 4 Input Files, 13mb total:"
    - "L0                                                                                                                 "
    - "L0.23[24,25] 23ns 1mb    |-----------------------------------------L0.23------------------------------------------|"
    - "L0.22[24,25] 22ns 1mb    |-----------------------------------------L0.22------------------------------------------|"
    - "L0.21[24,25] 21ns 1mb    |-----------------------------------------L0.21------------------------------------------|"
    - "L1                                                                                                                 "
    - "L1.13[24,25] 13ns 10mb   |-----------------------------------------L1.13------------------------------------------|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 13mb total:"
    - "L1, all files 13mb                                                                                                 "
    - "L1.?[24,25] 23ns         |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 4 files: L1.13, L0.21, L0.22, L0.23"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[20]). 20 Input Files, 203mb total:"
    - "L1                                                                                                                 "
    - "L1.14[26,27] 14ns 10mb                                                               |L1.14|                       "
    - "L1.15[28,29] 15ns 10mb                                                                   |L1.15|                   "
    - "L1.16[30,31] 16ns 10mb                                                                        |L1.16|              "
    - "L1.17[32,33] 17ns 10mb                                                                            |L1.17|          "
    - "L1.18[34,35] 18ns 10mb                                                                                 |L1.18|     "
    - "L1.19[36,37] 19ns 10mb                                                                                      |L1.19|"
    - "L1.20[38,39] 20ns 10mb                                                                                          |L1.20|"
    - "L1.12[22,23] 12ns 10mb                                                     |L1.12|                                 "
    - "L1.11[20,21] 11ns 10mb                                                 |L1.11|                                     "
    - "L1.10[18,19] 10ns 10mb                                            |L1.10|                                          "
    - "L1.9[16,17] 9ns 10mb                                         |L1.9|                                                "
    - "L1.8[14,15] 8ns 10mb                                     |L1.8|                                                    "
    - "L1.7[12,13] 7ns 10mb                                |L1.7|                                                         "
    - "L1.6[10,11] 6ns 10mb                            |L1.6|                                                             "
    - "L1.5[8,9] 5ns 10mb                         |L1.5|                                                                  "
    - "L1.4[6,7] 4ns 10mb                    |L1.4|                                                                       "
    - "L1.3[4,5] 3ns 10mb                |L1.3|                                                                           "
    - "L1.2[2,3] 2ns 10mb           |L1.2|                                                                                "
    - "L1.1[0,1] 1ns 10mb       |L1.1|                                                                                    "
    - "L1.24[24,25] 23ns 13mb                                                          |L1.24|                            "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 203mb total:"
    - "L2                                                                                                                 "
    - "L2.?[0,20] 23ns 104.1mb  |--------------------L2.?--------------------|                                            "
    - "L2.?[21,39] 23ns 98.9mb                                                  |-----------------L2.?------------------| "
    - "Committing partition 1:"
    - "  Soft Deleting 20 files: L1.1, L1.2, L1.3, L1.4, L1.5, L1.6, L1.7, L1.8, L1.9, L1.10, L1.11, L1.12, L1.14, L1.15, L1.16, L1.17, L1.18, L1.19, L1.20, L1.24"
    - "  Creating 2 files"
    - "**** Final Output Files (216mb written)"
    - "L2                                                                                                                 "
    - "L2.25[0,20] 23ns 104.1mb |-------------------L2.25--------------------|                                            "
    - "L2.26[21,39] 23ns 98.9mb                                                 |-----------------L2.26-----------------| "
    "###
    );
}

#[tokio::test]
async fn many_tiny_l0_files() {
    test_helpers::maybe_start_logging();

    // Observed size on production
    let seven_kb = 7 * 1024;

    let setup = layout_setup_builder().await.build().await;

    // models what happens if the compactor can't keep up for some
    // reason and gets to a partition where the ingster persistedq
    // every 5 minutes for an entire day with telegraf data, for
    // example
    let num_tiny_files = (24 * 60) / 5;

    for i in 0..num_tiny_files {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i)
                    .with_max_time(i + 1)
                    .with_compaction_level(CompactionLevel::Initial)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(seven_kb),
            )
            .await;
    }

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 7kb                                                                                                  "
    - "L0.1[0,1] 1ns            |L0.1|                                                                                    "
    - "L0.2[1,2] 2ns            |L0.2|                                                                                    "
    - "L0.3[2,3] 3ns            |L0.3|                                                                                    "
    - "L0.4[3,4] 4ns            |L0.4|                                                                                    "
    - "L0.5[4,5] 5ns             |L0.5|                                                                                   "
    - "L0.6[5,6] 6ns             |L0.6|                                                                                   "
    - "L0.7[6,7] 7ns             |L0.7|                                                                                   "
    - "L0.8[7,8] 8ns              |L0.8|                                                                                  "
    - "L0.9[8,9] 9ns              |L0.9|                                                                                  "
    - "L0.10[9,10] 10ns           |L0.10|                                                                                 "
    - "L0.11[10,11] 11ns           |L0.11|                                                                                "
    - "L0.12[11,12] 12ns           |L0.12|                                                                                "
    - "L0.13[12,13] 13ns           |L0.13|                                                                                "
    - "L0.14[13,14] 14ns            |L0.14|                                                                               "
    - "L0.15[14,15] 15ns            |L0.15|                                                                               "
    - "L0.16[15,16] 16ns            |L0.16|                                                                               "
    - "L0.17[16,17] 17ns             |L0.17|                                                                              "
    - "L0.18[17,18] 18ns             |L0.18|                                                                              "
    - "L0.19[18,19] 19ns             |L0.19|                                                                              "
    - "L0.20[19,20] 20ns             |L0.20|                                                                              "
    - "L0.21[20,21] 21ns              |L0.21|                                                                             "
    - "L0.22[21,22] 22ns              |L0.22|                                                                             "
    - "L0.23[22,23] 23ns              |L0.23|                                                                             "
    - "L0.24[23,24] 24ns               |L0.24|                                                                            "
    - "L0.25[24,25] 25ns               |L0.25|                                                                            "
    - "L0.26[25,26] 26ns               |L0.26|                                                                            "
    - "L0.27[26,27] 27ns                |L0.27|                                                                           "
    - "L0.28[27,28] 28ns                |L0.28|                                                                           "
    - "L0.29[28,29] 29ns                |L0.29|                                                                           "
    - "L0.30[29,30] 30ns                 |L0.30|                                                                          "
    - "L0.31[30,31] 31ns                 |L0.31|                                                                          "
    - "L0.32[31,32] 32ns                 |L0.32|                                                                          "
    - "L0.33[32,33] 33ns                  |L0.33|                                                                         "
    - "L0.34[33,34] 34ns                  |L0.34|                                                                         "
    - "L0.35[34,35] 35ns                  |L0.35|                                                                         "
    - "L0.36[35,36] 36ns                  |L0.36|                                                                         "
    - "L0.37[36,37] 37ns                   |L0.37|                                                                        "
    - "L0.38[37,38] 38ns                   |L0.38|                                                                        "
    - "L0.39[38,39] 39ns                   |L0.39|                                                                        "
    - "L0.40[39,40] 40ns                    |L0.40|                                                                       "
    - "L0.41[40,41] 41ns                    |L0.41|                                                                       "
    - "L0.42[41,42] 42ns                    |L0.42|                                                                       "
    - "L0.43[42,43] 43ns                     |L0.43|                                                                      "
    - "L0.44[43,44] 44ns                     |L0.44|                                                                      "
    - "L0.45[44,45] 45ns                     |L0.45|                                                                      "
    - "L0.46[45,46] 46ns                      |L0.46|                                                                     "
    - "L0.47[46,47] 47ns                      |L0.47|                                                                     "
    - "L0.48[47,48] 48ns                      |L0.48|                                                                     "
    - "L0.49[48,49] 49ns                       |L0.49|                                                                    "
    - "L0.50[49,50] 50ns                       |L0.50|                                                                    "
    - "L0.51[50,51] 51ns                       |L0.51|                                                                    "
    - "L0.52[51,52] 52ns                       |L0.52|                                                                    "
    - "L0.53[52,53] 53ns                        |L0.53|                                                                   "
    - "L0.54[53,54] 54ns                        |L0.54|                                                                   "
    - "L0.55[54,55] 55ns                        |L0.55|                                                                   "
    - "L0.56[55,56] 56ns                         |L0.56|                                                                  "
    - "L0.57[56,57] 57ns                         |L0.57|                                                                  "
    - "L0.58[57,58] 58ns                         |L0.58|                                                                  "
    - "L0.59[58,59] 59ns                          |L0.59|                                                                 "
    - "L0.60[59,60] 60ns                          |L0.60|                                                                 "
    - "L0.61[60,61] 61ns                          |L0.61|                                                                 "
    - "L0.62[61,62] 62ns                           |L0.62|                                                                "
    - "L0.63[62,63] 63ns                           |L0.63|                                                                "
    - "L0.64[63,64] 64ns                           |L0.64|                                                                "
    - "L0.65[64,65] 65ns                            |L0.65|                                                               "
    - "L0.66[65,66] 66ns                            |L0.66|                                                               "
    - "L0.67[66,67] 67ns                            |L0.67|                                                               "
    - "L0.68[67,68] 68ns                            |L0.68|                                                               "
    - "L0.69[68,69] 69ns                             |L0.69|                                                              "
    - "L0.70[69,70] 70ns                             |L0.70|                                                              "
    - "L0.71[70,71] 71ns                             |L0.71|                                                              "
    - "L0.72[71,72] 72ns                              |L0.72|                                                             "
    - "L0.73[72,73] 73ns                              |L0.73|                                                             "
    - "L0.74[73,74] 74ns                              |L0.74|                                                             "
    - "L0.75[74,75] 75ns                               |L0.75|                                                            "
    - "L0.76[75,76] 76ns                               |L0.76|                                                            "
    - "L0.77[76,77] 77ns                               |L0.77|                                                            "
    - "L0.78[77,78] 78ns                                |L0.78|                                                           "
    - "L0.79[78,79] 79ns                                |L0.79|                                                           "
    - "L0.80[79,80] 80ns                                |L0.80|                                                           "
    - "L0.81[80,81] 81ns                                 |L0.81|                                                          "
    - "L0.82[81,82] 82ns                                 |L0.82|                                                          "
    - "L0.83[82,83] 83ns                                 |L0.83|                                                          "
    - "L0.84[83,84] 84ns                                 |L0.84|                                                          "
    - "L0.85[84,85] 85ns                                  |L0.85|                                                         "
    - "L0.86[85,86] 86ns                                  |L0.86|                                                         "
    - "L0.87[86,87] 87ns                                  |L0.87|                                                         "
    - "L0.88[87,88] 88ns                                   |L0.88|                                                        "
    - "L0.89[88,89] 89ns                                   |L0.89|                                                        "
    - "L0.90[89,90] 90ns                                   |L0.90|                                                        "
    - "L0.91[90,91] 91ns                                    |L0.91|                                                       "
    - "L0.92[91,92] 92ns                                    |L0.92|                                                       "
    - "L0.93[92,93] 93ns                                    |L0.93|                                                       "
    - "L0.94[93,94] 94ns                                     |L0.94|                                                      "
    - "L0.95[94,95] 95ns                                     |L0.95|                                                      "
    - "L0.96[95,96] 96ns                                     |L0.96|                                                      "
    - "L0.97[96,97] 97ns                                      |L0.97|                                                     "
    - "L0.98[97,98] 98ns                                      |L0.98|                                                     "
    - "L0.99[98,99] 99ns                                      |L0.99|                                                     "
    - "L0.100[99,100] 100ns                                   |L0.100|                                                    "
    - "L0.101[100,101] 101ns                                   |L0.101|                                                   "
    - "L0.102[101,102] 102ns                                   |L0.102|                                                   "
    - "L0.103[102,103] 103ns                                   |L0.103|                                                   "
    - "L0.104[103,104] 104ns                                    |L0.104|                                                  "
    - "L0.105[104,105] 105ns                                    |L0.105|                                                  "
    - "L0.106[105,106] 106ns                                    |L0.106|                                                  "
    - "L0.107[106,107] 107ns                                     |L0.107|                                                 "
    - "L0.108[107,108] 108ns                                     |L0.108|                                                 "
    - "L0.109[108,109] 109ns                                     |L0.109|                                                 "
    - "L0.110[109,110] 110ns                                      |L0.110|                                                "
    - "L0.111[110,111] 111ns                                      |L0.111|                                                "
    - "L0.112[111,112] 112ns                                      |L0.112|                                                "
    - "L0.113[112,113] 113ns                                       |L0.113|                                               "
    - "L0.114[113,114] 114ns                                       |L0.114|                                               "
    - "L0.115[114,115] 115ns                                       |L0.115|                                               "
    - "L0.116[115,116] 116ns                                       |L0.116|                                               "
    - "L0.117[116,117] 117ns                                        |L0.117|                                              "
    - "L0.118[117,118] 118ns                                        |L0.118|                                              "
    - "L0.119[118,119] 119ns                                        |L0.119|                                              "
    - "L0.120[119,120] 120ns                                         |L0.120|                                             "
    - "L0.121[120,121] 121ns                                         |L0.121|                                             "
    - "L0.122[121,122] 122ns                                         |L0.122|                                             "
    - "L0.123[122,123] 123ns                                          |L0.123|                                            "
    - "L0.124[123,124] 124ns                                          |L0.124|                                            "
    - "L0.125[124,125] 125ns                                          |L0.125|                                            "
    - "L0.126[125,126] 126ns                                           |L0.126|                                           "
    - "L0.127[126,127] 127ns                                           |L0.127|                                           "
    - "L0.128[127,128] 128ns                                           |L0.128|                                           "
    - "L0.129[128,129] 129ns                                            |L0.129|                                          "
    - "L0.130[129,130] 130ns                                            |L0.130|                                          "
    - "L0.131[130,131] 131ns                                            |L0.131|                                          "
    - "L0.132[131,132] 132ns                                            |L0.132|                                          "
    - "L0.133[132,133] 133ns                                             |L0.133|                                         "
    - "L0.134[133,134] 134ns                                             |L0.134|                                         "
    - "L0.135[134,135] 135ns                                             |L0.135|                                         "
    - "L0.136[135,136] 136ns                                              |L0.136|                                        "
    - "L0.137[136,137] 137ns                                              |L0.137|                                        "
    - "L0.138[137,138] 138ns                                              |L0.138|                                        "
    - "L0.139[138,139] 139ns                                               |L0.139|                                       "
    - "L0.140[139,140] 140ns                                               |L0.140|                                       "
    - "L0.141[140,141] 141ns                                               |L0.141|                                       "
    - "L0.142[141,142] 142ns                                                |L0.142|                                      "
    - "L0.143[142,143] 143ns                                                |L0.143|                                      "
    - "L0.144[143,144] 144ns                                                |L0.144|                                      "
    - "L0.145[144,145] 145ns                                                 |L0.145|                                     "
    - "L0.146[145,146] 146ns                                                 |L0.146|                                     "
    - "L0.147[146,147] 147ns                                                 |L0.147|                                     "
    - "L0.148[147,148] 148ns                                                 |L0.148|                                     "
    - "L0.149[148,149] 149ns                                                  |L0.149|                                    "
    - "L0.150[149,150] 150ns                                                  |L0.150|                                    "
    - "L0.151[150,151] 151ns                                                  |L0.151|                                    "
    - "L0.152[151,152] 152ns                                                   |L0.152|                                   "
    - "L0.153[152,153] 153ns                                                   |L0.153|                                   "
    - "L0.154[153,154] 154ns                                                   |L0.154|                                   "
    - "L0.155[154,155] 155ns                                                    |L0.155|                                  "
    - "L0.156[155,156] 156ns                                                    |L0.156|                                  "
    - "L0.157[156,157] 157ns                                                    |L0.157|                                  "
    - "L0.158[157,158] 158ns                                                     |L0.158|                                 "
    - "L0.159[158,159] 159ns                                                     |L0.159|                                 "
    - "L0.160[159,160] 160ns                                                     |L0.160|                                 "
    - "L0.161[160,161] 161ns                                                      |L0.161|                                "
    - "L0.162[161,162] 162ns                                                      |L0.162|                                "
    - "L0.163[162,163] 163ns                                                      |L0.163|                                "
    - "L0.164[163,164] 164ns                                                      |L0.164|                                "
    - "L0.165[164,165] 165ns                                                       |L0.165|                               "
    - "L0.166[165,166] 166ns                                                       |L0.166|                               "
    - "L0.167[166,167] 167ns                                                       |L0.167|                               "
    - "L0.168[167,168] 168ns                                                        |L0.168|                              "
    - "L0.169[168,169] 169ns                                                        |L0.169|                              "
    - "L0.170[169,170] 170ns                                                        |L0.170|                              "
    - "L0.171[170,171] 171ns                                                         |L0.171|                             "
    - "L0.172[171,172] 172ns                                                         |L0.172|                             "
    - "L0.173[172,173] 173ns                                                         |L0.173|                             "
    - "L0.174[173,174] 174ns                                                          |L0.174|                            "
    - "L0.175[174,175] 175ns                                                          |L0.175|                            "
    - "L0.176[175,176] 176ns                                                          |L0.176|                            "
    - "L0.177[176,177] 177ns                                                           |L0.177|                           "
    - "L0.178[177,178] 178ns                                                           |L0.178|                           "
    - "L0.179[178,179] 179ns                                                           |L0.179|                           "
    - "L0.180[179,180] 180ns                                                           |L0.180|                           "
    - "L0.181[180,181] 181ns                                                            |L0.181|                          "
    - "L0.182[181,182] 182ns                                                            |L0.182|                          "
    - "L0.183[182,183] 183ns                                                            |L0.183|                          "
    - "L0.184[183,184] 184ns                                                             |L0.184|                         "
    - "L0.185[184,185] 185ns                                                             |L0.185|                         "
    - "L0.186[185,186] 186ns                                                             |L0.186|                         "
    - "L0.187[186,187] 187ns                                                              |L0.187|                        "
    - "L0.188[187,188] 188ns                                                              |L0.188|                        "
    - "L0.189[188,189] 189ns                                                              |L0.189|                        "
    - "L0.190[189,190] 190ns                                                               |L0.190|                       "
    - "L0.191[190,191] 191ns                                                               |L0.191|                       "
    - "L0.192[191,192] 192ns                                                               |L0.192|                       "
    - "L0.193[192,193] 193ns                                                                |L0.193|                      "
    - "L0.194[193,194] 194ns                                                                |L0.194|                      "
    - "L0.195[194,195] 195ns                                                                |L0.195|                      "
    - "L0.196[195,196] 196ns                                                                |L0.196|                      "
    - "L0.197[196,197] 197ns                                                                 |L0.197|                     "
    - "L0.198[197,198] 198ns                                                                 |L0.198|                     "
    - "L0.199[198,199] 199ns                                                                 |L0.199|                     "
    - "L0.200[199,200] 200ns                                                                  |L0.200|                    "
    - "L0.201[200,201] 201ns                                                                  |L0.201|                    "
    - "L0.202[201,202] 202ns                                                                  |L0.202|                    "
    - "L0.203[202,203] 203ns                                                                   |L0.203|                   "
    - "L0.204[203,204] 204ns                                                                   |L0.204|                   "
    - "L0.205[204,205] 205ns                                                                   |L0.205|                   "
    - "L0.206[205,206] 206ns                                                                    |L0.206|                  "
    - "L0.207[206,207] 207ns                                                                    |L0.207|                  "
    - "L0.208[207,208] 208ns                                                                    |L0.208|                  "
    - "L0.209[208,209] 209ns                                                                     |L0.209|                 "
    - "L0.210[209,210] 210ns                                                                     |L0.210|                 "
    - "L0.211[210,211] 211ns                                                                     |L0.211|                 "
    - "L0.212[211,212] 212ns                                                                     |L0.212|                 "
    - "L0.213[212,213] 213ns                                                                      |L0.213|                "
    - "L0.214[213,214] 214ns                                                                      |L0.214|                "
    - "L0.215[214,215] 215ns                                                                      |L0.215|                "
    - "L0.216[215,216] 216ns                                                                       |L0.216|               "
    - "L0.217[216,217] 217ns                                                                       |L0.217|               "
    - "L0.218[217,218] 218ns                                                                       |L0.218|               "
    - "L0.219[218,219] 219ns                                                                        |L0.219|              "
    - "L0.220[219,220] 220ns                                                                        |L0.220|              "
    - "L0.221[220,221] 221ns                                                                        |L0.221|              "
    - "L0.222[221,222] 222ns                                                                         |L0.222|             "
    - "L0.223[222,223] 223ns                                                                         |L0.223|             "
    - "L0.224[223,224] 224ns                                                                         |L0.224|             "
    - "L0.225[224,225] 225ns                                                                          |L0.225|            "
    - "L0.226[225,226] 226ns                                                                          |L0.226|            "
    - "L0.227[226,227] 227ns                                                                          |L0.227|            "
    - "L0.228[227,228] 228ns                                                                          |L0.228|            "
    - "L0.229[228,229] 229ns                                                                           |L0.229|           "
    - "L0.230[229,230] 230ns                                                                           |L0.230|           "
    - "L0.231[230,231] 231ns                                                                           |L0.231|           "
    - "L0.232[231,232] 232ns                                                                            |L0.232|          "
    - "L0.233[232,233] 233ns                                                                            |L0.233|          "
    - "L0.234[233,234] 234ns                                                                            |L0.234|          "
    - "L0.235[234,235] 235ns                                                                             |L0.235|         "
    - "L0.236[235,236] 236ns                                                                             |L0.236|         "
    - "L0.237[236,237] 237ns                                                                             |L0.237|         "
    - "L0.238[237,238] 238ns                                                                              |L0.238|        "
    - "L0.239[238,239] 239ns                                                                              |L0.239|        "
    - "L0.240[239,240] 240ns                                                                              |L0.240|        "
    - "L0.241[240,241] 241ns                                                                               |L0.241|       "
    - "L0.242[241,242] 242ns                                                                               |L0.242|       "
    - "L0.243[242,243] 243ns                                                                               |L0.243|       "
    - "L0.244[243,244] 244ns                                                                               |L0.244|       "
    - "L0.245[244,245] 245ns                                                                                |L0.245|      "
    - "L0.246[245,246] 246ns                                                                                |L0.246|      "
    - "L0.247[246,247] 247ns                                                                                |L0.247|      "
    - "L0.248[247,248] 248ns                                                                                 |L0.248|     "
    - "L0.249[248,249] 249ns                                                                                 |L0.249|     "
    - "L0.250[249,250] 250ns                                                                                 |L0.250|     "
    - "L0.251[250,251] 251ns                                                                                  |L0.251|    "
    - "L0.252[251,252] 252ns                                                                                  |L0.252|    "
    - "L0.253[252,253] 253ns                                                                                  |L0.253|    "
    - "L0.254[253,254] 254ns                                                                                   |L0.254|   "
    - "L0.255[254,255] 255ns                                                                                   |L0.255|   "
    - "L0.256[255,256] 256ns                                                                                   |L0.256|   "
    - "L0.257[256,257] 257ns                                                                                    |L0.257|  "
    - "L0.258[257,258] 258ns                                                                                    |L0.258|  "
    - "L0.259[258,259] 259ns                                                                                    |L0.259|  "
    - "L0.260[259,260] 260ns                                                                                    |L0.260|  "
    - "L0.261[260,261] 261ns                                                                                     |L0.261| "
    - "L0.262[261,262] 262ns                                                                                     |L0.262| "
    - "L0.263[262,263] 263ns                                                                                     |L0.263| "
    - "L0.264[263,264] 264ns                                                                                      |L0.264|"
    - "L0.265[264,265] 265ns                                                                                      |L0.265|"
    - "L0.266[265,266] 266ns                                                                                      |L0.266|"
    - "L0.267[266,267] 267ns                                                                                       |L0.267|"
    - "L0.268[267,268] 268ns                                                                                       |L0.268|"
    - "L0.269[268,269] 269ns                                                                                       |L0.269|"
    - "L0.270[269,270] 270ns                                                                                        |L0.270|"
    - "L0.271[270,271] 271ns                                                                                        |L0.271|"
    - "L0.272[271,272] 272ns                                                                                        |L0.272|"
    - "L0.273[272,273] 273ns                                                                                         |L0.273|"
    - "L0.274[273,274] 274ns                                                                                         |L0.274|"
    - "L0.275[274,275] 275ns                                                                                         |L0.275|"
    - "L0.276[275,276] 276ns                                                                                         |L0.276|"
    - "L0.277[276,277] 277ns                                                                                          |L0.277|"
    - "L0.278[277,278] 278ns                                                                                          |L0.278|"
    - "L0.279[278,279] 279ns                                                                                          |L0.279|"
    - "L0.280[279,280] 280ns                                                                                           |L0.280|"
    - "L0.281[280,281] 281ns                                                                                           |L0.281|"
    - "L0.282[281,282] 282ns                                                                                           |L0.282|"
    - "L0.283[282,283] 283ns                                                                                            |L0.283|"
    - "L0.284[283,284] 284ns                                                                                            |L0.284|"
    - "L0.285[284,285] 285ns                                                                                            |L0.285|"
    - "L0.286[285,286] 286ns                                                                                             |L0.286|"
    - "L0.287[286,287] 287ns                                                                                             |L0.287|"
    - "L0.288[287,288] 288ns                                                                                             |L0.288|"
    - "**** Simulation run 0, type=compact(ManySmallFiles). 200 Input Files, 1.37mb total:"
    - "L0, all files 7kb                                                                                                  "
    - "L0.1[0,1] 1ns            |L0.1|                                                                                    "
    - "L0.2[1,2] 2ns            |L0.2|                                                                                    "
    - "L0.3[2,3] 3ns            |L0.3|                                                                                    "
    - "L0.4[3,4] 4ns             |L0.4|                                                                                   "
    - "L0.5[4,5] 5ns             |L0.5|                                                                                   "
    - "L0.6[5,6] 6ns              |L0.6|                                                                                  "
    - "L0.7[6,7] 7ns              |L0.7|                                                                                  "
    - "L0.8[7,8] 8ns               |L0.8|                                                                                 "
    - "L0.9[8,9] 9ns               |L0.9|                                                                                 "
    - "L0.10[9,10] 10ns             |L0.10|                                                                               "
    - "L0.11[10,11] 11ns            |L0.11|                                                                               "
    - "L0.12[11,12] 12ns            |L0.12|                                                                               "
    - "L0.13[12,13] 13ns             |L0.13|                                                                              "
    - "L0.14[13,14] 14ns             |L0.14|                                                                              "
    - "L0.15[14,15] 15ns              |L0.15|                                                                             "
    - "L0.16[15,16] 16ns              |L0.16|                                                                             "
    - "L0.17[16,17] 17ns               |L0.17|                                                                            "
    - "L0.18[17,18] 18ns               |L0.18|                                                                            "
    - "L0.19[18,19] 19ns                |L0.19|                                                                           "
    - "L0.20[19,20] 20ns                |L0.20|                                                                           "
    - "L0.21[20,21] 21ns                 |L0.21|                                                                          "
    - "L0.22[21,22] 22ns                 |L0.22|                                                                          "
    - "L0.23[22,23] 23ns                 |L0.23|                                                                          "
    - "L0.24[23,24] 24ns                  |L0.24|                                                                         "
    - "L0.25[24,25] 25ns                  |L0.25|                                                                         "
    - "L0.26[25,26] 26ns                   |L0.26|                                                                        "
    - "L0.27[26,27] 27ns                   |L0.27|                                                                        "
    - "L0.28[27,28] 28ns                    |L0.28|                                                                       "
    - "L0.29[28,29] 29ns                    |L0.29|                                                                       "
    - "L0.30[29,30] 30ns                     |L0.30|                                                                      "
    - "L0.31[30,31] 31ns                     |L0.31|                                                                      "
    - "L0.32[31,32] 32ns                     |L0.32|                                                                      "
    - "L0.33[32,33] 33ns                      |L0.33|                                                                     "
    - "L0.34[33,34] 34ns                      |L0.34|                                                                     "
    - "L0.35[34,35] 35ns                       |L0.35|                                                                    "
    - "L0.36[35,36] 36ns                       |L0.36|                                                                    "
    - "L0.37[36,37] 37ns                        |L0.37|                                                                   "
    - "L0.38[37,38] 38ns                        |L0.38|                                                                   "
    - "L0.39[38,39] 39ns                         |L0.39|                                                                  "
    - "L0.40[39,40] 40ns                         |L0.40|                                                                  "
    - "L0.41[40,41] 41ns                          |L0.41|                                                                 "
    - "L0.42[41,42] 42ns                          |L0.42|                                                                 "
    - "L0.43[42,43] 43ns                          |L0.43|                                                                 "
    - "L0.44[43,44] 44ns                           |L0.44|                                                                "
    - "L0.45[44,45] 45ns                           |L0.45|                                                                "
    - "L0.46[45,46] 46ns                            |L0.46|                                                               "
    - "L0.47[46,47] 47ns                            |L0.47|                                                               "
    - "L0.48[47,48] 48ns                             |L0.48|                                                              "
    - "L0.49[48,49] 49ns                             |L0.49|                                                              "
    - "L0.50[49,50] 50ns                              |L0.50|                                                             "
    - "L0.51[50,51] 51ns                              |L0.51|                                                             "
    - "L0.52[51,52] 52ns                              |L0.52|                                                             "
    - "L0.53[52,53] 53ns                               |L0.53|                                                            "
    - "L0.54[53,54] 54ns                               |L0.54|                                                            "
    - "L0.55[54,55] 55ns                                |L0.55|                                                           "
    - "L0.56[55,56] 56ns                                |L0.56|                                                           "
    - "L0.57[56,57] 57ns                                 |L0.57|                                                          "
    - "L0.58[57,58] 58ns                                 |L0.58|                                                          "
    - "L0.59[58,59] 59ns                                  |L0.59|                                                         "
    - "L0.60[59,60] 60ns                                  |L0.60|                                                         "
    - "L0.61[60,61] 61ns                                   |L0.61|                                                        "
    - "L0.62[61,62] 62ns                                   |L0.62|                                                        "
    - "L0.63[62,63] 63ns                                   |L0.63|                                                        "
    - "L0.64[63,64] 64ns                                    |L0.64|                                                       "
    - "L0.65[64,65] 65ns                                    |L0.65|                                                       "
    - "L0.66[65,66] 66ns                                     |L0.66|                                                      "
    - "L0.67[66,67] 67ns                                     |L0.67|                                                      "
    - "L0.68[67,68] 68ns                                      |L0.68|                                                     "
    - "L0.69[68,69] 69ns                                      |L0.69|                                                     "
    - "L0.70[69,70] 70ns                                       |L0.70|                                                    "
    - "L0.71[70,71] 71ns                                       |L0.71|                                                    "
    - "L0.72[71,72] 72ns                                       |L0.72|                                                    "
    - "L0.73[72,73] 73ns                                        |L0.73|                                                   "
    - "L0.74[73,74] 74ns                                        |L0.74|                                                   "
    - "L0.75[74,75] 75ns                                         |L0.75|                                                  "
    - "L0.76[75,76] 76ns                                         |L0.76|                                                  "
    - "L0.77[76,77] 77ns                                          |L0.77|                                                 "
    - "L0.78[77,78] 78ns                                          |L0.78|                                                 "
    - "L0.79[78,79] 79ns                                           |L0.79|                                                "
    - "L0.80[79,80] 80ns                                           |L0.80|                                                "
    - "L0.81[80,81] 81ns                                            |L0.81|                                               "
    - "L0.82[81,82] 82ns                                            |L0.82|                                               "
    - "L0.83[82,83] 83ns                                            |L0.83|                                               "
    - "L0.84[83,84] 84ns                                             |L0.84|                                              "
    - "L0.85[84,85] 85ns                                             |L0.85|                                              "
    - "L0.86[85,86] 86ns                                              |L0.86|                                             "
    - "L0.87[86,87] 87ns                                              |L0.87|                                             "
    - "L0.88[87,88] 88ns                                               |L0.88|                                            "
    - "L0.89[88,89] 89ns                                               |L0.89|                                            "
    - "L0.90[89,90] 90ns                                                |L0.90|                                           "
    - "L0.91[90,91] 91ns                                                |L0.91|                                           "
    - "L0.92[91,92] 92ns                                                |L0.92|                                           "
    - "L0.93[92,93] 93ns                                                 |L0.93|                                          "
    - "L0.94[93,94] 94ns                                                 |L0.94|                                          "
    - "L0.95[94,95] 95ns                                                  |L0.95|                                         "
    - "L0.96[95,96] 96ns                                                  |L0.96|                                         "
    - "L0.97[96,97] 97ns                                                   |L0.97|                                        "
    - "L0.98[97,98] 98ns                                                   |L0.98|                                        "
    - "L0.99[98,99] 99ns                                                    |L0.99|                                       "
    - "L0.100[99,100] 100ns                                                 |L0.100|                                      "
    - "L0.101[100,101] 101ns                                                 |L0.101|                                     "
    - "L0.102[101,102] 102ns                                                 |L0.102|                                     "
    - "L0.103[102,103] 103ns                                                 |L0.103|                                     "
    - "L0.104[103,104] 104ns                                                  |L0.104|                                    "
    - "L0.105[104,105] 105ns                                                  |L0.105|                                    "
    - "L0.106[105,106] 106ns                                                   |L0.106|                                   "
    - "L0.107[106,107] 107ns                                                   |L0.107|                                   "
    - "L0.108[107,108] 108ns                                                    |L0.108|                                  "
    - "L0.109[108,109] 109ns                                                    |L0.109|                                  "
    - "L0.110[109,110] 110ns                                                     |L0.110|                                 "
    - "L0.111[110,111] 111ns                                                     |L0.111|                                 "
    - "L0.112[111,112] 112ns                                                     |L0.112|                                 "
    - "L0.113[112,113] 113ns                                                      |L0.113|                                "
    - "L0.114[113,114] 114ns                                                      |L0.114|                                "
    - "L0.115[114,115] 115ns                                                       |L0.115|                               "
    - "L0.116[115,116] 116ns                                                       |L0.116|                               "
    - "L0.117[116,117] 117ns                                                        |L0.117|                              "
    - "L0.118[117,118] 118ns                                                        |L0.118|                              "
    - "L0.119[118,119] 119ns                                                         |L0.119|                             "
    - "L0.120[119,120] 120ns                                                         |L0.120|                             "
    - "L0.121[120,121] 121ns                                                          |L0.121|                            "
    - "L0.122[121,122] 122ns                                                          |L0.122|                            "
    - "L0.123[122,123] 123ns                                                          |L0.123|                            "
    - "L0.124[123,124] 124ns                                                           |L0.124|                           "
    - "L0.125[124,125] 125ns                                                           |L0.125|                           "
    - "L0.126[125,126] 126ns                                                            |L0.126|                          "
    - "L0.127[126,127] 127ns                                                            |L0.127|                          "
    - "L0.128[127,128] 128ns                                                             |L0.128|                         "
    - "L0.129[128,129] 129ns                                                             |L0.129|                         "
    - "L0.130[129,130] 130ns                                                              |L0.130|                        "
    - "L0.131[130,131] 131ns                                                              |L0.131|                        "
    - "L0.132[131,132] 132ns                                                              |L0.132|                        "
    - "L0.133[132,133] 133ns                                                               |L0.133|                       "
    - "L0.134[133,134] 134ns                                                               |L0.134|                       "
    - "L0.135[134,135] 135ns                                                                |L0.135|                      "
    - "L0.136[135,136] 136ns                                                                |L0.136|                      "
    - "L0.137[136,137] 137ns                                                                 |L0.137|                     "
    - "L0.138[137,138] 138ns                                                                 |L0.138|                     "
    - "L0.139[138,139] 139ns                                                                  |L0.139|                    "
    - "L0.140[139,140] 140ns                                                                  |L0.140|                    "
    - "L0.141[140,141] 141ns                                                                   |L0.141|                   "
    - "L0.142[141,142] 142ns                                                                   |L0.142|                   "
    - "L0.143[142,143] 143ns                                                                   |L0.143|                   "
    - "L0.144[143,144] 144ns                                                                    |L0.144|                  "
    - "L0.145[144,145] 145ns                                                                    |L0.145|                  "
    - "L0.146[145,146] 146ns                                                                     |L0.146|                 "
    - "L0.147[146,147] 147ns                                                                     |L0.147|                 "
    - "L0.148[147,148] 148ns                                                                      |L0.148|                "
    - "L0.149[148,149] 149ns                                                                      |L0.149|                "
    - "L0.150[149,150] 150ns                                                                       |L0.150|               "
    - "L0.151[150,151] 151ns                                                                       |L0.151|               "
    - "L0.152[151,152] 152ns                                                                       |L0.152|               "
    - "L0.153[152,153] 153ns                                                                        |L0.153|              "
    - "L0.154[153,154] 154ns                                                                        |L0.154|              "
    - "L0.155[154,155] 155ns                                                                         |L0.155|             "
    - "L0.156[155,156] 156ns                                                                         |L0.156|             "
    - "L0.157[156,157] 157ns                                                                          |L0.157|            "
    - "L0.158[157,158] 158ns                                                                          |L0.158|            "
    - "L0.159[158,159] 159ns                                                                           |L0.159|           "
    - "L0.160[159,160] 160ns                                                                           |L0.160|           "
    - "L0.161[160,161] 161ns                                                                            |L0.161|          "
    - "L0.162[161,162] 162ns                                                                            |L0.162|          "
    - "L0.163[162,163] 163ns                                                                            |L0.163|          "
    - "L0.164[163,164] 164ns                                                                             |L0.164|         "
    - "L0.165[164,165] 165ns                                                                             |L0.165|         "
    - "L0.166[165,166] 166ns                                                                              |L0.166|        "
    - "L0.167[166,167] 167ns                                                                              |L0.167|        "
    - "L0.168[167,168] 168ns                                                                               |L0.168|       "
    - "L0.169[168,169] 169ns                                                                               |L0.169|       "
    - "L0.170[169,170] 170ns                                                                                |L0.170|      "
    - "L0.171[170,171] 171ns                                                                                |L0.171|      "
    - "L0.172[171,172] 172ns                                                                                |L0.172|      "
    - "L0.173[172,173] 173ns                                                                                 |L0.173|     "
    - "L0.174[173,174] 174ns                                                                                 |L0.174|     "
    - "L0.175[174,175] 175ns                                                                                  |L0.175|    "
    - "L0.176[175,176] 176ns                                                                                  |L0.176|    "
    - "L0.177[176,177] 177ns                                                                                   |L0.177|   "
    - "L0.178[177,178] 178ns                                                                                   |L0.178|   "
    - "L0.179[178,179] 179ns                                                                                    |L0.179|  "
    - "L0.180[179,180] 180ns                                                                                    |L0.180|  "
    - "L0.181[180,181] 181ns                                                                                     |L0.181| "
    - "L0.182[181,182] 182ns                                                                                     |L0.182| "
    - "L0.183[182,183] 183ns                                                                                     |L0.183| "
    - "L0.184[183,184] 184ns                                                                                      |L0.184|"
    - "L0.185[184,185] 185ns                                                                                      |L0.185|"
    - "L0.186[185,186] 186ns                                                                                       |L0.186|"
    - "L0.187[186,187] 187ns                                                                                       |L0.187|"
    - "L0.188[187,188] 188ns                                                                                        |L0.188|"
    - "L0.189[188,189] 189ns                                                                                        |L0.189|"
    - "L0.190[189,190] 190ns                                                                                         |L0.190|"
    - "L0.191[190,191] 191ns                                                                                         |L0.191|"
    - "L0.192[191,192] 192ns                                                                                         |L0.192|"
    - "L0.193[192,193] 193ns                                                                                          |L0.193|"
    - "L0.194[193,194] 194ns                                                                                          |L0.194|"
    - "L0.195[194,195] 195ns                                                                                           |L0.195|"
    - "L0.196[195,196] 196ns                                                                                           |L0.196|"
    - "L0.197[196,197] 197ns                                                                                            |L0.197|"
    - "L0.198[197,198] 198ns                                                                                            |L0.198|"
    - "L0.199[198,199] 199ns                                                                                             |L0.199|"
    - "L0.200[199,200] 200ns                                                                                             |L0.200|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 1.37mb total:"
    - "L0, all files 1.37mb                                                                                               "
    - "L0.?[0,200] 200ns        |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 200 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10, L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33, L0.34, L0.35, L0.36, L0.37, L0.38, L0.39, L0.40, L0.41, L0.42, L0.43, L0.44, L0.45, L0.46, L0.47, L0.48, L0.49, L0.50, L0.51, L0.52, L0.53, L0.54, L0.55, L0.56, L0.57, L0.58, L0.59, L0.60, L0.61, L0.62, L0.63, L0.64, L0.65, L0.66, L0.67, L0.68, L0.69, L0.70, L0.71, L0.72, L0.73, L0.74, L0.75, L0.76, L0.77, L0.78, L0.79, L0.80, L0.81, L0.82, L0.83, L0.84, L0.85, L0.86, L0.87, L0.88, L0.89, L0.90, L0.91, L0.92, L0.93, L0.94, L0.95, L0.96, L0.97, L0.98, L0.99, L0.100, L0.101, L0.102, L0.103, L0.104, L0.105, L0.106, L0.107, L0.108, L0.109, L0.110, L0.111, L0.112, L0.113, L0.114, L0.115, L0.116, L0.117, L0.118, L0.119, L0.120, L0.121, L0.122, L0.123, L0.124, L0.125, L0.126, L0.127, L0.128, L0.129, L0.130, L0.131, L0.132, L0.133, L0.134, L0.135, L0.136, L0.137, L0.138, L0.139, L0.140, L0.141, L0.142, L0.143, L0.144, L0.145, L0.146, L0.147, L0.148, L0.149, L0.150, L0.151, L0.152, L0.153, L0.154, L0.155, L0.156, L0.157, L0.158, L0.159, L0.160, L0.161, L0.162, L0.163, L0.164, L0.165, L0.166, L0.167, L0.168, L0.169, L0.170, L0.171, L0.172, L0.173, L0.174, L0.175, L0.176, L0.177, L0.178, L0.179, L0.180, L0.181, L0.182, L0.183, L0.184, L0.185, L0.186, L0.187, L0.188, L0.189, L0.190, L0.191, L0.192, L0.193, L0.194, L0.195, L0.196, L0.197, L0.198, L0.199, L0.200"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=compact(ManySmallFiles). 88 Input Files, 616kb total:"
    - "L0, all files 7kb                                                                                                  "
    - "L0.201[200,201] 201ns    |L0.201|                                                                                  "
    - "L0.202[201,202] 202ns     |L0.202|                                                                                 "
    - "L0.203[202,203] 203ns      |L0.203|                                                                                "
    - "L0.204[203,204] 204ns       |L0.204|                                                                               "
    - "L0.205[204,205] 205ns        |L0.205|                                                                              "
    - "L0.206[205,206] 206ns         |L0.206|                                                                             "
    - "L0.207[206,207] 207ns          |L0.207|                                                                            "
    - "L0.208[207,208] 208ns           |L0.208|                                                                           "
    - "L0.209[208,209] 209ns            |L0.209|                                                                          "
    - "L0.210[209,210] 210ns             |L0.210|                                                                         "
    - "L0.211[210,211] 211ns              |L0.211|                                                                        "
    - "L0.212[211,212] 212ns               |L0.212|                                                                       "
    - "L0.213[212,213] 213ns                |L0.213|                                                                      "
    - "L0.214[213,214] 214ns                 |L0.214|                                                                     "
    - "L0.215[214,215] 215ns                  |L0.215|                                                                    "
    - "L0.216[215,216] 216ns                   |L0.216|                                                                   "
    - "L0.217[216,217] 217ns                    |L0.217|                                                                  "
    - "L0.218[217,218] 218ns                     |L0.218|                                                                 "
    - "L0.219[218,219] 219ns                      |L0.219|                                                                "
    - "L0.220[219,220] 220ns                       |L0.220|                                                               "
    - "L0.221[220,221] 221ns                        |L0.221|                                                              "
    - "L0.222[221,222] 222ns                         |L0.222|                                                             "
    - "L0.223[222,223] 223ns                          |L0.223|                                                            "
    - "L0.224[223,224] 224ns                           |L0.224|                                                           "
    - "L0.225[224,225] 225ns                            |L0.225|                                                          "
    - "L0.226[225,226] 226ns                             |L0.226|                                                         "
    - "L0.227[226,227] 227ns                              |L0.227|                                                        "
    - "L0.228[227,228] 228ns                               |L0.228|                                                       "
    - "L0.229[228,229] 229ns                                |L0.229|                                                      "
    - "L0.230[229,230] 230ns                                 |L0.230|                                                     "
    - "L0.231[230,231] 231ns                                  |L0.231|                                                    "
    - "L0.232[231,232] 232ns                                   |L0.232|                                                   "
    - "L0.233[232,233] 233ns                                    |L0.233|                                                  "
    - "L0.234[233,234] 234ns                                     |L0.234|                                                 "
    - "L0.235[234,235] 235ns                                      |L0.235|                                                "
    - "L0.236[235,236] 236ns                                       |L0.236|                                               "
    - "L0.237[236,237] 237ns                                        |L0.237|                                              "
    - "L0.238[237,238] 238ns                                         |L0.238|                                             "
    - "L0.239[238,239] 239ns                                          |L0.239|                                            "
    - "L0.240[239,240] 240ns                                           |L0.240|                                           "
    - "L0.241[240,241] 241ns                                            |L0.241|                                          "
    - "L0.242[241,242] 242ns                                             |L0.242|                                         "
    - "L0.243[242,243] 243ns                                              |L0.243|                                        "
    - "L0.244[243,244] 244ns                                               |L0.244|                                       "
    - "L0.245[244,245] 245ns                                                 |L0.245|                                     "
    - "L0.246[245,246] 246ns                                                  |L0.246|                                    "
    - "L0.247[246,247] 247ns                                                   |L0.247|                                   "
    - "L0.248[247,248] 248ns                                                    |L0.248|                                  "
    - "L0.249[248,249] 249ns                                                     |L0.249|                                 "
    - "L0.250[249,250] 250ns                                                      |L0.250|                                "
    - "L0.251[250,251] 251ns                                                       |L0.251|                               "
    - "L0.252[251,252] 252ns                                                        |L0.252|                              "
    - "L0.253[252,253] 253ns                                                         |L0.253|                             "
    - "L0.254[253,254] 254ns                                                          |L0.254|                            "
    - "L0.255[254,255] 255ns                                                           |L0.255|                           "
    - "L0.256[255,256] 256ns                                                            |L0.256|                          "
    - "L0.257[256,257] 257ns                                                             |L0.257|                         "
    - "L0.258[257,258] 258ns                                                              |L0.258|                        "
    - "L0.259[258,259] 259ns                                                               |L0.259|                       "
    - "L0.260[259,260] 260ns                                                                |L0.260|                      "
    - "L0.261[260,261] 261ns                                                                 |L0.261|                     "
    - "L0.262[261,262] 262ns                                                                  |L0.262|                    "
    - "L0.263[262,263] 263ns                                                                   |L0.263|                   "
    - "L0.264[263,264] 264ns                                                                    |L0.264|                  "
    - "L0.265[264,265] 265ns                                                                     |L0.265|                 "
    - "L0.266[265,266] 266ns                                                                      |L0.266|                "
    - "L0.267[266,267] 267ns                                                                       |L0.267|               "
    - "L0.268[267,268] 268ns                                                                        |L0.268|              "
    - "L0.269[268,269] 269ns                                                                         |L0.269|             "
    - "L0.270[269,270] 270ns                                                                          |L0.270|            "
    - "L0.271[270,271] 271ns                                                                           |L0.271|           "
    - "L0.272[271,272] 272ns                                                                            |L0.272|          "
    - "L0.273[272,273] 273ns                                                                             |L0.273|         "
    - "L0.274[273,274] 274ns                                                                              |L0.274|        "
    - "L0.275[274,275] 275ns                                                                               |L0.275|       "
    - "L0.276[275,276] 276ns                                                                                |L0.276|      "
    - "L0.277[276,277] 277ns                                                                                 |L0.277|     "
    - "L0.278[277,278] 278ns                                                                                  |L0.278|    "
    - "L0.279[278,279] 279ns                                                                                   |L0.279|   "
    - "L0.280[279,280] 280ns                                                                                    |L0.280|  "
    - "L0.281[280,281] 281ns                                                                                     |L0.281| "
    - "L0.282[281,282] 282ns                                                                                      |L0.282|"
    - "L0.283[282,283] 283ns                                                                                       |L0.283|"
    - "L0.284[283,284] 284ns                                                                                        |L0.284|"
    - "L0.285[284,285] 285ns                                                                                         |L0.285|"
    - "L0.286[285,286] 286ns                                                                                          |L0.286|"
    - "L0.287[286,287] 287ns                                                                                           |L0.287|"
    - "L0.288[287,288] 288ns                                                                                            |L0.288|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 616kb total:"
    - "L0, all files 616kb                                                                                                "
    - "L0.?[200,288] 288ns      |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 88 files: L0.201, L0.202, L0.203, L0.204, L0.205, L0.206, L0.207, L0.208, L0.209, L0.210, L0.211, L0.212, L0.213, L0.214, L0.215, L0.216, L0.217, L0.218, L0.219, L0.220, L0.221, L0.222, L0.223, L0.224, L0.225, L0.226, L0.227, L0.228, L0.229, L0.230, L0.231, L0.232, L0.233, L0.234, L0.235, L0.236, L0.237, L0.238, L0.239, L0.240, L0.241, L0.242, L0.243, L0.244, L0.245, L0.246, L0.247, L0.248, L0.249, L0.250, L0.251, L0.252, L0.253, L0.254, L0.255, L0.256, L0.257, L0.258, L0.259, L0.260, L0.261, L0.262, L0.263, L0.264, L0.265, L0.266, L0.267, L0.268, L0.269, L0.270, L0.271, L0.272, L0.273, L0.274, L0.275, L0.276, L0.277, L0.278, L0.279, L0.280, L0.281, L0.282, L0.283, L0.284, L0.285, L0.286, L0.287, L0.288"
    - "  Creating 1 files"
    - "**** Simulation run 2, type=compact(TotalSizeLessThanMaxCompactSize). 2 Input Files, 1.97mb total:"
    - "L0                                                                                                                 "
    - "L0.290[200,288] 288ns 616kb                                                              |---------L0.290----------| "
    - "L0.289[0,200] 200ns 1.37mb|---------------------------L0.289---------------------------|                            "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 1.97mb total:"
    - "L1, all files 1.97mb                                                                                               "
    - "L1.?[0,288] 288ns        |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 2 files: L0.289, L0.290"
    - "  Creating 1 files"
    - "**** Final Output Files (3.94mb written)"
    - "L1, all files 1.97mb                                                                                               "
    - "L1.291[0,288] 288ns      |-----------------------------------------L1.291-----------------------------------------|"
    "###
    );
}

#[tokio::test]
async fn over_two_times_max_files_per_plan() {
    test_helpers::maybe_start_logging();

    // Observed size on production
    let seven_kb = 7 * 1024;

    let setup = layout_setup_builder().await.build().await;

    // Max over double the max files per plan
    // Expect to have 3 plans compacting L0s to L0, and then compact the resulted L0s  to L1
    let num_tiny_files = 410;

    for i in 0..num_tiny_files {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i)
                    .with_max_time(i + 1)
                    .with_compaction_level(CompactionLevel::Initial)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(seven_kb),
            )
            .await;
    }

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0, all files 7kb                                                                                                  "
    - "L0.1[0,1] 1ns            |L0.1|                                                                                    "
    - "L0.2[1,2] 2ns            |L0.2|                                                                                    "
    - "L0.3[2,3] 3ns            |L0.3|                                                                                    "
    - "L0.4[3,4] 4ns            |L0.4|                                                                                    "
    - "L0.5[4,5] 5ns            |L0.5|                                                                                    "
    - "L0.6[5,6] 6ns             |L0.6|                                                                                   "
    - "L0.7[6,7] 7ns             |L0.7|                                                                                   "
    - "L0.8[7,8] 8ns             |L0.8|                                                                                   "
    - "L0.9[8,9] 9ns             |L0.9|                                                                                   "
    - "L0.10[9,10] 10ns          |L0.10|                                                                                  "
    - "L0.11[10,11] 11ns          |L0.11|                                                                                 "
    - "L0.12[11,12] 12ns          |L0.12|                                                                                 "
    - "L0.13[12,13] 13ns          |L0.13|                                                                                 "
    - "L0.14[13,14] 14ns          |L0.14|                                                                                 "
    - "L0.15[14,15] 15ns           |L0.15|                                                                                "
    - "L0.16[15,16] 16ns           |L0.16|                                                                                "
    - "L0.17[16,17] 17ns           |L0.17|                                                                                "
    - "L0.18[17,18] 18ns           |L0.18|                                                                                "
    - "L0.19[18,19] 19ns           |L0.19|                                                                                "
    - "L0.20[19,20] 20ns            |L0.20|                                                                               "
    - "L0.21[20,21] 21ns            |L0.21|                                                                               "
    - "L0.22[21,22] 22ns            |L0.22|                                                                               "
    - "L0.23[22,23] 23ns            |L0.23|                                                                               "
    - "L0.24[23,24] 24ns             |L0.24|                                                                              "
    - "L0.25[24,25] 25ns             |L0.25|                                                                              "
    - "L0.26[25,26] 26ns             |L0.26|                                                                              "
    - "L0.27[26,27] 27ns             |L0.27|                                                                              "
    - "L0.28[27,28] 28ns             |L0.28|                                                                              "
    - "L0.29[28,29] 29ns              |L0.29|                                                                             "
    - "L0.30[29,30] 30ns              |L0.30|                                                                             "
    - "L0.31[30,31] 31ns              |L0.31|                                                                             "
    - "L0.32[31,32] 32ns              |L0.32|                                                                             "
    - "L0.33[32,33] 33ns               |L0.33|                                                                            "
    - "L0.34[33,34] 34ns               |L0.34|                                                                            "
    - "L0.35[34,35] 35ns               |L0.35|                                                                            "
    - "L0.36[35,36] 36ns               |L0.36|                                                                            "
    - "L0.37[36,37] 37ns               |L0.37|                                                                            "
    - "L0.38[37,38] 38ns                |L0.38|                                                                           "
    - "L0.39[38,39] 39ns                |L0.39|                                                                           "
    - "L0.40[39,40] 40ns                |L0.40|                                                                           "
    - "L0.41[40,41] 41ns                |L0.41|                                                                           "
    - "L0.42[41,42] 42ns                 |L0.42|                                                                          "
    - "L0.43[42,43] 43ns                 |L0.43|                                                                          "
    - "L0.44[43,44] 44ns                 |L0.44|                                                                          "
    - "L0.45[44,45] 45ns                 |L0.45|                                                                          "
    - "L0.46[45,46] 46ns                 |L0.46|                                                                          "
    - "L0.47[46,47] 47ns                  |L0.47|                                                                         "
    - "L0.48[47,48] 48ns                  |L0.48|                                                                         "
    - "L0.49[48,49] 49ns                  |L0.49|                                                                         "
    - "L0.50[49,50] 50ns                  |L0.50|                                                                         "
    - "L0.51[50,51] 51ns                  |L0.51|                                                                         "
    - "L0.52[51,52] 52ns                   |L0.52|                                                                        "
    - "L0.53[52,53] 53ns                   |L0.53|                                                                        "
    - "L0.54[53,54] 54ns                   |L0.54|                                                                        "
    - "L0.55[54,55] 55ns                   |L0.55|                                                                        "
    - "L0.56[55,56] 56ns                    |L0.56|                                                                       "
    - "L0.57[56,57] 57ns                    |L0.57|                                                                       "
    - "L0.58[57,58] 58ns                    |L0.58|                                                                       "
    - "L0.59[58,59] 59ns                    |L0.59|                                                                       "
    - "L0.60[59,60] 60ns                    |L0.60|                                                                       "
    - "L0.61[60,61] 61ns                     |L0.61|                                                                      "
    - "L0.62[61,62] 62ns                     |L0.62|                                                                      "
    - "L0.63[62,63] 63ns                     |L0.63|                                                                      "
    - "L0.64[63,64] 64ns                     |L0.64|                                                                      "
    - "L0.65[64,65] 65ns                      |L0.65|                                                                     "
    - "L0.66[65,66] 66ns                      |L0.66|                                                                     "
    - "L0.67[66,67] 67ns                      |L0.67|                                                                     "
    - "L0.68[67,68] 68ns                      |L0.68|                                                                     "
    - "L0.69[68,69] 69ns                      |L0.69|                                                                     "
    - "L0.70[69,70] 70ns                       |L0.70|                                                                    "
    - "L0.71[70,71] 71ns                       |L0.71|                                                                    "
    - "L0.72[71,72] 72ns                       |L0.72|                                                                    "
    - "L0.73[72,73] 73ns                       |L0.73|                                                                    "
    - "L0.74[73,74] 74ns                        |L0.74|                                                                   "
    - "L0.75[74,75] 75ns                        |L0.75|                                                                   "
    - "L0.76[75,76] 76ns                        |L0.76|                                                                   "
    - "L0.77[76,77] 77ns                        |L0.77|                                                                   "
    - "L0.78[77,78] 78ns                        |L0.78|                                                                   "
    - "L0.79[78,79] 79ns                         |L0.79|                                                                  "
    - "L0.80[79,80] 80ns                         |L0.80|                                                                  "
    - "L0.81[80,81] 81ns                         |L0.81|                                                                  "
    - "L0.82[81,82] 82ns                         |L0.82|                                                                  "
    - "L0.83[82,83] 83ns                          |L0.83|                                                                 "
    - "L0.84[83,84] 84ns                          |L0.84|                                                                 "
    - "L0.85[84,85] 85ns                          |L0.85|                                                                 "
    - "L0.86[85,86] 86ns                          |L0.86|                                                                 "
    - "L0.87[86,87] 87ns                          |L0.87|                                                                 "
    - "L0.88[87,88] 88ns                           |L0.88|                                                                "
    - "L0.89[88,89] 89ns                           |L0.89|                                                                "
    - "L0.90[89,90] 90ns                           |L0.90|                                                                "
    - "L0.91[90,91] 91ns                           |L0.91|                                                                "
    - "L0.92[91,92] 92ns                           |L0.92|                                                                "
    - "L0.93[92,93] 93ns                            |L0.93|                                                               "
    - "L0.94[93,94] 94ns                            |L0.94|                                                               "
    - "L0.95[94,95] 95ns                            |L0.95|                                                               "
    - "L0.96[95,96] 96ns                            |L0.96|                                                               "
    - "L0.97[96,97] 97ns                             |L0.97|                                                              "
    - "L0.98[97,98] 98ns                             |L0.98|                                                              "
    - "L0.99[98,99] 99ns                             |L0.99|                                                              "
    - "L0.100[99,100] 100ns                          |L0.100|                                                             "
    - "L0.101[100,101] 101ns                         |L0.101|                                                             "
    - "L0.102[101,102] 102ns                          |L0.102|                                                            "
    - "L0.103[102,103] 103ns                          |L0.103|                                                            "
    - "L0.104[103,104] 104ns                          |L0.104|                                                            "
    - "L0.105[104,105] 105ns                          |L0.105|                                                            "
    - "L0.106[105,106] 106ns                           |L0.106|                                                           "
    - "L0.107[106,107] 107ns                           |L0.107|                                                           "
    - "L0.108[107,108] 108ns                           |L0.108|                                                           "
    - "L0.109[108,109] 109ns                           |L0.109|                                                           "
    - "L0.110[109,110] 110ns                           |L0.110|                                                           "
    - "L0.111[110,111] 111ns                            |L0.111|                                                          "
    - "L0.112[111,112] 112ns                            |L0.112|                                                          "
    - "L0.113[112,113] 113ns                            |L0.113|                                                          "
    - "L0.114[113,114] 114ns                            |L0.114|                                                          "
    - "L0.115[114,115] 115ns                             |L0.115|                                                         "
    - "L0.116[115,116] 116ns                             |L0.116|                                                         "
    - "L0.117[116,117] 117ns                             |L0.117|                                                         "
    - "L0.118[117,118] 118ns                             |L0.118|                                                         "
    - "L0.119[118,119] 119ns                             |L0.119|                                                         "
    - "L0.120[119,120] 120ns                              |L0.120|                                                        "
    - "L0.121[120,121] 121ns                              |L0.121|                                                        "
    - "L0.122[121,122] 122ns                              |L0.122|                                                        "
    - "L0.123[122,123] 123ns                              |L0.123|                                                        "
    - "L0.124[123,124] 124ns                               |L0.124|                                                       "
    - "L0.125[124,125] 125ns                               |L0.125|                                                       "
    - "L0.126[125,126] 126ns                               |L0.126|                                                       "
    - "L0.127[126,127] 127ns                               |L0.127|                                                       "
    - "L0.128[127,128] 128ns                               |L0.128|                                                       "
    - "L0.129[128,129] 129ns                                |L0.129|                                                      "
    - "L0.130[129,130] 130ns                                |L0.130|                                                      "
    - "L0.131[130,131] 131ns                                |L0.131|                                                      "
    - "L0.132[131,132] 132ns                                |L0.132|                                                      "
    - "L0.133[132,133] 133ns                                |L0.133|                                                      "
    - "L0.134[133,134] 134ns                                 |L0.134|                                                     "
    - "L0.135[134,135] 135ns                                 |L0.135|                                                     "
    - "L0.136[135,136] 136ns                                 |L0.136|                                                     "
    - "L0.137[136,137] 137ns                                 |L0.137|                                                     "
    - "L0.138[137,138] 138ns                                  |L0.138|                                                    "
    - "L0.139[138,139] 139ns                                  |L0.139|                                                    "
    - "L0.140[139,140] 140ns                                  |L0.140|                                                    "
    - "L0.141[140,141] 141ns                                  |L0.141|                                                    "
    - "L0.142[141,142] 142ns                                  |L0.142|                                                    "
    - "L0.143[142,143] 143ns                                   |L0.143|                                                   "
    - "L0.144[143,144] 144ns                                   |L0.144|                                                   "
    - "L0.145[144,145] 145ns                                   |L0.145|                                                   "
    - "L0.146[145,146] 146ns                                   |L0.146|                                                   "
    - "L0.147[146,147] 147ns                                    |L0.147|                                                  "
    - "L0.148[147,148] 148ns                                    |L0.148|                                                  "
    - "L0.149[148,149] 149ns                                    |L0.149|                                                  "
    - "L0.150[149,150] 150ns                                    |L0.150|                                                  "
    - "L0.151[150,151] 151ns                                    |L0.151|                                                  "
    - "L0.152[151,152] 152ns                                     |L0.152|                                                 "
    - "L0.153[152,153] 153ns                                     |L0.153|                                                 "
    - "L0.154[153,154] 154ns                                     |L0.154|                                                 "
    - "L0.155[154,155] 155ns                                     |L0.155|                                                 "
    - "L0.156[155,156] 156ns                                      |L0.156|                                                "
    - "L0.157[156,157] 157ns                                      |L0.157|                                                "
    - "L0.158[157,158] 158ns                                      |L0.158|                                                "
    - "L0.159[158,159] 159ns                                      |L0.159|                                                "
    - "L0.160[159,160] 160ns                                      |L0.160|                                                "
    - "L0.161[160,161] 161ns                                       |L0.161|                                               "
    - "L0.162[161,162] 162ns                                       |L0.162|                                               "
    - "L0.163[162,163] 163ns                                       |L0.163|                                               "
    - "L0.164[163,164] 164ns                                       |L0.164|                                               "
    - "L0.165[164,165] 165ns                                        |L0.165|                                              "
    - "L0.166[165,166] 166ns                                        |L0.166|                                              "
    - "L0.167[166,167] 167ns                                        |L0.167|                                              "
    - "L0.168[167,168] 168ns                                        |L0.168|                                              "
    - "L0.169[168,169] 169ns                                        |L0.169|                                              "
    - "L0.170[169,170] 170ns                                         |L0.170|                                             "
    - "L0.171[170,171] 171ns                                         |L0.171|                                             "
    - "L0.172[171,172] 172ns                                         |L0.172|                                             "
    - "L0.173[172,173] 173ns                                         |L0.173|                                             "
    - "L0.174[173,174] 174ns                                         |L0.174|                                             "
    - "L0.175[174,175] 175ns                                          |L0.175|                                            "
    - "L0.176[175,176] 176ns                                          |L0.176|                                            "
    - "L0.177[176,177] 177ns                                          |L0.177|                                            "
    - "L0.178[177,178] 178ns                                          |L0.178|                                            "
    - "L0.179[178,179] 179ns                                           |L0.179|                                           "
    - "L0.180[179,180] 180ns                                           |L0.180|                                           "
    - "L0.181[180,181] 181ns                                           |L0.181|                                           "
    - "L0.182[181,182] 182ns                                           |L0.182|                                           "
    - "L0.183[182,183] 183ns                                           |L0.183|                                           "
    - "L0.184[183,184] 184ns                                            |L0.184|                                          "
    - "L0.185[184,185] 185ns                                            |L0.185|                                          "
    - "L0.186[185,186] 186ns                                            |L0.186|                                          "
    - "L0.187[186,187] 187ns                                            |L0.187|                                          "
    - "L0.188[187,188] 188ns                                             |L0.188|                                         "
    - "L0.189[188,189] 189ns                                             |L0.189|                                         "
    - "L0.190[189,190] 190ns                                             |L0.190|                                         "
    - "L0.191[190,191] 191ns                                             |L0.191|                                         "
    - "L0.192[191,192] 192ns                                             |L0.192|                                         "
    - "L0.193[192,193] 193ns                                              |L0.193|                                        "
    - "L0.194[193,194] 194ns                                              |L0.194|                                        "
    - "L0.195[194,195] 195ns                                              |L0.195|                                        "
    - "L0.196[195,196] 196ns                                              |L0.196|                                        "
    - "L0.197[196,197] 197ns                                               |L0.197|                                       "
    - "L0.198[197,198] 198ns                                               |L0.198|                                       "
    - "L0.199[198,199] 199ns                                               |L0.199|                                       "
    - "L0.200[199,200] 200ns                                               |L0.200|                                       "
    - "L0.201[200,201] 201ns                                               |L0.201|                                       "
    - "L0.202[201,202] 202ns                                                |L0.202|                                      "
    - "L0.203[202,203] 203ns                                                |L0.203|                                      "
    - "L0.204[203,204] 204ns                                                |L0.204|                                      "
    - "L0.205[204,205] 205ns                                                |L0.205|                                      "
    - "L0.206[205,206] 206ns                                                 |L0.206|                                     "
    - "L0.207[206,207] 207ns                                                 |L0.207|                                     "
    - "L0.208[207,208] 208ns                                                 |L0.208|                                     "
    - "L0.209[208,209] 209ns                                                 |L0.209|                                     "
    - "L0.210[209,210] 210ns                                                 |L0.210|                                     "
    - "L0.211[210,211] 211ns                                                  |L0.211|                                    "
    - "L0.212[211,212] 212ns                                                  |L0.212|                                    "
    - "L0.213[212,213] 213ns                                                  |L0.213|                                    "
    - "L0.214[213,214] 214ns                                                  |L0.214|                                    "
    - "L0.215[214,215] 215ns                                                  |L0.215|                                    "
    - "L0.216[215,216] 216ns                                                   |L0.216|                                   "
    - "L0.217[216,217] 217ns                                                   |L0.217|                                   "
    - "L0.218[217,218] 218ns                                                   |L0.218|                                   "
    - "L0.219[218,219] 219ns                                                   |L0.219|                                   "
    - "L0.220[219,220] 220ns                                                    |L0.220|                                  "
    - "L0.221[220,221] 221ns                                                    |L0.221|                                  "
    - "L0.222[221,222] 222ns                                                    |L0.222|                                  "
    - "L0.223[222,223] 223ns                                                    |L0.223|                                  "
    - "L0.224[223,224] 224ns                                                    |L0.224|                                  "
    - "L0.225[224,225] 225ns                                                     |L0.225|                                 "
    - "L0.226[225,226] 226ns                                                     |L0.226|                                 "
    - "L0.227[226,227] 227ns                                                     |L0.227|                                 "
    - "L0.228[227,228] 228ns                                                     |L0.228|                                 "
    - "L0.229[228,229] 229ns                                                      |L0.229|                                "
    - "L0.230[229,230] 230ns                                                      |L0.230|                                "
    - "L0.231[230,231] 231ns                                                      |L0.231|                                "
    - "L0.232[231,232] 232ns                                                      |L0.232|                                "
    - "L0.233[232,233] 233ns                                                      |L0.233|                                "
    - "L0.234[233,234] 234ns                                                       |L0.234|                               "
    - "L0.235[234,235] 235ns                                                       |L0.235|                               "
    - "L0.236[235,236] 236ns                                                       |L0.236|                               "
    - "L0.237[236,237] 237ns                                                       |L0.237|                               "
    - "L0.238[237,238] 238ns                                                        |L0.238|                              "
    - "L0.239[238,239] 239ns                                                        |L0.239|                              "
    - "L0.240[239,240] 240ns                                                        |L0.240|                              "
    - "L0.241[240,241] 241ns                                                        |L0.241|                              "
    - "L0.242[241,242] 242ns                                                        |L0.242|                              "
    - "L0.243[242,243] 243ns                                                         |L0.243|                             "
    - "L0.244[243,244] 244ns                                                         |L0.244|                             "
    - "L0.245[244,245] 245ns                                                         |L0.245|                             "
    - "L0.246[245,246] 246ns                                                         |L0.246|                             "
    - "L0.247[246,247] 247ns                                                          |L0.247|                            "
    - "L0.248[247,248] 248ns                                                          |L0.248|                            "
    - "L0.249[248,249] 249ns                                                          |L0.249|                            "
    - "L0.250[249,250] 250ns                                                          |L0.250|                            "
    - "L0.251[250,251] 251ns                                                          |L0.251|                            "
    - "L0.252[251,252] 252ns                                                           |L0.252|                           "
    - "L0.253[252,253] 253ns                                                           |L0.253|                           "
    - "L0.254[253,254] 254ns                                                           |L0.254|                           "
    - "L0.255[254,255] 255ns                                                           |L0.255|                           "
    - "L0.256[255,256] 256ns                                                           |L0.256|                           "
    - "L0.257[256,257] 257ns                                                            |L0.257|                          "
    - "L0.258[257,258] 258ns                                                            |L0.258|                          "
    - "L0.259[258,259] 259ns                                                            |L0.259|                          "
    - "L0.260[259,260] 260ns                                                            |L0.260|                          "
    - "L0.261[260,261] 261ns                                                             |L0.261|                         "
    - "L0.262[261,262] 262ns                                                             |L0.262|                         "
    - "L0.263[262,263] 263ns                                                             |L0.263|                         "
    - "L0.264[263,264] 264ns                                                             |L0.264|                         "
    - "L0.265[264,265] 265ns                                                             |L0.265|                         "
    - "L0.266[265,266] 266ns                                                              |L0.266|                        "
    - "L0.267[266,267] 267ns                                                              |L0.267|                        "
    - "L0.268[267,268] 268ns                                                              |L0.268|                        "
    - "L0.269[268,269] 269ns                                                              |L0.269|                        "
    - "L0.270[269,270] 270ns                                                               |L0.270|                       "
    - "L0.271[270,271] 271ns                                                               |L0.271|                       "
    - "L0.272[271,272] 272ns                                                               |L0.272|                       "
    - "L0.273[272,273] 273ns                                                               |L0.273|                       "
    - "L0.274[273,274] 274ns                                                               |L0.274|                       "
    - "L0.275[274,275] 275ns                                                                |L0.275|                      "
    - "L0.276[275,276] 276ns                                                                |L0.276|                      "
    - "L0.277[276,277] 277ns                                                                |L0.277|                      "
    - "L0.278[277,278] 278ns                                                                |L0.278|                      "
    - "L0.279[278,279] 279ns                                                                 |L0.279|                     "
    - "L0.280[279,280] 280ns                                                                 |L0.280|                     "
    - "L0.281[280,281] 281ns                                                                 |L0.281|                     "
    - "L0.282[281,282] 282ns                                                                 |L0.282|                     "
    - "L0.283[282,283] 283ns                                                                 |L0.283|                     "
    - "L0.284[283,284] 284ns                                                                  |L0.284|                    "
    - "L0.285[284,285] 285ns                                                                  |L0.285|                    "
    - "L0.286[285,286] 286ns                                                                  |L0.286|                    "
    - "L0.287[286,287] 287ns                                                                  |L0.287|                    "
    - "L0.288[287,288] 288ns                                                                   |L0.288|                   "
    - "L0.289[288,289] 289ns                                                                   |L0.289|                   "
    - "L0.290[289,290] 290ns                                                                   |L0.290|                   "
    - "L0.291[290,291] 291ns                                                                   |L0.291|                   "
    - "L0.292[291,292] 292ns                                                                   |L0.292|                   "
    - "L0.293[292,293] 293ns                                                                    |L0.293|                  "
    - "L0.294[293,294] 294ns                                                                    |L0.294|                  "
    - "L0.295[294,295] 295ns                                                                    |L0.295|                  "
    - "L0.296[295,296] 296ns                                                                    |L0.296|                  "
    - "L0.297[296,297] 297ns                                                                    |L0.297|                  "
    - "L0.298[297,298] 298ns                                                                     |L0.298|                 "
    - "L0.299[298,299] 299ns                                                                     |L0.299|                 "
    - "L0.300[299,300] 300ns                                                                     |L0.300|                 "
    - "L0.301[300,301] 301ns                                                                     |L0.301|                 "
    - "L0.302[301,302] 302ns                                                                      |L0.302|                "
    - "L0.303[302,303] 303ns                                                                      |L0.303|                "
    - "L0.304[303,304] 304ns                                                                      |L0.304|                "
    - "L0.305[304,305] 305ns                                                                      |L0.305|                "
    - "L0.306[305,306] 306ns                                                                      |L0.306|                "
    - "L0.307[306,307] 307ns                                                                       |L0.307|               "
    - "L0.308[307,308] 308ns                                                                       |L0.308|               "
    - "L0.309[308,309] 309ns                                                                       |L0.309|               "
    - "L0.310[309,310] 310ns                                                                       |L0.310|               "
    - "L0.311[310,311] 311ns                                                                        |L0.311|              "
    - "L0.312[311,312] 312ns                                                                        |L0.312|              "
    - "L0.313[312,313] 313ns                                                                        |L0.313|              "
    - "L0.314[313,314] 314ns                                                                        |L0.314|              "
    - "L0.315[314,315] 315ns                                                                        |L0.315|              "
    - "L0.316[315,316] 316ns                                                                         |L0.316|             "
    - "L0.317[316,317] 317ns                                                                         |L0.317|             "
    - "L0.318[317,318] 318ns                                                                         |L0.318|             "
    - "L0.319[318,319] 319ns                                                                         |L0.319|             "
    - "L0.320[319,320] 320ns                                                                          |L0.320|            "
    - "L0.321[320,321] 321ns                                                                          |L0.321|            "
    - "L0.322[321,322] 322ns                                                                          |L0.322|            "
    - "L0.323[322,323] 323ns                                                                          |L0.323|            "
    - "L0.324[323,324] 324ns                                                                          |L0.324|            "
    - "L0.325[324,325] 325ns                                                                           |L0.325|           "
    - "L0.326[325,326] 326ns                                                                           |L0.326|           "
    - "L0.327[326,327] 327ns                                                                           |L0.327|           "
    - "L0.328[327,328] 328ns                                                                           |L0.328|           "
    - "L0.329[328,329] 329ns                                                                            |L0.329|          "
    - "L0.330[329,330] 330ns                                                                            |L0.330|          "
    - "L0.331[330,331] 331ns                                                                            |L0.331|          "
    - "L0.332[331,332] 332ns                                                                            |L0.332|          "
    - "L0.333[332,333] 333ns                                                                            |L0.333|          "
    - "L0.334[333,334] 334ns                                                                             |L0.334|         "
    - "L0.335[334,335] 335ns                                                                             |L0.335|         "
    - "L0.336[335,336] 336ns                                                                             |L0.336|         "
    - "L0.337[336,337] 337ns                                                                             |L0.337|         "
    - "L0.338[337,338] 338ns                                                                             |L0.338|         "
    - "L0.339[338,339] 339ns                                                                              |L0.339|        "
    - "L0.340[339,340] 340ns                                                                              |L0.340|        "
    - "L0.341[340,341] 341ns                                                                              |L0.341|        "
    - "L0.342[341,342] 342ns                                                                              |L0.342|        "
    - "L0.343[342,343] 343ns                                                                               |L0.343|       "
    - "L0.344[343,344] 344ns                                                                               |L0.344|       "
    - "L0.345[344,345] 345ns                                                                               |L0.345|       "
    - "L0.346[345,346] 346ns                                                                               |L0.346|       "
    - "L0.347[346,347] 347ns                                                                               |L0.347|       "
    - "L0.348[347,348] 348ns                                                                                |L0.348|      "
    - "L0.349[348,349] 349ns                                                                                |L0.349|      "
    - "L0.350[349,350] 350ns                                                                                |L0.350|      "
    - "L0.351[350,351] 351ns                                                                                |L0.351|      "
    - "L0.352[351,352] 352ns                                                                                 |L0.352|     "
    - "L0.353[352,353] 353ns                                                                                 |L0.353|     "
    - "L0.354[353,354] 354ns                                                                                 |L0.354|     "
    - "L0.355[354,355] 355ns                                                                                 |L0.355|     "
    - "L0.356[355,356] 356ns                                                                                 |L0.356|     "
    - "L0.357[356,357] 357ns                                                                                  |L0.357|    "
    - "L0.358[357,358] 358ns                                                                                  |L0.358|    "
    - "L0.359[358,359] 359ns                                                                                  |L0.359|    "
    - "L0.360[359,360] 360ns                                                                                  |L0.360|    "
    - "L0.361[360,361] 361ns                                                                                   |L0.361|   "
    - "L0.362[361,362] 362ns                                                                                   |L0.362|   "
    - "L0.363[362,363] 363ns                                                                                   |L0.363|   "
    - "L0.364[363,364] 364ns                                                                                   |L0.364|   "
    - "L0.365[364,365] 365ns                                                                                   |L0.365|   "
    - "L0.366[365,366] 366ns                                                                                    |L0.366|  "
    - "L0.367[366,367] 367ns                                                                                    |L0.367|  "
    - "L0.368[367,368] 368ns                                                                                    |L0.368|  "
    - "L0.369[368,369] 369ns                                                                                    |L0.369|  "
    - "L0.370[369,370] 370ns                                                                                     |L0.370| "
    - "L0.371[370,371] 371ns                                                                                     |L0.371| "
    - "L0.372[371,372] 372ns                                                                                     |L0.372| "
    - "L0.373[372,373] 373ns                                                                                     |L0.373| "
    - "L0.374[373,374] 374ns                                                                                     |L0.374| "
    - "L0.375[374,375] 375ns                                                                                      |L0.375|"
    - "L0.376[375,376] 376ns                                                                                      |L0.376|"
    - "L0.377[376,377] 377ns                                                                                      |L0.377|"
    - "L0.378[377,378] 378ns                                                                                      |L0.378|"
    - "L0.379[378,379] 379ns                                                                                      |L0.379|"
    - "L0.380[379,380] 380ns                                                                                       |L0.380|"
    - "L0.381[380,381] 381ns                                                                                       |L0.381|"
    - "L0.382[381,382] 382ns                                                                                       |L0.382|"
    - "L0.383[382,383] 383ns                                                                                       |L0.383|"
    - "L0.384[383,384] 384ns                                                                                        |L0.384|"
    - "L0.385[384,385] 385ns                                                                                        |L0.385|"
    - "L0.386[385,386] 386ns                                                                                        |L0.386|"
    - "L0.387[386,387] 387ns                                                                                        |L0.387|"
    - "L0.388[387,388] 388ns                                                                                        |L0.388|"
    - "L0.389[388,389] 389ns                                                                                         |L0.389|"
    - "L0.390[389,390] 390ns                                                                                         |L0.390|"
    - "L0.391[390,391] 391ns                                                                                         |L0.391|"
    - "L0.392[391,392] 392ns                                                                                         |L0.392|"
    - "L0.393[392,393] 393ns                                                                                          |L0.393|"
    - "L0.394[393,394] 394ns                                                                                          |L0.394|"
    - "L0.395[394,395] 395ns                                                                                          |L0.395|"
    - "L0.396[395,396] 396ns                                                                                          |L0.396|"
    - "L0.397[396,397] 397ns                                                                                          |L0.397|"
    - "L0.398[397,398] 398ns                                                                                           |L0.398|"
    - "L0.399[398,399] 399ns                                                                                           |L0.399|"
    - "L0.400[399,400] 400ns                                                                                           |L0.400|"
    - "L0.401[400,401] 401ns                                                                                           |L0.401|"
    - "L0.402[401,402] 402ns                                                                                            |L0.402|"
    - "L0.403[402,403] 403ns                                                                                            |L0.403|"
    - "L0.404[403,404] 404ns                                                                                            |L0.404|"
    - "L0.405[404,405] 405ns                                                                                            |L0.405|"
    - "L0.406[405,406] 406ns                                                                                            |L0.406|"
    - "L0.407[406,407] 407ns                                                                                             |L0.407|"
    - "L0.408[407,408] 408ns                                                                                             |L0.408|"
    - "L0.409[408,409] 409ns                                                                                             |L0.409|"
    - "L0.410[409,410] 410ns                                                                                             |L0.410|"
    - "**** Simulation run 0, type=compact(ManySmallFiles). 200 Input Files, 1.37mb total:"
    - "L0, all files 7kb                                                                                                  "
    - "L0.1[0,1] 1ns            |L0.1|                                                                                    "
    - "L0.2[1,2] 2ns            |L0.2|                                                                                    "
    - "L0.3[2,3] 3ns            |L0.3|                                                                                    "
    - "L0.4[3,4] 4ns             |L0.4|                                                                                   "
    - "L0.5[4,5] 5ns             |L0.5|                                                                                   "
    - "L0.6[5,6] 6ns              |L0.6|                                                                                  "
    - "L0.7[6,7] 7ns              |L0.7|                                                                                  "
    - "L0.8[7,8] 8ns               |L0.8|                                                                                 "
    - "L0.9[8,9] 9ns               |L0.9|                                                                                 "
    - "L0.10[9,10] 10ns             |L0.10|                                                                               "
    - "L0.11[10,11] 11ns            |L0.11|                                                                               "
    - "L0.12[11,12] 12ns            |L0.12|                                                                               "
    - "L0.13[12,13] 13ns             |L0.13|                                                                              "
    - "L0.14[13,14] 14ns             |L0.14|                                                                              "
    - "L0.15[14,15] 15ns              |L0.15|                                                                             "
    - "L0.16[15,16] 16ns              |L0.16|                                                                             "
    - "L0.17[16,17] 17ns               |L0.17|                                                                            "
    - "L0.18[17,18] 18ns               |L0.18|                                                                            "
    - "L0.19[18,19] 19ns                |L0.19|                                                                           "
    - "L0.20[19,20] 20ns                |L0.20|                                                                           "
    - "L0.21[20,21] 21ns                 |L0.21|                                                                          "
    - "L0.22[21,22] 22ns                 |L0.22|                                                                          "
    - "L0.23[22,23] 23ns                 |L0.23|                                                                          "
    - "L0.24[23,24] 24ns                  |L0.24|                                                                         "
    - "L0.25[24,25] 25ns                  |L0.25|                                                                         "
    - "L0.26[25,26] 26ns                   |L0.26|                                                                        "
    - "L0.27[26,27] 27ns                   |L0.27|                                                                        "
    - "L0.28[27,28] 28ns                    |L0.28|                                                                       "
    - "L0.29[28,29] 29ns                    |L0.29|                                                                       "
    - "L0.30[29,30] 30ns                     |L0.30|                                                                      "
    - "L0.31[30,31] 31ns                     |L0.31|                                                                      "
    - "L0.32[31,32] 32ns                     |L0.32|                                                                      "
    - "L0.33[32,33] 33ns                      |L0.33|                                                                     "
    - "L0.34[33,34] 34ns                      |L0.34|                                                                     "
    - "L0.35[34,35] 35ns                       |L0.35|                                                                    "
    - "L0.36[35,36] 36ns                       |L0.36|                                                                    "
    - "L0.37[36,37] 37ns                        |L0.37|                                                                   "
    - "L0.38[37,38] 38ns                        |L0.38|                                                                   "
    - "L0.39[38,39] 39ns                         |L0.39|                                                                  "
    - "L0.40[39,40] 40ns                         |L0.40|                                                                  "
    - "L0.41[40,41] 41ns                          |L0.41|                                                                 "
    - "L0.42[41,42] 42ns                          |L0.42|                                                                 "
    - "L0.43[42,43] 43ns                          |L0.43|                                                                 "
    - "L0.44[43,44] 44ns                           |L0.44|                                                                "
    - "L0.45[44,45] 45ns                           |L0.45|                                                                "
    - "L0.46[45,46] 46ns                            |L0.46|                                                               "
    - "L0.47[46,47] 47ns                            |L0.47|                                                               "
    - "L0.48[47,48] 48ns                             |L0.48|                                                              "
    - "L0.49[48,49] 49ns                             |L0.49|                                                              "
    - "L0.50[49,50] 50ns                              |L0.50|                                                             "
    - "L0.51[50,51] 51ns                              |L0.51|                                                             "
    - "L0.52[51,52] 52ns                              |L0.52|                                                             "
    - "L0.53[52,53] 53ns                               |L0.53|                                                            "
    - "L0.54[53,54] 54ns                               |L0.54|                                                            "
    - "L0.55[54,55] 55ns                                |L0.55|                                                           "
    - "L0.56[55,56] 56ns                                |L0.56|                                                           "
    - "L0.57[56,57] 57ns                                 |L0.57|                                                          "
    - "L0.58[57,58] 58ns                                 |L0.58|                                                          "
    - "L0.59[58,59] 59ns                                  |L0.59|                                                         "
    - "L0.60[59,60] 60ns                                  |L0.60|                                                         "
    - "L0.61[60,61] 61ns                                   |L0.61|                                                        "
    - "L0.62[61,62] 62ns                                   |L0.62|                                                        "
    - "L0.63[62,63] 63ns                                   |L0.63|                                                        "
    - "L0.64[63,64] 64ns                                    |L0.64|                                                       "
    - "L0.65[64,65] 65ns                                    |L0.65|                                                       "
    - "L0.66[65,66] 66ns                                     |L0.66|                                                      "
    - "L0.67[66,67] 67ns                                     |L0.67|                                                      "
    - "L0.68[67,68] 68ns                                      |L0.68|                                                     "
    - "L0.69[68,69] 69ns                                      |L0.69|                                                     "
    - "L0.70[69,70] 70ns                                       |L0.70|                                                    "
    - "L0.71[70,71] 71ns                                       |L0.71|                                                    "
    - "L0.72[71,72] 72ns                                       |L0.72|                                                    "
    - "L0.73[72,73] 73ns                                        |L0.73|                                                   "
    - "L0.74[73,74] 74ns                                        |L0.74|                                                   "
    - "L0.75[74,75] 75ns                                         |L0.75|                                                  "
    - "L0.76[75,76] 76ns                                         |L0.76|                                                  "
    - "L0.77[76,77] 77ns                                          |L0.77|                                                 "
    - "L0.78[77,78] 78ns                                          |L0.78|                                                 "
    - "L0.79[78,79] 79ns                                           |L0.79|                                                "
    - "L0.80[79,80] 80ns                                           |L0.80|                                                "
    - "L0.81[80,81] 81ns                                            |L0.81|                                               "
    - "L0.82[81,82] 82ns                                            |L0.82|                                               "
    - "L0.83[82,83] 83ns                                            |L0.83|                                               "
    - "L0.84[83,84] 84ns                                             |L0.84|                                              "
    - "L0.85[84,85] 85ns                                             |L0.85|                                              "
    - "L0.86[85,86] 86ns                                              |L0.86|                                             "
    - "L0.87[86,87] 87ns                                              |L0.87|                                             "
    - "L0.88[87,88] 88ns                                               |L0.88|                                            "
    - "L0.89[88,89] 89ns                                               |L0.89|                                            "
    - "L0.90[89,90] 90ns                                                |L0.90|                                           "
    - "L0.91[90,91] 91ns                                                |L0.91|                                           "
    - "L0.92[91,92] 92ns                                                |L0.92|                                           "
    - "L0.93[92,93] 93ns                                                 |L0.93|                                          "
    - "L0.94[93,94] 94ns                                                 |L0.94|                                          "
    - "L0.95[94,95] 95ns                                                  |L0.95|                                         "
    - "L0.96[95,96] 96ns                                                  |L0.96|                                         "
    - "L0.97[96,97] 97ns                                                   |L0.97|                                        "
    - "L0.98[97,98] 98ns                                                   |L0.98|                                        "
    - "L0.99[98,99] 99ns                                                    |L0.99|                                       "
    - "L0.100[99,100] 100ns                                                 |L0.100|                                      "
    - "L0.101[100,101] 101ns                                                 |L0.101|                                     "
    - "L0.102[101,102] 102ns                                                 |L0.102|                                     "
    - "L0.103[102,103] 103ns                                                 |L0.103|                                     "
    - "L0.104[103,104] 104ns                                                  |L0.104|                                    "
    - "L0.105[104,105] 105ns                                                  |L0.105|                                    "
    - "L0.106[105,106] 106ns                                                   |L0.106|                                   "
    - "L0.107[106,107] 107ns                                                   |L0.107|                                   "
    - "L0.108[107,108] 108ns                                                    |L0.108|                                  "
    - "L0.109[108,109] 109ns                                                    |L0.109|                                  "
    - "L0.110[109,110] 110ns                                                     |L0.110|                                 "
    - "L0.111[110,111] 111ns                                                     |L0.111|                                 "
    - "L0.112[111,112] 112ns                                                     |L0.112|                                 "
    - "L0.113[112,113] 113ns                                                      |L0.113|                                "
    - "L0.114[113,114] 114ns                                                      |L0.114|                                "
    - "L0.115[114,115] 115ns                                                       |L0.115|                               "
    - "L0.116[115,116] 116ns                                                       |L0.116|                               "
    - "L0.117[116,117] 117ns                                                        |L0.117|                              "
    - "L0.118[117,118] 118ns                                                        |L0.118|                              "
    - "L0.119[118,119] 119ns                                                         |L0.119|                             "
    - "L0.120[119,120] 120ns                                                         |L0.120|                             "
    - "L0.121[120,121] 121ns                                                          |L0.121|                            "
    - "L0.122[121,122] 122ns                                                          |L0.122|                            "
    - "L0.123[122,123] 123ns                                                          |L0.123|                            "
    - "L0.124[123,124] 124ns                                                           |L0.124|                           "
    - "L0.125[124,125] 125ns                                                           |L0.125|                           "
    - "L0.126[125,126] 126ns                                                            |L0.126|                          "
    - "L0.127[126,127] 127ns                                                            |L0.127|                          "
    - "L0.128[127,128] 128ns                                                             |L0.128|                         "
    - "L0.129[128,129] 129ns                                                             |L0.129|                         "
    - "L0.130[129,130] 130ns                                                              |L0.130|                        "
    - "L0.131[130,131] 131ns                                                              |L0.131|                        "
    - "L0.132[131,132] 132ns                                                              |L0.132|                        "
    - "L0.133[132,133] 133ns                                                               |L0.133|                       "
    - "L0.134[133,134] 134ns                                                               |L0.134|                       "
    - "L0.135[134,135] 135ns                                                                |L0.135|                      "
    - "L0.136[135,136] 136ns                                                                |L0.136|                      "
    - "L0.137[136,137] 137ns                                                                 |L0.137|                     "
    - "L0.138[137,138] 138ns                                                                 |L0.138|                     "
    - "L0.139[138,139] 139ns                                                                  |L0.139|                    "
    - "L0.140[139,140] 140ns                                                                  |L0.140|                    "
    - "L0.141[140,141] 141ns                                                                   |L0.141|                   "
    - "L0.142[141,142] 142ns                                                                   |L0.142|                   "
    - "L0.143[142,143] 143ns                                                                   |L0.143|                   "
    - "L0.144[143,144] 144ns                                                                    |L0.144|                  "
    - "L0.145[144,145] 145ns                                                                    |L0.145|                  "
    - "L0.146[145,146] 146ns                                                                     |L0.146|                 "
    - "L0.147[146,147] 147ns                                                                     |L0.147|                 "
    - "L0.148[147,148] 148ns                                                                      |L0.148|                "
    - "L0.149[148,149] 149ns                                                                      |L0.149|                "
    - "L0.150[149,150] 150ns                                                                       |L0.150|               "
    - "L0.151[150,151] 151ns                                                                       |L0.151|               "
    - "L0.152[151,152] 152ns                                                                       |L0.152|               "
    - "L0.153[152,153] 153ns                                                                        |L0.153|              "
    - "L0.154[153,154] 154ns                                                                        |L0.154|              "
    - "L0.155[154,155] 155ns                                                                         |L0.155|             "
    - "L0.156[155,156] 156ns                                                                         |L0.156|             "
    - "L0.157[156,157] 157ns                                                                          |L0.157|            "
    - "L0.158[157,158] 158ns                                                                          |L0.158|            "
    - "L0.159[158,159] 159ns                                                                           |L0.159|           "
    - "L0.160[159,160] 160ns                                                                           |L0.160|           "
    - "L0.161[160,161] 161ns                                                                            |L0.161|          "
    - "L0.162[161,162] 162ns                                                                            |L0.162|          "
    - "L0.163[162,163] 163ns                                                                            |L0.163|          "
    - "L0.164[163,164] 164ns                                                                             |L0.164|         "
    - "L0.165[164,165] 165ns                                                                             |L0.165|         "
    - "L0.166[165,166] 166ns                                                                              |L0.166|        "
    - "L0.167[166,167] 167ns                                                                              |L0.167|        "
    - "L0.168[167,168] 168ns                                                                               |L0.168|       "
    - "L0.169[168,169] 169ns                                                                               |L0.169|       "
    - "L0.170[169,170] 170ns                                                                                |L0.170|      "
    - "L0.171[170,171] 171ns                                                                                |L0.171|      "
    - "L0.172[171,172] 172ns                                                                                |L0.172|      "
    - "L0.173[172,173] 173ns                                                                                 |L0.173|     "
    - "L0.174[173,174] 174ns                                                                                 |L0.174|     "
    - "L0.175[174,175] 175ns                                                                                  |L0.175|    "
    - "L0.176[175,176] 176ns                                                                                  |L0.176|    "
    - "L0.177[176,177] 177ns                                                                                   |L0.177|   "
    - "L0.178[177,178] 178ns                                                                                   |L0.178|   "
    - "L0.179[178,179] 179ns                                                                                    |L0.179|  "
    - "L0.180[179,180] 180ns                                                                                    |L0.180|  "
    - "L0.181[180,181] 181ns                                                                                     |L0.181| "
    - "L0.182[181,182] 182ns                                                                                     |L0.182| "
    - "L0.183[182,183] 183ns                                                                                     |L0.183| "
    - "L0.184[183,184] 184ns                                                                                      |L0.184|"
    - "L0.185[184,185] 185ns                                                                                      |L0.185|"
    - "L0.186[185,186] 186ns                                                                                       |L0.186|"
    - "L0.187[186,187] 187ns                                                                                       |L0.187|"
    - "L0.188[187,188] 188ns                                                                                        |L0.188|"
    - "L0.189[188,189] 189ns                                                                                        |L0.189|"
    - "L0.190[189,190] 190ns                                                                                         |L0.190|"
    - "L0.191[190,191] 191ns                                                                                         |L0.191|"
    - "L0.192[191,192] 192ns                                                                                         |L0.192|"
    - "L0.193[192,193] 193ns                                                                                          |L0.193|"
    - "L0.194[193,194] 194ns                                                                                          |L0.194|"
    - "L0.195[194,195] 195ns                                                                                           |L0.195|"
    - "L0.196[195,196] 196ns                                                                                           |L0.196|"
    - "L0.197[196,197] 197ns                                                                                            |L0.197|"
    - "L0.198[197,198] 198ns                                                                                            |L0.198|"
    - "L0.199[198,199] 199ns                                                                                             |L0.199|"
    - "L0.200[199,200] 200ns                                                                                             |L0.200|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 1.37mb total:"
    - "L0, all files 1.37mb                                                                                               "
    - "L0.?[0,200] 200ns        |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 200 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10, L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33, L0.34, L0.35, L0.36, L0.37, L0.38, L0.39, L0.40, L0.41, L0.42, L0.43, L0.44, L0.45, L0.46, L0.47, L0.48, L0.49, L0.50, L0.51, L0.52, L0.53, L0.54, L0.55, L0.56, L0.57, L0.58, L0.59, L0.60, L0.61, L0.62, L0.63, L0.64, L0.65, L0.66, L0.67, L0.68, L0.69, L0.70, L0.71, L0.72, L0.73, L0.74, L0.75, L0.76, L0.77, L0.78, L0.79, L0.80, L0.81, L0.82, L0.83, L0.84, L0.85, L0.86, L0.87, L0.88, L0.89, L0.90, L0.91, L0.92, L0.93, L0.94, L0.95, L0.96, L0.97, L0.98, L0.99, L0.100, L0.101, L0.102, L0.103, L0.104, L0.105, L0.106, L0.107, L0.108, L0.109, L0.110, L0.111, L0.112, L0.113, L0.114, L0.115, L0.116, L0.117, L0.118, L0.119, L0.120, L0.121, L0.122, L0.123, L0.124, L0.125, L0.126, L0.127, L0.128, L0.129, L0.130, L0.131, L0.132, L0.133, L0.134, L0.135, L0.136, L0.137, L0.138, L0.139, L0.140, L0.141, L0.142, L0.143, L0.144, L0.145, L0.146, L0.147, L0.148, L0.149, L0.150, L0.151, L0.152, L0.153, L0.154, L0.155, L0.156, L0.157, L0.158, L0.159, L0.160, L0.161, L0.162, L0.163, L0.164, L0.165, L0.166, L0.167, L0.168, L0.169, L0.170, L0.171, L0.172, L0.173, L0.174, L0.175, L0.176, L0.177, L0.178, L0.179, L0.180, L0.181, L0.182, L0.183, L0.184, L0.185, L0.186, L0.187, L0.188, L0.189, L0.190, L0.191, L0.192, L0.193, L0.194, L0.195, L0.196, L0.197, L0.198, L0.199, L0.200"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=compact(ManySmallFiles). 200 Input Files, 1.37mb total:"
    - "L0, all files 7kb                                                                                                  "
    - "L0.201[200,201] 201ns    |L0.201|                                                                                  "
    - "L0.202[201,202] 202ns    |L0.202|                                                                                  "
    - "L0.203[202,203] 203ns    |L0.203|                                                                                  "
    - "L0.204[203,204] 204ns     |L0.204|                                                                                 "
    - "L0.205[204,205] 205ns     |L0.205|                                                                                 "
    - "L0.206[205,206] 206ns      |L0.206|                                                                                "
    - "L0.207[206,207] 207ns      |L0.207|                                                                                "
    - "L0.208[207,208] 208ns       |L0.208|                                                                               "
    - "L0.209[208,209] 209ns       |L0.209|                                                                               "
    - "L0.210[209,210] 210ns        |L0.210|                                                                              "
    - "L0.211[210,211] 211ns        |L0.211|                                                                              "
    - "L0.212[211,212] 212ns        |L0.212|                                                                              "
    - "L0.213[212,213] 213ns         |L0.213|                                                                             "
    - "L0.214[213,214] 214ns         |L0.214|                                                                             "
    - "L0.215[214,215] 215ns          |L0.215|                                                                            "
    - "L0.216[215,216] 216ns          |L0.216|                                                                            "
    - "L0.217[216,217] 217ns           |L0.217|                                                                           "
    - "L0.218[217,218] 218ns           |L0.218|                                                                           "
    - "L0.219[218,219] 219ns            |L0.219|                                                                          "
    - "L0.220[219,220] 220ns            |L0.220|                                                                          "
    - "L0.221[220,221] 221ns             |L0.221|                                                                         "
    - "L0.222[221,222] 222ns             |L0.222|                                                                         "
    - "L0.223[222,223] 223ns             |L0.223|                                                                         "
    - "L0.224[223,224] 224ns              |L0.224|                                                                        "
    - "L0.225[224,225] 225ns              |L0.225|                                                                        "
    - "L0.226[225,226] 226ns               |L0.226|                                                                       "
    - "L0.227[226,227] 227ns               |L0.227|                                                                       "
    - "L0.228[227,228] 228ns                |L0.228|                                                                      "
    - "L0.229[228,229] 229ns                |L0.229|                                                                      "
    - "L0.230[229,230] 230ns                 |L0.230|                                                                     "
    - "L0.231[230,231] 231ns                 |L0.231|                                                                     "
    - "L0.232[231,232] 232ns                 |L0.232|                                                                     "
    - "L0.233[232,233] 233ns                  |L0.233|                                                                    "
    - "L0.234[233,234] 234ns                  |L0.234|                                                                    "
    - "L0.235[234,235] 235ns                   |L0.235|                                                                   "
    - "L0.236[235,236] 236ns                   |L0.236|                                                                   "
    - "L0.237[236,237] 237ns                    |L0.237|                                                                  "
    - "L0.238[237,238] 238ns                    |L0.238|                                                                  "
    - "L0.239[238,239] 239ns                     |L0.239|                                                                 "
    - "L0.240[239,240] 240ns                     |L0.240|                                                                 "
    - "L0.241[240,241] 241ns                      |L0.241|                                                                "
    - "L0.242[241,242] 242ns                      |L0.242|                                                                "
    - "L0.243[242,243] 243ns                      |L0.243|                                                                "
    - "L0.244[243,244] 244ns                       |L0.244|                                                               "
    - "L0.245[244,245] 245ns                       |L0.245|                                                               "
    - "L0.246[245,246] 246ns                        |L0.246|                                                              "
    - "L0.247[246,247] 247ns                        |L0.247|                                                              "
    - "L0.248[247,248] 248ns                         |L0.248|                                                             "
    - "L0.249[248,249] 249ns                         |L0.249|                                                             "
    - "L0.250[249,250] 250ns                          |L0.250|                                                            "
    - "L0.251[250,251] 251ns                          |L0.251|                                                            "
    - "L0.252[251,252] 252ns                          |L0.252|                                                            "
    - "L0.253[252,253] 253ns                           |L0.253|                                                           "
    - "L0.254[253,254] 254ns                           |L0.254|                                                           "
    - "L0.255[254,255] 255ns                            |L0.255|                                                          "
    - "L0.256[255,256] 256ns                            |L0.256|                                                          "
    - "L0.257[256,257] 257ns                             |L0.257|                                                         "
    - "L0.258[257,258] 258ns                             |L0.258|                                                         "
    - "L0.259[258,259] 259ns                              |L0.259|                                                        "
    - "L0.260[259,260] 260ns                              |L0.260|                                                        "
    - "L0.261[260,261] 261ns                               |L0.261|                                                       "
    - "L0.262[261,262] 262ns                               |L0.262|                                                       "
    - "L0.263[262,263] 263ns                               |L0.263|                                                       "
    - "L0.264[263,264] 264ns                                |L0.264|                                                      "
    - "L0.265[264,265] 265ns                                |L0.265|                                                      "
    - "L0.266[265,266] 266ns                                 |L0.266|                                                     "
    - "L0.267[266,267] 267ns                                 |L0.267|                                                     "
    - "L0.268[267,268] 268ns                                  |L0.268|                                                    "
    - "L0.269[268,269] 269ns                                  |L0.269|                                                    "
    - "L0.270[269,270] 270ns                                   |L0.270|                                                   "
    - "L0.271[270,271] 271ns                                   |L0.271|                                                   "
    - "L0.272[271,272] 272ns                                   |L0.272|                                                   "
    - "L0.273[272,273] 273ns                                    |L0.273|                                                  "
    - "L0.274[273,274] 274ns                                    |L0.274|                                                  "
    - "L0.275[274,275] 275ns                                     |L0.275|                                                 "
    - "L0.276[275,276] 276ns                                     |L0.276|                                                 "
    - "L0.277[276,277] 277ns                                      |L0.277|                                                "
    - "L0.278[277,278] 278ns                                      |L0.278|                                                "
    - "L0.279[278,279] 279ns                                       |L0.279|                                               "
    - "L0.280[279,280] 280ns                                       |L0.280|                                               "
    - "L0.281[280,281] 281ns                                        |L0.281|                                              "
    - "L0.282[281,282] 282ns                                        |L0.282|                                              "
    - "L0.283[282,283] 283ns                                        |L0.283|                                              "
    - "L0.284[283,284] 284ns                                         |L0.284|                                             "
    - "L0.285[284,285] 285ns                                         |L0.285|                                             "
    - "L0.286[285,286] 286ns                                          |L0.286|                                            "
    - "L0.287[286,287] 287ns                                          |L0.287|                                            "
    - "L0.288[287,288] 288ns                                           |L0.288|                                           "
    - "L0.289[288,289] 289ns                                           |L0.289|                                           "
    - "L0.290[289,290] 290ns                                            |L0.290|                                          "
    - "L0.291[290,291] 291ns                                            |L0.291|                                          "
    - "L0.292[291,292] 292ns                                            |L0.292|                                          "
    - "L0.293[292,293] 293ns                                             |L0.293|                                         "
    - "L0.294[293,294] 294ns                                             |L0.294|                                         "
    - "L0.295[294,295] 295ns                                              |L0.295|                                        "
    - "L0.296[295,296] 296ns                                              |L0.296|                                        "
    - "L0.297[296,297] 297ns                                               |L0.297|                                       "
    - "L0.298[297,298] 298ns                                               |L0.298|                                       "
    - "L0.299[298,299] 299ns                                                |L0.299|                                      "
    - "L0.300[299,300] 300ns                                                |L0.300|                                      "
    - "L0.301[300,301] 301ns                                                 |L0.301|                                     "
    - "L0.302[301,302] 302ns                                                 |L0.302|                                     "
    - "L0.303[302,303] 303ns                                                 |L0.303|                                     "
    - "L0.304[303,304] 304ns                                                  |L0.304|                                    "
    - "L0.305[304,305] 305ns                                                  |L0.305|                                    "
    - "L0.306[305,306] 306ns                                                   |L0.306|                                   "
    - "L0.307[306,307] 307ns                                                   |L0.307|                                   "
    - "L0.308[307,308] 308ns                                                    |L0.308|                                  "
    - "L0.309[308,309] 309ns                                                    |L0.309|                                  "
    - "L0.310[309,310] 310ns                                                     |L0.310|                                 "
    - "L0.311[310,311] 311ns                                                     |L0.311|                                 "
    - "L0.312[311,312] 312ns                                                     |L0.312|                                 "
    - "L0.313[312,313] 313ns                                                      |L0.313|                                "
    - "L0.314[313,314] 314ns                                                      |L0.314|                                "
    - "L0.315[314,315] 315ns                                                       |L0.315|                               "
    - "L0.316[315,316] 316ns                                                       |L0.316|                               "
    - "L0.317[316,317] 317ns                                                        |L0.317|                              "
    - "L0.318[317,318] 318ns                                                        |L0.318|                              "
    - "L0.319[318,319] 319ns                                                         |L0.319|                             "
    - "L0.320[319,320] 320ns                                                         |L0.320|                             "
    - "L0.321[320,321] 321ns                                                          |L0.321|                            "
    - "L0.322[321,322] 322ns                                                          |L0.322|                            "
    - "L0.323[322,323] 323ns                                                          |L0.323|                            "
    - "L0.324[323,324] 324ns                                                           |L0.324|                           "
    - "L0.325[324,325] 325ns                                                           |L0.325|                           "
    - "L0.326[325,326] 326ns                                                            |L0.326|                          "
    - "L0.327[326,327] 327ns                                                            |L0.327|                          "
    - "L0.328[327,328] 328ns                                                             |L0.328|                         "
    - "L0.329[328,329] 329ns                                                             |L0.329|                         "
    - "L0.330[329,330] 330ns                                                              |L0.330|                        "
    - "L0.331[330,331] 331ns                                                              |L0.331|                        "
    - "L0.332[331,332] 332ns                                                              |L0.332|                        "
    - "L0.333[332,333] 333ns                                                               |L0.333|                       "
    - "L0.334[333,334] 334ns                                                               |L0.334|                       "
    - "L0.335[334,335] 335ns                                                                |L0.335|                      "
    - "L0.336[335,336] 336ns                                                                |L0.336|                      "
    - "L0.337[336,337] 337ns                                                                 |L0.337|                     "
    - "L0.338[337,338] 338ns                                                                 |L0.338|                     "
    - "L0.339[338,339] 339ns                                                                  |L0.339|                    "
    - "L0.340[339,340] 340ns                                                                  |L0.340|                    "
    - "L0.341[340,341] 341ns                                                                   |L0.341|                   "
    - "L0.342[341,342] 342ns                                                                   |L0.342|                   "
    - "L0.343[342,343] 343ns                                                                   |L0.343|                   "
    - "L0.344[343,344] 344ns                                                                    |L0.344|                  "
    - "L0.345[344,345] 345ns                                                                    |L0.345|                  "
    - "L0.346[345,346] 346ns                                                                     |L0.346|                 "
    - "L0.347[346,347] 347ns                                                                     |L0.347|                 "
    - "L0.348[347,348] 348ns                                                                      |L0.348|                "
    - "L0.349[348,349] 349ns                                                                      |L0.349|                "
    - "L0.350[349,350] 350ns                                                                       |L0.350|               "
    - "L0.351[350,351] 351ns                                                                       |L0.351|               "
    - "L0.352[351,352] 352ns                                                                       |L0.352|               "
    - "L0.353[352,353] 353ns                                                                        |L0.353|              "
    - "L0.354[353,354] 354ns                                                                        |L0.354|              "
    - "L0.355[354,355] 355ns                                                                         |L0.355|             "
    - "L0.356[355,356] 356ns                                                                         |L0.356|             "
    - "L0.357[356,357] 357ns                                                                          |L0.357|            "
    - "L0.358[357,358] 358ns                                                                          |L0.358|            "
    - "L0.359[358,359] 359ns                                                                           |L0.359|           "
    - "L0.360[359,360] 360ns                                                                           |L0.360|           "
    - "L0.361[360,361] 361ns                                                                            |L0.361|          "
    - "L0.362[361,362] 362ns                                                                            |L0.362|          "
    - "L0.363[362,363] 363ns                                                                            |L0.363|          "
    - "L0.364[363,364] 364ns                                                                             |L0.364|         "
    - "L0.365[364,365] 365ns                                                                             |L0.365|         "
    - "L0.366[365,366] 366ns                                                                              |L0.366|        "
    - "L0.367[366,367] 367ns                                                                              |L0.367|        "
    - "L0.368[367,368] 368ns                                                                               |L0.368|       "
    - "L0.369[368,369] 369ns                                                                               |L0.369|       "
    - "L0.370[369,370] 370ns                                                                                |L0.370|      "
    - "L0.371[370,371] 371ns                                                                                |L0.371|      "
    - "L0.372[371,372] 372ns                                                                                |L0.372|      "
    - "L0.373[372,373] 373ns                                                                                 |L0.373|     "
    - "L0.374[373,374] 374ns                                                                                 |L0.374|     "
    - "L0.375[374,375] 375ns                                                                                  |L0.375|    "
    - "L0.376[375,376] 376ns                                                                                  |L0.376|    "
    - "L0.377[376,377] 377ns                                                                                   |L0.377|   "
    - "L0.378[377,378] 378ns                                                                                   |L0.378|   "
    - "L0.379[378,379] 379ns                                                                                    |L0.379|  "
    - "L0.380[379,380] 380ns                                                                                    |L0.380|  "
    - "L0.381[380,381] 381ns                                                                                     |L0.381| "
    - "L0.382[381,382] 382ns                                                                                     |L0.382| "
    - "L0.383[382,383] 383ns                                                                                     |L0.383| "
    - "L0.384[383,384] 384ns                                                                                      |L0.384|"
    - "L0.385[384,385] 385ns                                                                                      |L0.385|"
    - "L0.386[385,386] 386ns                                                                                       |L0.386|"
    - "L0.387[386,387] 387ns                                                                                       |L0.387|"
    - "L0.388[387,388] 388ns                                                                                        |L0.388|"
    - "L0.389[388,389] 389ns                                                                                        |L0.389|"
    - "L0.390[389,390] 390ns                                                                                         |L0.390|"
    - "L0.391[390,391] 391ns                                                                                         |L0.391|"
    - "L0.392[391,392] 392ns                                                                                         |L0.392|"
    - "L0.393[392,393] 393ns                                                                                          |L0.393|"
    - "L0.394[393,394] 394ns                                                                                          |L0.394|"
    - "L0.395[394,395] 395ns                                                                                           |L0.395|"
    - "L0.396[395,396] 396ns                                                                                           |L0.396|"
    - "L0.397[396,397] 397ns                                                                                            |L0.397|"
    - "L0.398[397,398] 398ns                                                                                            |L0.398|"
    - "L0.399[398,399] 399ns                                                                                             |L0.399|"
    - "L0.400[399,400] 400ns                                                                                             |L0.400|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 1.37mb total:"
    - "L0, all files 1.37mb                                                                                               "
    - "L0.?[200,400] 400ns      |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 200 files: L0.201, L0.202, L0.203, L0.204, L0.205, L0.206, L0.207, L0.208, L0.209, L0.210, L0.211, L0.212, L0.213, L0.214, L0.215, L0.216, L0.217, L0.218, L0.219, L0.220, L0.221, L0.222, L0.223, L0.224, L0.225, L0.226, L0.227, L0.228, L0.229, L0.230, L0.231, L0.232, L0.233, L0.234, L0.235, L0.236, L0.237, L0.238, L0.239, L0.240, L0.241, L0.242, L0.243, L0.244, L0.245, L0.246, L0.247, L0.248, L0.249, L0.250, L0.251, L0.252, L0.253, L0.254, L0.255, L0.256, L0.257, L0.258, L0.259, L0.260, L0.261, L0.262, L0.263, L0.264, L0.265, L0.266, L0.267, L0.268, L0.269, L0.270, L0.271, L0.272, L0.273, L0.274, L0.275, L0.276, L0.277, L0.278, L0.279, L0.280, L0.281, L0.282, L0.283, L0.284, L0.285, L0.286, L0.287, L0.288, L0.289, L0.290, L0.291, L0.292, L0.293, L0.294, L0.295, L0.296, L0.297, L0.298, L0.299, L0.300, L0.301, L0.302, L0.303, L0.304, L0.305, L0.306, L0.307, L0.308, L0.309, L0.310, L0.311, L0.312, L0.313, L0.314, L0.315, L0.316, L0.317, L0.318, L0.319, L0.320, L0.321, L0.322, L0.323, L0.324, L0.325, L0.326, L0.327, L0.328, L0.329, L0.330, L0.331, L0.332, L0.333, L0.334, L0.335, L0.336, L0.337, L0.338, L0.339, L0.340, L0.341, L0.342, L0.343, L0.344, L0.345, L0.346, L0.347, L0.348, L0.349, L0.350, L0.351, L0.352, L0.353, L0.354, L0.355, L0.356, L0.357, L0.358, L0.359, L0.360, L0.361, L0.362, L0.363, L0.364, L0.365, L0.366, L0.367, L0.368, L0.369, L0.370, L0.371, L0.372, L0.373, L0.374, L0.375, L0.376, L0.377, L0.378, L0.379, L0.380, L0.381, L0.382, L0.383, L0.384, L0.385, L0.386, L0.387, L0.388, L0.389, L0.390, L0.391, L0.392, L0.393, L0.394, L0.395, L0.396, L0.397, L0.398, L0.399, L0.400"
    - "  Creating 1 files"
    - "**** Simulation run 2, type=compact(ManySmallFiles). 10 Input Files, 70kb total:"
    - "L0, all files 7kb                                                                                                  "
    - "L0.401[400,401] 401ns    |L0.401-|                                                                                 "
    - "L0.402[401,402] 402ns             |L0.402-|                                                                        "
    - "L0.403[402,403] 403ns                      |L0.403-|                                                               "
    - "L0.404[403,404] 404ns                               |L0.404-|                                                      "
    - "L0.405[404,405] 405ns                                        |L0.405-|                                             "
    - "L0.406[405,406] 406ns                                                 |L0.406-|                                    "
    - "L0.407[406,407] 407ns                                                          |L0.407-|                           "
    - "L0.408[407,408] 408ns                                                                   |L0.408-|                  "
    - "L0.409[408,409] 409ns                                                                            |L0.409-|         "
    - "L0.410[409,410] 410ns                                                                                     |L0.410-|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 70kb total:"
    - "L0, all files 70kb                                                                                                 "
    - "L0.?[400,410] 410ns      |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 10 files: L0.401, L0.402, L0.403, L0.404, L0.405, L0.406, L0.407, L0.408, L0.409, L0.410"
    - "  Creating 1 files"
    - "**** Simulation run 3, type=compact(TotalSizeLessThanMaxCompactSize). 3 Input Files, 2.8mb total:"
    - "L0                                                                                                                 "
    - "L0.413[400,410] 410ns 70kb                                                                                       |L0.413|"
    - "L0.412[200,400] 400ns 1.37mb                                           |-----------------L0.412------------------|    "
    - "L0.411[0,200] 200ns 1.37mb|-----------------L0.411------------------|                                               "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 2.8mb total:"
    - "L1, all files 2.8mb                                                                                                "
    - "L1.?[0,410] 410ns        |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 3 files: L0.411, L0.412, L0.413"
    - "  Creating 1 files"
    - "**** Final Output Files (5.61mb written)"
    - "L1, all files 2.8mb                                                                                                "
    - "L1.414[0,410] 410ns      |-----------------------------------------L1.414-----------------------------------------|"
    "###
    );
}

#[tokio::test]
async fn many_tiny_l1_files() {
    test_helpers::maybe_start_logging();

    // Observed size on production
    let seven_kb = 7 * 1024;

    let setup = layout_setup_builder().await.build().await;

    // models what happens if the compactor can't keep up for some
    // reason and gets to a partition where the ingster persisted
    // every 5 minutes for an entire day with telegraf data, for
    // example
    let num_tiny_files = (24 * 60) / 5;

    for i in 0..num_tiny_files {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(2 * i)
                    .with_max_time(2 * i + 1)
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(seven_kb),
            )
            .await;
    }

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L1, all files 7kb                                                                                                  "
    - "L1.1[0,1] 1ns            |L1.1|                                                                                    "
    - "L1.2[2,3] 2ns            |L1.2|                                                                                    "
    - "L1.3[4,5] 3ns            |L1.3|                                                                                    "
    - "L1.4[6,7] 4ns            |L1.4|                                                                                    "
    - "L1.5[8,9] 5ns             |L1.5|                                                                                   "
    - "L1.6[10,11] 6ns           |L1.6|                                                                                   "
    - "L1.7[12,13] 7ns           |L1.7|                                                                                   "
    - "L1.8[14,15] 8ns            |L1.8|                                                                                  "
    - "L1.9[16,17] 9ns            |L1.9|                                                                                  "
    - "L1.10[18,19] 10ns          |L1.10|                                                                                 "
    - "L1.11[20,21] 11ns           |L1.11|                                                                                "
    - "L1.12[22,23] 12ns           |L1.12|                                                                                "
    - "L1.13[24,25] 13ns           |L1.13|                                                                                "
    - "L1.14[26,27] 14ns            |L1.14|                                                                               "
    - "L1.15[28,29] 15ns            |L1.15|                                                                               "
    - "L1.16[30,31] 16ns            |L1.16|                                                                               "
    - "L1.17[32,33] 17ns             |L1.17|                                                                              "
    - "L1.18[34,35] 18ns             |L1.18|                                                                              "
    - "L1.19[36,37] 19ns             |L1.19|                                                                              "
    - "L1.20[38,39] 20ns             |L1.20|                                                                              "
    - "L1.21[40,41] 21ns              |L1.21|                                                                             "
    - "L1.22[42,43] 22ns              |L1.22|                                                                             "
    - "L1.23[44,45] 23ns              |L1.23|                                                                             "
    - "L1.24[46,47] 24ns               |L1.24|                                                                            "
    - "L1.25[48,49] 25ns               |L1.25|                                                                            "
    - "L1.26[50,51] 26ns               |L1.26|                                                                            "
    - "L1.27[52,53] 27ns                |L1.27|                                                                           "
    - "L1.28[54,55] 28ns                |L1.28|                                                                           "
    - "L1.29[56,57] 29ns                |L1.29|                                                                           "
    - "L1.30[58,59] 30ns                 |L1.30|                                                                          "
    - "L1.31[60,61] 31ns                 |L1.31|                                                                          "
    - "L1.32[62,63] 32ns                 |L1.32|                                                                          "
    - "L1.33[64,65] 33ns                  |L1.33|                                                                         "
    - "L1.34[66,67] 34ns                  |L1.34|                                                                         "
    - "L1.35[68,69] 35ns                  |L1.35|                                                                         "
    - "L1.36[70,71] 36ns                  |L1.36|                                                                         "
    - "L1.37[72,73] 37ns                   |L1.37|                                                                        "
    - "L1.38[74,75] 38ns                   |L1.38|                                                                        "
    - "L1.39[76,77] 39ns                   |L1.39|                                                                        "
    - "L1.40[78,79] 40ns                    |L1.40|                                                                       "
    - "L1.41[80,81] 41ns                    |L1.41|                                                                       "
    - "L1.42[82,83] 42ns                    |L1.42|                                                                       "
    - "L1.43[84,85] 43ns                     |L1.43|                                                                      "
    - "L1.44[86,87] 44ns                     |L1.44|                                                                      "
    - "L1.45[88,89] 45ns                     |L1.45|                                                                      "
    - "L1.46[90,91] 46ns                      |L1.46|                                                                     "
    - "L1.47[92,93] 47ns                      |L1.47|                                                                     "
    - "L1.48[94,95] 48ns                      |L1.48|                                                                     "
    - "L1.49[96,97] 49ns                       |L1.49|                                                                    "
    - "L1.50[98,99] 50ns                       |L1.50|                                                                    "
    - "L1.51[100,101] 51ns                     |L1.51|                                                                    "
    - "L1.52[102,103] 52ns                     |L1.52|                                                                    "
    - "L1.53[104,105] 53ns                      |L1.53|                                                                   "
    - "L1.54[106,107] 54ns                      |L1.54|                                                                   "
    - "L1.55[108,109] 55ns                      |L1.55|                                                                   "
    - "L1.56[110,111] 56ns                       |L1.56|                                                                  "
    - "L1.57[112,113] 57ns                       |L1.57|                                                                  "
    - "L1.58[114,115] 58ns                       |L1.58|                                                                  "
    - "L1.59[116,117] 59ns                        |L1.59|                                                                 "
    - "L1.60[118,119] 60ns                        |L1.60|                                                                 "
    - "L1.61[120,121] 61ns                        |L1.61|                                                                 "
    - "L1.62[122,123] 62ns                         |L1.62|                                                                "
    - "L1.63[124,125] 63ns                         |L1.63|                                                                "
    - "L1.64[126,127] 64ns                         |L1.64|                                                                "
    - "L1.65[128,129] 65ns                          |L1.65|                                                               "
    - "L1.66[130,131] 66ns                          |L1.66|                                                               "
    - "L1.67[132,133] 67ns                          |L1.67|                                                               "
    - "L1.68[134,135] 68ns                          |L1.68|                                                               "
    - "L1.69[136,137] 69ns                           |L1.69|                                                              "
    - "L1.70[138,139] 70ns                           |L1.70|                                                              "
    - "L1.71[140,141] 71ns                           |L1.71|                                                              "
    - "L1.72[142,143] 72ns                            |L1.72|                                                             "
    - "L1.73[144,145] 73ns                            |L1.73|                                                             "
    - "L1.74[146,147] 74ns                            |L1.74|                                                             "
    - "L1.75[148,149] 75ns                             |L1.75|                                                            "
    - "L1.76[150,151] 76ns                             |L1.76|                                                            "
    - "L1.77[152,153] 77ns                             |L1.77|                                                            "
    - "L1.78[154,155] 78ns                              |L1.78|                                                           "
    - "L1.79[156,157] 79ns                              |L1.79|                                                           "
    - "L1.80[158,159] 80ns                              |L1.80|                                                           "
    - "L1.81[160,161] 81ns                               |L1.81|                                                          "
    - "L1.82[162,163] 82ns                               |L1.82|                                                          "
    - "L1.83[164,165] 83ns                               |L1.83|                                                          "
    - "L1.84[166,167] 84ns                               |L1.84|                                                          "
    - "L1.85[168,169] 85ns                                |L1.85|                                                         "
    - "L1.86[170,171] 86ns                                |L1.86|                                                         "
    - "L1.87[172,173] 87ns                                |L1.87|                                                         "
    - "L1.88[174,175] 88ns                                 |L1.88|                                                        "
    - "L1.89[176,177] 89ns                                 |L1.89|                                                        "
    - "L1.90[178,179] 90ns                                 |L1.90|                                                        "
    - "L1.91[180,181] 91ns                                  |L1.91|                                                       "
    - "L1.92[182,183] 92ns                                  |L1.92|                                                       "
    - "L1.93[184,185] 93ns                                  |L1.93|                                                       "
    - "L1.94[186,187] 94ns                                   |L1.94|                                                      "
    - "L1.95[188,189] 95ns                                   |L1.95|                                                      "
    - "L1.96[190,191] 96ns                                   |L1.96|                                                      "
    - "L1.97[192,193] 97ns                                    |L1.97|                                                     "
    - "L1.98[194,195] 98ns                                    |L1.98|                                                     "
    - "L1.99[196,197] 99ns                                    |L1.99|                                                     "
    - "L1.100[198,199] 100ns                                  |L1.100|                                                    "
    - "L1.101[200,201] 101ns                                   |L1.101|                                                   "
    - "L1.102[202,203] 102ns                                   |L1.102|                                                   "
    - "L1.103[204,205] 103ns                                   |L1.103|                                                   "
    - "L1.104[206,207] 104ns                                    |L1.104|                                                  "
    - "L1.105[208,209] 105ns                                    |L1.105|                                                  "
    - "L1.106[210,211] 106ns                                    |L1.106|                                                  "
    - "L1.107[212,213] 107ns                                     |L1.107|                                                 "
    - "L1.108[214,215] 108ns                                     |L1.108|                                                 "
    - "L1.109[216,217] 109ns                                     |L1.109|                                                 "
    - "L1.110[218,219] 110ns                                      |L1.110|                                                "
    - "L1.111[220,221] 111ns                                      |L1.111|                                                "
    - "L1.112[222,223] 112ns                                      |L1.112|                                                "
    - "L1.113[224,225] 113ns                                       |L1.113|                                               "
    - "L1.114[226,227] 114ns                                       |L1.114|                                               "
    - "L1.115[228,229] 115ns                                       |L1.115|                                               "
    - "L1.116[230,231] 116ns                                        |L1.116|                                              "
    - "L1.117[232,233] 117ns                                        |L1.117|                                              "
    - "L1.118[234,235] 118ns                                        |L1.118|                                              "
    - "L1.119[236,237] 119ns                                        |L1.119|                                              "
    - "L1.120[238,239] 120ns                                         |L1.120|                                             "
    - "L1.121[240,241] 121ns                                         |L1.121|                                             "
    - "L1.122[242,243] 122ns                                         |L1.122|                                             "
    - "L1.123[244,245] 123ns                                          |L1.123|                                            "
    - "L1.124[246,247] 124ns                                          |L1.124|                                            "
    - "L1.125[248,249] 125ns                                          |L1.125|                                            "
    - "L1.126[250,251] 126ns                                           |L1.126|                                           "
    - "L1.127[252,253] 127ns                                           |L1.127|                                           "
    - "L1.128[254,255] 128ns                                           |L1.128|                                           "
    - "L1.129[256,257] 129ns                                            |L1.129|                                          "
    - "L1.130[258,259] 130ns                                            |L1.130|                                          "
    - "L1.131[260,261] 131ns                                            |L1.131|                                          "
    - "L1.132[262,263] 132ns                                             |L1.132|                                         "
    - "L1.133[264,265] 133ns                                             |L1.133|                                         "
    - "L1.134[266,267] 134ns                                             |L1.134|                                         "
    - "L1.135[268,269] 135ns                                             |L1.135|                                         "
    - "L1.136[270,271] 136ns                                              |L1.136|                                        "
    - "L1.137[272,273] 137ns                                              |L1.137|                                        "
    - "L1.138[274,275] 138ns                                              |L1.138|                                        "
    - "L1.139[276,277] 139ns                                               |L1.139|                                       "
    - "L1.140[278,279] 140ns                                               |L1.140|                                       "
    - "L1.141[280,281] 141ns                                               |L1.141|                                       "
    - "L1.142[282,283] 142ns                                                |L1.142|                                      "
    - "L1.143[284,285] 143ns                                                |L1.143|                                      "
    - "L1.144[286,287] 144ns                                                |L1.144|                                      "
    - "L1.145[288,289] 145ns                                                 |L1.145|                                     "
    - "L1.146[290,291] 146ns                                                 |L1.146|                                     "
    - "L1.147[292,293] 147ns                                                 |L1.147|                                     "
    - "L1.148[294,295] 148ns                                                  |L1.148|                                    "
    - "L1.149[296,297] 149ns                                                  |L1.149|                                    "
    - "L1.150[298,299] 150ns                                                  |L1.150|                                    "
    - "L1.151[300,301] 151ns                                                  |L1.151|                                    "
    - "L1.152[302,303] 152ns                                                   |L1.152|                                   "
    - "L1.153[304,305] 153ns                                                   |L1.153|                                   "
    - "L1.154[306,307] 154ns                                                   |L1.154|                                   "
    - "L1.155[308,309] 155ns                                                    |L1.155|                                  "
    - "L1.156[310,311] 156ns                                                    |L1.156|                                  "
    - "L1.157[312,313] 157ns                                                    |L1.157|                                  "
    - "L1.158[314,315] 158ns                                                     |L1.158|                                 "
    - "L1.159[316,317] 159ns                                                     |L1.159|                                 "
    - "L1.160[318,319] 160ns                                                     |L1.160|                                 "
    - "L1.161[320,321] 161ns                                                      |L1.161|                                "
    - "L1.162[322,323] 162ns                                                      |L1.162|                                "
    - "L1.163[324,325] 163ns                                                      |L1.163|                                "
    - "L1.164[326,327] 164ns                                                       |L1.164|                               "
    - "L1.165[328,329] 165ns                                                       |L1.165|                               "
    - "L1.166[330,331] 166ns                                                       |L1.166|                               "
    - "L1.167[332,333] 167ns                                                       |L1.167|                               "
    - "L1.168[334,335] 168ns                                                        |L1.168|                              "
    - "L1.169[336,337] 169ns                                                        |L1.169|                              "
    - "L1.170[338,339] 170ns                                                        |L1.170|                              "
    - "L1.171[340,341] 171ns                                                         |L1.171|                             "
    - "L1.172[342,343] 172ns                                                         |L1.172|                             "
    - "L1.173[344,345] 173ns                                                         |L1.173|                             "
    - "L1.174[346,347] 174ns                                                          |L1.174|                            "
    - "L1.175[348,349] 175ns                                                          |L1.175|                            "
    - "L1.176[350,351] 176ns                                                          |L1.176|                            "
    - "L1.177[352,353] 177ns                                                           |L1.177|                           "
    - "L1.178[354,355] 178ns                                                           |L1.178|                           "
    - "L1.179[356,357] 179ns                                                           |L1.179|                           "
    - "L1.180[358,359] 180ns                                                            |L1.180|                          "
    - "L1.181[360,361] 181ns                                                            |L1.181|                          "
    - "L1.182[362,363] 182ns                                                            |L1.182|                          "
    - "L1.183[364,365] 183ns                                                            |L1.183|                          "
    - "L1.184[366,367] 184ns                                                             |L1.184|                         "
    - "L1.185[368,369] 185ns                                                             |L1.185|                         "
    - "L1.186[370,371] 186ns                                                             |L1.186|                         "
    - "L1.187[372,373] 187ns                                                              |L1.187|                        "
    - "L1.188[374,375] 188ns                                                              |L1.188|                        "
    - "L1.189[376,377] 189ns                                                              |L1.189|                        "
    - "L1.190[378,379] 190ns                                                               |L1.190|                       "
    - "L1.191[380,381] 191ns                                                               |L1.191|                       "
    - "L1.192[382,383] 192ns                                                               |L1.192|                       "
    - "L1.193[384,385] 193ns                                                                |L1.193|                      "
    - "L1.194[386,387] 194ns                                                                |L1.194|                      "
    - "L1.195[388,389] 195ns                                                                |L1.195|                      "
    - "L1.196[390,391] 196ns                                                                 |L1.196|                     "
    - "L1.197[392,393] 197ns                                                                 |L1.197|                     "
    - "L1.198[394,395] 198ns                                                                 |L1.198|                     "
    - "L1.199[396,397] 199ns                                                                 |L1.199|                     "
    - "L1.200[398,399] 200ns                                                                  |L1.200|                    "
    - "L1.201[400,401] 201ns                                                                  |L1.201|                    "
    - "L1.202[402,403] 202ns                                                                  |L1.202|                    "
    - "L1.203[404,405] 203ns                                                                   |L1.203|                   "
    - "L1.204[406,407] 204ns                                                                   |L1.204|                   "
    - "L1.205[408,409] 205ns                                                                   |L1.205|                   "
    - "L1.206[410,411] 206ns                                                                    |L1.206|                  "
    - "L1.207[412,413] 207ns                                                                    |L1.207|                  "
    - "L1.208[414,415] 208ns                                                                    |L1.208|                  "
    - "L1.209[416,417] 209ns                                                                     |L1.209|                 "
    - "L1.210[418,419] 210ns                                                                     |L1.210|                 "
    - "L1.211[420,421] 211ns                                                                     |L1.211|                 "
    - "L1.212[422,423] 212ns                                                                      |L1.212|                "
    - "L1.213[424,425] 213ns                                                                      |L1.213|                "
    - "L1.214[426,427] 214ns                                                                      |L1.214|                "
    - "L1.215[428,429] 215ns                                                                      |L1.215|                "
    - "L1.216[430,431] 216ns                                                                       |L1.216|               "
    - "L1.217[432,433] 217ns                                                                       |L1.217|               "
    - "L1.218[434,435] 218ns                                                                       |L1.218|               "
    - "L1.219[436,437] 219ns                                                                        |L1.219|              "
    - "L1.220[438,439] 220ns                                                                        |L1.220|              "
    - "L1.221[440,441] 221ns                                                                        |L1.221|              "
    - "L1.222[442,443] 222ns                                                                         |L1.222|             "
    - "L1.223[444,445] 223ns                                                                         |L1.223|             "
    - "L1.224[446,447] 224ns                                                                         |L1.224|             "
    - "L1.225[448,449] 225ns                                                                          |L1.225|            "
    - "L1.226[450,451] 226ns                                                                          |L1.226|            "
    - "L1.227[452,453] 227ns                                                                          |L1.227|            "
    - "L1.228[454,455] 228ns                                                                           |L1.228|           "
    - "L1.229[456,457] 229ns                                                                           |L1.229|           "
    - "L1.230[458,459] 230ns                                                                           |L1.230|           "
    - "L1.231[460,461] 231ns                                                                            |L1.231|          "
    - "L1.232[462,463] 232ns                                                                            |L1.232|          "
    - "L1.233[464,465] 233ns                                                                            |L1.233|          "
    - "L1.234[466,467] 234ns                                                                            |L1.234|          "
    - "L1.235[468,469] 235ns                                                                             |L1.235|         "
    - "L1.236[470,471] 236ns                                                                             |L1.236|         "
    - "L1.237[472,473] 237ns                                                                             |L1.237|         "
    - "L1.238[474,475] 238ns                                                                              |L1.238|        "
    - "L1.239[476,477] 239ns                                                                              |L1.239|        "
    - "L1.240[478,479] 240ns                                                                              |L1.240|        "
    - "L1.241[480,481] 241ns                                                                               |L1.241|       "
    - "L1.242[482,483] 242ns                                                                               |L1.242|       "
    - "L1.243[484,485] 243ns                                                                               |L1.243|       "
    - "L1.244[486,487] 244ns                                                                                |L1.244|      "
    - "L1.245[488,489] 245ns                                                                                |L1.245|      "
    - "L1.246[490,491] 246ns                                                                                |L1.246|      "
    - "L1.247[492,493] 247ns                                                                                 |L1.247|     "
    - "L1.248[494,495] 248ns                                                                                 |L1.248|     "
    - "L1.249[496,497] 249ns                                                                                 |L1.249|     "
    - "L1.250[498,499] 250ns                                                                                 |L1.250|     "
    - "L1.251[500,501] 251ns                                                                                  |L1.251|    "
    - "L1.252[502,503] 252ns                                                                                  |L1.252|    "
    - "L1.253[504,505] 253ns                                                                                  |L1.253|    "
    - "L1.254[506,507] 254ns                                                                                   |L1.254|   "
    - "L1.255[508,509] 255ns                                                                                   |L1.255|   "
    - "L1.256[510,511] 256ns                                                                                   |L1.256|   "
    - "L1.257[512,513] 257ns                                                                                    |L1.257|  "
    - "L1.258[514,515] 258ns                                                                                    |L1.258|  "
    - "L1.259[516,517] 259ns                                                                                    |L1.259|  "
    - "L1.260[518,519] 260ns                                                                                     |L1.260| "
    - "L1.261[520,521] 261ns                                                                                     |L1.261| "
    - "L1.262[522,523] 262ns                                                                                     |L1.262| "
    - "L1.263[524,525] 263ns                                                                                      |L1.263|"
    - "L1.264[526,527] 264ns                                                                                      |L1.264|"
    - "L1.265[528,529] 265ns                                                                                      |L1.265|"
    - "L1.266[530,531] 266ns                                                                                      |L1.266|"
    - "L1.267[532,533] 267ns                                                                                       |L1.267|"
    - "L1.268[534,535] 268ns                                                                                       |L1.268|"
    - "L1.269[536,537] 269ns                                                                                       |L1.269|"
    - "L1.270[538,539] 270ns                                                                                        |L1.270|"
    - "L1.271[540,541] 271ns                                                                                        |L1.271|"
    - "L1.272[542,543] 272ns                                                                                        |L1.272|"
    - "L1.273[544,545] 273ns                                                                                         |L1.273|"
    - "L1.274[546,547] 274ns                                                                                         |L1.274|"
    - "L1.275[548,549] 275ns                                                                                         |L1.275|"
    - "L1.276[550,551] 276ns                                                                                          |L1.276|"
    - "L1.277[552,553] 277ns                                                                                          |L1.277|"
    - "L1.278[554,555] 278ns                                                                                          |L1.278|"
    - "L1.279[556,557] 279ns                                                                                           |L1.279|"
    - "L1.280[558,559] 280ns                                                                                           |L1.280|"
    - "L1.281[560,561] 281ns                                                                                           |L1.281|"
    - "L1.282[562,563] 282ns                                                                                           |L1.282|"
    - "L1.283[564,565] 283ns                                                                                            |L1.283|"
    - "L1.284[566,567] 284ns                                                                                            |L1.284|"
    - "L1.285[568,569] 285ns                                                                                            |L1.285|"
    - "L1.286[570,571] 286ns                                                                                             |L1.286|"
    - "L1.287[572,573] 287ns                                                                                             |L1.287|"
    - "L1.288[574,575] 288ns                                                                                             |L1.288|"
    - "**** Simulation run 0, type=compact(ManySmallFiles). 200 Input Files, 1.37mb total:"
    - "L1, all files 7kb                                                                                                  "
    - "L1.1[0,1] 1ns            |L1.1|                                                                                    "
    - "L1.2[2,3] 2ns            |L1.2|                                                                                    "
    - "L1.3[4,5] 3ns            |L1.3|                                                                                    "
    - "L1.4[6,7] 4ns             |L1.4|                                                                                   "
    - "L1.5[8,9] 5ns             |L1.5|                                                                                   "
    - "L1.6[10,11] 6ns            |L1.6|                                                                                  "
    - "L1.7[12,13] 7ns            |L1.7|                                                                                  "
    - "L1.8[14,15] 8ns             |L1.8|                                                                                 "
    - "L1.9[16,17] 9ns             |L1.9|                                                                                 "
    - "L1.10[18,19] 10ns            |L1.10|                                                                               "
    - "L1.11[20,21] 11ns            |L1.11|                                                                               "
    - "L1.12[22,23] 12ns            |L1.12|                                                                               "
    - "L1.13[24,25] 13ns             |L1.13|                                                                              "
    - "L1.14[26,27] 14ns             |L1.14|                                                                              "
    - "L1.15[28,29] 15ns              |L1.15|                                                                             "
    - "L1.16[30,31] 16ns              |L1.16|                                                                             "
    - "L1.17[32,33] 17ns               |L1.17|                                                                            "
    - "L1.18[34,35] 18ns               |L1.18|                                                                            "
    - "L1.19[36,37] 19ns                |L1.19|                                                                           "
    - "L1.20[38,39] 20ns                |L1.20|                                                                           "
    - "L1.21[40,41] 21ns                 |L1.21|                                                                          "
    - "L1.22[42,43] 22ns                 |L1.22|                                                                          "
    - "L1.23[44,45] 23ns                 |L1.23|                                                                          "
    - "L1.24[46,47] 24ns                  |L1.24|                                                                         "
    - "L1.25[48,49] 25ns                  |L1.25|                                                                         "
    - "L1.26[50,51] 26ns                   |L1.26|                                                                        "
    - "L1.27[52,53] 27ns                   |L1.27|                                                                        "
    - "L1.28[54,55] 28ns                    |L1.28|                                                                       "
    - "L1.29[56,57] 29ns                    |L1.29|                                                                       "
    - "L1.30[58,59] 30ns                     |L1.30|                                                                      "
    - "L1.31[60,61] 31ns                     |L1.31|                                                                      "
    - "L1.32[62,63] 32ns                     |L1.32|                                                                      "
    - "L1.33[64,65] 33ns                      |L1.33|                                                                     "
    - "L1.34[66,67] 34ns                      |L1.34|                                                                     "
    - "L1.35[68,69] 35ns                       |L1.35|                                                                    "
    - "L1.36[70,71] 36ns                       |L1.36|                                                                    "
    - "L1.37[72,73] 37ns                        |L1.37|                                                                   "
    - "L1.38[74,75] 38ns                        |L1.38|                                                                   "
    - "L1.39[76,77] 39ns                         |L1.39|                                                                  "
    - "L1.40[78,79] 40ns                         |L1.40|                                                                  "
    - "L1.41[80,81] 41ns                          |L1.41|                                                                 "
    - "L1.42[82,83] 42ns                          |L1.42|                                                                 "
    - "L1.43[84,85] 43ns                          |L1.43|                                                                 "
    - "L1.44[86,87] 44ns                           |L1.44|                                                                "
    - "L1.45[88,89] 45ns                           |L1.45|                                                                "
    - "L1.46[90,91] 46ns                            |L1.46|                                                               "
    - "L1.47[92,93] 47ns                            |L1.47|                                                               "
    - "L1.48[94,95] 48ns                             |L1.48|                                                              "
    - "L1.49[96,97] 49ns                             |L1.49|                                                              "
    - "L1.50[98,99] 50ns                              |L1.50|                                                             "
    - "L1.51[100,101] 51ns                            |L1.51|                                                             "
    - "L1.52[102,103] 52ns                             |L1.52|                                                            "
    - "L1.53[104,105] 53ns                             |L1.53|                                                            "
    - "L1.54[106,107] 54ns                             |L1.54|                                                            "
    - "L1.55[108,109] 55ns                              |L1.55|                                                           "
    - "L1.56[110,111] 56ns                              |L1.56|                                                           "
    - "L1.57[112,113] 57ns                               |L1.57|                                                          "
    - "L1.58[114,115] 58ns                               |L1.58|                                                          "
    - "L1.59[116,117] 59ns                                |L1.59|                                                         "
    - "L1.60[118,119] 60ns                                |L1.60|                                                         "
    - "L1.61[120,121] 61ns                                 |L1.61|                                                        "
    - "L1.62[122,123] 62ns                                 |L1.62|                                                        "
    - "L1.63[124,125] 63ns                                 |L1.63|                                                        "
    - "L1.64[126,127] 64ns                                  |L1.64|                                                       "
    - "L1.65[128,129] 65ns                                  |L1.65|                                                       "
    - "L1.66[130,131] 66ns                                   |L1.66|                                                      "
    - "L1.67[132,133] 67ns                                   |L1.67|                                                      "
    - "L1.68[134,135] 68ns                                    |L1.68|                                                     "
    - "L1.69[136,137] 69ns                                    |L1.69|                                                     "
    - "L1.70[138,139] 70ns                                     |L1.70|                                                    "
    - "L1.71[140,141] 71ns                                     |L1.71|                                                    "
    - "L1.72[142,143] 72ns                                      |L1.72|                                                   "
    - "L1.73[144,145] 73ns                                      |L1.73|                                                   "
    - "L1.74[146,147] 74ns                                      |L1.74|                                                   "
    - "L1.75[148,149] 75ns                                       |L1.75|                                                  "
    - "L1.76[150,151] 76ns                                       |L1.76|                                                  "
    - "L1.77[152,153] 77ns                                        |L1.77|                                                 "
    - "L1.78[154,155] 78ns                                        |L1.78|                                                 "
    - "L1.79[156,157] 79ns                                         |L1.79|                                                "
    - "L1.80[158,159] 80ns                                         |L1.80|                                                "
    - "L1.81[160,161] 81ns                                          |L1.81|                                               "
    - "L1.82[162,163] 82ns                                          |L1.82|                                               "
    - "L1.83[164,165] 83ns                                          |L1.83|                                               "
    - "L1.84[166,167] 84ns                                           |L1.84|                                              "
    - "L1.85[168,169] 85ns                                           |L1.85|                                              "
    - "L1.86[170,171] 86ns                                            |L1.86|                                             "
    - "L1.87[172,173] 87ns                                            |L1.87|                                             "
    - "L1.88[174,175] 88ns                                             |L1.88|                                            "
    - "L1.89[176,177] 89ns                                             |L1.89|                                            "
    - "L1.90[178,179] 90ns                                              |L1.90|                                           "
    - "L1.91[180,181] 91ns                                              |L1.91|                                           "
    - "L1.92[182,183] 92ns                                               |L1.92|                                          "
    - "L1.93[184,185] 93ns                                               |L1.93|                                          "
    - "L1.94[186,187] 94ns                                               |L1.94|                                          "
    - "L1.95[188,189] 95ns                                                |L1.95|                                         "
    - "L1.96[190,191] 96ns                                                |L1.96|                                         "
    - "L1.97[192,193] 97ns                                                 |L1.97|                                        "
    - "L1.98[194,195] 98ns                                                 |L1.98|                                        "
    - "L1.99[196,197] 99ns                                                  |L1.99|                                       "
    - "L1.100[198,199] 100ns                                                |L1.100|                                      "
    - "L1.101[200,201] 101ns                                                 |L1.101|                                     "
    - "L1.102[202,203] 102ns                                                 |L1.102|                                     "
    - "L1.103[204,205] 103ns                                                  |L1.103|                                    "
    - "L1.104[206,207] 104ns                                                  |L1.104|                                    "
    - "L1.105[208,209] 105ns                                                  |L1.105|                                    "
    - "L1.106[210,211] 106ns                                                   |L1.106|                                   "
    - "L1.107[212,213] 107ns                                                   |L1.107|                                   "
    - "L1.108[214,215] 108ns                                                    |L1.108|                                  "
    - "L1.109[216,217] 109ns                                                    |L1.109|                                  "
    - "L1.110[218,219] 110ns                                                     |L1.110|                                 "
    - "L1.111[220,221] 111ns                                                     |L1.111|                                 "
    - "L1.112[222,223] 112ns                                                      |L1.112|                                "
    - "L1.113[224,225] 113ns                                                      |L1.113|                                "
    - "L1.114[226,227] 114ns                                                      |L1.114|                                "
    - "L1.115[228,229] 115ns                                                       |L1.115|                               "
    - "L1.116[230,231] 116ns                                                       |L1.116|                               "
    - "L1.117[232,233] 117ns                                                        |L1.117|                              "
    - "L1.118[234,235] 118ns                                                        |L1.118|                              "
    - "L1.119[236,237] 119ns                                                         |L1.119|                             "
    - "L1.120[238,239] 120ns                                                         |L1.120|                             "
    - "L1.121[240,241] 121ns                                                          |L1.121|                            "
    - "L1.122[242,243] 122ns                                                          |L1.122|                            "
    - "L1.123[244,245] 123ns                                                           |L1.123|                           "
    - "L1.124[246,247] 124ns                                                           |L1.124|                           "
    - "L1.125[248,249] 125ns                                                           |L1.125|                           "
    - "L1.126[250,251] 126ns                                                            |L1.126|                          "
    - "L1.127[252,253] 127ns                                                            |L1.127|                          "
    - "L1.128[254,255] 128ns                                                             |L1.128|                         "
    - "L1.129[256,257] 129ns                                                             |L1.129|                         "
    - "L1.130[258,259] 130ns                                                              |L1.130|                        "
    - "L1.131[260,261] 131ns                                                              |L1.131|                        "
    - "L1.132[262,263] 132ns                                                               |L1.132|                       "
    - "L1.133[264,265] 133ns                                                               |L1.133|                       "
    - "L1.134[266,267] 134ns                                                                |L1.134|                      "
    - "L1.135[268,269] 135ns                                                                |L1.135|                      "
    - "L1.136[270,271] 136ns                                                                |L1.136|                      "
    - "L1.137[272,273] 137ns                                                                 |L1.137|                     "
    - "L1.138[274,275] 138ns                                                                 |L1.138|                     "
    - "L1.139[276,277] 139ns                                                                  |L1.139|                    "
    - "L1.140[278,279] 140ns                                                                  |L1.140|                    "
    - "L1.141[280,281] 141ns                                                                   |L1.141|                   "
    - "L1.142[282,283] 142ns                                                                   |L1.142|                   "
    - "L1.143[284,285] 143ns                                                                    |L1.143|                  "
    - "L1.144[286,287] 144ns                                                                    |L1.144|                  "
    - "L1.145[288,289] 145ns                                                                    |L1.145|                  "
    - "L1.146[290,291] 146ns                                                                     |L1.146|                 "
    - "L1.147[292,293] 147ns                                                                     |L1.147|                 "
    - "L1.148[294,295] 148ns                                                                      |L1.148|                "
    - "L1.149[296,297] 149ns                                                                      |L1.149|                "
    - "L1.150[298,299] 150ns                                                                       |L1.150|               "
    - "L1.151[300,301] 151ns                                                                       |L1.151|               "
    - "L1.152[302,303] 152ns                                                                        |L1.152|              "
    - "L1.153[304,305] 153ns                                                                        |L1.153|              "
    - "L1.154[306,307] 154ns                                                                         |L1.154|             "
    - "L1.155[308,309] 155ns                                                                         |L1.155|             "
    - "L1.156[310,311] 156ns                                                                         |L1.156|             "
    - "L1.157[312,313] 157ns                                                                          |L1.157|            "
    - "L1.158[314,315] 158ns                                                                          |L1.158|            "
    - "L1.159[316,317] 159ns                                                                           |L1.159|           "
    - "L1.160[318,319] 160ns                                                                           |L1.160|           "
    - "L1.161[320,321] 161ns                                                                            |L1.161|          "
    - "L1.162[322,323] 162ns                                                                            |L1.162|          "
    - "L1.163[324,325] 163ns                                                                             |L1.163|         "
    - "L1.164[326,327] 164ns                                                                             |L1.164|         "
    - "L1.165[328,329] 165ns                                                                             |L1.165|         "
    - "L1.166[330,331] 166ns                                                                              |L1.166|        "
    - "L1.167[332,333] 167ns                                                                              |L1.167|        "
    - "L1.168[334,335] 168ns                                                                               |L1.168|       "
    - "L1.169[336,337] 169ns                                                                               |L1.169|       "
    - "L1.170[338,339] 170ns                                                                                |L1.170|      "
    - "L1.171[340,341] 171ns                                                                                |L1.171|      "
    - "L1.172[342,343] 172ns                                                                                 |L1.172|     "
    - "L1.173[344,345] 173ns                                                                                 |L1.173|     "
    - "L1.174[346,347] 174ns                                                                                  |L1.174|    "
    - "L1.175[348,349] 175ns                                                                                  |L1.175|    "
    - "L1.176[350,351] 176ns                                                                                  |L1.176|    "
    - "L1.177[352,353] 177ns                                                                                   |L1.177|   "
    - "L1.178[354,355] 178ns                                                                                   |L1.178|   "
    - "L1.179[356,357] 179ns                                                                                    |L1.179|  "
    - "L1.180[358,359] 180ns                                                                                    |L1.180|  "
    - "L1.181[360,361] 181ns                                                                                     |L1.181| "
    - "L1.182[362,363] 182ns                                                                                     |L1.182| "
    - "L1.183[364,365] 183ns                                                                                      |L1.183|"
    - "L1.184[366,367] 184ns                                                                                      |L1.184|"
    - "L1.185[368,369] 185ns                                                                                       |L1.185|"
    - "L1.186[370,371] 186ns                                                                                       |L1.186|"
    - "L1.187[372,373] 187ns                                                                                       |L1.187|"
    - "L1.188[374,375] 188ns                                                                                        |L1.188|"
    - "L1.189[376,377] 189ns                                                                                        |L1.189|"
    - "L1.190[378,379] 190ns                                                                                         |L1.190|"
    - "L1.191[380,381] 191ns                                                                                         |L1.191|"
    - "L1.192[382,383] 192ns                                                                                          |L1.192|"
    - "L1.193[384,385] 193ns                                                                                          |L1.193|"
    - "L1.194[386,387] 194ns                                                                                           |L1.194|"
    - "L1.195[388,389] 195ns                                                                                           |L1.195|"
    - "L1.196[390,391] 196ns                                                                                           |L1.196|"
    - "L1.197[392,393] 197ns                                                                                            |L1.197|"
    - "L1.198[394,395] 198ns                                                                                            |L1.198|"
    - "L1.199[396,397] 199ns                                                                                             |L1.199|"
    - "L1.200[398,399] 200ns                                                                                             |L1.200|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 1.37mb total:"
    - "L1, all files 1.37mb                                                                                               "
    - "L1.?[0,399] 200ns        |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 200 files: L1.1, L1.2, L1.3, L1.4, L1.5, L1.6, L1.7, L1.8, L1.9, L1.10, L1.11, L1.12, L1.13, L1.14, L1.15, L1.16, L1.17, L1.18, L1.19, L1.20, L1.21, L1.22, L1.23, L1.24, L1.25, L1.26, L1.27, L1.28, L1.29, L1.30, L1.31, L1.32, L1.33, L1.34, L1.35, L1.36, L1.37, L1.38, L1.39, L1.40, L1.41, L1.42, L1.43, L1.44, L1.45, L1.46, L1.47, L1.48, L1.49, L1.50, L1.51, L1.52, L1.53, L1.54, L1.55, L1.56, L1.57, L1.58, L1.59, L1.60, L1.61, L1.62, L1.63, L1.64, L1.65, L1.66, L1.67, L1.68, L1.69, L1.70, L1.71, L1.72, L1.73, L1.74, L1.75, L1.76, L1.77, L1.78, L1.79, L1.80, L1.81, L1.82, L1.83, L1.84, L1.85, L1.86, L1.87, L1.88, L1.89, L1.90, L1.91, L1.92, L1.93, L1.94, L1.95, L1.96, L1.97, L1.98, L1.99, L1.100, L1.101, L1.102, L1.103, L1.104, L1.105, L1.106, L1.107, L1.108, L1.109, L1.110, L1.111, L1.112, L1.113, L1.114, L1.115, L1.116, L1.117, L1.118, L1.119, L1.120, L1.121, L1.122, L1.123, L1.124, L1.125, L1.126, L1.127, L1.128, L1.129, L1.130, L1.131, L1.132, L1.133, L1.134, L1.135, L1.136, L1.137, L1.138, L1.139, L1.140, L1.141, L1.142, L1.143, L1.144, L1.145, L1.146, L1.147, L1.148, L1.149, L1.150, L1.151, L1.152, L1.153, L1.154, L1.155, L1.156, L1.157, L1.158, L1.159, L1.160, L1.161, L1.162, L1.163, L1.164, L1.165, L1.166, L1.167, L1.168, L1.169, L1.170, L1.171, L1.172, L1.173, L1.174, L1.175, L1.176, L1.177, L1.178, L1.179, L1.180, L1.181, L1.182, L1.183, L1.184, L1.185, L1.186, L1.187, L1.188, L1.189, L1.190, L1.191, L1.192, L1.193, L1.194, L1.195, L1.196, L1.197, L1.198, L1.199, L1.200"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=compact(ManySmallFiles). 88 Input Files, 616kb total:"
    - "L1, all files 7kb                                                                                                  "
    - "L1.201[400,401] 201ns    |L1.201|                                                                                  "
    - "L1.202[402,403] 202ns     |L1.202|                                                                                 "
    - "L1.203[404,405] 203ns      |L1.203|                                                                                "
    - "L1.204[406,407] 204ns       |L1.204|                                                                               "
    - "L1.205[408,409] 205ns        |L1.205|                                                                              "
    - "L1.206[410,411] 206ns         |L1.206|                                                                             "
    - "L1.207[412,413] 207ns          |L1.207|                                                                            "
    - "L1.208[414,415] 208ns           |L1.208|                                                                           "
    - "L1.209[416,417] 209ns            |L1.209|                                                                          "
    - "L1.210[418,419] 210ns             |L1.210|                                                                         "
    - "L1.211[420,421] 211ns              |L1.211|                                                                        "
    - "L1.212[422,423] 212ns               |L1.212|                                                                       "
    - "L1.213[424,425] 213ns                |L1.213|                                                                      "
    - "L1.214[426,427] 214ns                 |L1.214|                                                                     "
    - "L1.215[428,429] 215ns                  |L1.215|                                                                    "
    - "L1.216[430,431] 216ns                   |L1.216|                                                                   "
    - "L1.217[432,433] 217ns                    |L1.217|                                                                  "
    - "L1.218[434,435] 218ns                     |L1.218|                                                                 "
    - "L1.219[436,437] 219ns                      |L1.219|                                                                "
    - "L1.220[438,439] 220ns                       |L1.220|                                                               "
    - "L1.221[440,441] 221ns                        |L1.221|                                                              "
    - "L1.222[442,443] 222ns                         |L1.222|                                                             "
    - "L1.223[444,445] 223ns                          |L1.223|                                                            "
    - "L1.224[446,447] 224ns                           |L1.224|                                                           "
    - "L1.225[448,449] 225ns                            |L1.225|                                                          "
    - "L1.226[450,451] 226ns                             |L1.226|                                                         "
    - "L1.227[452,453] 227ns                              |L1.227|                                                        "
    - "L1.228[454,455] 228ns                               |L1.228|                                                       "
    - "L1.229[456,457] 229ns                                |L1.229|                                                      "
    - "L1.230[458,459] 230ns                                 |L1.230|                                                     "
    - "L1.231[460,461] 231ns                                  |L1.231|                                                    "
    - "L1.232[462,463] 232ns                                   |L1.232|                                                   "
    - "L1.233[464,465] 233ns                                    |L1.233|                                                  "
    - "L1.234[466,467] 234ns                                     |L1.234|                                                 "
    - "L1.235[468,469] 235ns                                      |L1.235|                                                "
    - "L1.236[470,471] 236ns                                        |L1.236|                                              "
    - "L1.237[472,473] 237ns                                         |L1.237|                                             "
    - "L1.238[474,475] 238ns                                          |L1.238|                                            "
    - "L1.239[476,477] 239ns                                           |L1.239|                                           "
    - "L1.240[478,479] 240ns                                            |L1.240|                                          "
    - "L1.241[480,481] 241ns                                             |L1.241|                                         "
    - "L1.242[482,483] 242ns                                              |L1.242|                                        "
    - "L1.243[484,485] 243ns                                               |L1.243|                                       "
    - "L1.244[486,487] 244ns                                                |L1.244|                                      "
    - "L1.245[488,489] 245ns                                                 |L1.245|                                     "
    - "L1.246[490,491] 246ns                                                  |L1.246|                                    "
    - "L1.247[492,493] 247ns                                                   |L1.247|                                   "
    - "L1.248[494,495] 248ns                                                    |L1.248|                                  "
    - "L1.249[496,497] 249ns                                                     |L1.249|                                 "
    - "L1.250[498,499] 250ns                                                      |L1.250|                                "
    - "L1.251[500,501] 251ns                                                       |L1.251|                               "
    - "L1.252[502,503] 252ns                                                        |L1.252|                              "
    - "L1.253[504,505] 253ns                                                         |L1.253|                             "
    - "L1.254[506,507] 254ns                                                          |L1.254|                            "
    - "L1.255[508,509] 255ns                                                           |L1.255|                           "
    - "L1.256[510,511] 256ns                                                            |L1.256|                          "
    - "L1.257[512,513] 257ns                                                             |L1.257|                         "
    - "L1.258[514,515] 258ns                                                              |L1.258|                        "
    - "L1.259[516,517] 259ns                                                               |L1.259|                       "
    - "L1.260[518,519] 260ns                                                                |L1.260|                      "
    - "L1.261[520,521] 261ns                                                                 |L1.261|                     "
    - "L1.262[522,523] 262ns                                                                  |L1.262|                    "
    - "L1.263[524,525] 263ns                                                                   |L1.263|                   "
    - "L1.264[526,527] 264ns                                                                    |L1.264|                  "
    - "L1.265[528,529] 265ns                                                                     |L1.265|                 "
    - "L1.266[530,531] 266ns                                                                      |L1.266|                "
    - "L1.267[532,533] 267ns                                                                       |L1.267|               "
    - "L1.268[534,535] 268ns                                                                        |L1.268|              "
    - "L1.269[536,537] 269ns                                                                         |L1.269|             "
    - "L1.270[538,539] 270ns                                                                          |L1.270|            "
    - "L1.271[540,541] 271ns                                                                            |L1.271|          "
    - "L1.272[542,543] 272ns                                                                             |L1.272|         "
    - "L1.273[544,545] 273ns                                                                              |L1.273|        "
    - "L1.274[546,547] 274ns                                                                               |L1.274|       "
    - "L1.275[548,549] 275ns                                                                                |L1.275|      "
    - "L1.276[550,551] 276ns                                                                                 |L1.276|     "
    - "L1.277[552,553] 277ns                                                                                  |L1.277|    "
    - "L1.278[554,555] 278ns                                                                                   |L1.278|   "
    - "L1.279[556,557] 279ns                                                                                    |L1.279|  "
    - "L1.280[558,559] 280ns                                                                                     |L1.280| "
    - "L1.281[560,561] 281ns                                                                                      |L1.281|"
    - "L1.282[562,563] 282ns                                                                                       |L1.282|"
    - "L1.283[564,565] 283ns                                                                                        |L1.283|"
    - "L1.284[566,567] 284ns                                                                                         |L1.284|"
    - "L1.285[568,569] 285ns                                                                                          |L1.285|"
    - "L1.286[570,571] 286ns                                                                                           |L1.286|"
    - "L1.287[572,573] 287ns                                                                                            |L1.287|"
    - "L1.288[574,575] 288ns                                                                                             |L1.288|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 616kb total:"
    - "L1, all files 616kb                                                                                                "
    - "L1.?[400,575] 288ns      |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 88 files: L1.201, L1.202, L1.203, L1.204, L1.205, L1.206, L1.207, L1.208, L1.209, L1.210, L1.211, L1.212, L1.213, L1.214, L1.215, L1.216, L1.217, L1.218, L1.219, L1.220, L1.221, L1.222, L1.223, L1.224, L1.225, L1.226, L1.227, L1.228, L1.229, L1.230, L1.231, L1.232, L1.233, L1.234, L1.235, L1.236, L1.237, L1.238, L1.239, L1.240, L1.241, L1.242, L1.243, L1.244, L1.245, L1.246, L1.247, L1.248, L1.249, L1.250, L1.251, L1.252, L1.253, L1.254, L1.255, L1.256, L1.257, L1.258, L1.259, L1.260, L1.261, L1.262, L1.263, L1.264, L1.265, L1.266, L1.267, L1.268, L1.269, L1.270, L1.271, L1.272, L1.273, L1.274, L1.275, L1.276, L1.277, L1.278, L1.279, L1.280, L1.281, L1.282, L1.283, L1.284, L1.285, L1.286, L1.287, L1.288"
    - "  Creating 1 files"
    - "**** Final Output Files (1.97mb written)"
    - "L1                                                                                                                 "
    - "L1.289[0,399] 200ns 1.37mb|---------------------------L1.289---------------------------|                            "
    - "L1.290[400,575] 288ns 616kb                                                              |---------L1.290----------| "
    "###
    );
}

#[tokio::test]
async fn many_l0_and_overlapped_l1_files() {
    test_helpers::maybe_start_logging();

    // Observed size on production
    let seven_kb = 7 * 1024;
    let one_mb = 1024 * 1024;

    let setup = layout_setup_builder().await.build().await;

    // create 190 L0 files & 20 L1 files
    // The rationale here is that the number of files in each level
    // is lower than the limit of 200 files per plan
    // however, since the compactor may, in the worst case, compact all files
    // together in two adjacent levels in a single plan, ensure it can handle this case
    let num_tiny_l0_files = 190;
    let num_l1_files = 20;

    for i in 0..num_tiny_l0_files {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i)
                    .with_max_time(i + 1)
                    .with_compaction_level(CompactionLevel::Initial)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(num_l1_files + i + 1))
                    .with_file_size_bytes(seven_kb),
            )
            .await;
    }

    // create 20 L1 files that overlap with the L0 files
    for i in 0..num_l1_files {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i * 10)
                    .with_max_time(i * 10 + 9)
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    // the content of L1s created before L0s
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(one_mb),
            )
            .await;
    }

    // Compact 190 L0 files into one L0 file
    // Then compact the result L0 file with overlapped L1 files into two L1 files
    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0                                                                                                                 "
    - "L0.1[0,1] 21ns 7kb       |L0.1|                                                                                    "
    - "L0.2[1,2] 22ns 7kb       |L0.2|                                                                                    "
    - "L0.3[2,3] 23ns 7kb       |L0.3|                                                                                    "
    - "L0.4[3,4] 24ns 7kb        |L0.4|                                                                                   "
    - "L0.5[4,5] 25ns 7kb        |L0.5|                                                                                   "
    - "L0.6[5,6] 26ns 7kb         |L0.6|                                                                                  "
    - "L0.7[6,7] 27ns 7kb         |L0.7|                                                                                  "
    - "L0.8[7,8] 28ns 7kb          |L0.8|                                                                                 "
    - "L0.9[8,9] 29ns 7kb          |L0.9|                                                                                 "
    - "L0.10[9,10] 30ns 7kb         |L0.10|                                                                               "
    - "L0.11[10,11] 31ns 7kb        |L0.11|                                                                               "
    - "L0.12[11,12] 32ns 7kb        |L0.12|                                                                               "
    - "L0.13[12,13] 33ns 7kb         |L0.13|                                                                              "
    - "L0.14[13,14] 34ns 7kb         |L0.14|                                                                              "
    - "L0.15[14,15] 35ns 7kb          |L0.15|                                                                             "
    - "L0.16[15,16] 36ns 7kb          |L0.16|                                                                             "
    - "L0.17[16,17] 37ns 7kb           |L0.17|                                                                            "
    - "L0.18[17,18] 38ns 7kb           |L0.18|                                                                            "
    - "L0.19[18,19] 39ns 7kb            |L0.19|                                                                           "
    - "L0.20[19,20] 40ns 7kb            |L0.20|                                                                           "
    - "L0.21[20,21] 41ns 7kb             |L0.21|                                                                          "
    - "L0.22[21,22] 42ns 7kb             |L0.22|                                                                          "
    - "L0.23[22,23] 43ns 7kb             |L0.23|                                                                          "
    - "L0.24[23,24] 44ns 7kb              |L0.24|                                                                         "
    - "L0.25[24,25] 45ns 7kb              |L0.25|                                                                         "
    - "L0.26[25,26] 46ns 7kb               |L0.26|                                                                        "
    - "L0.27[26,27] 47ns 7kb               |L0.27|                                                                        "
    - "L0.28[27,28] 48ns 7kb                |L0.28|                                                                       "
    - "L0.29[28,29] 49ns 7kb                |L0.29|                                                                       "
    - "L0.30[29,30] 50ns 7kb                 |L0.30|                                                                      "
    - "L0.31[30,31] 51ns 7kb                 |L0.31|                                                                      "
    - "L0.32[31,32] 52ns 7kb                  |L0.32|                                                                     "
    - "L0.33[32,33] 53ns 7kb                  |L0.33|                                                                     "
    - "L0.34[33,34] 54ns 7kb                  |L0.34|                                                                     "
    - "L0.35[34,35] 55ns 7kb                   |L0.35|                                                                    "
    - "L0.36[35,36] 56ns 7kb                   |L0.36|                                                                    "
    - "L0.37[36,37] 57ns 7kb                    |L0.37|                                                                   "
    - "L0.38[37,38] 58ns 7kb                    |L0.38|                                                                   "
    - "L0.39[38,39] 59ns 7kb                     |L0.39|                                                                  "
    - "L0.40[39,40] 60ns 7kb                     |L0.40|                                                                  "
    - "L0.41[40,41] 61ns 7kb                      |L0.41|                                                                 "
    - "L0.42[41,42] 62ns 7kb                      |L0.42|                                                                 "
    - "L0.43[42,43] 63ns 7kb                      |L0.43|                                                                 "
    - "L0.44[43,44] 64ns 7kb                       |L0.44|                                                                "
    - "L0.45[44,45] 65ns 7kb                       |L0.45|                                                                "
    - "L0.46[45,46] 66ns 7kb                        |L0.46|                                                               "
    - "L0.47[46,47] 67ns 7kb                        |L0.47|                                                               "
    - "L0.48[47,48] 68ns 7kb                         |L0.48|                                                              "
    - "L0.49[48,49] 69ns 7kb                         |L0.49|                                                              "
    - "L0.50[49,50] 70ns 7kb                          |L0.50|                                                             "
    - "L0.51[50,51] 71ns 7kb                          |L0.51|                                                             "
    - "L0.52[51,52] 72ns 7kb                           |L0.52|                                                            "
    - "L0.53[52,53] 73ns 7kb                           |L0.53|                                                            "
    - "L0.54[53,54] 74ns 7kb                           |L0.54|                                                            "
    - "L0.55[54,55] 75ns 7kb                            |L0.55|                                                           "
    - "L0.56[55,56] 76ns 7kb                            |L0.56|                                                           "
    - "L0.57[56,57] 77ns 7kb                             |L0.57|                                                          "
    - "L0.58[57,58] 78ns 7kb                             |L0.58|                                                          "
    - "L0.59[58,59] 79ns 7kb                              |L0.59|                                                         "
    - "L0.60[59,60] 80ns 7kb                              |L0.60|                                                         "
    - "L0.61[60,61] 81ns 7kb                               |L0.61|                                                        "
    - "L0.62[61,62] 82ns 7kb                               |L0.62|                                                        "
    - "L0.63[62,63] 83ns 7kb                                |L0.63|                                                       "
    - "L0.64[63,64] 84ns 7kb                                |L0.64|                                                       "
    - "L0.65[64,65] 85ns 7kb                                |L0.65|                                                       "
    - "L0.66[65,66] 86ns 7kb                                 |L0.66|                                                      "
    - "L0.67[66,67] 87ns 7kb                                 |L0.67|                                                      "
    - "L0.68[67,68] 88ns 7kb                                  |L0.68|                                                     "
    - "L0.69[68,69] 89ns 7kb                                  |L0.69|                                                     "
    - "L0.70[69,70] 90ns 7kb                                   |L0.70|                                                    "
    - "L0.71[70,71] 91ns 7kb                                   |L0.71|                                                    "
    - "L0.72[71,72] 92ns 7kb                                    |L0.72|                                                   "
    - "L0.73[72,73] 93ns 7kb                                    |L0.73|                                                   "
    - "L0.74[73,74] 94ns 7kb                                     |L0.74|                                                  "
    - "L0.75[74,75] 95ns 7kb                                     |L0.75|                                                  "
    - "L0.76[75,76] 96ns 7kb                                     |L0.76|                                                  "
    - "L0.77[76,77] 97ns 7kb                                      |L0.77|                                                 "
    - "L0.78[77,78] 98ns 7kb                                      |L0.78|                                                 "
    - "L0.79[78,79] 99ns 7kb                                       |L0.79|                                                "
    - "L0.80[79,80] 100ns 7kb                                      |L0.80|                                                "
    - "L0.81[80,81] 101ns 7kb                                       |L0.81|                                               "
    - "L0.82[81,82] 102ns 7kb                                       |L0.82|                                               "
    - "L0.83[82,83] 103ns 7kb                                        |L0.83|                                              "
    - "L0.84[83,84] 104ns 7kb                                        |L0.84|                                              "
    - "L0.85[84,85] 105ns 7kb                                        |L0.85|                                              "
    - "L0.86[85,86] 106ns 7kb                                         |L0.86|                                             "
    - "L0.87[86,87] 107ns 7kb                                         |L0.87|                                             "
    - "L0.88[87,88] 108ns 7kb                                          |L0.88|                                            "
    - "L0.89[88,89] 109ns 7kb                                          |L0.89|                                            "
    - "L0.90[89,90] 110ns 7kb                                           |L0.90|                                           "
    - "L0.91[90,91] 111ns 7kb                                           |L0.91|                                           "
    - "L0.92[91,92] 112ns 7kb                                            |L0.92|                                          "
    - "L0.93[92,93] 113ns 7kb                                            |L0.93|                                          "
    - "L0.94[93,94] 114ns 7kb                                             |L0.94|                                         "
    - "L0.95[94,95] 115ns 7kb                                             |L0.95|                                         "
    - "L0.96[95,96] 116ns 7kb                                             |L0.96|                                         "
    - "L0.97[96,97] 117ns 7kb                                              |L0.97|                                        "
    - "L0.98[97,98] 118ns 7kb                                              |L0.98|                                        "
    - "L0.99[98,99] 119ns 7kb                                               |L0.99|                                       "
    - "L0.100[99,100] 120ns 7kb                                             |L0.100|                                      "
    - "L0.101[100,101] 121ns 7kb                                             |L0.101|                                     "
    - "L0.102[101,102] 122ns 7kb                                             |L0.102|                                     "
    - "L0.103[102,103] 123ns 7kb                                              |L0.103|                                    "
    - "L0.104[103,104] 124ns 7kb                                              |L0.104|                                    "
    - "L0.105[104,105] 125ns 7kb                                               |L0.105|                                   "
    - "L0.106[105,106] 126ns 7kb                                               |L0.106|                                   "
    - "L0.107[106,107] 127ns 7kb                                               |L0.107|                                   "
    - "L0.108[107,108] 128ns 7kb                                                |L0.108|                                  "
    - "L0.109[108,109] 129ns 7kb                                                |L0.109|                                  "
    - "L0.110[109,110] 130ns 7kb                                                 |L0.110|                                 "
    - "L0.111[110,111] 131ns 7kb                                                 |L0.111|                                 "
    - "L0.112[111,112] 132ns 7kb                                                  |L0.112|                                "
    - "L0.113[112,113] 133ns 7kb                                                  |L0.113|                                "
    - "L0.114[113,114] 134ns 7kb                                                   |L0.114|                               "
    - "L0.115[114,115] 135ns 7kb                                                   |L0.115|                               "
    - "L0.116[115,116] 136ns 7kb                                                    |L0.116|                              "
    - "L0.117[116,117] 137ns 7kb                                                    |L0.117|                              "
    - "L0.118[117,118] 138ns 7kb                                                    |L0.118|                              "
    - "L0.119[118,119] 139ns 7kb                                                     |L0.119|                             "
    - "L0.120[119,120] 140ns 7kb                                                     |L0.120|                             "
    - "L0.121[120,121] 141ns 7kb                                                      |L0.121|                            "
    - "L0.122[121,122] 142ns 7kb                                                      |L0.122|                            "
    - "L0.123[122,123] 143ns 7kb                                                       |L0.123|                           "
    - "L0.124[123,124] 144ns 7kb                                                       |L0.124|                           "
    - "L0.125[124,125] 145ns 7kb                                                        |L0.125|                          "
    - "L0.126[125,126] 146ns 7kb                                                        |L0.126|                          "
    - "L0.127[126,127] 147ns 7kb                                                        |L0.127|                          "
    - "L0.128[127,128] 148ns 7kb                                                         |L0.128|                         "
    - "L0.129[128,129] 149ns 7kb                                                         |L0.129|                         "
    - "L0.130[129,130] 150ns 7kb                                                          |L0.130|                        "
    - "L0.131[130,131] 151ns 7kb                                                          |L0.131|                        "
    - "L0.132[131,132] 152ns 7kb                                                           |L0.132|                       "
    - "L0.133[132,133] 153ns 7kb                                                           |L0.133|                       "
    - "L0.134[133,134] 154ns 7kb                                                            |L0.134|                      "
    - "L0.135[134,135] 155ns 7kb                                                            |L0.135|                      "
    - "L0.136[135,136] 156ns 7kb                                                             |L0.136|                     "
    - "L0.137[136,137] 157ns 7kb                                                             |L0.137|                     "
    - "L0.138[137,138] 158ns 7kb                                                             |L0.138|                     "
    - "L0.139[138,139] 159ns 7kb                                                              |L0.139|                    "
    - "L0.140[139,140] 160ns 7kb                                                              |L0.140|                    "
    - "L0.141[140,141] 161ns 7kb                                                               |L0.141|                   "
    - "L0.142[141,142] 162ns 7kb                                                               |L0.142|                   "
    - "L0.143[142,143] 163ns 7kb                                                                |L0.143|                  "
    - "L0.144[143,144] 164ns 7kb                                                                |L0.144|                  "
    - "L0.145[144,145] 165ns 7kb                                                                 |L0.145|                 "
    - "L0.146[145,146] 166ns 7kb                                                                 |L0.146|                 "
    - "L0.147[146,147] 167ns 7kb                                                                  |L0.147|                "
    - "L0.148[147,148] 168ns 7kb                                                                  |L0.148|                "
    - "L0.149[148,149] 169ns 7kb                                                                  |L0.149|                "
    - "L0.150[149,150] 170ns 7kb                                                                   |L0.150|               "
    - "L0.151[150,151] 171ns 7kb                                                                   |L0.151|               "
    - "L0.152[151,152] 172ns 7kb                                                                    |L0.152|              "
    - "L0.153[152,153] 173ns 7kb                                                                    |L0.153|              "
    - "L0.154[153,154] 174ns 7kb                                                                     |L0.154|             "
    - "L0.155[154,155] 175ns 7kb                                                                     |L0.155|             "
    - "L0.156[155,156] 176ns 7kb                                                                      |L0.156|            "
    - "L0.157[156,157] 177ns 7kb                                                                      |L0.157|            "
    - "L0.158[157,158] 178ns 7kb                                                                       |L0.158|           "
    - "L0.159[158,159] 179ns 7kb                                                                       |L0.159|           "
    - "L0.160[159,160] 180ns 7kb                                                                       |L0.160|           "
    - "L0.161[160,161] 181ns 7kb                                                                        |L0.161|          "
    - "L0.162[161,162] 182ns 7kb                                                                        |L0.162|          "
    - "L0.163[162,163] 183ns 7kb                                                                         |L0.163|         "
    - "L0.164[163,164] 184ns 7kb                                                                         |L0.164|         "
    - "L0.165[164,165] 185ns 7kb                                                                          |L0.165|        "
    - "L0.166[165,166] 186ns 7kb                                                                          |L0.166|        "
    - "L0.167[166,167] 187ns 7kb                                                                           |L0.167|       "
    - "L0.168[167,168] 188ns 7kb                                                                           |L0.168|       "
    - "L0.169[168,169] 189ns 7kb                                                                           |L0.169|       "
    - "L0.170[169,170] 190ns 7kb                                                                            |L0.170|      "
    - "L0.171[170,171] 191ns 7kb                                                                            |L0.171|      "
    - "L0.172[171,172] 192ns 7kb                                                                             |L0.172|     "
    - "L0.173[172,173] 193ns 7kb                                                                             |L0.173|     "
    - "L0.174[173,174] 194ns 7kb                                                                              |L0.174|    "
    - "L0.175[174,175] 195ns 7kb                                                                              |L0.175|    "
    - "L0.176[175,176] 196ns 7kb                                                                               |L0.176|   "
    - "L0.177[176,177] 197ns 7kb                                                                               |L0.177|   "
    - "L0.178[177,178] 198ns 7kb                                                                                |L0.178|  "
    - "L0.179[178,179] 199ns 7kb                                                                                |L0.179|  "
    - "L0.180[179,180] 200ns 7kb                                                                                |L0.180|  "
    - "L0.181[180,181] 201ns 7kb                                                                                 |L0.181| "
    - "L0.182[181,182] 202ns 7kb                                                                                 |L0.182| "
    - "L0.183[182,183] 203ns 7kb                                                                                  |L0.183|"
    - "L0.184[183,184] 204ns 7kb                                                                                  |L0.184|"
    - "L0.185[184,185] 205ns 7kb                                                                                   |L0.185|"
    - "L0.186[185,186] 206ns 7kb                                                                                   |L0.186|"
    - "L0.187[186,187] 207ns 7kb                                                                                    |L0.187|"
    - "L0.188[187,188] 208ns 7kb                                                                                    |L0.188|"
    - "L0.189[188,189] 209ns 7kb                                                                                     |L0.189|"
    - "L0.190[189,190] 210ns 7kb                                                                                     |L0.190|"
    - "L1                                                                                                                 "
    - "L1.191[0,9] 1ns 1mb      |L1.191|                                                                                  "
    - "L1.192[10,19] 2ns 1mb        |L1.192|                                                                              "
    - "L1.193[20,29] 3ns 1mb             |L1.193|                                                                         "
    - "L1.194[30,39] 4ns 1mb                 |L1.194|                                                                     "
    - "L1.195[40,49] 5ns 1mb                      |L1.195|                                                                "
    - "L1.196[50,59] 6ns 1mb                          |L1.196|                                                            "
    - "L1.197[60,69] 7ns 1mb                               |L1.197|                                                       "
    - "L1.198[70,79] 8ns 1mb                                   |L1.198|                                                   "
    - "L1.199[80,89] 9ns 1mb                                        |L1.199|                                              "
    - "L1.200[90,99] 10ns 1mb                                           |L1.200|                                          "
    - "L1.201[100,109] 11ns 1mb                                              |L1.201|                                     "
    - "L1.202[110,119] 12ns 1mb                                                  |L1.202|                                 "
    - "L1.203[120,129] 13ns 1mb                                                       |L1.203|                            "
    - "L1.204[130,139] 14ns 1mb                                                           |L1.204|                        "
    - "L1.205[140,149] 15ns 1mb                                                                |L1.205|                   "
    - "L1.206[150,159] 16ns 1mb                                                                    |L1.206|               "
    - "L1.207[160,169] 17ns 1mb                                                                         |L1.207|          "
    - "L1.208[170,179] 18ns 1mb                                                                             |L1.208|      "
    - "L1.209[180,189] 19ns 1mb                                                                                  |L1.209| "
    - "L1.210[190,199] 20ns 1mb                                                                                      |L1.210|"
    - "**** Simulation run 0, type=compact(ManySmallFiles). 190 Input Files, 1.3mb total:"
    - "L0, all files 7kb                                                                                                  "
    - "L0.1[0,1] 21ns           |L0.1|                                                                                    "
    - "L0.2[1,2] 22ns           |L0.2|                                                                                    "
    - "L0.3[2,3] 23ns           |L0.3|                                                                                    "
    - "L0.4[3,4] 24ns            |L0.4|                                                                                   "
    - "L0.5[4,5] 25ns            |L0.5|                                                                                   "
    - "L0.6[5,6] 26ns             |L0.6|                                                                                  "
    - "L0.7[6,7] 27ns             |L0.7|                                                                                  "
    - "L0.8[7,8] 28ns              |L0.8|                                                                                 "
    - "L0.9[8,9] 29ns              |L0.9|                                                                                 "
    - "L0.10[9,10] 30ns             |L0.10|                                                                               "
    - "L0.11[10,11] 31ns            |L0.11|                                                                               "
    - "L0.12[11,12] 32ns             |L0.12|                                                                              "
    - "L0.13[12,13] 33ns             |L0.13|                                                                              "
    - "L0.14[13,14] 34ns              |L0.14|                                                                             "
    - "L0.15[14,15] 35ns              |L0.15|                                                                             "
    - "L0.16[15,16] 36ns               |L0.16|                                                                            "
    - "L0.17[16,17] 37ns               |L0.17|                                                                            "
    - "L0.18[17,18] 38ns                |L0.18|                                                                           "
    - "L0.19[18,19] 39ns                |L0.19|                                                                           "
    - "L0.20[19,20] 40ns                 |L0.20|                                                                          "
    - "L0.21[20,21] 41ns                 |L0.21|                                                                          "
    - "L0.22[21,22] 42ns                 |L0.22|                                                                          "
    - "L0.23[22,23] 43ns                  |L0.23|                                                                         "
    - "L0.24[23,24] 44ns                  |L0.24|                                                                         "
    - "L0.25[24,25] 45ns                   |L0.25|                                                                        "
    - "L0.26[25,26] 46ns                   |L0.26|                                                                        "
    - "L0.27[26,27] 47ns                    |L0.27|                                                                       "
    - "L0.28[27,28] 48ns                    |L0.28|                                                                       "
    - "L0.29[28,29] 49ns                     |L0.29|                                                                      "
    - "L0.30[29,30] 50ns                     |L0.30|                                                                      "
    - "L0.31[30,31] 51ns                      |L0.31|                                                                     "
    - "L0.32[31,32] 52ns                      |L0.32|                                                                     "
    - "L0.33[32,33] 53ns                       |L0.33|                                                                    "
    - "L0.34[33,34] 54ns                       |L0.34|                                                                    "
    - "L0.35[34,35] 55ns                        |L0.35|                                                                   "
    - "L0.36[35,36] 56ns                        |L0.36|                                                                   "
    - "L0.37[36,37] 57ns                         |L0.37|                                                                  "
    - "L0.38[37,38] 58ns                         |L0.38|                                                                  "
    - "L0.39[38,39] 59ns                          |L0.39|                                                                 "
    - "L0.40[39,40] 60ns                          |L0.40|                                                                 "
    - "L0.41[40,41] 61ns                          |L0.41|                                                                 "
    - "L0.42[41,42] 62ns                           |L0.42|                                                                "
    - "L0.43[42,43] 63ns                           |L0.43|                                                                "
    - "L0.44[43,44] 64ns                            |L0.44|                                                               "
    - "L0.45[44,45] 65ns                            |L0.45|                                                               "
    - "L0.46[45,46] 66ns                             |L0.46|                                                              "
    - "L0.47[46,47] 67ns                             |L0.47|                                                              "
    - "L0.48[47,48] 68ns                              |L0.48|                                                             "
    - "L0.49[48,49] 69ns                              |L0.49|                                                             "
    - "L0.50[49,50] 70ns                               |L0.50|                                                            "
    - "L0.51[50,51] 71ns                               |L0.51|                                                            "
    - "L0.52[51,52] 72ns                                |L0.52|                                                           "
    - "L0.53[52,53] 73ns                                |L0.53|                                                           "
    - "L0.54[53,54] 74ns                                 |L0.54|                                                          "
    - "L0.55[54,55] 75ns                                 |L0.55|                                                          "
    - "L0.56[55,56] 76ns                                  |L0.56|                                                         "
    - "L0.57[56,57] 77ns                                  |L0.57|                                                         "
    - "L0.58[57,58] 78ns                                   |L0.58|                                                        "
    - "L0.59[58,59] 79ns                                   |L0.59|                                                        "
    - "L0.60[59,60] 80ns                                   |L0.60|                                                        "
    - "L0.61[60,61] 81ns                                    |L0.61|                                                       "
    - "L0.62[61,62] 82ns                                    |L0.62|                                                       "
    - "L0.63[62,63] 83ns                                     |L0.63|                                                      "
    - "L0.64[63,64] 84ns                                     |L0.64|                                                      "
    - "L0.65[64,65] 85ns                                      |L0.65|                                                     "
    - "L0.66[65,66] 86ns                                      |L0.66|                                                     "
    - "L0.67[66,67] 87ns                                       |L0.67|                                                    "
    - "L0.68[67,68] 88ns                                       |L0.68|                                                    "
    - "L0.69[68,69] 89ns                                        |L0.69|                                                   "
    - "L0.70[69,70] 90ns                                        |L0.70|                                                   "
    - "L0.71[70,71] 91ns                                         |L0.71|                                                  "
    - "L0.72[71,72] 92ns                                         |L0.72|                                                  "
    - "L0.73[72,73] 93ns                                          |L0.73|                                                 "
    - "L0.74[73,74] 94ns                                          |L0.74|                                                 "
    - "L0.75[74,75] 95ns                                           |L0.75|                                                "
    - "L0.76[75,76] 96ns                                           |L0.76|                                                "
    - "L0.77[76,77] 97ns                                            |L0.77|                                               "
    - "L0.78[77,78] 98ns                                            |L0.78|                                               "
    - "L0.79[78,79] 99ns                                            |L0.79|                                               "
    - "L0.80[79,80] 100ns                                            |L0.80|                                              "
    - "L0.81[80,81] 101ns                                            |L0.81|                                              "
    - "L0.82[81,82] 102ns                                             |L0.82|                                             "
    - "L0.83[82,83] 103ns                                             |L0.83|                                             "
    - "L0.84[83,84] 104ns                                              |L0.84|                                            "
    - "L0.85[84,85] 105ns                                              |L0.85|                                            "
    - "L0.86[85,86] 106ns                                               |L0.86|                                           "
    - "L0.87[86,87] 107ns                                               |L0.87|                                           "
    - "L0.88[87,88] 108ns                                                |L0.88|                                          "
    - "L0.89[88,89] 109ns                                                |L0.89|                                          "
    - "L0.90[89,90] 110ns                                                 |L0.90|                                         "
    - "L0.91[90,91] 111ns                                                 |L0.91|                                         "
    - "L0.92[91,92] 112ns                                                  |L0.92|                                        "
    - "L0.93[92,93] 113ns                                                  |L0.93|                                        "
    - "L0.94[93,94] 114ns                                                   |L0.94|                                       "
    - "L0.95[94,95] 115ns                                                   |L0.95|                                       "
    - "L0.96[95,96] 116ns                                                    |L0.96|                                      "
    - "L0.97[96,97] 117ns                                                    |L0.97|                                      "
    - "L0.98[97,98] 118ns                                                    |L0.98|                                      "
    - "L0.99[98,99] 119ns                                                     |L0.99|                                     "
    - "L0.100[99,100] 120ns                                                   |L0.100|                                    "
    - "L0.101[100,101] 121ns                                                   |L0.101|                                   "
    - "L0.102[101,102] 122ns                                                   |L0.102|                                   "
    - "L0.103[102,103] 123ns                                                    |L0.103|                                  "
    - "L0.104[103,104] 124ns                                                    |L0.104|                                  "
    - "L0.105[104,105] 125ns                                                     |L0.105|                                 "
    - "L0.106[105,106] 126ns                                                     |L0.106|                                 "
    - "L0.107[106,107] 127ns                                                      |L0.107|                                "
    - "L0.108[107,108] 128ns                                                      |L0.108|                                "
    - "L0.109[108,109] 129ns                                                       |L0.109|                               "
    - "L0.110[109,110] 130ns                                                       |L0.110|                               "
    - "L0.111[110,111] 131ns                                                        |L0.111|                              "
    - "L0.112[111,112] 132ns                                                        |L0.112|                              "
    - "L0.113[112,113] 133ns                                                         |L0.113|                             "
    - "L0.114[113,114] 134ns                                                         |L0.114|                             "
    - "L0.115[114,115] 135ns                                                          |L0.115|                            "
    - "L0.116[115,116] 136ns                                                          |L0.116|                            "
    - "L0.117[116,117] 137ns                                                          |L0.117|                            "
    - "L0.118[117,118] 138ns                                                           |L0.118|                           "
    - "L0.119[118,119] 139ns                                                           |L0.119|                           "
    - "L0.120[119,120] 140ns                                                            |L0.120|                          "
    - "L0.121[120,121] 141ns                                                            |L0.121|                          "
    - "L0.122[121,122] 142ns                                                             |L0.122|                         "
    - "L0.123[122,123] 143ns                                                             |L0.123|                         "
    - "L0.124[123,124] 144ns                                                              |L0.124|                        "
    - "L0.125[124,125] 145ns                                                              |L0.125|                        "
    - "L0.126[125,126] 146ns                                                               |L0.126|                       "
    - "L0.127[126,127] 147ns                                                               |L0.127|                       "
    - "L0.128[127,128] 148ns                                                                |L0.128|                      "
    - "L0.129[128,129] 149ns                                                                |L0.129|                      "
    - "L0.130[129,130] 150ns                                                                 |L0.130|                     "
    - "L0.131[130,131] 151ns                                                                 |L0.131|                     "
    - "L0.132[131,132] 152ns                                                                  |L0.132|                    "
    - "L0.133[132,133] 153ns                                                                  |L0.133|                    "
    - "L0.134[133,134] 154ns                                                                   |L0.134|                   "
    - "L0.135[134,135] 155ns                                                                   |L0.135|                   "
    - "L0.136[135,136] 156ns                                                                   |L0.136|                   "
    - "L0.137[136,137] 157ns                                                                    |L0.137|                  "
    - "L0.138[137,138] 158ns                                                                    |L0.138|                  "
    - "L0.139[138,139] 159ns                                                                     |L0.139|                 "
    - "L0.140[139,140] 160ns                                                                     |L0.140|                 "
    - "L0.141[140,141] 161ns                                                                      |L0.141|                "
    - "L0.142[141,142] 162ns                                                                      |L0.142|                "
    - "L0.143[142,143] 163ns                                                                       |L0.143|               "
    - "L0.144[143,144] 164ns                                                                       |L0.144|               "
    - "L0.145[144,145] 165ns                                                                        |L0.145|              "
    - "L0.146[145,146] 166ns                                                                        |L0.146|              "
    - "L0.147[146,147] 167ns                                                                         |L0.147|             "
    - "L0.148[147,148] 168ns                                                                         |L0.148|             "
    - "L0.149[148,149] 169ns                                                                          |L0.149|            "
    - "L0.150[149,150] 170ns                                                                          |L0.150|            "
    - "L0.151[150,151] 171ns                                                                           |L0.151|           "
    - "L0.152[151,152] 172ns                                                                           |L0.152|           "
    - "L0.153[152,153] 173ns                                                                            |L0.153|          "
    - "L0.154[153,154] 174ns                                                                            |L0.154|          "
    - "L0.155[154,155] 175ns                                                                            |L0.155|          "
    - "L0.156[155,156] 176ns                                                                             |L0.156|         "
    - "L0.157[156,157] 177ns                                                                             |L0.157|         "
    - "L0.158[157,158] 178ns                                                                              |L0.158|        "
    - "L0.159[158,159] 179ns                                                                              |L0.159|        "
    - "L0.160[159,160] 180ns                                                                               |L0.160|       "
    - "L0.161[160,161] 181ns                                                                               |L0.161|       "
    - "L0.162[161,162] 182ns                                                                                |L0.162|      "
    - "L0.163[162,163] 183ns                                                                                |L0.163|      "
    - "L0.164[163,164] 184ns                                                                                 |L0.164|     "
    - "L0.165[164,165] 185ns                                                                                 |L0.165|     "
    - "L0.166[165,166] 186ns                                                                                  |L0.166|    "
    - "L0.167[166,167] 187ns                                                                                  |L0.167|    "
    - "L0.168[167,168] 188ns                                                                                   |L0.168|   "
    - "L0.169[168,169] 189ns                                                                                   |L0.169|   "
    - "L0.170[169,170] 190ns                                                                                    |L0.170|  "
    - "L0.171[170,171] 191ns                                                                                    |L0.171|  "
    - "L0.172[171,172] 192ns                                                                                     |L0.172| "
    - "L0.173[172,173] 193ns                                                                                     |L0.173| "
    - "L0.174[173,174] 194ns                                                                                     |L0.174| "
    - "L0.175[174,175] 195ns                                                                                      |L0.175|"
    - "L0.176[175,176] 196ns                                                                                      |L0.176|"
    - "L0.177[176,177] 197ns                                                                                       |L0.177|"
    - "L0.178[177,178] 198ns                                                                                       |L0.178|"
    - "L0.179[178,179] 199ns                                                                                        |L0.179|"
    - "L0.180[179,180] 200ns                                                                                        |L0.180|"
    - "L0.181[180,181] 201ns                                                                                         |L0.181|"
    - "L0.182[181,182] 202ns                                                                                         |L0.182|"
    - "L0.183[182,183] 203ns                                                                                          |L0.183|"
    - "L0.184[183,184] 204ns                                                                                          |L0.184|"
    - "L0.185[184,185] 205ns                                                                                           |L0.185|"
    - "L0.186[185,186] 206ns                                                                                           |L0.186|"
    - "L0.187[186,187] 207ns                                                                                            |L0.187|"
    - "L0.188[187,188] 208ns                                                                                            |L0.188|"
    - "L0.189[188,189] 209ns                                                                                             |L0.189|"
    - "L0.190[189,190] 210ns                                                                                             |L0.190|"
    - "**** 1 Output Files (parquet_file_id not yet assigned), 1.3mb total:"
    - "L0, all files 1.3mb                                                                                                "
    - "L0.?[0,190] 210ns        |------------------------------------------L0.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 190 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10, L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33, L0.34, L0.35, L0.36, L0.37, L0.38, L0.39, L0.40, L0.41, L0.42, L0.43, L0.44, L0.45, L0.46, L0.47, L0.48, L0.49, L0.50, L0.51, L0.52, L0.53, L0.54, L0.55, L0.56, L0.57, L0.58, L0.59, L0.60, L0.61, L0.62, L0.63, L0.64, L0.65, L0.66, L0.67, L0.68, L0.69, L0.70, L0.71, L0.72, L0.73, L0.74, L0.75, L0.76, L0.77, L0.78, L0.79, L0.80, L0.81, L0.82, L0.83, L0.84, L0.85, L0.86, L0.87, L0.88, L0.89, L0.90, L0.91, L0.92, L0.93, L0.94, L0.95, L0.96, L0.97, L0.98, L0.99, L0.100, L0.101, L0.102, L0.103, L0.104, L0.105, L0.106, L0.107, L0.108, L0.109, L0.110, L0.111, L0.112, L0.113, L0.114, L0.115, L0.116, L0.117, L0.118, L0.119, L0.120, L0.121, L0.122, L0.123, L0.124, L0.125, L0.126, L0.127, L0.128, L0.129, L0.130, L0.131, L0.132, L0.133, L0.134, L0.135, L0.136, L0.137, L0.138, L0.139, L0.140, L0.141, L0.142, L0.143, L0.144, L0.145, L0.146, L0.147, L0.148, L0.149, L0.150, L0.151, L0.152, L0.153, L0.154, L0.155, L0.156, L0.157, L0.158, L0.159, L0.160, L0.161, L0.162, L0.163, L0.164, L0.165, L0.166, L0.167, L0.168, L0.169, L0.170, L0.171, L0.172, L0.173, L0.174, L0.175, L0.176, L0.177, L0.178, L0.179, L0.180, L0.181, L0.182, L0.183, L0.184, L0.185, L0.186, L0.187, L0.188, L0.189, L0.190"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[159]). 21 Input Files, 21.3mb total:"
    - "L0                                                                                                                 "
    - "L0.211[0,190] 210ns 1.3mb|--------------------------------------L0.211---------------------------------------|     "
    - "L1                                                                                                                 "
    - "L1.191[0,9] 1ns 1mb      |L1.191|                                                                                  "
    - "L1.210[190,199] 20ns 1mb                                                                                      |L1.210|"
    - "L1.192[10,19] 2ns 1mb        |L1.192|                                                                              "
    - "L1.193[20,29] 3ns 1mb             |L1.193|                                                                         "
    - "L1.194[30,39] 4ns 1mb                 |L1.194|                                                                     "
    - "L1.195[40,49] 5ns 1mb                      |L1.195|                                                                "
    - "L1.196[50,59] 6ns 1mb                          |L1.196|                                                            "
    - "L1.197[60,69] 7ns 1mb                               |L1.197|                                                       "
    - "L1.198[70,79] 8ns 1mb                                   |L1.198|                                                   "
    - "L1.199[80,89] 9ns 1mb                                        |L1.199|                                              "
    - "L1.200[90,99] 10ns 1mb                                           |L1.200|                                          "
    - "L1.201[100,109] 11ns 1mb                                              |L1.201|                                     "
    - "L1.202[110,119] 12ns 1mb                                                  |L1.202|                                 "
    - "L1.203[120,129] 13ns 1mb                                                       |L1.203|                            "
    - "L1.204[130,139] 14ns 1mb                                                           |L1.204|                        "
    - "L1.205[140,149] 15ns 1mb                                                                |L1.205|                   "
    - "L1.206[150,159] 16ns 1mb                                                                    |L1.206|               "
    - "L1.207[160,169] 17ns 1mb                                                                         |L1.207|          "
    - "L1.208[170,179] 18ns 1mb                                                                             |L1.208|      "
    - "L1.209[180,189] 19ns 1mb                                                                                  |L1.209| "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 21.3mb total:"
    - "L1                                                                                                                 "
    - "L1.?[0,159] 210ns 17.02mb|--------------------------------L1.?---------------------------------|                   "
    - "L1.?[160,199] 210ns 4.28mb                                                                        |-----L1.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 21 files: L1.191, L1.192, L1.193, L1.194, L1.195, L1.196, L1.197, L1.198, L1.199, L1.200, L1.201, L1.202, L1.203, L1.204, L1.205, L1.206, L1.207, L1.208, L1.209, L1.210, L0.211"
    - "  Creating 2 files"
    - "**** Final Output Files (22.6mb written)"
    - "L1                                                                                                                 "
    - "L1.212[0,159] 210ns 17.02mb|-------------------------------L1.212--------------------------------|                   "
    - "L1.213[160,199] 210ns 4.28mb                                                                        |----L1.213-----| "
    "###
    );
}

#[tokio::test]
async fn not_many_l0_and_overlapped_l1_files() {
    test_helpers::maybe_start_logging();

    // Observed size on production
    // num L0 files + L1 files > max files per plan but
    // num L0 files + num non-overlapped L1 files < max files per plan
    let seven_kb = 7 * 1024;
    let one_mb = 1024 * 1024;

    let setup = layout_setup_builder().await.build().await;

    // create 190 L0 files & 20 L1 files
    let num_tiny_l0_files = 190;
    let num_overlapped_l1_files = 5;
    let num_non_overlaped_l1_files = 15;

    for i in 0..num_tiny_l0_files {
        let i = i as i64;
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i)
                    .with_max_time(i + 1)
                    .with_compaction_level(CompactionLevel::Initial)
                    // L0s created after L1s
                    .with_max_l0_created_at(Time::from_timestamp_nanos(
                        num_overlapped_l1_files + num_non_overlaped_l1_files + i + 1,
                    ))
                    .with_file_size_bytes(seven_kb),
            )
            .await;
    }

    // create 5 overlapped L1
    for i in 0..num_overlapped_l1_files {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time(i * 10)
                    .with_max_time(i * 10 + 9)
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(i + 1))
                    .with_file_size_bytes(one_mb),
            )
            .await;
    }

    // create 15 non-overlapped L1
    for i in 0..num_non_overlaped_l1_files {
        setup
            .partition
            .create_parquet_file(
                parquet_builder()
                    .with_min_time((i + 1) * 200)
                    .with_max_time((i + 1) * 200 + 9)
                    .with_compaction_level(CompactionLevel::FileNonOverlapped)
                    .with_max_l0_created_at(Time::from_timestamp_nanos(
                        num_overlapped_l1_files + i + 1,
                    ))
                    .with_file_size_bytes(one_mb),
            )
            .await;
    }

    // Will cpmpact all 190 L0s with 5 overlapped L1s first.
    // Then compact the resulted L1 with the other 15 non-overlapped L1s into L2s

    insta::assert_yaml_snapshot!(
        run_layout_scenario(&setup).await,
        @r###"
    ---
    - "**** Input Files "
    - "L0                                                                                                                 "
    - "L0.1[0,1] 21ns 7kb       |L0.1|                                                                                    "
    - "L0.2[1,2] 22ns 7kb       |L0.2|                                                                                    "
    - "L0.3[2,3] 23ns 7kb       |L0.3|                                                                                    "
    - "L0.4[3,4] 24ns 7kb       |L0.4|                                                                                    "
    - "L0.5[4,5] 25ns 7kb       |L0.5|                                                                                    "
    - "L0.6[5,6] 26ns 7kb       |L0.6|                                                                                    "
    - "L0.7[6,7] 27ns 7kb       |L0.7|                                                                                    "
    - "L0.8[7,8] 28ns 7kb       |L0.8|                                                                                    "
    - "L0.9[8,9] 29ns 7kb       |L0.9|                                                                                    "
    - "L0.10[9,10] 30ns 7kb     |L0.10|                                                                                   "
    - "L0.11[10,11] 31ns 7kb    |L0.11|                                                                                   "
    - "L0.12[11,12] 32ns 7kb    |L0.12|                                                                                   "
    - "L0.13[12,13] 33ns 7kb    |L0.13|                                                                                   "
    - "L0.14[13,14] 34ns 7kb    |L0.14|                                                                                   "
    - "L0.15[14,15] 35ns 7kb    |L0.15|                                                                                   "
    - "L0.16[15,16] 36ns 7kb    |L0.16|                                                                                   "
    - "L0.17[16,17] 37ns 7kb    |L0.17|                                                                                   "
    - "L0.18[17,18] 38ns 7kb    |L0.18|                                                                                   "
    - "L0.19[18,19] 39ns 7kb    |L0.19|                                                                                   "
    - "L0.20[19,20] 40ns 7kb    |L0.20|                                                                                   "
    - "L0.21[20,21] 41ns 7kb    |L0.21|                                                                                   "
    - "L0.22[21,22] 42ns 7kb    |L0.22|                                                                                   "
    - "L0.23[22,23] 43ns 7kb    |L0.23|                                                                                   "
    - "L0.24[23,24] 44ns 7kb    |L0.24|                                                                                   "
    - "L0.25[24,25] 45ns 7kb    |L0.25|                                                                                   "
    - "L0.26[25,26] 46ns 7kb    |L0.26|                                                                                   "
    - "L0.27[26,27] 47ns 7kb    |L0.27|                                                                                   "
    - "L0.28[27,28] 48ns 7kb    |L0.28|                                                                                   "
    - "L0.29[28,29] 49ns 7kb    |L0.29|                                                                                   "
    - "L0.30[29,30] 50ns 7kb    |L0.30|                                                                                   "
    - "L0.31[30,31] 51ns 7kb    |L0.31|                                                                                   "
    - "L0.32[31,32] 52ns 7kb    |L0.32|                                                                                   "
    - "L0.33[32,33] 53ns 7kb    |L0.33|                                                                                   "
    - "L0.34[33,34] 54ns 7kb    |L0.34|                                                                                   "
    - "L0.35[34,35] 55ns 7kb     |L0.35|                                                                                  "
    - "L0.36[35,36] 56ns 7kb     |L0.36|                                                                                  "
    - "L0.37[36,37] 57ns 7kb     |L0.37|                                                                                  "
    - "L0.38[37,38] 58ns 7kb     |L0.38|                                                                                  "
    - "L0.39[38,39] 59ns 7kb     |L0.39|                                                                                  "
    - "L0.40[39,40] 60ns 7kb     |L0.40|                                                                                  "
    - "L0.41[40,41] 61ns 7kb     |L0.41|                                                                                  "
    - "L0.42[41,42] 62ns 7kb     |L0.42|                                                                                  "
    - "L0.43[42,43] 63ns 7kb     |L0.43|                                                                                  "
    - "L0.44[43,44] 64ns 7kb     |L0.44|                                                                                  "
    - "L0.45[44,45] 65ns 7kb     |L0.45|                                                                                  "
    - "L0.46[45,46] 66ns 7kb     |L0.46|                                                                                  "
    - "L0.47[46,47] 67ns 7kb     |L0.47|                                                                                  "
    - "L0.48[47,48] 68ns 7kb     |L0.48|                                                                                  "
    - "L0.49[48,49] 69ns 7kb     |L0.49|                                                                                  "
    - "L0.50[49,50] 70ns 7kb     |L0.50|                                                                                  "
    - "L0.51[50,51] 71ns 7kb     |L0.51|                                                                                  "
    - "L0.52[51,52] 72ns 7kb     |L0.52|                                                                                  "
    - "L0.53[52,53] 73ns 7kb     |L0.53|                                                                                  "
    - "L0.54[53,54] 74ns 7kb     |L0.54|                                                                                  "
    - "L0.55[54,55] 75ns 7kb     |L0.55|                                                                                  "
    - "L0.56[55,56] 76ns 7kb     |L0.56|                                                                                  "
    - "L0.57[56,57] 77ns 7kb     |L0.57|                                                                                  "
    - "L0.58[57,58] 78ns 7kb     |L0.58|                                                                                  "
    - "L0.59[58,59] 79ns 7kb     |L0.59|                                                                                  "
    - "L0.60[59,60] 80ns 7kb     |L0.60|                                                                                  "
    - "L0.61[60,61] 81ns 7kb     |L0.61|                                                                                  "
    - "L0.62[61,62] 82ns 7kb     |L0.62|                                                                                  "
    - "L0.63[62,63] 83ns 7kb     |L0.63|                                                                                  "
    - "L0.64[63,64] 84ns 7kb     |L0.64|                                                                                  "
    - "L0.65[64,65] 85ns 7kb     |L0.65|                                                                                  "
    - "L0.66[65,66] 86ns 7kb     |L0.66|                                                                                  "
    - "L0.67[66,67] 87ns 7kb     |L0.67|                                                                                  "
    - "L0.68[67,68] 88ns 7kb      |L0.68|                                                                                 "
    - "L0.69[68,69] 89ns 7kb      |L0.69|                                                                                 "
    - "L0.70[69,70] 90ns 7kb      |L0.70|                                                                                 "
    - "L0.71[70,71] 91ns 7kb      |L0.71|                                                                                 "
    - "L0.72[71,72] 92ns 7kb      |L0.72|                                                                                 "
    - "L0.73[72,73] 93ns 7kb      |L0.73|                                                                                 "
    - "L0.74[73,74] 94ns 7kb      |L0.74|                                                                                 "
    - "L0.75[74,75] 95ns 7kb      |L0.75|                                                                                 "
    - "L0.76[75,76] 96ns 7kb      |L0.76|                                                                                 "
    - "L0.77[76,77] 97ns 7kb      |L0.77|                                                                                 "
    - "L0.78[77,78] 98ns 7kb      |L0.78|                                                                                 "
    - "L0.79[78,79] 99ns 7kb      |L0.79|                                                                                 "
    - "L0.80[79,80] 100ns 7kb     |L0.80|                                                                                 "
    - "L0.81[80,81] 101ns 7kb     |L0.81|                                                                                 "
    - "L0.82[81,82] 102ns 7kb     |L0.82|                                                                                 "
    - "L0.83[82,83] 103ns 7kb     |L0.83|                                                                                 "
    - "L0.84[83,84] 104ns 7kb     |L0.84|                                                                                 "
    - "L0.85[84,85] 105ns 7kb     |L0.85|                                                                                 "
    - "L0.86[85,86] 106ns 7kb     |L0.86|                                                                                 "
    - "L0.87[86,87] 107ns 7kb     |L0.87|                                                                                 "
    - "L0.88[87,88] 108ns 7kb     |L0.88|                                                                                 "
    - "L0.89[88,89] 109ns 7kb     |L0.89|                                                                                 "
    - "L0.90[89,90] 110ns 7kb     |L0.90|                                                                                 "
    - "L0.91[90,91] 111ns 7kb     |L0.91|                                                                                 "
    - "L0.92[91,92] 112ns 7kb     |L0.92|                                                                                 "
    - "L0.93[92,93] 113ns 7kb     |L0.93|                                                                                 "
    - "L0.94[93,94] 114ns 7kb     |L0.94|                                                                                 "
    - "L0.95[94,95] 115ns 7kb     |L0.95|                                                                                 "
    - "L0.96[95,96] 116ns 7kb     |L0.96|                                                                                 "
    - "L0.97[96,97] 117ns 7kb     |L0.97|                                                                                 "
    - "L0.98[97,98] 118ns 7kb     |L0.98|                                                                                 "
    - "L0.99[98,99] 119ns 7kb     |L0.99|                                                                                 "
    - "L0.100[99,100] 120ns 7kb   |L0.100|                                                                                "
    - "L0.101[100,101] 121ns 7kb  |L0.101|                                                                                "
    - "L0.102[101,102] 122ns 7kb   |L0.102|                                                                               "
    - "L0.103[102,103] 123ns 7kb   |L0.103|                                                                               "
    - "L0.104[103,104] 124ns 7kb   |L0.104|                                                                               "
    - "L0.105[104,105] 125ns 7kb   |L0.105|                                                                               "
    - "L0.106[105,106] 126ns 7kb   |L0.106|                                                                               "
    - "L0.107[106,107] 127ns 7kb   |L0.107|                                                                               "
    - "L0.108[107,108] 128ns 7kb   |L0.108|                                                                               "
    - "L0.109[108,109] 129ns 7kb   |L0.109|                                                                               "
    - "L0.110[109,110] 130ns 7kb   |L0.110|                                                                               "
    - "L0.111[110,111] 131ns 7kb   |L0.111|                                                                               "
    - "L0.112[111,112] 132ns 7kb   |L0.112|                                                                               "
    - "L0.113[112,113] 133ns 7kb   |L0.113|                                                                               "
    - "L0.114[113,114] 134ns 7kb   |L0.114|                                                                               "
    - "L0.115[114,115] 135ns 7kb   |L0.115|                                                                               "
    - "L0.116[115,116] 136ns 7kb   |L0.116|                                                                               "
    - "L0.117[116,117] 137ns 7kb   |L0.117|                                                                               "
    - "L0.118[117,118] 138ns 7kb   |L0.118|                                                                               "
    - "L0.119[118,119] 139ns 7kb   |L0.119|                                                                               "
    - "L0.120[119,120] 140ns 7kb   |L0.120|                                                                               "
    - "L0.121[120,121] 141ns 7kb   |L0.121|                                                                               "
    - "L0.122[121,122] 142ns 7kb   |L0.122|                                                                               "
    - "L0.123[122,123] 143ns 7kb   |L0.123|                                                                               "
    - "L0.124[123,124] 144ns 7kb   |L0.124|                                                                               "
    - "L0.125[124,125] 145ns 7kb   |L0.125|                                                                               "
    - "L0.126[125,126] 146ns 7kb   |L0.126|                                                                               "
    - "L0.127[126,127] 147ns 7kb   |L0.127|                                                                               "
    - "L0.128[127,128] 148ns 7kb   |L0.128|                                                                               "
    - "L0.129[128,129] 149ns 7kb   |L0.129|                                                                               "
    - "L0.130[129,130] 150ns 7kb   |L0.130|                                                                               "
    - "L0.131[130,131] 151ns 7kb   |L0.131|                                                                               "
    - "L0.132[131,132] 152ns 7kb   |L0.132|                                                                               "
    - "L0.133[132,133] 153ns 7kb   |L0.133|                                                                               "
    - "L0.134[133,134] 154ns 7kb   |L0.134|                                                                               "
    - "L0.135[134,135] 155ns 7kb    |L0.135|                                                                              "
    - "L0.136[135,136] 156ns 7kb    |L0.136|                                                                              "
    - "L0.137[136,137] 157ns 7kb    |L0.137|                                                                              "
    - "L0.138[137,138] 158ns 7kb    |L0.138|                                                                              "
    - "L0.139[138,139] 159ns 7kb    |L0.139|                                                                              "
    - "L0.140[139,140] 160ns 7kb    |L0.140|                                                                              "
    - "L0.141[140,141] 161ns 7kb    |L0.141|                                                                              "
    - "L0.142[141,142] 162ns 7kb    |L0.142|                                                                              "
    - "L0.143[142,143] 163ns 7kb    |L0.143|                                                                              "
    - "L0.144[143,144] 164ns 7kb    |L0.144|                                                                              "
    - "L0.145[144,145] 165ns 7kb    |L0.145|                                                                              "
    - "L0.146[145,146] 166ns 7kb    |L0.146|                                                                              "
    - "L0.147[146,147] 167ns 7kb    |L0.147|                                                                              "
    - "L0.148[147,148] 168ns 7kb    |L0.148|                                                                              "
    - "L0.149[148,149] 169ns 7kb    |L0.149|                                                                              "
    - "L0.150[149,150] 170ns 7kb    |L0.150|                                                                              "
    - "L0.151[150,151] 171ns 7kb    |L0.151|                                                                              "
    - "L0.152[151,152] 172ns 7kb    |L0.152|                                                                              "
    - "L0.153[152,153] 173ns 7kb    |L0.153|                                                                              "
    - "L0.154[153,154] 174ns 7kb    |L0.154|                                                                              "
    - "L0.155[154,155] 175ns 7kb    |L0.155|                                                                              "
    - "L0.156[155,156] 176ns 7kb    |L0.156|                                                                              "
    - "L0.157[156,157] 177ns 7kb    |L0.157|                                                                              "
    - "L0.158[157,158] 178ns 7kb    |L0.158|                                                                              "
    - "L0.159[158,159] 179ns 7kb    |L0.159|                                                                              "
    - "L0.160[159,160] 180ns 7kb    |L0.160|                                                                              "
    - "L0.161[160,161] 181ns 7kb    |L0.161|                                                                              "
    - "L0.162[161,162] 182ns 7kb    |L0.162|                                                                              "
    - "L0.163[162,163] 183ns 7kb    |L0.163|                                                                              "
    - "L0.164[163,164] 184ns 7kb    |L0.164|                                                                              "
    - "L0.165[164,165] 185ns 7kb    |L0.165|                                                                              "
    - "L0.166[165,166] 186ns 7kb    |L0.166|                                                                              "
    - "L0.167[166,167] 187ns 7kb    |L0.167|                                                                              "
    - "L0.168[167,168] 188ns 7kb    |L0.168|                                                                              "
    - "L0.169[168,169] 189ns 7kb     |L0.169|                                                                             "
    - "L0.170[169,170] 190ns 7kb     |L0.170|                                                                             "
    - "L0.171[170,171] 191ns 7kb     |L0.171|                                                                             "
    - "L0.172[171,172] 192ns 7kb     |L0.172|                                                                             "
    - "L0.173[172,173] 193ns 7kb     |L0.173|                                                                             "
    - "L0.174[173,174] 194ns 7kb     |L0.174|                                                                             "
    - "L0.175[174,175] 195ns 7kb     |L0.175|                                                                             "
    - "L0.176[175,176] 196ns 7kb     |L0.176|                                                                             "
    - "L0.177[176,177] 197ns 7kb     |L0.177|                                                                             "
    - "L0.178[177,178] 198ns 7kb     |L0.178|                                                                             "
    - "L0.179[178,179] 199ns 7kb     |L0.179|                                                                             "
    - "L0.180[179,180] 200ns 7kb     |L0.180|                                                                             "
    - "L0.181[180,181] 201ns 7kb     |L0.181|                                                                             "
    - "L0.182[181,182] 202ns 7kb     |L0.182|                                                                             "
    - "L0.183[182,183] 203ns 7kb     |L0.183|                                                                             "
    - "L0.184[183,184] 204ns 7kb     |L0.184|                                                                             "
    - "L0.185[184,185] 205ns 7kb     |L0.185|                                                                             "
    - "L0.186[185,186] 206ns 7kb     |L0.186|                                                                             "
    - "L0.187[186,187] 207ns 7kb     |L0.187|                                                                             "
    - "L0.188[187,188] 208ns 7kb     |L0.188|                                                                             "
    - "L0.189[188,189] 209ns 7kb     |L0.189|                                                                             "
    - "L0.190[189,190] 210ns 7kb     |L0.190|                                                                             "
    - "L1                                                                                                                 "
    - "L1.191[0,9] 1ns 1mb      |L1.191|                                                                                  "
    - "L1.192[10,19] 2ns 1mb    |L1.192|                                                                                  "
    - "L1.193[20,29] 3ns 1mb    |L1.193|                                                                                  "
    - "L1.194[30,39] 4ns 1mb    |L1.194|                                                                                  "
    - "L1.195[40,49] 5ns 1mb     |L1.195|                                                                                 "
    - "L1.196[200,209] 6ns 1mb       |L1.196|                                                                             "
    - "L1.197[400,409] 7ns 1mb             |L1.197|                                                                       "
    - "L1.198[600,609] 8ns 1mb                   |L1.198|                                                                 "
    - "L1.199[800,809] 9ns 1mb                         |L1.199|                                                           "
    - "L1.200[1000,1009] 10ns 1mb                             |L1.200|                                                     "
    - "L1.201[1200,1209] 11ns 1mb                                   |L1.201|                                               "
    - "L1.202[1400,1409] 12ns 1mb                                         |L1.202|                                         "
    - "L1.203[1600,1609] 13ns 1mb                                               |L1.203|                                   "
    - "L1.204[1800,1809] 14ns 1mb                                                     |L1.204|                             "
    - "L1.205[2000,2009] 15ns 1mb                                                           |L1.205|                       "
    - "L1.206[2200,2209] 16ns 1mb                                                                 |L1.206|                 "
    - "L1.207[2400,2409] 17ns 1mb                                                                       |L1.207|           "
    - "L1.208[2600,2609] 18ns 1mb                                                                             |L1.208|     "
    - "L1.209[2800,2809] 19ns 1mb                                                                                   |L1.209|"
    - "L1.210[3000,3009] 20ns 1mb                                                                                         |L1.210|"
    - "**** Simulation run 0, type=compact(TotalSizeLessThanMaxCompactSize). 195 Input Files, 6.3mb total:"
    - "L0                                                                                                                 "
    - "L0.190[189,190] 210ns 7kb                                                                                         |L0.190|"
    - "L0.189[188,189] 209ns 7kb                                                                                         |L0.189|"
    - "L0.188[187,188] 208ns 7kb                                                                                        |L0.188|"
    - "L0.187[186,187] 207ns 7kb                                                                                        |L0.187|"
    - "L0.186[185,186] 206ns 7kb                                                                                       |L0.186|"
    - "L0.185[184,185] 205ns 7kb                                                                                       |L0.185|"
    - "L0.184[183,184] 204ns 7kb                                                                                      |L0.184|"
    - "L0.183[182,183] 203ns 7kb                                                                                      |L0.183|"
    - "L0.182[181,182] 202ns 7kb                                                                                     |L0.182|"
    - "L0.181[180,181] 201ns 7kb                                                                                     |L0.181|"
    - "L0.180[179,180] 200ns 7kb                                                                                    |L0.180|"
    - "L0.179[178,179] 199ns 7kb                                                                                    |L0.179|"
    - "L0.178[177,178] 198ns 7kb                                                                                   |L0.178|"
    - "L0.177[176,177] 197ns 7kb                                                                                   |L0.177|"
    - "L0.176[175,176] 196ns 7kb                                                                                  |L0.176|"
    - "L0.175[174,175] 195ns 7kb                                                                                  |L0.175|"
    - "L0.174[173,174] 194ns 7kb                                                                                 |L0.174| "
    - "L0.173[172,173] 193ns 7kb                                                                                 |L0.173| "
    - "L0.172[171,172] 192ns 7kb                                                                                 |L0.172| "
    - "L0.171[170,171] 191ns 7kb                                                                                |L0.171|  "
    - "L0.170[169,170] 190ns 7kb                                                                                |L0.170|  "
    - "L0.169[168,169] 189ns 7kb                                                                               |L0.169|   "
    - "L0.168[167,168] 188ns 7kb                                                                               |L0.168|   "
    - "L0.167[166,167] 187ns 7kb                                                                              |L0.167|    "
    - "L0.166[165,166] 186ns 7kb                                                                              |L0.166|    "
    - "L0.165[164,165] 185ns 7kb                                                                             |L0.165|     "
    - "L0.164[163,164] 184ns 7kb                                                                             |L0.164|     "
    - "L0.163[162,163] 183ns 7kb                                                                            |L0.163|      "
    - "L0.162[161,162] 182ns 7kb                                                                            |L0.162|      "
    - "L0.161[160,161] 181ns 7kb                                                                           |L0.161|       "
    - "L0.160[159,160] 180ns 7kb                                                                           |L0.160|       "
    - "L0.159[158,159] 179ns 7kb                                                                          |L0.159|        "
    - "L0.158[157,158] 178ns 7kb                                                                          |L0.158|        "
    - "L0.157[156,157] 177ns 7kb                                                                         |L0.157|         "
    - "L0.156[155,156] 176ns 7kb                                                                         |L0.156|         "
    - "L0.155[154,155] 175ns 7kb                                                                        |L0.155|          "
    - "L0.154[153,154] 174ns 7kb                                                                        |L0.154|          "
    - "L0.153[152,153] 173ns 7kb                                                                        |L0.153|          "
    - "L0.152[151,152] 172ns 7kb                                                                       |L0.152|           "
    - "L0.151[150,151] 171ns 7kb                                                                       |L0.151|           "
    - "L0.150[149,150] 170ns 7kb                                                                      |L0.150|            "
    - "L0.149[148,149] 169ns 7kb                                                                      |L0.149|            "
    - "L0.148[147,148] 168ns 7kb                                                                     |L0.148|             "
    - "L0.147[146,147] 167ns 7kb                                                                     |L0.147|             "
    - "L0.146[145,146] 166ns 7kb                                                                    |L0.146|              "
    - "L0.145[144,145] 165ns 7kb                                                                    |L0.145|              "
    - "L0.144[143,144] 164ns 7kb                                                                   |L0.144|               "
    - "L0.143[142,143] 163ns 7kb                                                                   |L0.143|               "
    - "L0.142[141,142] 162ns 7kb                                                                  |L0.142|                "
    - "L0.141[140,141] 161ns 7kb                                                                  |L0.141|                "
    - "L0.140[139,140] 160ns 7kb                                                                 |L0.140|                 "
    - "L0.139[138,139] 159ns 7kb                                                                 |L0.139|                 "
    - "L0.138[137,138] 158ns 7kb                                                                |L0.138|                  "
    - "L0.137[136,137] 157ns 7kb                                                                |L0.137|                  "
    - "L0.136[135,136] 156ns 7kb                                                               |L0.136|                   "
    - "L0.135[134,135] 155ns 7kb                                                               |L0.135|                   "
    - "L0.134[133,134] 154ns 7kb                                                               |L0.134|                   "
    - "L0.133[132,133] 153ns 7kb                                                              |L0.133|                    "
    - "L0.132[131,132] 152ns 7kb                                                              |L0.132|                    "
    - "L0.131[130,131] 151ns 7kb                                                             |L0.131|                     "
    - "L0.130[129,130] 150ns 7kb                                                             |L0.130|                     "
    - "L0.129[128,129] 149ns 7kb                                                            |L0.129|                      "
    - "L0.128[127,128] 148ns 7kb                                                            |L0.128|                      "
    - "L0.127[126,127] 147ns 7kb                                                           |L0.127|                       "
    - "L0.126[125,126] 146ns 7kb                                                           |L0.126|                       "
    - "L0.125[124,125] 145ns 7kb                                                          |L0.125|                        "
    - "L0.124[123,124] 144ns 7kb                                                          |L0.124|                        "
    - "L0.123[122,123] 143ns 7kb                                                         |L0.123|                         "
    - "L0.122[121,122] 142ns 7kb                                                         |L0.122|                         "
    - "L0.121[120,121] 141ns 7kb                                                        |L0.121|                          "
    - "L0.120[119,120] 140ns 7kb                                                        |L0.120|                          "
    - "L0.119[118,119] 139ns 7kb                                                       |L0.119|                           "
    - "L0.118[117,118] 138ns 7kb                                                       |L0.118|                           "
    - "L0.117[116,117] 137ns 7kb                                                      |L0.117|                            "
    - "L0.116[115,116] 136ns 7kb                                                      |L0.116|                            "
    - "L0.115[114,115] 135ns 7kb                                                      |L0.115|                            "
    - "L0.114[113,114] 134ns 7kb                                                     |L0.114|                             "
    - "L0.113[112,113] 133ns 7kb                                                     |L0.113|                             "
    - "L0.112[111,112] 132ns 7kb                                                    |L0.112|                              "
    - "L0.111[110,111] 131ns 7kb                                                    |L0.111|                              "
    - "L0.110[109,110] 130ns 7kb                                                   |L0.110|                               "
    - "L0.109[108,109] 129ns 7kb                                                   |L0.109|                               "
    - "L0.108[107,108] 128ns 7kb                                                  |L0.108|                                "
    - "L0.107[106,107] 127ns 7kb                                                  |L0.107|                                "
    - "L0.106[105,106] 126ns 7kb                                                 |L0.106|                                 "
    - "L0.105[104,105] 125ns 7kb                                                 |L0.105|                                 "
    - "L0.104[103,104] 124ns 7kb                                                |L0.104|                                  "
    - "L0.103[102,103] 123ns 7kb                                                |L0.103|                                  "
    - "L0.102[101,102] 122ns 7kb                                               |L0.102|                                   "
    - "L0.101[100,101] 121ns 7kb                                               |L0.101|                                   "
    - "L0.100[99,100] 120ns 7kb                                               |L0.100|                                    "
    - "L0.99[98,99] 119ns 7kb                                                 |L0.99|                                     "
    - "L0.98[97,98] 118ns 7kb                                                |L0.98|                                      "
    - "L0.97[96,97] 117ns 7kb                                                |L0.97|                                      "
    - "L0.96[95,96] 116ns 7kb                                                |L0.96|                                      "
    - "L0.95[94,95] 115ns 7kb                                               |L0.95|                                       "
    - "L0.94[93,94] 114ns 7kb                                               |L0.94|                                       "
    - "L0.93[92,93] 113ns 7kb                                              |L0.93|                                        "
    - "L0.92[91,92] 112ns 7kb                                              |L0.92|                                        "
    - "L0.91[90,91] 111ns 7kb                                             |L0.91|                                         "
    - "L0.90[89,90] 110ns 7kb                                             |L0.90|                                         "
    - "L0.89[88,89] 109ns 7kb                                            |L0.89|                                          "
    - "L0.88[87,88] 108ns 7kb                                            |L0.88|                                          "
    - "L0.87[86,87] 107ns 7kb                                           |L0.87|                                           "
    - "L0.86[85,86] 106ns 7kb                                           |L0.86|                                           "
    - "L0.85[84,85] 105ns 7kb                                          |L0.85|                                            "
    - "L0.84[83,84] 104ns 7kb                                          |L0.84|                                            "
    - "L0.83[82,83] 103ns 7kb                                         |L0.83|                                             "
    - "L0.82[81,82] 102ns 7kb                                         |L0.82|                                             "
    - "L0.81[80,81] 101ns 7kb                                        |L0.81|                                              "
    - "L0.80[79,80] 100ns 7kb                                        |L0.80|                                              "
    - "L0.79[78,79] 99ns 7kb                                        |L0.79|                                               "
    - "L0.78[77,78] 98ns 7kb                                        |L0.78|                                               "
    - "L0.77[76,77] 97ns 7kb                                        |L0.77|                                               "
    - "L0.76[75,76] 96ns 7kb                                       |L0.76|                                                "
    - "L0.75[74,75] 95ns 7kb                                       |L0.75|                                                "
    - "L0.74[73,74] 94ns 7kb                                      |L0.74|                                                 "
    - "L0.73[72,73] 93ns 7kb                                      |L0.73|                                                 "
    - "L0.72[71,72] 92ns 7kb                                     |L0.72|                                                  "
    - "L0.71[70,71] 91ns 7kb                                     |L0.71|                                                  "
    - "L0.70[69,70] 90ns 7kb                                    |L0.70|                                                   "
    - "L0.69[68,69] 89ns 7kb                                    |L0.69|                                                   "
    - "L0.68[67,68] 88ns 7kb                                   |L0.68|                                                    "
    - "L0.67[66,67] 87ns 7kb                                   |L0.67|                                                    "
    - "L0.66[65,66] 86ns 7kb                                  |L0.66|                                                     "
    - "L0.65[64,65] 85ns 7kb                                  |L0.65|                                                     "
    - "L0.64[63,64] 84ns 7kb                                 |L0.64|                                                      "
    - "L0.63[62,63] 83ns 7kb                                 |L0.63|                                                      "
    - "L0.62[61,62] 82ns 7kb                                |L0.62|                                                       "
    - "L0.61[60,61] 81ns 7kb                                |L0.61|                                                       "
    - "L0.60[59,60] 80ns 7kb                               |L0.60|                                                        "
    - "L0.59[58,59] 79ns 7kb                               |L0.59|                                                        "
    - "L0.58[57,58] 78ns 7kb                               |L0.58|                                                        "
    - "L0.57[56,57] 77ns 7kb                              |L0.57|                                                         "
    - "L0.56[55,56] 76ns 7kb                              |L0.56|                                                         "
    - "L0.55[54,55] 75ns 7kb                             |L0.55|                                                          "
    - "L0.54[53,54] 74ns 7kb                             |L0.54|                                                          "
    - "L0.53[52,53] 73ns 7kb                            |L0.53|                                                           "
    - "L0.52[51,52] 72ns 7kb                            |L0.52|                                                           "
    - "L0.51[50,51] 71ns 7kb                           |L0.51|                                                            "
    - "L0.50[49,50] 70ns 7kb                           |L0.50|                                                            "
    - "L0.49[48,49] 69ns 7kb                          |L0.49|                                                             "
    - "L0.48[47,48] 68ns 7kb                          |L0.48|                                                             "
    - "L0.47[46,47] 67ns 7kb                         |L0.47|                                                              "
    - "L0.46[45,46] 66ns 7kb                         |L0.46|                                                              "
    - "L0.45[44,45] 65ns 7kb                        |L0.45|                                                               "
    - "L0.44[43,44] 64ns 7kb                        |L0.44|                                                               "
    - "L0.43[42,43] 63ns 7kb                       |L0.43|                                                                "
    - "L0.42[41,42] 62ns 7kb                       |L0.42|                                                                "
    - "L0.41[40,41] 61ns 7kb                      |L0.41|                                                                 "
    - "L0.40[39,40] 60ns 7kb                      |L0.40|                                                                 "
    - "L0.39[38,39] 59ns 7kb                      |L0.39|                                                                 "
    - "L0.38[37,38] 58ns 7kb                     |L0.38|                                                                  "
    - "L0.37[36,37] 57ns 7kb                     |L0.37|                                                                  "
    - "L0.36[35,36] 56ns 7kb                    |L0.36|                                                                   "
    - "L0.35[34,35] 55ns 7kb                    |L0.35|                                                                   "
    - "L0.34[33,34] 54ns 7kb                   |L0.34|                                                                    "
    - "L0.33[32,33] 53ns 7kb                   |L0.33|                                                                    "
    - "L0.32[31,32] 52ns 7kb                  |L0.32|                                                                     "
    - "L0.31[30,31] 51ns 7kb                  |L0.31|                                                                     "
    - "L0.30[29,30] 50ns 7kb                 |L0.30|                                                                      "
    - "L0.29[28,29] 49ns 7kb                 |L0.29|                                                                      "
    - "L0.28[27,28] 48ns 7kb                |L0.28|                                                                       "
    - "L0.27[26,27] 47ns 7kb                |L0.27|                                                                       "
    - "L0.26[25,26] 46ns 7kb               |L0.26|                                                                        "
    - "L0.25[24,25] 45ns 7kb               |L0.25|                                                                        "
    - "L0.24[23,24] 44ns 7kb              |L0.24|                                                                         "
    - "L0.23[22,23] 43ns 7kb              |L0.23|                                                                         "
    - "L0.22[21,22] 42ns 7kb             |L0.22|                                                                          "
    - "L0.21[20,21] 41ns 7kb             |L0.21|                                                                          "
    - "L0.20[19,20] 40ns 7kb             |L0.20|                                                                          "
    - "L0.19[18,19] 39ns 7kb            |L0.19|                                                                           "
    - "L0.18[17,18] 38ns 7kb            |L0.18|                                                                           "
    - "L0.17[16,17] 37ns 7kb           |L0.17|                                                                            "
    - "L0.16[15,16] 36ns 7kb           |L0.16|                                                                            "
    - "L0.15[14,15] 35ns 7kb          |L0.15|                                                                             "
    - "L0.14[13,14] 34ns 7kb          |L0.14|                                                                             "
    - "L0.13[12,13] 33ns 7kb         |L0.13|                                                                              "
    - "L0.12[11,12] 32ns 7kb         |L0.12|                                                                              "
    - "L0.11[10,11] 31ns 7kb        |L0.11|                                                                               "
    - "L0.10[9,10] 30ns 7kb         |L0.10|                                                                               "
    - "L0.9[8,9] 29ns 7kb          |L0.9|                                                                                 "
    - "L0.8[7,8] 28ns 7kb          |L0.8|                                                                                 "
    - "L0.7[6,7] 27ns 7kb         |L0.7|                                                                                  "
    - "L0.6[5,6] 26ns 7kb         |L0.6|                                                                                  "
    - "L0.5[4,5] 25ns 7kb        |L0.5|                                                                                   "
    - "L0.4[3,4] 24ns 7kb        |L0.4|                                                                                   "
    - "L0.3[2,3] 23ns 7kb       |L0.3|                                                                                    "
    - "L0.2[1,2] 22ns 7kb       |L0.2|                                                                                    "
    - "L0.1[0,1] 21ns 7kb       |L0.1|                                                                                    "
    - "L1                                                                                                                 "
    - "L1.191[0,9] 1ns 1mb      |L1.191|                                                                                  "
    - "L1.195[40,49] 5ns 1mb                      |L1.195|                                                                "
    - "L1.192[10,19] 2ns 1mb        |L1.192|                                                                              "
    - "L1.193[20,29] 3ns 1mb             |L1.193|                                                                         "
    - "L1.194[30,39] 4ns 1mb                  |L1.194|                                                                    "
    - "**** 1 Output Files (parquet_file_id not yet assigned), 6.3mb total:"
    - "L1, all files 6.3mb                                                                                                "
    - "L1.?[0,190] 210ns        |------------------------------------------L1.?------------------------------------------|"
    - "Committing partition 1:"
    - "  Soft Deleting 195 files: L0.1, L0.2, L0.3, L0.4, L0.5, L0.6, L0.7, L0.8, L0.9, L0.10, L0.11, L0.12, L0.13, L0.14, L0.15, L0.16, L0.17, L0.18, L0.19, L0.20, L0.21, L0.22, L0.23, L0.24, L0.25, L0.26, L0.27, L0.28, L0.29, L0.30, L0.31, L0.32, L0.33, L0.34, L0.35, L0.36, L0.37, L0.38, L0.39, L0.40, L0.41, L0.42, L0.43, L0.44, L0.45, L0.46, L0.47, L0.48, L0.49, L0.50, L0.51, L0.52, L0.53, L0.54, L0.55, L0.56, L0.57, L0.58, L0.59, L0.60, L0.61, L0.62, L0.63, L0.64, L0.65, L0.66, L0.67, L0.68, L0.69, L0.70, L0.71, L0.72, L0.73, L0.74, L0.75, L0.76, L0.77, L0.78, L0.79, L0.80, L0.81, L0.82, L0.83, L0.84, L0.85, L0.86, L0.87, L0.88, L0.89, L0.90, L0.91, L0.92, L0.93, L0.94, L0.95, L0.96, L0.97, L0.98, L0.99, L0.100, L0.101, L0.102, L0.103, L0.104, L0.105, L0.106, L0.107, L0.108, L0.109, L0.110, L0.111, L0.112, L0.113, L0.114, L0.115, L0.116, L0.117, L0.118, L0.119, L0.120, L0.121, L0.122, L0.123, L0.124, L0.125, L0.126, L0.127, L0.128, L0.129, L0.130, L0.131, L0.132, L0.133, L0.134, L0.135, L0.136, L0.137, L0.138, L0.139, L0.140, L0.141, L0.142, L0.143, L0.144, L0.145, L0.146, L0.147, L0.148, L0.149, L0.150, L0.151, L0.152, L0.153, L0.154, L0.155, L0.156, L0.157, L0.158, L0.159, L0.160, L0.161, L0.162, L0.163, L0.164, L0.165, L0.166, L0.167, L0.168, L0.169, L0.170, L0.171, L0.172, L0.173, L0.174, L0.175, L0.176, L0.177, L0.178, L0.179, L0.180, L0.181, L0.182, L0.183, L0.184, L0.185, L0.186, L0.187, L0.188, L0.189, L0.190, L1.191, L1.192, L1.193, L1.194, L1.195"
    - "  Creating 1 files"
    - "**** Simulation run 1, type=split(CompactAndSplitOutput(TotalSizeLessThanMaxCompactSize))(split_times=[2407]). 16 Input Files, 21.3mb total:"
    - "L1                                                                                                                 "
    - "L1.196[200,209] 6ns 1mb       |L1.196|                                                                             "
    - "L1.197[400,409] 7ns 1mb             |L1.197|                                                                       "
    - "L1.198[600,609] 8ns 1mb                   |L1.198|                                                                 "
    - "L1.199[800,809] 9ns 1mb                         |L1.199|                                                           "
    - "L1.200[1000,1009] 10ns 1mb                             |L1.200|                                                     "
    - "L1.201[1200,1209] 11ns 1mb                                   |L1.201|                                               "
    - "L1.202[1400,1409] 12ns 1mb                                         |L1.202|                                         "
    - "L1.203[1600,1609] 13ns 1mb                                               |L1.203|                                   "
    - "L1.204[1800,1809] 14ns 1mb                                                     |L1.204|                             "
    - "L1.205[2000,2009] 15ns 1mb                                                           |L1.205|                       "
    - "L1.206[2200,2209] 16ns 1mb                                                                 |L1.206|                 "
    - "L1.207[2400,2409] 17ns 1mb                                                                       |L1.207|           "
    - "L1.208[2600,2609] 18ns 1mb                                                                             |L1.208|     "
    - "L1.209[2800,2809] 19ns 1mb                                                                                   |L1.209|"
    - "L1.210[3000,3009] 20ns 1mb                                                                                         |L1.210|"
    - "L1.211[0,190] 210ns 6.3mb|L1.211|                                                                                  "
    - "**** 2 Output Files (parquet_file_id not yet assigned), 21.3mb total:"
    - "L2                                                                                                                 "
    - "L2.?[0,2407] 210ns 17.04mb|--------------------------------L2.?---------------------------------|                   "
    - "L2.?[2408,3009] 210ns 4.26mb                                                                        |-----L2.?------| "
    - "Committing partition 1:"
    - "  Soft Deleting 16 files: L1.196, L1.197, L1.198, L1.199, L1.200, L1.201, L1.202, L1.203, L1.204, L1.205, L1.206, L1.207, L1.208, L1.209, L1.210, L1.211"
    - "  Creating 2 files"
    - "**** Final Output Files (27.6mb written)"
    - "L2                                                                                                                 "
    - "L2.212[0,2407] 210ns 17.04mb|-------------------------------L2.212--------------------------------|                   "
    - "L2.213[2408,3009] 210ns 4.26mb                                                                        |----L2.213-----| "
    "###
    );
}
